/**
 * ASPEED 2050 I2C driver.
 * Copyright (C) 2008 Avocent Corp.
 * Avocent Corporation. Proprietary Information.
 *
 * This file is subject to the terms and conditions of the GNU 
 * General Public License. This program is distributed in the hope 
 * that it will be useful, but WITHOUT ANY WARRANTY; without even 
 * the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 *
 *
 *----------------------------------------------------------------------------\n
 *  MODULES     Linux device Driver                                           \n
 *----------------------------------------------------------------------------\n
 *  @file   aess_i2cdrv.c
 *  @brief  i2c driver and interrupt service routine
 *
 *  @internal
 * $RCSfile: aess_i2cdrv.h,v $
 * $Revision:  $
 * $Date:  $
 * $Author:  $
 * $Source:  $
 * $Name:  $
 *----------------------------------------------------------------------------*/
/******************************************************************************
* Content
* ----------------
*   aess_i2c_disabled_interrupt()   - The function is used to disable I2C interrupt.
*   aess_i2c_enabled_interrupt()    - The function is used to enable I2C interrupt.
*   aess_i2c_check_bus_status()     - The function is to check I2C bus status.
*   aess_i2c_bus_recover_polling_func() - The function pooling I2C bust status.
*   aess_i2c_init()                 - The function is invoked when I2C module is loaded into the kernel.
*   aess_i2c_exit()                 - The function is invoked when I2C module is removed from the kernel.
*   aess_i2c_open()                 - The function is invoked when user space opens the Linux I2C driver.
*   aess_i2c_release()              - The API is invoked when user space closed the Linux I2C driver.
*   aess_i2c_ioctl()                - The API is invoked when user space issues an ioctl command. 
*   aess_i2c_reset()                - The function is an internal function to reset an I2C bus.
*   aess_i2c_config()               - The function is an internal function to configure an I2C bus.
*   aess_i2c_wr()                   - The function is an internal function to start an I2C transaction. 
*                                     After generating a start condition with slave address, 
*                                     it will wait until interrupt handler wrote/read the rest of data.
*   aess_i2c_ctrlhw()               - The function is an internal function to allow controlling hardware directly.
*                                     Two supported actions, generating clock pulses with a stop condition 
*                                     (I2C_DRV_CLK_PULSE), generating a stop condition (I2C_DRV_FORCE_STOP) and
*                                     issuing a SMBus reset (I2C_DRV_RESET_SMBUS).
*   aess_i2c_isr()                  - The function is the interrupt handler.
*   aess_i2c_isr_handler()          - The function is an internal function for interrupt handler.
*   aess_i2c_polling_func()         - This function polling I2C bus status.
*   aess_i2c_set_smbus_scl()        - This function pull SMBUS CLK line to low
*   aess_i2c_clear_smbus_scl()      - This function release SMBUS CLK line
*   aess_i2c_reset_smbus()          - This function pull SCL pin to low for 35ms to reset smbus device on the bus
******************************************************************************/

#define AESSI2CDRV_C
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <linux/wait.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/spinlock.h>
#include <asm/bitops.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>
#include <asm/arch/hardware.h>
#include <linux/aess_debugfs.h>
#include <asm/arch/hw_regs.h>

static u8   u8tmpcounter = 0;
static u16 u16tmpcounter = 0;
static u32 u32tmpcounter = 0;

#include "aess_i2cdrv.h"
#include "../../eventhandler/eventhandler_driver/aess_eventhandlerdrv.h"

#include <linux/timer.h>
#include <linux/jiffies.h>

/* #include <linux/signal.h> */

/* change to the highest message level */
#if 0
#define I2C_TESTING
#else
#define I2C_TESTING KERN_DEBUG
#endif

/* enable to print out message */
#if 0
#define DEBUGP(format, args...) printk(I2C_TESTING format, ## args)
#else
#define DEBUGP(format, args...)
#endif

/* enable a background timer to poll statuses of all initiated I2C buses */
#if 0
#define I2C_POLLING_STATUS
#endif

#if 0
#define I2C_BUS_RECOVER
#endif

/* Address Definition */
#ifdef AST2000
#define BASE_ADDRESS_REG      (ASPEED_I2C_VIRTUAL+(ASPEED_I2C_BASE & 0x00000fff))/* (0x1E6E0800L) */
#endif

#ifdef AST2100
#include <linux/spinlock.h>
#include <asm/arch/irqs.h>

#define ASPEED_SCU_VIRTUAL      IO_ADDRESS(ASPEED_SCU_BASE)
#define ASPEED_VIC_VIRTUAL      IO_ADDRESS(ASPEED_VIC_BASE)


/* Values for SCU_WR_PROT_KEY */
#define SCU_WR_KEY_EN           0x1688A8A8
#define SCU_WR_KEY_DIS          0x00000000

/* Bits of SCU_RESET_CTRL */
#define SCU_RESET_I2C           (1<<2)

#define __REG(x)    (*((volatile u32 *)x))

#define IRQ_I2C             INT_I2C
#define BASE_ADDRESS_REG    IO_ADDRESS(ASPEED_I2C_BASE)
#endif

/*spin lock*/
spinlock_t my_lock; 

/* Multi-function Pin Control Register */
#define SCU_MFPC_I2C5_ENABLED   0x00001000
#define SCU_MFPC_I2C6_ENABLED   0x00002000
#define SCU_MFPC_I2C7_ENABLED   0x00004000

/* Device Interrupt Status Register */
#define I2C_SMB_DEV7_INT        0x40
#define I2C_SMB_DEV6_INT        0x20
#define I2C_SMB_DEV5_INT        0x10
#define I2C_SMB_DEV4_INT        0x08
#define I2C_SMB_DEV3_INT        0x04
#define I2C_SMB_DEV2_INT        0x02
#define I2C_SMB_DEV1_INT        0x01

/* Clear Reg32 to zero */
#define REG32_CLEAR_TO_ZERO             0x00000000

/* Function Control Register */
#define FUNC_CTRL_REG_RESERVED_MASK     0x0000F800
#define RECV_DATA_FAST_MODE_ENABLE      0x00000400
#define DATA_SEQ_MSB_LSB_SELECT         0x00000200
#define SDA_ACTIVE_PULL_HIGH_ENABLE     0x00000100
#define SCL_ACTIVE_PULL_HIGH_ENABLE     0x00000080
#define P2P_MODE_ENBALE                 0x00000040
#define SMBUS_DEV_DEFAULT_RSP_ENABLE    0X00000020
#define SMBUS_DEV_ALERT_RSP_ENABLE      0x00000010
#define SMBUS_ARP_HOST_RSP              0x00000008
#define SLAVE_GENERAL_CALL_RSP          0x00000004
#define SLAVE_FUNCTION_ENABLE           0x00000002
#define MASTER_FUNCTION_ENABLE          0x00000001

/* Clock and AC Timing Control Register 1 */
#define BUS_FREE_TIME_STOP_N_START_MASK     0xF0000000
#define MASTER_HDSTA_MASK                   0x0F000000
#define MASTER_ACST_MASK                    0x00F00000
#define MASTER_CKHIGH_MASK                  0x000D0000
#define MASTER_CKLOW_MASK                   0x00007000
#define MASTER_SLAVE_HDDAT_MASK             0x00000C00
#define TIMEOUT_BASE_CLK_DIV_MASK           0x00000300
#define AC_TIMING_BASE_CLK_2_DIV_MASK       0x000000F0
#define AC_TIMING_BASE_CLK_1_DIV_MASK       0x0000000F

/* Clock and AC Timing Control Register 2 */
#define CLK_N_AC_TIMING_REG2_RESERVED_MASK  0xFFFFF000
#define CLK_LOW_TIMEOUT                     0x00000007

/* Interrupt Control Register */
#define INT_CTRL_REG_RESERVED_MASK          0xFFFFE000
#define BUS_RECOVER_DONE_ENABLE             0x00002000
#define SMBUS_DEV_ALERT_INT_ENABLE          0x00001000
#define SMBUS_ARP_HOST_DETECT_INT_ENABLE    0x00000800
#define SMBUS_DEV_ALERT_RSP_INT_ENABLE      0x00000400
#define SMBUS_DEV_DEFAULT_DETECT_INT_ENABLE 0x00000200
#define GENERAL_CALL_DETECT_INT_ENABLE      0x00000100
#define SLAVE_ADDR_RECV_MATCH_INT_ENABLE    0x00000080
#define SCL_LOW_TIMEOUT_INT_ENABLE          0x00000040
#define ABNORMAL_START_N_STOP_INT_ENABLE    0x00000020
#define NORMAL_STOP_INT_ENABLE              0x00000010
#define MASTER_ARBITRATION_LOSS_INT_ENABLE  0x00000008
#define RECV_DONE_INT_ENABLE                0x00000004
#define TRANS_NACK_RET_INT_ENABLE           0x00000002
#define TRANS_DONE_ACK_RET_INT_ENABLE       0x00000001

/* Interrupt Status Register */
#define INT_STATUS_REG_RESERVED_MASK        0xFFFFC000
#define BUS_RECOVER_INT_STATUS              0x00002000
#define SMBUS_DEV_ALERT_INT_STATUS          0x00001000
#define SMBUS_ARP_HOST_DETECT_INT_STATUS    0x00000800
#define SMBUS_DEV_ALERT_RSP_INT_STATUS      0x00000400
#define SMBUS_DEV_DEFAULT_DETECT_INT_STATUS 0x00000200
#define GENERAL_CALL_DETECT_INT_STATUS      0x00000100
#define SLAVE_ADDR_RECV_MATCH_INT_STATUS    0x00000080
#define SCL_LOW_TIMEOUT_INT_STATUS          0x00000040
#define ABNORMAL_START_N_STOP_INT_STATUS    0x00000020
#define NORMAL_STOP_INT_STATUS              0x00000010
#define MASTER_ARBITRATION_LOSS_INT_STATUS  0x00000008
#define RECV_DONE_INT_STATUS                0x00000004
#define TRANS_NACK_RET_INT_STATUS           0x00000002
#define TRANS_DONE_ACK_RET_INT_STATUS       0x00000001

/* Command/Status Register */
#define CMD_STATUS_REG_RESERVED             0xC0000000
#define SDA_OE                              0x10000000
#define SDA_O                               0x08000000
#define SCL_OE                              0x04000000
#define SCL_O                               0x02000000
#define TRANS_MODE_TIMING_STAGE             0x01800000
#define TRANS_MODE_STATE_MACHINE            0x00780000
#define SAMPLED_SCL_LINE_STATE              0x00040000
#define SAMPLED_SDA_LINE_STATE              0x00020000
#define BUS_BUSY_STATUS                     0x00010000
#define SDA_OE_OUTPUT                       0x00008000
#define SDA_O_OUTPUT                        0x00004000
#define SCL_OE_OUTPUT                       0x00002000
#define SCL_O_OUTPUT                        0x00001000
#define BUS_RECOVER                         0x00000800
#define SMBUS_SLAVE_ALERT_ENABLE            0x00000400
#define MASTER_SLAVE_RECV_DMA_BUF_ENABLE    0x00000200
#define MASTER_SLAVE_TRANS_DMA_BUF_ENABLE   0x00000100
#define MASTER_SLAVE_RECV_DATA_BUF_ENABLE   0x00000080
#define MASTER_SLAVE_TRANS_DATA_BUF_ENABLE  0x00000040
#define MASTER_STOP_CMD                     0x00000020
#define MASTER_SLAVE_RECV_CMD_LAST          0x00000010
#define MASTER_RECV_CMD                     0x00000008
#define SLAVE_TRANS_CMD                     0x00000004
#define MASTER_TRANS_CMD                    0x00000002
#define MASTER_START_CMD                    0x00000001

/* Device Address */
#define DEV_ADDR_RESERVED           0xFFFFFF80  
#define SLAVE_DEV_ADDR              0x0000007F

/* Buffer Control Register */
#define REAL_RECV_DATA_BUF_ADDR_PTR     0xFF000000
#define RECV_DATA_BUF_ENDED_ADDR        0x00FF0000
#define TRANS_DATA_BUF_ENDED_ADDR       0x0000FF00
#define BUF_CTRL_REG_RESERVED           0x000000C0
#define BUF_BASE_ADDR_PTR               0x0000003F

/* Transmit/Receive Byte Buffer */
#define TRANS_RECV_BYTE_BUF_RESERVED    0xFFFF0000
#define RECV_BYTE_BUF                   0x0000FF00
#define TRANS_BYTE_BUF                  0x000000FF

/* (Device1/2 Only) DMA Mode Control Register */
#define DMA_MODE_CTRL_REG_RESERVED  0xF0000000
#define DMA_BUF_BASE_ADDR           0x0FFFF000
#define DMA_BUF_SIZE                0x00000FFF

/* (Device1/2 Only) DMA Mode Status Register */
#define DMA_MODE_STATUS_REG_RESERVED    0xFFFFF000
#define DMA_LAST_ACCESSED_ADDR          0x00000FFF

/* NULL Address */
#define NULL_ADDRESS    0x00000000

/* definitions for bus status */
#define I2C_SET_BUS_ERROR_BIT       0x08
#define I2C_SET_BUS_BUSY_BIT        0x04
#define I2C_SET_BUS_SCL_HIGH_BIT    0x02
#define I2C_SET_BUS_SDA_HIGH_BIT    0x01


/******************************************************************************
*   Enum      :   eI2CDrvStateType
******************************************************************************/
/**
 *  @brief   I2C driver state machine enumeration
 *
 *****************************************************************************/
typedef enum
{
    /** Idle state */
    I2C_DRV_IDLE_STATE,
    
    /** Slave receive */
    I2C_DRV_SLV_RCV_STATE,
    
    /** Slave receive FIFO */
    I2C_DRV_SLV_RCV_FIFO_STATE,
    
    /** Slave transmit */
    I2C_DRV_SLV_XMIT_STATE,
    
    /** Master receive */
    I2C_DRV_MST_RCV_STATE,
    
    /** Master receive dummy read */
    I2C_DRV_MST_RCV_DUMMY_STATE,
    
    /** Master transmit */
    I2C_DRV_MST_XMIT_STATE,
    
    /** Start state */
    I2C_DRV_START_STATE,
    
    /** Stop state */
    I2C_DRV_STOP_STATE
    
} eI2CDrvStateType;

#define I2C_DRV_MAX_FIFO_SIZE       8

/******************************************************************************
*   STRUCT      :   sI2CDrvFIFOType
******************************************************************************/
/**
 *  @brief   structure of I2C receive message FIFO
 *
 *****************************************************************************/
typedef struct
{
    /** Spin lock */
    spinlock_t lock;   
    
    /** Head of FIFO */
    UINT8 u8Head;
    
    /** Tail of FIFO */
    UINT8 u8Tail;
    
    /** Message count */
    UINT8 u8MsgCount;
    
    /** Buffer */
    UINT8 au8buffer[I2C_DRV_MAX_FIFO_SIZE][I2C_DRV_MAX_MSG_SIZE];
    
    /** Message length */
    UINT8 au8MsgLen[I2C_DRV_MAX_FIFO_SIZE];
    
} sI2CDrvFIFOType;


/******************************************************************************
*   STRUCT      :   sI2CDrvBufferType
******************************************************************************/
/**
 *  @brief   structure of I2C driver buffer
 *
 *****************************************************************************/
typedef struct
{
    /** IPMB msg received flag */
    UINT32 u32RecFlag;

    /** Msg receive FIFO */
    sI2CDrvFIFOType *psFIFO;

    /** Access semaphore */
    struct semaphore sem;
        
    /** Spinlock */
    spinlock_t data_lock;

    /** Wait queue */
    wait_queue_head_t wq;

    /** Driver ID for kernel event handler */
    UINT16 u16DriverID;
    
    /** Current start tracking count */
    UINT16 u16CurStartCount;

    /** Current stop tracking count */
    UINT16 u16CurStopCount;    
    
    /** Reserve */
    UINT16 Rev;
        
    /** I2C bus channel */
    UINT8 u8Channel;
    
    /** Initial mode (IPMB or PI2C) */
    UINT8 u8InitMode;
    
    /** Current mode (IPMB or PI2C) */
    UINT8 u8CurMode;
    
    /** IPMB slave address */
    UINT8 u8SlaveAddr;
    
    /** I2C bus frequency selection */
    UINT8 u8Frequency;
    
    /** Bus initialization status */
    UINT8 u8HasInit;
    
    /** State machine variable */
    UINT8 u8State;
    
    /** Bus error status */
    UINT8 u8ErrorStatus;
    
    /** I2C bus action finished flag */
    UINT8 u8FinishedFlag;
    
    /** Msg transmit buffer */
    UINT8 u8MsgSendBuffer[I2C_DRV_MAX_MSG_SIZE];
    
    /** Msg receive buffer */
    UINT8 u8MsgRecBuffer[I2C_DRV_MAX_MSG_SIZE];

    /** Msg transmit counter */
    UINT8 u8MsgSendCounter;
    
    /** Msg transmit data length */
    UINT8 u8MsgSendDataSize;
    
    /** Msg receive counter */
    UINT8 u8MsgRecCounter;
    
    /** Msg receive data length */
    UINT8 u8MsgRecDataSize;
    
    /** Condition checking for wait queue */
    UINT8 wq_flag;
    
    /** lock mode */
    UINT8 tLockMode;
    
    /** Transaction type: 0:Normal, 1:Mux, 2:SMBus Blk */
    UINT8 u8TransType;    
    
} sI2CDrvBufferType;

#define I2C_DRV_LOCK_MODE_LOCK      0
#define I2C_DRV_LOCK_MODE_UNLOCK    1

/******************************************************************************
*   STRUCT      :   i2c_reg_type
******************************************************************************/
/**
 *  @brief   structure of I2C hardware register
 *
 *****************************************************************************/
typedef struct 
{
    /** I2C control Reg. */
    volatile UINT32 func_ctrl;

    /** I2C Clock and AC Timing Control Reg.#1 */
    volatile UINT32 clk_n_ac_timing_1;

    /** I2C Clock and AC Timing Control Reg.#2 */
    volatile UINT32 clk_n_ac_timing_2;

    /** I2C Channel Interrupt Control Reg. */
    volatile UINT32 int_ctrl;

    /** I2C Channel Interrupt Status */
    volatile UINT32 int_status;

    /** I2C Command Status */
    volatile UINT32 cmd_status;

    /** I2C Slave Device Address Reg. */
    volatile UINT32 dev_addr;

    /** I2C Pool buffer Control Reg. */
    volatile UINT32 buf_ctrl;

    /** I2C Transmit/Receive Byte Buffer Reg. */
    volatile UINT32 trans_recv_byte_buf;

    /** DMA Mode Comtrol Reg. */
    volatile UINT32 dma_mode_ctrl_reg;

    /** DMA mode status Reg. */
    volatile UINT32 dma_mode_status_reg;
} i2c_reg_type;

/* i2c hardware register structure array */
volatile static i2c_reg_type i2c_reg[] = 
{
    {
        I2C0_FUNC_CTRL_REG,
        I2C0_CLK_N_AC_TIMING_CTRL_REG_1,
        I2C0_CLK_N_AC_TIMING_CTRL_REG_2,
        I2C0_INT_CTRL_REG,
        I2C0_INT_STATUS_REG,
        I2C0_CMD_STATUS_REG,
        I2C0_DEV_ADDR,
        I2C0_BUF_CTRL_REG,
        I2C0_TRANS_RECV_BYTE_BUF,
        I2C0_DMA_MODE_CTRL_REG,
        I2C0_DMA_MODE_STATUS_REG
    },
    {
        I2C1_FUNC_CTRL_REG,
        I2C1_CLK_N_AC_TIMING_CTRL_REG_1,
        I2C1_CLK_N_AC_TIMING_CTRL_REG_2,
        I2C1_INT_CTRL_REG,
        I2C1_INT_STATUS_REG,
        I2C1_CMD_STATUS_REG,
        I2C1_DEV_ADDR,
        I2C1_BUF_CTRL_REG,
        I2C1_TRANS_RECV_BYTE_BUF,
        I2C1_DMA_MODE_CTRL_REG,
        I2C1_DMA_MODE_STATUS_REG
    },
    {
        I2C2_FUNC_CTRL_REG,
        I2C2_CLK_N_AC_TIMING_CTRL_REG_1,
        I2C2_CLK_N_AC_TIMING_CTRL_REG_2,
        I2C2_INT_CTRL_REG,
        I2C2_INT_STATUS_REG,
        I2C2_CMD_STATUS_REG,
        I2C2_DEV_ADDR,
        I2C2_BUF_CTRL_REG,
        I2C2_TRANS_RECV_BYTE_BUF,
        NULL_ADDRESS,
        NULL_ADDRESS
    },
    {
        I2C3_FUNC_CTRL_REG,
        I2C3_CLK_N_AC_TIMING_CTRL_REG_1,
        I2C3_CLK_N_AC_TIMING_CTRL_REG_2,
        I2C3_INT_CTRL_REG,
        I2C3_INT_STATUS_REG,
        I2C3_CMD_STATUS_REG,
        I2C3_DEV_ADDR,
        I2C3_BUF_CTRL_REG,
        I2C3_TRANS_RECV_BYTE_BUF,
        NULL_ADDRESS,
        NULL_ADDRESS
    },
    {
        I2C4_FUNC_CTRL_REG,
        I2C4_CLK_N_AC_TIMING_CTRL_REG_1,
        I2C4_CLK_N_AC_TIMING_CTRL_REG_2,
        I2C4_INT_CTRL_REG,
        I2C4_INT_STATUS_REG,
        I2C4_CMD_STATUS_REG,
        I2C4_DEV_ADDR,
        I2C4_BUF_CTRL_REG,
        I2C4_TRANS_RECV_BYTE_BUF,
        NULL_ADDRESS,
        NULL_ADDRESS
    },
    {
        I2C5_FUNC_CTRL_REG,
        I2C5_CLK_N_AC_TIMING_CTRL_REG_1,
        I2C5_CLK_N_AC_TIMING_CTRL_REG_2,
        I2C5_INT_CTRL_REG,
        I2C5_INT_STATUS_REG,
        I2C5_CMD_STATUS_REG,
        I2C5_DEV_ADDR,
        I2C5_BUF_CTRL_REG,
        I2C5_TRANS_RECV_BYTE_BUF,
        NULL_ADDRESS,
        NULL_ADDRESS
    },
    {
        I2C6_FUNC_CTRL_REG,
        I2C6_CLK_N_AC_TIMING_CTRL_REG_1,
        I2C6_CLK_N_AC_TIMING_CTRL_REG_2,
        I2C6_INT_CTRL_REG,
        I2C6_INT_STATUS_REG,
        I2C6_CMD_STATUS_REG,
        I2C6_DEV_ADDR,
        I2C6_BUF_CTRL_REG,
        I2C6_TRANS_RECV_BYTE_BUF,
        NULL_ADDRESS,
        NULL_ADDRESS
    }
};

/* driver name will passed by insmod command */
static char *driver_name="aess_i2cdrv";

/* record the kernel i2c initial state */
volatile static int i2c_init_state = 0;

sI2CDrvBufferType i2c_buffer[I2C_DRV_MAX_BUS];
UINT32 u32RecBusCount[I2C_DRV_MAX_BUS];

#ifdef I2C_BUS_RECOVER
volatile struct timer_list i2c_bus_recover;
#endif

#ifdef I2C_POLLING_STATUS
volatile struct timer_list i2c_polling;
#endif

dev_t i2c_dev;
struct cdev *i2c_cdev;
static int aess_i2c_reset(sI2CDrvBufferType *);
static int aess_i2c_config(sI2CDrvBufferType *);

#ifdef I2C_BUS_RECOVER

/******************************************************************************
*   FUNCTION        :   aess_i2c_disabled_interrupt
******************************************************************************/
/**
 *  @brief      The function is used to disable I2C interrupt.
 *
 *  @return     None
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    none
 *
 *  @note       None
 *
 *  @internal   None \n
 *
 *****************************************************************************/
void aess_i2c_disabled_interrupt(
                                  /** I2C Bus Channel Number */
                                  UINT8 bus_num
                                )
{
    volatile i2c_reg_type *current_reg;

    current_reg = &i2c_reg[bus_num];
            
    /* clearing interrupt control */
    iowrite32((UINT32) (REG32_CLEAR_TO_ZERO),
              (void *) current_reg->int_ctrl);          
}

/******************************************************************************
*   FUNCTION        :   aess_i2c_enabled_interrupt
******************************************************************************/
/**
 *  @brief      The function is used to enable I2C interrupt.
 *
 *  @return     None
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   None \n
 *
 *****************************************************************************/
void aess_i2c_enabled_interrupt(
                                 /** I2C Bus Channel Number */
                                 UINT8 bus_num
                               )
{
    volatile i2c_reg_type *current_reg;

    current_reg = &i2c_reg[bus_num];
            
    iowrite32((UINT32) (SLAVE_ADDR_RECV_MATCH_INT_ENABLE
#ifdef I2C_BUS_RECOVER    
                        | BUS_RECOVER_DONE_ENABLE   
#endif                        
                        | ABNORMAL_START_N_STOP_INT_ENABLE
                        | NORMAL_STOP_INT_ENABLE        
                        | MASTER_ARBITRATION_LOSS_INT_ENABLE
                        | RECV_DONE_INT_ENABLE              
                        | TRANS_NACK_RET_INT_ENABLE         
                        | TRANS_DONE_ACK_RET_INT_ENABLE),       
              (void *) current_reg->int_ctrl);     
}

/******************************************************************************
*   FUNCTION        :   aess_i2c_check_bus_status
******************************************************************************/
/**
 *  @brief      The function is used to check I2C bus status.
 *
 *  @return     None
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   None \n
 *
 *****************************************************************************/
void aess_i2c_check_bus_status(
                                /** I2C Bus Channel Number */
                                UINT8 bus_num
                              )
{
#define SDA_LINE_STATUS_MASK 0x00020000
#define SCL_LINE_STATUS_MASK 0x00040000

    static UINT32 i;
    static UINT32 current_ctrl_reg_reading;
    static UINT32 int_reg_reading; 
    static UINT32 temp_reg_reading;    
    static UINT32 reg_reading;    
    i2c_reg_type *current_reg;  
    volatile sI2CDrvBufferType *psI2CDrvBuffer;
    
    /* get semaphore */
    if(down_interruptible(&i2c_buffer[bus_num].sem))
    {
        DRV_ERROR(" bus #%d get semaphore error!\n",bus_num);
        return;
    }  
    
    DRV_MSG(" bus #%d aess_i2c_check_bus_status!\n",bus_num);
    
    DRV_MSG("enter bus recovery function aess_i2c_check_bus_status()...");
    current_reg = &i2c_reg[bus_num];
    int_reg_reading = ioread32((void *) current_reg->int_ctrl);     
    current_ctrl_reg_reading = ioread32((void *) current_reg->func_ctrl); 
    
    iowrite32((UINT32) (BUS_RECOVER_DONE_ENABLE),(void *) current_reg->int_ctrl); 
    iowrite32((UINT32) (0),(void *) current_reg->func_ctrl);      

    reg_reading = ioread32((void *) current_reg->cmd_status);  
    if ((0 == (reg_reading & SDA_LINE_STATUS_MASK)) || (0 == (reg_reading & SCL_LINE_STATUS_MASK)))
    {
        DRV_INFO(" bus #%d reg_reading is 0x%08x!\n",bus_num,reg_reading);
        DRV_INFO(" bus #%d SDA_LINE_STATUS_MASK is 0x%08x!\n",bus_num,SDA_LINE_STATUS_MASK);
        DRV_INFO(" bus #%d (reg_reading & SDA_LINE_STATUS_MASK) is 0x%08x!\n",bus_num,(reg_reading & SDA_LINE_STATUS_MASK));
        DRV_INFO(" bus #%d current_reg is 0x%08x!\n",bus_num,current_reg);
        u32RecBusCount[bus_num] = u32RecBusCount[bus_num] + 1;
        DRV_INFO(" bus #%d u32RecBusCount is 0x%08x!\n",bus_num,u32RecBusCount[bus_num]);
        DRV_INFO(" bus #%d current_ctrl_reg_reading is 0x%08x!\n",bus_num,current_ctrl_reg_reading);
        DRV_INFO(" bus #%d current_reg->func_ctrl is 0x%08x!\n",bus_num,ioread32((void *) current_reg->func_ctrl));
        DRV_INFO(" bus #%d current_reg->cmd_status is 0x%08x!\n",bus_num,ioread32((void *) current_reg->cmd_status));
        DRV_INFO(" bus #%d isuue bus recover signel!\n",bus_num);
        iowrite32((UINT32) (BUS_RECOVER),(void *) current_reg->cmd_status);
 
        for ( i=0; i <= 1000 ; i++)
        {
        	/*add a delay*/
        }

        iowrite32((UINT32) int_reg_reading, (void *) current_reg->int_ctrl);    
        iowrite32((UINT32) current_ctrl_reg_reading, (void *) current_reg->func_ctrl);    
        psI2CDrvBuffer = &i2c_buffer[bus_num];
        DRV_MSG("aess_i2c_check_bus_status bus recovery done....");
//      aess_i2c_reset(psI2CDrvBuffer);
//      DEBUGP(KERN_DEBUG " bus #%d isuue start and stop!\n",bus_num);
//      iowrite32((UINT32) (MASTER_START_CMD | MASTER_STOP_CMD), (void *) current_reg->cmd_status);                   
    }   
                
    /* release semaphore */
    up(&i2c_buffer[bus_num].sem);      
    
    return;
    
#undef SDA_LINE_STATUS_MASK
#undef SCL_LINE_STATUS_MASK
}
#endif

#ifdef I2C_BUS_RECOVER
/******************************************************************************
*   FUNCTION        :   aess_i2c_bus_recover_polling_func
******************************************************************************/
/**
 *  @brief      The function is used to polling I2C bus status.
 *
 *  @return     None
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   None \n
 *
 *****************************************************************************/
void aess_i2c_bus_recover_polling_func(
                                       /** Dummy data*/ 
                                        unsigned long data
                                       )
{
    UINT8 bus_num;
    
    DRV_MSG("aess_i2c_bus_recover_polling_func\n");
    for (bus_num = 0; bus_num < I2C_DRV_MAX_BUS; bus_num++)
    {
        if (i2c_buffer[bus_num].u8HasInit == 1)
        {
            if (I2C_DRV_IDLE_STATE == i2c_buffer[bus_num].u8State)
            {
                aess_i2c_check_bus_status(bus_num);
            }
        }
    }
    
    /* timer interval is 10s */
    i2c_bus_recover.expires = jiffies + 5000;
    add_timer(&i2c_bus_recover);    
                    
    return;
}
#endif

/******************************************************************************
*   FUNCTION        :   aess_i2c_set_smbus_scl
******************************************************************************/
/**
 *  @brief      Set SMBUS CLK pin to low
 *
 *  @return     OK
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   
 *
 *****************************************************************************/
static int aess_i2c_set_smbus_scl(

                /** I2C Channel Number */
                UINT8 bus_num
                
                                     )
{
    UINT32 u32RegReading; 
    i2c_reg_type *current_reg;
    current_reg = (i2c_reg_type *)&i2c_reg[bus_num];
    
    DRV_MSG("KN:Bus(%d) aess_i2c_set_multifun_bit() enter! \n", bus_num);	
   
    /* Enable I2C BUS Multi-function pin */

        if (4 == bus_num)
        {
            spin_lock(&my_lock);

            DRV_MSG("Write enable key Bus 4\n");
            __REG(SCU_WR_PROT_KEY) = SCU_WR_KEY_EN;    /* Write enable key */   
            u32RegReading = ioread32((void *) SCU_MUTI_FUN_PIN_CTRL1);
            iowrite32((UINT32) (u32RegReading | SCU_MFPC_I2C5_ENABLED),
            (void *) SCU_MUTI_FUN_PIN_CTRL1);           
            __REG(SCU_WR_PROT_KEY) = SCU_WR_KEY_DIS;   /* Write disable key */      
            DRV_MSG("Write disable key\n");          

            spin_unlock(&my_lock);    
        }
        else if (5 == bus_num)
        {
            spin_lock(&my_lock);    

            DRV_MSG("Write enable key Bus 5\n");
            __REG(SCU_WR_PROT_KEY) = SCU_WR_KEY_EN;    /* Write enable key */       
            u32RegReading = ioread32((void *) SCU_MUTI_FUN_PIN_CTRL1);
            iowrite32((UINT32) (u32RegReading | SCU_MFPC_I2C6_ENABLED),
            (void *) SCU_MUTI_FUN_PIN_CTRL1);   
            __REG(SCU_WR_PROT_KEY) = SCU_WR_KEY_DIS;   /* Write disable key */      
            DRV_MSG("Write disable key\n");          

            spin_unlock(&my_lock);    
        }
        else if (6 == bus_num)
        {
            spin_lock(&my_lock);    

            DRV_MSG("Write enable key Bus 6\n");
            __REG(SCU_WR_PROT_KEY) = SCU_WR_KEY_EN;    /* Write enable key */       
            u32RegReading = ioread32((void *) SCU_MUTI_FUN_PIN_CTRL1);
            iowrite32((UINT32) (u32RegReading | SCU_MFPC_I2C7_ENABLED),
            (void *) SCU_MUTI_FUN_PIN_CTRL1);    
            __REG(SCU_WR_PROT_KEY) = SCU_WR_KEY_DIS;   /* Write disable key */      
            DRV_MSG("Write disable key\n");              

            spin_unlock(&my_lock); 
        }

    /* disable device */
    u32RegReading = ioread32((void *) current_reg->func_ctrl);
    u32RegReading = u32RegReading & ~(I2C_MASTER_ENABLE | I2C_SLAVE_ENABLE);
    DRV_INFO ("aess_i2c_set_multifun_bit set func_ctrl u32RegReading = %08x\n", u32RegReading);
    iowrite32((UINT32) u32RegReading, (void *) current_reg->func_ctrl);

    /* set I2C CLK to low */
    u32RegReading = ioread32((void *) current_reg->cmd_status);
    u32RegReading = (u32RegReading | I2C_SCL_OE) & (~I2C_SCL_O);
    DRV_INFO ("aess_i2c_set_multifun_bit set cmd_status u32RegReading = %08x\n", u32RegReading);
    iowrite32((UINT32) u32RegReading, (void *) current_reg->cmd_status);
            
    return 0;
    
}

/******************************************************************************
*   FUNCTION        :   aess_i2c_clear_smbus_scl
******************************************************************************/
/**
 *  @brief      Release SMBUS CLK pin. 
 *
 *  @return     OK
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   
 *
 *****************************************************************************/
static int aess_i2c_clear_smbus_scl(

                /** I2C Channel Number */
                UINT8 bus_num
                                    )
{
    UINT32 u32RegReading; 
    i2c_reg_type *current_reg;
    current_reg = (i2c_reg_type *)&i2c_reg[bus_num];

    DRV_MSG("KN:Bus(%d) aess_i2c_clear_multifun_bit() enter! \n", bus_num);	

    /* Release I2C CLK */
    u32RegReading = ioread32((void *) current_reg->cmd_status);
    u32RegReading = u32RegReading & ~I2C_SCL_OE;
    DRV_INFO ("aess_i2c_clear_multifun_bit set cmd_status u32RegReading = %08x\n", u32RegReading);
    iowrite32((UINT32) u32RegReading, (void *) current_reg->cmd_status);
    
    /* Enable device */
    u32RegReading = ioread32((void *) current_reg->func_ctrl);
    u32RegReading = u32RegReading | I2C_MASTER_ENABLE | I2C_SLAVE_ENABLE;
    DRV_INFO ("aess_i2c_clear_multifun_bit set func_ctrl u32RegReading = %08x\n", u32RegReading);
    iowrite32((UINT32) u32RegReading, (void *) current_reg->func_ctrl);

    DRV_MSG("KN:Bus(%d) aess_i2c_clear_multifun_bit() data 0x%x\n", bus_num, u32RegReading);	
    
    return 0;
    
}

/******************************************************************************
*   FUNCTION        :   aess_i2c_reset_smbus
******************************************************************************/
/**
 *  @brief      pull SCL more than 35ms to reset SMBus. 
 *
 *  @return     OK
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   None\n
 *
 *****************************************************************************/
static int aess_i2c_reset_smbus(

                /** I2C Channel Number */
                UINT8 bus_num
                
                                                )
{
    DRV_MSG("KN:Bus(%d) aess_i2c_reset_smbus() enter!\n", bus_num);	
    
    aess_i2c_set_smbus_scl(bus_num);
    mdelay(35);			      
    aess_i2c_clear_smbus_scl(bus_num);
    
    return 0;
    
}

/******************************************************************************
*   FUNCTION        :   aess_i2c_isr_handler
******************************************************************************/
/**
 *  @brief      The function is an internal function for interrupt handler.
 *
 *  @return     IRQ_NONE: IRQ not handled
 *              IRQ_HANDLED: IRQ handled
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   None \n
 *
 *****************************************************************************/
irqreturn_t aess_i2c_isr_handler(
                                 /** I2C Bus Channel Number */ 
                                 UINT8 bus_num
                                )
{
    volatile i2c_reg_type *current_reg;
    volatile UINT32 int_reading;
    volatile UINT32 reg_reading;
    
    volatile sI2CDrvBufferType *psI2CDrvBuffer;
    volatile sI2CDrvFIFOType *psI2CDrvFIFO;
    volatile UINT8 u8TotalByteCount = 0;

    DRV_MSG("In ISR handler.........................\n");

    if (bus_num < I2C_DRV_MAX_BUS)
    {
        current_reg = &i2c_reg[bus_num];
        psI2CDrvBuffer = &i2c_buffer[bus_num];
    }
    else
    {
        DRV_ERROR("Bus:%d IRQ_NONE \n", bus_num);
        return IRQ_NONE;
    }

    /* */
    psI2CDrvFIFO = psI2CDrvBuffer->psFIFO;

    /* read the interrupt status register */
    int_reading = ioread32((void *) current_reg->int_status); 

    if (int_reading & BUS_RECOVER_INT_STATUS)
    {
        
        DRV_INFO(" Bus #%d recover done!\n",bus_num);
                    
        /* clearing interrupt status */
        iowrite32((UINT32) BUS_RECOVER_INT_STATUS,
                            (void *) current_reg->int_status);        
    }    

    if (int_reading & ABNORMAL_START_N_STOP_INT_STATUS)
    {
        /* soft reset, turn off both the Master and Slave function enable, 
           then HW will reset all state machine and status */
        iowrite32((UINT32) (0), (void *) current_reg->func_ctrl);   

        aess_i2c_reset((sI2CDrvBufferType *)psI2CDrvBuffer);

        aess_i2c_config((sI2CDrvBufferType *)psI2CDrvBuffer);
        
        /* recode error status */
        psI2CDrvBuffer->u8ErrorStatus = I2C_DRV_ERROR_START;
        
        /* setup next state */
        psI2CDrvBuffer->u8State = I2C_DRV_IDLE_STATE;    

        /* clearing Master Arbitration Loss Interrupt status */
        iowrite32((UINT32) ABNORMAL_START_N_STOP_INT_STATUS,
                  (void *) current_reg->int_status);

        DRV_ERROR("Bus: %d ABNORMAL_START_N_STOP_INT_STATUS\n", bus_num);
        
    }

    if (int_reading & NORMAL_STOP_INT_STATUS)
    {
        switch (psI2CDrvBuffer->u8State)
        {
            case I2C_DRV_STOP_STATE:
            {
                DRV_INFO("Bus: %d NORMAL_STOP_INT_STATUS\n", bus_num);

                /* increase stop tracing count */
                psI2CDrvBuffer->u16CurStopCount++;

                /* setup next state */
                    psI2CDrvBuffer->u8State = I2C_DRV_IDLE_STATE;
                    i2c_buffer[bus_num].u8State = I2C_DRV_IDLE_STATE;
                break;
            }
            
            case I2C_DRV_SLV_RCV_STATE:
            {
                DRV_INFO("Slave Receive State enter...\n");

                psI2CDrvFIFO->u8MsgCount++;
                
                DRV_MSG("Message Count:%d\n", psI2CDrvFIFO->u8MsgCount);
                
                if (psI2CDrvFIFO->u8MsgCount <= I2C_DRV_MAX_FIFO_SIZE)
                {
                    /* if the receive count is big enough */
                    if (psI2CDrvFIFO->au8MsgLen[psI2CDrvFIFO->u8Head]
                        >= I2C_DRV_IPMB_MSG_MIN)
                    {
                        /* get spin lock */
                        spin_lock(&psI2CDrvFIFO->lock);
                                        
                        /* set the next buffer */
                        if (psI2CDrvFIFO->u8Head == (I2C_DRV_MAX_FIFO_SIZE - 1))
                        {
                             psI2CDrvFIFO->u8Head = 0;
                        }
                        else
                        {
                             psI2CDrvFIFO->u8Head++;
                        }
                          
                        DRV_MSG("Bus:%d Buffer Head: u8Head=%d\n", bus_num, psI2CDrvFIFO->u8Head);
                    
                        /* release spin lock */
                        spin_unlock(&psI2CDrvFIFO->lock);
                    
                        DRV_MSG("generate event to notify uplayer AP!\n");

                        /* generate event to notify */
                        aess_generate_driver_event(psI2CDrvBuffer->u16DriverID, 
                                                   psI2CDrvBuffer->u32RecFlag);
                    }
                
                    /* if the message length is less than I2C_DRV_IPMB_MSG_MIN bytes */
                    else
                    {
                        /* reset the message length */
                        psI2CDrvFIFO->au8MsgLen[psI2CDrvFIFO->u8Head] = 0;
                        psI2CDrvFIFO->u8MsgCount--;  
                        
                        DRV_MSG(" Bus:%d receive length less than %d bytes\n", bus_num, I2C_DRV_IPMB_MSG_MIN);
                    }
                }
                else
                {
                    psI2CDrvFIFO->u8MsgCount--;                    
                }
                                    
                /* setup next state */
                psI2CDrvBuffer->u8State = I2C_DRV_IDLE_STATE;

                /* increase start tracing count */
                psI2CDrvBuffer->u16CurStartCount++;
                
                /* increase stop tracing count */
                psI2CDrvBuffer->u16CurStopCount++;
                
                DRV_MSG("Bus:%d exit S SLV_RCV_STATE\n", bus_num);
                
                break;
            }
            
            case I2C_DRV_SLV_RCV_FIFO_STATE:
            {
                DRV_MSG("Bus: %d exit S SLV_RCV_STATE\n", bus_num);

                break;
            }
            
            case I2C_DRV_MST_RCV_STATE:
            {
                DRV_MSG("Bus:%d enter I2C_DRV_MST_RCV_STATE\n", bus_num);

                /* set slave address as READ */
                iowrite32((UINT32) psI2CDrvBuffer->u8MsgSendBuffer[0] | 1, 
                          (void *) current_reg->trans_recv_byte_buf);

                /* generate a start and receive condition */
                iowrite32((UINT32) (MASTER_START_CMD | MASTER_RECV_CMD), 
                          (void *) current_reg->cmd_status);

                DRV_MSG("Bus:%d exit I2C_DRV_MST_RCV_STATE\n", bus_num);       

                break;
            }
            
            case I2C_DRV_IDLE_STATE:
            {
                DRV_MSG("Bus:%d From STOP_STATE to IDLE_STATE\n", bus_num);
                
                break;
            }
            default:
            {
                /* error occurred TBD */
                DRV_ERROR("Bus:%d S default_STATE\n", bus_num);
                
                break;
            }
        }
        
        /* clearing interrupt status */
        iowrite32((UINT32) NORMAL_STOP_INT_STATUS,
                  (void *) current_reg->int_status);
    }

    if (int_reading & TRANS_DONE_ACK_RET_INT_STATUS)
    {   
        DRV_MSG("Bus:%d S Transmit ACK return\n", bus_num);

        iowrite32((UINT32) TRANS_DONE_ACK_RET_INT_STATUS,
                  (void *) current_reg->int_status);
                  
        if (psI2CDrvBuffer->u8State == I2C_DRV_MST_XMIT_STATE)
        {
            /* if has sending data */
            if (psI2CDrvBuffer->u8MsgSendDataSize - psI2CDrvBuffer->u8MsgSendCounter - 1)
            {
                psI2CDrvBuffer->u8MsgSendCounter++;
            
                /******************************************** 
                 * Add delay after first write command *
                 ********************************************/ 
                udelay(50);

                DRV_INFO("Send Buffer[0x%02x]\n",
                psI2CDrvBuffer->u8MsgSendBuffer[psI2CDrvBuffer->u8MsgSendCounter]);

                /* set data */
                iowrite32((UINT32) psI2CDrvBuffer->u8MsgSendBuffer[psI2CDrvBuffer->u8MsgSendCounter], 
                          (void *) current_reg->trans_recv_byte_buf);

                /* generate a transmit condition */
                iowrite32((UINT32) (MASTER_TRANS_CMD), (void *) current_reg->cmd_status);
            
                DRV_INFO("Bus:%d send data:0x%02x counter:%d\n", 
                        bus_num, 
                        psI2CDrvBuffer->u8MsgSendBuffer[psI2CDrvBuffer->u8MsgSendCounter],
                        psI2CDrvBuffer->u8MsgSendCounter);
            }
            else
            {
                /* transfer to master receiver mode */
                if (psI2CDrvBuffer->u8MsgRecDataSize)
                {
                    DRV_MSG("Bus:%d setup next state -> I2C_DRV_MST_RCV_STATE\n", bus_num);
                    
                    /* setup next state */
                    psI2CDrvBuffer->u8State = I2C_DRV_MST_RCV_STATE;

                    /* set slave address as READ */
                    iowrite32((UINT32) psI2CDrvBuffer->u8MsgSendBuffer[0] | 1, 
                              (void *) current_reg->trans_recv_byte_buf);

                    DRV_MSG("Bus:%d repeated start\n", bus_num);

                    if (1 == psI2CDrvBuffer->u8MsgRecDataSize)
                    {
                        /* generate a Sr + SLAVEADDR + READ Data byte (NAK) */
                        iowrite32((UINT32) (MASTER_START_CMD | MASTER_TRANS_CMD 
                                            | MASTER_SLAVE_RECV_CMD_LAST | MASTER_RECV_CMD), 
                                  (void *) current_reg->cmd_status);    
                    }
                    else
                    {
                        /* There are more one byte need to receive */
                        /* generate a Sr + SLAVEADDR + READ Data byte (ACK) */
                        iowrite32((UINT32) (MASTER_START_CMD | MASTER_TRANS_CMD 
                                            | MASTER_RECV_CMD), 
                                  (void *) current_reg->cmd_status);    
                    }
                }
            
                /* no data needs to be read */
                else
                {
                    DRV_MSG("Bus:%d setup next state -> I2C_DRV_STOP_STATE\n", bus_num);
                    
                    /* setup next state */
                    psI2CDrvBuffer->u8State = I2C_DRV_STOP_STATE;

                    DRV_MSG("Bus:%d gen stop\n", bus_num);  
                    
                    /* generate a stop condition */
                    iowrite32((UINT32) (MASTER_STOP_CMD), 
                              (void *) current_reg->cmd_status);
                }
            }
                  
            /* recode error status */
            psI2CDrvBuffer->u8ErrorStatus = I2C_DRV_ERROR_NONE;
            DRV_MSG("Record error status:I2C_DRV_ERROR_NONE\n");
        }
        else if (psI2CDrvBuffer->u8State == I2C_DRV_MST_RCV_STATE)
        {
            DRV_INFO("Bus:%d Write I2C_DRV_MST_RCV_STATE mode slave address ok!\n", bus_num); 
        }
        else if (psI2CDrvBuffer->u8State == I2C_DRV_SLV_RCV_STATE)
        {
            psI2CDrvBuffer->u8State = I2C_DRV_MST_XMIT_STATE;

            /* if has sending data */
            if (psI2CDrvBuffer->u8MsgSendDataSize - psI2CDrvBuffer->u8MsgSendCounter - 1)
            {
                psI2CDrvBuffer->u8MsgSendCounter++;

                /* set data */
                iowrite32((UINT32) psI2CDrvBuffer->u8MsgSendBuffer[psI2CDrvBuffer->u8MsgSendCounter], 
                          (void *) current_reg->trans_recv_byte_buf);

                /* generate a transmit condition */
                iowrite32((UINT32) (MASTER_TRANS_CMD), (void *) current_reg->cmd_status);
            
                DRV_INFO("Bus:%d send data:0x%02x counter:%d\n", 
                        bus_num, 
                        psI2CDrvBuffer->u8MsgSendBuffer[psI2CDrvBuffer->u8MsgSendCounter],
                        psI2CDrvBuffer->u8MsgSendCounter);
            }
            else
            {
                /* transfer to master receiver mode */
                if (psI2CDrvBuffer->u8MsgRecDataSize)
                {
                    DRV_MSG("Bus:%d setup next state -> I2C_DRV_MST_RCV_STATE\n", bus_num);
                    
                    /* setup next state */
                    psI2CDrvBuffer->u8State = I2C_DRV_MST_RCV_STATE;

                    /* set slave address as READ */
                    iowrite32((UINT32) psI2CDrvBuffer->u8MsgSendBuffer[0] | 1, 
                              (void *) current_reg->trans_recv_byte_buf);

                    DRV_MSG("Bus:%d repeated start\n", bus_num);

                    /* generate a start and transmit condition */
                    iowrite32((UINT32) (MASTER_START_CMD | MASTER_TRANS_CMD | MASTER_RECV_CMD), 
                              (void *) current_reg->cmd_status);                                      
                          
                }
            
                /* no data needs to be read */
                else
                {
                    DRV_MSG("Bus:%d setup next state -> I2C_DRV_STOP_STATE\n", bus_num);
                    /* setup next state */
                    psI2CDrvBuffer->u8State = I2C_DRV_STOP_STATE;

                    DRV_MSG("Bus:%d gen stop\n", bus_num);             
                    /* generate a stop condition */
                    iowrite32((UINT32) (MASTER_STOP_CMD), 
                              (void *) current_reg->cmd_status);
                }
            }
                  
            /* recode error status */
            psI2CDrvBuffer->u8ErrorStatus = I2C_DRV_ERROR_NONE;
        }       
        else
        {
            DRV_ERROR("Bus:%d Unknown State here! Error!\n", bus_num);  

            DRV_MSG("Bus:%d gen stop\n", bus_num);        
                
            /* generate a stop condition */
            iowrite32((UINT32) (MASTER_STOP_CMD), 
                      (void *) current_reg->cmd_status);            
        }                        
                
        DRV_MSG("Bus:%d ERROR_NONE\n", bus_num);                
    }

    if (int_reading & TRANS_NACK_RET_INT_STATUS)
    {
        DRV_ERROR("Bus:%d S Transmit NACK return\n", bus_num);
        
        /* recode error status */
        psI2CDrvBuffer->u8ErrorStatus = I2C_DRV_ERROR_NACK;
        
        /* setup next state */
        psI2CDrvBuffer->u8State = I2C_DRV_STOP_STATE;

        iowrite32((UINT32) (MASTER_STOP_CMD), 
                  (void *) current_reg->cmd_status);

        /* clear TRANS_NACK_RET_INT_STATUS */
        iowrite32((UINT32) TRANS_NACK_RET_INT_STATUS,
                  (void *) current_reg->int_status); 
    }

    if (int_reading & RECV_DONE_INT_STATUS)
    {
        DRV_MSG("Bus:%d enter RECV DONE INT STATE\n", bus_num);     
        
        if (I2C_DRV_MST_RCV_STATE == psI2CDrvBuffer->u8State)
        {
            /* Clear the interrupt status */
            iowrite32((UINT32) RECV_DONE_INT_STATUS, (void *) current_reg->int_status);                 

            /* Receive the interrupt, it express we have received one byte. 
               So add the receive counter here. */
            psI2CDrvBuffer->u8MsgRecCounter++;
            
            /* setup next state */
            psI2CDrvBuffer->u8State = I2C_DRV_MST_RCV_STATE;
            
            DRV_MSG("Bus:%d MST_RCV_STATE\n", bus_num);
            DRV_INFO("Bus:%d u8MsgRecDataSize[%d] u8MsgRecCounter[%d]\n", bus_num,psI2CDrvBuffer->u8MsgRecDataSize,psI2CDrvBuffer->u8MsgRecCounter);

            /* If this is the SMBUS Block read protocol, 
               update the read counter in the first reading */
            if ((psI2CDrvBuffer->u8MsgRecCounter == 1)
                && (I2C_DRV_TRANS_SMBUS_BLK == psI2CDrvBuffer->u8TransType))
            {
            	  DRV_MSG("Bus:%d Block read mode\n" , bus_num);
                
                /* Read the data byte in the data register */
                reg_reading = ioread32((void *) current_reg->trans_recv_byte_buf);

                psI2CDrvBuffer->u8MsgRecBuffer[0] = (UINT8) ((reg_reading & RECV_BYTE_BUF) >> 8);

                /* Recieve data include counter so we add 1 for this counter*/
                u8TotalByteCount = psI2CDrvBuffer->u8MsgRecBuffer[0] + 1;

                /* The minmux byte count need to greater and equal than 3 for the most of device */
                if ((u8TotalByteCount < 3) || (u8TotalByteCount > psI2CDrvBuffer->u8MsgRecDataSize))
                {
                    DRV_ERROR("Bus:%d SMBus Read Error, read count=0x%x\n", bus_num, u8TotalByteCount);
                    
                    /* setup next state */
                    psI2CDrvBuffer->u8State = I2C_DRV_STOP_STATE;
                    
                    /* generate a stop condition */
                    iowrite32((UINT32) (MASTER_STOP_CMD), 
                              (void *) current_reg->cmd_status);
                }
                else
                {
                    /* update the receive counter */
                    psI2CDrvBuffer->u8MsgRecDataSize = u8TotalByteCount;
                }
            }

            /* If the next byte is the last byte, 
               we need to send NAK when we read the last byte */
            if (1 == (psI2CDrvBuffer->u8MsgRecDataSize - psI2CDrvBuffer->u8MsgRecCounter))
            {
                /* Read the data byte in the data register */
                reg_reading = ioread32((void *) current_reg->trans_recv_byte_buf);

                /* Issue command to read the last byte and generate a NAK */
                iowrite32((UINT32) (MASTER_SLAVE_RECV_CMD_LAST | MASTER_RECV_CMD), 
                          (void *) current_reg->cmd_status);      
                
                /* The index of psI2CDrvBuffer->u8MsgRecBuffer is the zero-base, 
                   so the index should be equal 
                   (psI2CDrvBuffer->u8MsgRecCounter -1) */
                psI2CDrvBuffer->u8MsgRecBuffer[psI2CDrvBuffer->u8MsgRecCounter - 1] 
                    = (UINT8) ((reg_reading & RECV_BYTE_BUF) >> 8);
                
                DRV_MSG("Bus:%d Reading 0x%02x, current receive Counter %d \n", 
                       bus_num, 
                       psI2CDrvBuffer ->u8MsgRecBuffer[psI2CDrvBuffer->u8MsgRecCounter-1] , 
                       psI2CDrvBuffer->u8MsgRecCounter);  
                          
                DRV_MSG("Bus:%d GEN NACK\n", bus_num);
            }    
            
            /* If this is the last byte, we need to send STOP right now  */
            else if (0 == (psI2CDrvBuffer->u8MsgRecDataSize - psI2CDrvBuffer->u8MsgRecCounter))   
            {
                DRV_MSG("Bus:%d Read after STOP\n", bus_num);
                
                /* read the last byte in the data register */
                reg_reading = ioread32((void *) current_reg->trans_recv_byte_buf);

                /* generate a stop condition */
                iowrite32((UINT32) (MASTER_STOP_CMD), 
                          (void *) current_reg->cmd_status);

                /* setup next state */
                psI2CDrvBuffer->u8State = I2C_DRV_STOP_STATE;

                DRV_MSG("Nest Stage:I2C_DRV_STOP_STATE\n");
                DRV_MSG("reg_reading=0x%08x\n", reg_reading);
                
                psI2CDrvBuffer->u8MsgRecBuffer[psI2CDrvBuffer->u8MsgRecCounter - 1] 
                    = (UINT8) ((reg_reading & RECV_BYTE_BUF) >> 8);
                
                DRV_INFO("bus: %d Reading 0x%02x, Counter %d last byte\n", 
                         bus_num, 
                         psI2CDrvBuffer ->u8MsgRecBuffer[psI2CDrvBuffer->u8MsgRecCounter-1] , 
                         psI2CDrvBuffer->u8MsgRecCounter);          
            }          
            else
            {
                /* Read the data byte in the data register */
                reg_reading = ioread32((void *) current_reg->trans_recv_byte_buf);

                /* Issue command to read the next byte data */
                iowrite32((UINT32) (MASTER_RECV_CMD), 
                          (void *) current_reg->cmd_status);

                psI2CDrvBuffer->u8MsgRecBuffer[psI2CDrvBuffer->u8MsgRecCounter - 1] 
                = (UINT8) ((reg_reading & RECV_BYTE_BUF) >> 8);
                
                DRV_INFO("Bus:%d Reading 0x%02x, Counter %d\n", 
                         bus_num, 
                         psI2CDrvBuffer ->u8MsgRecBuffer[psI2CDrvBuffer->u8MsgRecCounter-1] , 
                         psI2CDrvBuffer->u8MsgRecCounter);  
            }
        }
        else if (I2C_DRV_SLV_RCV_STATE == psI2CDrvBuffer->u8State)
        {
            DRV_MSG("Bus:%d Slave Receive Mode\n", bus_num);
            
            if (int_reading & SLAVE_ADDR_RECV_MATCH_INT_STATUS)
            {
                DRV_MSG("SLAVE_ADDR_RECV_MATCH_INT_STATUS\n");
                
                /* read address */
                reg_reading = ioread32((void *) current_reg->trans_recv_byte_buf); 
        
                DRV_INFO("Bus:%d reg_reading is 0x%02x\n", bus_num, reg_reading);

                psI2CDrvFIFO->au8buffer[psI2CDrvFIFO->u8Head][psI2CDrvFIFO->au8MsgLen[psI2CDrvFIFO->u8Head]]
                    = (UINT8) ((reg_reading & RECV_BYTE_BUF) >> 8);
    
                DRV_INFO("Bus:%d Reading 0x%02x, Data counter %d\n", 
                         bus_num, (UINT8) ((reg_reading & RECV_BYTE_BUF) >> 8), 
                         psI2CDrvFIFO->au8MsgLen[psI2CDrvFIFO->u8Head]);
    
                /* increase message length */
                psI2CDrvFIFO->au8MsgLen[psI2CDrvFIFO->u8Head]++;                
                
                /* clearing interrupt status */
                iowrite32((UINT32) (RECV_DONE_INT_STATUS | SLAVE_ADDR_RECV_MATCH_INT_STATUS),
                          (void *) current_reg->int_status);                   
            }
            else
            {
                /* if buffer is not full */
                if (psI2CDrvFIFO->u8MsgCount < I2C_DRV_MAX_FIFO_SIZE)
                {
                    if (psI2CDrvFIFO->au8MsgLen[psI2CDrvFIFO->u8Head] < I2C_DRV_MAX_MSG_SIZE)
                    {                
                        /* get data from data register */
                        reg_reading = ioread32((void *) current_reg->trans_recv_byte_buf); 
                                               
                        iowrite32((UINT32) RECV_DONE_INT_STATUS,
                                  (void *) current_reg->int_status);                                            

                        psI2CDrvFIFO->au8buffer[psI2CDrvFIFO->u8Head][psI2CDrvFIFO->au8MsgLen[psI2CDrvFIFO->u8Head]]
                            = (UINT8) ((reg_reading & RECV_BYTE_BUF) >> 8);
        
                        DRV_INFO("Bus:%d Reading 0x%02x, Data counter %d\n", 
                                 bus_num, (UINT8) ((reg_reading & RECV_BYTE_BUF) >> 8), 
                                 psI2CDrvFIFO->au8MsgLen[psI2CDrvFIFO->u8Head]);
        
                        /* increase message length */
                        psI2CDrvFIFO->au8MsgLen[psI2CDrvFIFO->u8Head]++;                    

                    }
                    else
                    {
                        /* dummy read */
                        reg_reading = ioread32((void *) current_reg->trans_recv_byte_buf); 

                        iowrite32((UINT32) RECV_DONE_INT_STATUS,
                                  (void *) current_reg->int_status); 

                        /* buffer overflowed */
                        psI2CDrvBuffer->u8MsgRecCounter = 0xFF;

                        /* reset the message length */
                        psI2CDrvFIFO->au8MsgLen[psI2CDrvFIFO->u8Head] = 0;
                    
                        DRV_ERROR("Bus:%d receive buffer overflowed\n", bus_num);
                    }               
                }
                else
                {                
                    /* dummy read */
                    reg_reading = ioread32((void *) current_reg->trans_recv_byte_buf); 
                    
                    DRV_ERROR("Bus:%d Buffer full Dummy Reading 0x%02x\n", 
                               bus_num, (UINT8) ((reg_reading & RECV_BYTE_BUF) >> 8));
                    
                    iowrite32((UINT32) RECV_DONE_INT_STATUS,
                                (void *) current_reg->int_status);                        
                }
            }
        }
        else
        {
            DRV_WARN("Bus:%d at Unkonwn State %d\n", bus_num,psI2CDrvBuffer->u8State);
            
            /* dummy read */
            reg_reading = ioread32((void *) current_reg->trans_recv_byte_buf);          

            iowrite32((UINT32) RECV_DONE_INT_STATUS,
                      (void *) current_reg->int_status); 

            aess_i2c_reset((sI2CDrvBufferType *)psI2CDrvBuffer);

            aess_i2c_config((sI2CDrvBufferType *)psI2CDrvBuffer);
        }  
               
        DRV_MSG("Bus:%d exit RECV_DONE_INT_STATUS\n", bus_num);        
    }

    if (int_reading & MASTER_ARBITRATION_LOSS_INT_STATUS)
    {
        DRV_MSG("Bus:%d S Master Arbitration loss\n", bus_num);

        /* soft reset, turn off both the Master and Slave function enable, 
           then HW will reset all state machine and status */
        iowrite32((UINT32) (0), (void *) current_reg->func_ctrl);   

        aess_i2c_reset((sI2CDrvBufferType *)psI2CDrvBuffer);

        aess_i2c_config((sI2CDrvBufferType *)psI2CDrvBuffer);
        
        /* recode error status */
        psI2CDrvBuffer->u8ErrorStatus = I2C_DRV_ERROR_AL;
        
        /* setup next state */
        psI2CDrvBuffer->u8State = I2C_DRV_IDLE_STATE;    

        /* clearing Master Arbitration Loss Interrupt status */
        iowrite32((UINT32) MASTER_ARBITRATION_LOSS_INT_STATUS,
                  (void *) current_reg->int_status);

        DRV_ERROR("Bus:%d ERROR statue: I2C_DRV_ERROR_AL\n", bus_num);
    }

    /* if the transaction is started by aess_i2c_wr function */
    if (psI2CDrvBuffer->u8State == I2C_DRV_IDLE_STATE)
    {
        /* wake up */
        psI2CDrvBuffer->wq_flag = 1;
        
        wake_up_interruptible( (wait_queue_head_t *)&psI2CDrvBuffer->wq);
        
        if (I2C_DRV_MODE_IPMB == psI2CDrvBuffer->u8InitMode)
        {
            DRV_MSG("Mode is IPMB Set to I2C_DRV_SLV_RCV_STATE\n");
            
            psI2CDrvBuffer->u8State = I2C_DRV_SLV_RCV_STATE;
            i2c_buffer[bus_num].u8State = I2C_DRV_SLV_RCV_STATE;
        }
        else
        {
            DRV_MSG("Mode is NOT IPMB Set to I2C_DRV_IDLE_STATE\n");
            
            psI2CDrvBuffer->u8State = I2C_DRV_IDLE_STATE;
            i2c_buffer[bus_num].u8State = I2C_DRV_IDLE_STATE;
        }
    }

    return IRQ_HANDLED;
}

/******************************************************************************
*   FUNCTION        :   aess_i2c_isr
******************************************************************************/
/**
 *  @brief      The function is the interrupt handler
 *
 *  @return     IRQ_NONE: IRQ not handled
 *              IRQ_HANDLED: IRQ handled
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   None \n
 *
 *****************************************************************************/
irqreturn_t aess_i2c_isr(
                          /** IRQ number */
                          int irq,
                          /** Device ID */
                          void *dev_id
                         )
{
    volatile irqreturn_t irqreturn = IRQ_NONE;
    volatile UINT8 i = 0;
    volatile UINT8 u8IntMask = 0x01;
    volatile UINT32 reg_reading = ioread32((void *) I2C_GLB_REG);
    
    DRV_MSG("aess_i2c_isr GR 0x%08x\n",reg_reading);
    
#ifdef AST2100
    IRQ_EDGE_CLEAR(IRQ_I2C);
#endif

    /* if the interrupt is generated by I2C 0 */
    for(i = 0; i < I2C_DRV_MAX_BUS; i++)
    {
        if (reg_reading & u8IntMask) 
        {            
            if (IRQ_HANDLED == aess_i2c_isr_handler(i))
            {
            	  DRV_INFO("Found I2C IRQ:%d\n", i);
                irqreturn = IRQ_HANDLED;
            }
        }
        u8IntMask = u8IntMask << 1;
    }

    return irqreturn;
}

/******************************************************************************
*   FUNCTION        :   aess_i2c_reset
******************************************************************************/
/**
 *  @brief      The function is an internal function to reset an I2C bus.
 *
 *  @return     0 never fail.
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   None \n
 *
 *****************************************************************************/
static int aess_i2c_reset(
                          /** I2C Driver Buffer */
                          sI2CDrvBufferType *psI2CDrvBuffer
                         )
{
    UINT8 bus_num;
    
    i2c_reg_type *current_reg;
    static UINT32 reg_reading; 
    
    bus_num = psI2CDrvBuffer->u8Channel;
    current_reg = (i2c_reg_type *)&i2c_reg[bus_num];
    
    DRV_MSG("Bus:%d Enter aess_i2c_reset()...\n", bus_num);
    /* soft reset, turn off both the Master and Slave function enable, 
       then HW will reset all state machine and status */
    reg_reading = ioread32((void *) current_reg->func_ctrl); 
    
    iowrite32((UINT32) REG32_CLEAR_TO_ZERO, (void *) current_reg->func_ctrl); 
    iowrite32((UINT32) reg_reading, (void *) current_reg->func_ctrl);               
    
    /* reset variable */
    if (i2c_buffer[bus_num].u8InitMode == I2C_DRV_MODE_IPMB)
    {
        DRV_MSG("IPMB mode reset to I2C_DRV_SLV_RCV_STATE\n");
        i2c_buffer[bus_num].u8State = I2C_DRV_SLV_RCV_STATE;

       
        /* clear the memory */
        memset(i2c_buffer[bus_num].psFIFO, 0, sizeof(sI2CDrvFIFOType));
    }
    else
    {
    	  DRV_MSG("NOT IPMB mode reset to I2C_DRV_IDLE_STATE\n");
        i2c_buffer[bus_num].u8State = I2C_DRV_IDLE_STATE;
    }
    
    i2c_buffer[bus_num].u8ErrorStatus = I2C_DRV_ERROR_NONE;
    i2c_buffer[bus_num].u8MsgSendDataSize = 0;
    i2c_buffer[bus_num].u8MsgSendCounter = 0;
    i2c_buffer[bus_num].u8MsgRecDataSize = 0;
    i2c_buffer[bus_num].u8MsgRecCounter = 0;
    
    return 0;
}

/******************************************************************************
*   FUNCTION        :   aess_i2c_config
******************************************************************************/
/**
 *  @brief      The function is an internal function to configure an I2C bus.
 *
 *  @return     0 never fail.
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   None \n
 *
 *****************************************************************************/
static int aess_i2c_config(
                            /** I2C Driver Buffer */
                            sI2CDrvBufferType *psI2CDrvBuffer
                          )
{
    UINT8 bus_num;
    i2c_reg_type *current_reg;
    
    bus_num = psI2CDrvBuffer->u8Channel;
    current_reg = (i2c_reg_type *)&i2c_reg[bus_num];
    
    DRV_MSG("Bus:%d Enter aess_i2c_config()\n",bus_num);

    iowrite32((UINT32) REG32_CLEAR_TO_ZERO, 
              (void *) current_reg->clk_n_ac_timing_1);
              
    iowrite32((UINT32) REG32_CLEAR_TO_ZERO, 
              (void *) current_reg->clk_n_ac_timing_2);

    iowrite32((UINT32) REG32_CLEAR_TO_ZERO, 
              (void *) current_reg->func_ctrl);                       

    /* clearing interrupt control */
    iowrite32((UINT32) (REG32_CLEAR_TO_ZERO),
              (void *) current_reg->int_ctrl);            
              
    aess_i2c_reset(&i2c_buffer[bus_num]);             

    /* Set Slave address */
    iowrite32((UINT32) ((psI2CDrvBuffer->u8SlaveAddr) >> 1),
                        (void *) current_reg->dev_addr);

    /* Set AC Timing: Assume System Core Clock = 50MHz, 
       and I2C Clock rate = 400KHz. */
    
    if (psI2CDrvBuffer->u8Frequency == I2C_DRV_FREQ_400)
    {
        iowrite32((UINT32) (0x77733413),
                  (void *) current_reg->clk_n_ac_timing_1);        
    }
    else if (psI2CDrvBuffer->u8Frequency == I2C_DRV_FREQ_100)
    {
        iowrite32((UINT32) (0x77777414),
                  (void *) current_reg->clk_n_ac_timing_1);        
    }
    else
    {
        iowrite32((UINT32) (0x77777415),
                  (void *) current_reg->clk_n_ac_timing_1);        
    }     

    iowrite32((UINT32) REG32_CLEAR_TO_ZERO, 
              (void *) current_reg->clk_n_ac_timing_2);

    /* clearing interrupt status */
    iowrite32((UINT32) (~REG32_CLEAR_TO_ZERO),
              (void *) current_reg->int_status);

    if (psI2CDrvBuffer->u8CurMode == I2C_DRV_MODE_IPMB)
    {
        /* enable function mode Master Function Enable */
        iowrite32((UINT32) (MASTER_FUNCTION_ENABLE | SLAVE_FUNCTION_ENABLE),
                  (void *) current_reg->func_ctrl);        
    }
    else
    {
        /* enable function mode Master Function Enable */
        iowrite32((UINT32) (MASTER_FUNCTION_ENABLE),
                  (void *) current_reg->func_ctrl);
    }

    /* enable interrupt */
    iowrite32((UINT32) (ABNORMAL_START_N_STOP_INT_ENABLE
#ifdef I2C_BUS_RECOVER    
                        | BUS_RECOVER_DONE_ENABLE   
#endif                        
//                      | SLAVE_ADDR_RECV_MATCH_INT_ENABLE
                        | NORMAL_STOP_INT_ENABLE        
                        | MASTER_ARBITRATION_LOSS_INT_ENABLE
                        | RECV_DONE_INT_ENABLE              
                        | TRANS_NACK_RET_INT_ENABLE         
                        | TRANS_DONE_ACK_RET_INT_ENABLE),       
              (void *) current_reg->int_ctrl); 
     
    DRV_INFO("Bus:%d slave address 0x%x\n", bus_num, psI2CDrvBuffer->u8SlaveAddr);
    
    /* set flag to indicate this bus has been initiated */
    psI2CDrvBuffer->u8HasInit = 1;
    
    DRV_MSG("Bus:%d exit aess_i2c_config()\n", bus_num);
    return 0;
}

/******************************************************************************
*   FUNCTION        :   aess_i2c_wr
******************************************************************************/
/**
 *  @brief      The function is an internal function to start an I2C transaction.
 *              After generating a start condition with slave address, 
 *              it will wait until interrupt handler wrote/read the rest of data.
 *
 *  @return     0 if success.
 *              minus if fail.
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   None \n
 *
 *****************************************************************************/
static int aess_i2c_wr(
                        /** I2C Driver Buffer */
                        sI2CDrvBufferType *psI2CDrvBuffer
                      )
{
    UINT8 bus_num;
    i2c_reg_type *current_reg;
    UINT16 i ;
    
    bus_num = psI2CDrvBuffer->u8Channel;
    current_reg = (i2c_reg_type *)&i2c_reg[bus_num];
    
    DRV_MSG("I2C Bus[%d] enter aess_i2c_wr()\n", bus_num);
        
    /* check for transmitter or receiver mode */
    if (psI2CDrvBuffer->u8MsgSendBuffer[0] & 0x01)
    {
    	  DRV_MSG("aess_i2c_wr receiver mode\n");
        if (psI2CDrvBuffer->u8MsgRecDataSize != 0)
        {
            DRV_MSG("I2C Bus[%d] Rec Data Size %d\n", 
                   bus_num, 
                   psI2CDrvBuffer->u8MsgRecDataSize);

            DRV_MSG("set to I2C_DRV_MST_RCV_STATE\n"); 
            
            /* setup next state */
            psI2CDrvBuffer->u8State = I2C_DRV_MST_RCV_STATE;
            i2c_buffer[bus_num].u8State = I2C_DRV_MST_RCV_STATE;
        }
        else
        {
            DRV_MSG("set to I2C_DRV_IDLE_STATE\n");
            
            /* setup next state */
            psI2CDrvBuffer->u8State = I2C_DRV_IDLE_STATE;
            i2c_buffer[bus_num].u8State = I2C_DRV_IDLE_STATE;
            
            return 0;
        }
    }
    
    /* transmitter mode */
    else
    {
        DRV_MSG("aess_i2c_wr transmitter mode set to I2C_DRV_MST_XMIT_STATE\n");          
        /* setup next state */
        psI2CDrvBuffer->u8State = I2C_DRV_MST_XMIT_STATE;
        i2c_buffer[bus_num].u8State = I2C_DRV_MST_XMIT_STATE;
    }

    /* set data */
    iowrite32((UINT32) psI2CDrvBuffer->u8MsgSendBuffer[psI2CDrvBuffer->u8MsgSendCounter], 
              (void *) current_reg->trans_recv_byte_buf);

    DRV_INFO("   ==>trans_buf:0x%x\n",
           ioread32((void *) i2c_reg[bus_num].trans_recv_byte_buf));
    
    DRV_INFO("psI2CDrvBuffer->u8State is %d\n",psI2CDrvBuffer->u8State);
    DRV_INFO("i2c_buffer[bus_num].u8State is %d\n",i2c_buffer[bus_num].u8State);  
    
    /* check for transmitter or receiver mode */
    if (psI2CDrvBuffer->u8MsgSendBuffer[0] & 0x01)
    {
        DRV_MSG("I2C Bus[%d] generate a start and recive condition\n", bus_num); 

        /* TBD: NEED to determine if only one byte will to read */
        if (1 == psI2CDrvBuffer->u8MsgRecDataSize)
        {
            /* generate a Sr + SLAVEADDR + READ Data byte (NAK) */
            iowrite32((UINT32) (MASTER_START_CMD | MASTER_TRANS_CMD 
                                | MASTER_SLAVE_RECV_CMD_LAST | MASTER_RECV_CMD), 
                      (void *) current_reg->cmd_status);    
        }
        else
        {
            /* There are more one byte need to receive */
            /* generate a Sr + SLAVEADDR + READ Data byte (ACK) */
            iowrite32((UINT32) (MASTER_START_CMD | MASTER_TRANS_CMD 
                                | MASTER_RECV_CMD), 
                      (void *) current_reg->cmd_status);    
        }
    }
    else
    {
        DRV_MSG("I2C Bus[%d] generate a start and send slave address and transmit condition\n", bus_num);
        
        /* generate a start and transmit condition */
        iowrite32((UINT32) (MASTER_START_CMD | MASTER_TRANS_CMD), 
                            (void *) current_reg->cmd_status);
    }
    
    /* wait for I2C transaction */
    if (0 == wait_event_interruptible_timeout(psI2CDrvBuffer->wq, 
                                              psI2CDrvBuffer->wq_flag == 1, 
                                              20))
    {
        /* If timeout, we still need to double the flag whether is setting or not.
           If the flag is set, the I2C transaction is still success to finish */
        if (0 == psI2CDrvBuffer->wq_flag)
        {
            /* I2C Timeout */
            
            /* reset i2c bus to try recoverying a hung bus */
            aess_i2c_reset(psI2CDrvBuffer);

            /* reconfig i2c bus */
            aess_i2c_config(psI2CDrvBuffer);

            /* recode error status */
            psI2CDrvBuffer->u8ErrorStatus = I2C_DRV_ERROR_TIMEOUT;
            DRV_ERROR("Bus:%d I2C_DRV_ERROR_TIMEOUT error\n",psI2CDrvBuffer->u8Channel);

            if (psI2CDrvBuffer->u8CurMode == I2C_DRV_MODE_IPMB)
            {
                /* enable function mode Master Function Enable */
                iowrite32((UINT32) (MASTER_FUNCTION_ENABLE | SLAVE_FUNCTION_ENABLE),
                          (void *) current_reg->func_ctrl);        
            }

            /* error occur */
            return -EFAULT;     
        }
    }
    
    /* clear the flag of condition checking */
    psI2CDrvBuffer->wq_flag = 0;
    /* Add a workround for some PSU device will hold stop pin cause the bus hang */
    for ( i = 0; i < 1000; i++)
    {
      /* if i2c bus is busy */
      if (BUS_BUSY_STATUS & ioread32((void *) current_reg->cmd_status))
         udelay(1);
      else
         break;
    }

      DRV_MSG("I2C Bus[%d] exit wq\n", bus_num);
    
    return 0;
}

/******************************************************************************
*   FUNCTION        :   aess_i2c_ctrlhw
******************************************************************************/
/**
 *  @brief      The function is an internal function to recovery I2C Bus.
 *
 *  @return     always return 0
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   None \n
 *
 *****************************************************************************/
static int aess_i2c_ctrlhw(
                            /** I2C Driver Buffer */
                            sI2CDrvBusInfoType *psI2CDrvBusInfo
                          )
{
  UINT8 bus_num;
  unsigned long spin_lock_flag;
  i2c_reg_type *current_reg;

  /* get spin lock */
  spin_lock_irqsave(&my_lock, spin_lock_flag);
  bus_num = psI2CDrvBusInfo->u8Channel;	
  current_reg = (i2c_reg_type *)&i2c_reg[bus_num];

  /* recode error status */
  i2c_buffer[bus_num].u8ErrorStatus = I2C_DRV_ERROR_BUS;

  switch (psI2CDrvBusInfo->u8CtrlHW)
  {
          case I2C_DRV_CLK_PULSE:

               DRV_MSG("Bus:%d aess_i2c_ctrlhw: I2C_DRV_CLK_PULSE\n", bus_num);
               /*Issue 1~8 clock to recover the bus*/
               iowrite32((UINT32) (BUS_RECOVER),(void *) current_reg->cmd_status);
               /* Wait 8 clock plus */
               udelay(100);
               break;

          case I2C_DRV_FORCE_STOP:

               DRV_MSG("Bus:%d aess_i2c_ctrlhw: I2C_DRV_FORCE_STOP\n", bus_num);				
               /* generate STOP condition */                  
               iowrite32((UINT32) (MASTER_STOP_CMD), (void *) current_reg->cmd_status);

                break;

           case I2C_DRV_RESET_SMBUS:

                DRV_MSG("Bus:%d I2C_DRV_RESET_SMBUS\n", bus_num);

                break;

           default:

                DRV_ERROR("Bus:%d not support type\n", bus_num);

                break;
  }
	
  /* reset variable */
  i2c_buffer[bus_num].u16CurStartCount = 0;
  i2c_buffer[bus_num].u16CurStopCount = 0;			

  if (i2c_buffer[bus_num].u8InitMode == I2C_DRV_MODE_IPMB)
  {
      i2c_buffer[bus_num].u8State = I2C_DRV_SLV_RCV_STATE;
  }
  else
  {
      i2c_buffer[bus_num].u8State = I2C_DRV_IDLE_STATE;
  }
	
	/* recode error status */                             
      i2c_buffer[bus_num].u8ErrorStatus = I2C_DRV_ERROR_NONE;

      /* release spin lock */
      spin_unlock_irqrestore(&my_lock, spin_lock_flag);

      return 0;
}

/******************************************************************************
*   FUNCTION        :   aess_i2c_ioctl
******************************************************************************/
/**
 *  @brief      The function is an internal function to handle I2C command.
 *
 *  @return     return 0 if success
 *              return minus if fail
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   None \n
 *
 *****************************************************************************/
static int aess_i2c_ioctl(
                          /** inode number */ 
                          struct inode * inode,
                          /** file pointer */
                          struct file *filp, 
                          /** Command number */
                          unsigned int cmd, 
                          /** Argument */
                          unsigned long arg
                         )
{
    int result;
    UINT8 bus_num = 0;
    UINT8 u8BusStatus;
    UINT16 i = 0;
    UINT32 reg_reading;
    unsigned long spin_lock_flag;
    i2c_reg_type *current_reg;
    sI2CDrvBusInfoType *psI2CDrvBusInfo;
    sI2CDrvBufInfoType *psI2CDrvBufInfo;
    sI2CDrvFIFOType *psI2CDrvFIFO;
    
    DRV_MSG("aess_i2c_ioctl\n");
    
    switch (cmd)
    {
        case AESS_I2CDRV_INIT:
        {
            
            DRV_MSG("ioctl AESS_I2CDRV_INIT\n");
            
            psI2CDrvBusInfo = (sI2CDrvBusInfoType *) arg;
            
            /* Check the structure address is OK */
            if(!access_ok(VERIFY_WRITE, (void *) psI2CDrvBusInfo, 
                          sizeof(sI2CDrvBusInfoType)))
            {
                DRV_ERROR("aess_i2c_ioctl struct access error!\n");
                
                /* error occur */
                return -EFAULT;
            }
            
            /* retrieve bus number */
            bus_num = psI2CDrvBusInfo->u8Channel;

            if (bus_num >= I2C_DRV_MAX_BUS)
                return -EFAULT;
            
            /* get semaphore */
            if(down_interruptible(&i2c_buffer[bus_num].sem))
            {
                DRV_ERROR("aess_i2c_ioctl AESS_I2CDRV_INIT get semaphore fail!\n");
                return -ERESTARTSYS;
            }
            
            if (i2c_buffer[bus_num].u8HasInit == 0)
            {
                /* Issue a SMBUS timeout to reset all device on this bus*/
                aess_i2c_reset_smbus(bus_num);

                i2c_buffer[bus_num].u8Channel = bus_num;
                i2c_buffer[bus_num].u8InitMode = psI2CDrvBusInfo->u8InitMode;
                i2c_buffer[bus_num].u8CurMode = psI2CDrvBusInfo->u8InitMode;
                i2c_buffer[bus_num].u8SlaveAddr = psI2CDrvBusInfo->u8SlaveAddr;
                i2c_buffer[bus_num].u8Frequency = psI2CDrvBusInfo->u8Frequency;
                i2c_buffer[bus_num].u16DriverID = psI2CDrvBusInfo->u16DriverID;
                i2c_buffer[bus_num].u32RecFlag = psI2CDrvBusInfo->u32RecFlag;
                i2c_buffer[bus_num].tLockMode = I2C_DRV_LOCK_MODE_UNLOCK ; 
                
                spin_lock_init(&i2c_buffer[bus_num].data_lock);
                
                if (i2c_buffer[bus_num].u8CurMode == I2C_DRV_MODE_IPMB)
                {
                    DRV_MSG("IPMB bus init to I2C_DRV_SLV_RCV_STATE\n");
                    i2c_buffer[bus_num].u8State = I2C_DRV_SLV_RCV_STATE;
                }
                else
                {
                    i2c_buffer[bus_num].u8State = I2C_DRV_IDLE_STATE;
                }
                
                i2c_buffer[bus_num].u8ErrorStatus = I2C_DRV_ERROR_NONE;
                
                /* clear the flag of condition checking */
                i2c_buffer[bus_num].wq_flag = 0;
                
                /* initiate wait queue */
                init_waitqueue_head(&i2c_buffer[bus_num].wq);
                
                if (i2c_buffer[bus_num].u8CurMode == I2C_DRV_MODE_IPMB)
                {                                           
                    /* allocate memory for receive message FIFO */
                    i2c_buffer[bus_num].psFIFO 
                        = kmalloc(sizeof(sI2CDrvFIFOType), GFP_KERNEL);
                    
                    /* clear the memory */
                    memset(i2c_buffer[bus_num].psFIFO, 0, sizeof(sI2CDrvFIFOType));

                    /* initiate spin lock */
                    spin_lock_init(&psI2CDrvFIFO->lock);
                }
                
                result = aess_i2c_config(&i2c_buffer[bus_num]);
                
                if (result)
                {
                    DRV_ERROR("Bus:%d AESS_I2CDRV_INIT error\n", 
                           bus_num);
                    
                    /* release semaphore */
                    up(&i2c_buffer[bus_num].sem);
                    
                    return (result);
                }
                
                /* if i2c bus is busy */
                reg_reading = ioread32((void *) i2c_reg[bus_num].cmd_status);
            
                if (reg_reading & BUS_BUSY_STATUS)
                {
                	 for ( i = 0; i < 1000; i++)
                   {
                     /* if i2c bus is busy */
                     if (BUS_BUSY_STATUS & ioread32((void *) i2c_reg[bus_num].cmd_status))
                     udelay(100);
                     else
                     break;
                   }

                   if (i == 1000)
                   {
                     /* release semaphore */
                     up(&i2c_buffer[bus_num].sem);
                
                     DRV_ERROR("Bus:%d ERROR_BUSY\n", bus_num);
                
                     return -EFAULT;
                   }
                }
            }
            /* release semaphore */
            up(&i2c_buffer[bus_num].sem);
            
            break;
        }
        
        case AESS_I2CDRV_CONFIG:
        {
            DRV_MSG("ioctl AESS_I2CDRV_CONFIG\n");
            
            psI2CDrvBusInfo = (sI2CDrvBusInfoType *) arg;
            
            /* Check the structure address is OK */
            if(!access_ok(VERIFY_WRITE, (void *) psI2CDrvBusInfo, 
                          sizeof(sI2CDrvBusInfoType)))
            {
                DRV_ERROR("Bus:%d AESS_I2CDRV_CONFIG struct access error!\n", bus_num);
                
                /* error occur */
                return -EFAULT;
            }
            
            /* retrieve bus number */
            bus_num = psI2CDrvBusInfo->u8Channel;

            if (bus_num >= I2C_DRV_MAX_BUS)
                return -EFAULT;
            
            /* get semaphore */
            if(down_interruptible(&i2c_buffer[bus_num].sem))
            {
                return -ERESTARTSYS;
            } 
            
            if (i2c_buffer[bus_num].u8HasInit == 1)
            {
                /* allow to change the current mode */
                i2c_buffer[bus_num].u8CurMode = psI2CDrvBusInfo->u8CurMode;
                
                /* invoke the sub function to initiate the I2C bus */
                result = aess_i2c_config(&i2c_buffer[bus_num]);
                
                DRV_MSG("Bus:%d ioctl AESS_I2CDRV_CONFIG reconfig\n", bus_num);
                
                if (result)
                {
                    /* release semaphore */
                    up(&i2c_buffer[bus_num].sem);
                    DRV_ERROR("aess_i2c_config() fail!\n");
                    return (result);
                }
            }
            
            /* release semaphore */
            up(&i2c_buffer[bus_num].sem); 
            
            break;
        }
        
        case AESS_I2CDRV_WR:
        {
            DRV_MSG("ioctl AESS_I2CDRV_WR\n");
            
            /* */
            psI2CDrvBufInfo = (sI2CDrvBufInfoType *) arg;
            
            /* Check the structure address is OK */
            if(!access_ok(VERIFY_WRITE, (void *) psI2CDrvBufInfo, 
                          sizeof(sI2CDrvBufInfoType)))
            {
                DRV_ERROR("struct access write error!\n");
                
                /* error occur */
                return -EFAULT;
            }

            /* Check related parameter */
            if (I2C_DRV_TRANS_SMBUS_BLK == psI2CDrvBufInfo->u8TransType)
            {
                if ((psI2CDrvBufInfo->u8MsgRecDataSize > 0)
                    && psI2CDrvBufInfo->u8MsgRecDataSize < 3)
                {
                    DRV_ERROR("ioctl AESS_I2CDRV_WR receive data lengeth of SMBUS Block read need greater than 2\n");
                    
                    /* error occur */
                    return -EFAULT;
                }
            }

            /* retrieve bus number */
            bus_num = psI2CDrvBufInfo->u8Channel;

            if (bus_num >= I2C_DRV_MAX_BUS)
                return -EFAULT;

            
            /* if i2c bus is busy */
            reg_reading = ioread32((void *) i2c_reg[bus_num].cmd_status);
            
            if (reg_reading & BUS_BUSY_STATUS)
            {
                /* report error status */
                psI2CDrvBufInfo->u8ErrorStatus = I2C_DRV_ERROR_BUSY;
                
                //aess_i2c_check_bus_status(bus_num);
                //aess_i2c_config(&i2c_buffer[bus_num]);
                
                /* release semaphore */
                up(&i2c_buffer[bus_num].sem);
                
                DRV_ERROR("Bus:%d ERROR_BUSY\n", bus_num);
                
                return 0;
            }
            
            /* get semaphore */
            if(down_interruptible(&i2c_buffer[bus_num].sem))
            {
                return -ERESTARTSYS;
            }

            i2c_buffer[bus_num].u8ErrorStatus = I2C_DRV_ERROR_NONE;
            i2c_buffer[bus_num].u8TransType = psI2CDrvBufInfo->u8TransType;
            
            i2c_buffer[bus_num].u8MsgSendBuffer[0] = psI2CDrvBufInfo->u8DeviceAddr;
            
            if (psI2CDrvBufInfo->u8MsgSendDataSize == 0)
            {
                DRV_ERROR("Bus:%d No sending data\n", bus_num);
                i2c_buffer[bus_num].u8MsgSendBuffer[0] |= 0x01;
            }
            else
            {
                if (copy_from_user ((void *) &i2c_buffer[bus_num].u8MsgSendBuffer[1],
                                    (void *) psI2CDrvBufInfo->pu8MsgSendBuffer,
                                    psI2CDrvBufInfo->u8MsgSendDataSize))
                {
                	  DRV_ERROR("Bus:%d ioctl AESS_I2CDRV_WR copy_from_user fail!!\n", bus_num);
                    /* release semaphore */
                    up(&i2c_buffer[bus_num].sem);
                    
                    return -EFAULT;
                }
                i2c_buffer[bus_num].u8MsgSendBuffer[0] &= 0xfe;
            }
            
            i2c_buffer[bus_num].u8MsgSendDataSize = (psI2CDrvBufInfo->u8MsgSendDataSize + 1);
            i2c_buffer[bus_num].u8MsgSendCounter = 0;
            i2c_buffer[bus_num].u8MsgRecDataSize = psI2CDrvBufInfo->u8MsgRecDataSize;
            i2c_buffer[bus_num].u8MsgRecCounter = 0;
            
            /* clear the flag of condition checking */
            i2c_buffer[bus_num].wq_flag = 0;

            DRV_MSG("Bus:%d u8MsgSendData %x\n", 
                   bus_num, i2c_buffer[bus_num].u8MsgSendBuffer[0]);
            
            DRV_MSG("Bus:%d u8MsgSendDataSize %d\n", 
                   bus_num, i2c_buffer[bus_num].u8MsgSendDataSize);
            
            DRV_MSG("Bus:%d u8MsgSendCounter %d\n", 
                   bus_num, i2c_buffer[bus_num].u8MsgSendCounter);
            
            DRV_MSG("Bus:%d u8MsgRecDataSize %d\n", 
                   bus_num, i2c_buffer[bus_num].u8MsgRecDataSize);
            
            DRV_MSG("Bus:%d u8MsgRecCounter %d\n", 
                   bus_num, i2c_buffer[bus_num].u8MsgRecCounter);
            
            DRV_MSG("Bus:%d Enter wr func\n", bus_num);

            /* invoke the sub function to execute */
            result = aess_i2c_wr(&i2c_buffer[bus_num]);
            
            DRV_MSG("Bus:%d Exit wr func, result=0x%x:0 = success, i2c_buffer[bus_num].u8ErrorStatus=0x%x\n",
                    bus_num, result, i2c_buffer[bus_num].u8ErrorStatus);
            
            /** Store Error Status for sysfs use */
            if (i2c_buffer[bus_num].u8ErrorStatus != I2C_DRV_ERROR_NONE)
            {
                psI2CDrvBufInfo->u8ErrorStatus = i2c_buffer[bus_num].u8ErrorStatus;
                /* reset error status */
                i2c_buffer[bus_num].u8ErrorStatus = I2C_DRV_ERROR_NONE;
                
                if (i2c_buffer[bus_num].u8CurMode == I2C_DRV_MODE_IPMB)
                {
                    DRV_ERROR("aess_i2c_wr fail set back to I2C_DRV_SLV_RCV_STATE\n"); 
                    i2c_buffer[bus_num].u8State = I2C_DRV_SLV_RCV_STATE;
                }
                
                /* release semaphore */
                up(&i2c_buffer[bus_num].sem);
                
                return -EFAULT;
            }
            
                if (i2c_buffer[bus_num].u8MsgRecCounter > 0)
                {  
                    DRV_MSG("Bus:%d Copy data to user mode\n", bus_num);
                    
                    if (copy_to_user((void *) psI2CDrvBufInfo->pu8MsgRecBuffer,
                                     (void *) &i2c_buffer[bus_num].u8MsgRecBuffer[0],
                                     i2c_buffer[bus_num].u8MsgRecCounter))
                    {
                        i2c_buffer[bus_num].u8MsgRecCounter = 0;
                        
                        if (i2c_buffer[bus_num].u8CurMode == I2C_DRV_MODE_IPMB)
                        {
                            i2c_buffer[bus_num].u8State 
                            = I2C_DRV_SLV_RCV_STATE;
                        }
                        
                        /* release semaphore */
                        up(&i2c_buffer[bus_num].sem);
                        
                        DRV_ERROR("ioctl AESS_I2CDRV_WR copy_to_user fail!!\n");
                        return -EFAULT;
                    }
                    
                    psI2CDrvBufInfo->u8MsgRecDataSize = i2c_buffer[bus_num].u8MsgRecCounter;
                    
                    i2c_buffer[bus_num].u8MsgRecCounter = 0;
                }
            
            /* release semaphore */
            up(&i2c_buffer[bus_num].sem);
            
            break;
        }
        
        case AESS_I2CDRV_GET_MSG:
        {
            DRV_MSG("ioctl AESS_I2CDRV_GET_MSG\n");
            
            psI2CDrvBufInfo = (sI2CDrvBufInfoType *) arg;

            /* Check the structure address is OK */
            if(!access_ok(VERIFY_WRITE, (void *) psI2CDrvBufInfo, 
                          sizeof(sI2CDrvBufInfoType)))
            {
                DRV_ERROR("struct access error!\n");
                
                /* error occur */
                return -EFAULT;
            }
            
            /* retrieve bus number */
            bus_num = psI2CDrvBufInfo->u8Channel;

            if (bus_num >= I2C_DRV_MAX_BUS)
                return -EFAULT;
            
            psI2CDrvFIFO = i2c_buffer[bus_num].psFIFO;

            /* if there is no received message */
            if (psI2CDrvFIFO->u8MsgCount == 0)
            {
                /* set data size to 0 */
                psI2CDrvBufInfo->u8MsgRecDataSize = 0;
                
                DRV_MSG("Bus:%d No received data\n", bus_num);
                
                return 0;
            }
            
            /* if the receive count is big enough */
            if (psI2CDrvFIFO->au8MsgLen[psI2CDrvFIFO->u8Tail] >= I2C_DRV_IPMB_MSG_MIN)
            {
                DRV_MSG("bus:%d Copy data to user mode\n", bus_num);
                
                /* IPMB packet includes slave address */
                if (copy_to_user((void *) psI2CDrvBufInfo->pu8MsgRecBuffer,
                                 (void *) &psI2CDrvFIFO->au8buffer[psI2CDrvFIFO->u8Tail][0],
                                 psI2CDrvFIFO->au8MsgLen[psI2CDrvFIFO->u8Tail]))
                {
                	  DRV_ERROR("ioctl AESS_I2CDRV_GET_MSG copy_to_user fail!!\n");
                    return -EFAULT;
                }
            }
            
            /* IPMB message size includes the slave address */
            psI2CDrvBufInfo->u8MsgRecDataSize = psI2CDrvFIFO->au8MsgLen[psI2CDrvFIFO->u8Tail];
                            
            DRV_INFO("Bus:%d u8Tail %d au8MsgLen %d u8MsgCount %d\n", 
                     bus_num,
                     psI2CDrvFIFO->u8Tail,
                     psI2CDrvBufInfo->u8MsgRecDataSize,
                     psI2CDrvFIFO->u8MsgCount);
            
            /* reset the message length */
            psI2CDrvFIFO->au8MsgLen[psI2CDrvFIFO->u8Tail] = 0;
            
            /* get spin lock */
            spin_lock_irqsave(&psI2CDrvFIFO->lock, spin_lock_flag);
            
            /* move the tail to the next */
            if (psI2CDrvFIFO->u8Tail == (I2C_DRV_MAX_FIFO_SIZE - 1))
            {
                psI2CDrvFIFO->u8Tail = 0;
            }
            else
            {
                psI2CDrvFIFO->u8Tail++;
            }
            
            /* decrease the message count */
            psI2CDrvFIFO->u8MsgCount--;
            
            /* release spin lock */
            spin_unlock_irqrestore(&psI2CDrvFIFO->lock, spin_lock_flag);

            DRV_INFO("Bus:%d u8Tail %d u8MsgCount %d\n", 
                     bus_num, psI2CDrvFIFO->u8Tail, psI2CDrvFIFO->u8MsgCount);
            
            /* if FIFO is not empty */
            if (psI2CDrvFIFO->u8MsgCount)
            {
                DRV_MSG("generate event to notify uplayer AP!\n");
                
                /* generate event to notify */
                aess_generate_driver_event(i2c_buffer[bus_num].u16DriverID, 
                                           i2c_buffer[bus_num].u32RecFlag);
            }
            
            break;
        }
        
        case AESS_I2CDRV_RESET:
        {
            DRV_MSG("ioctl AESS_I2CDRV_RESET\n");
            
            /* */
            psI2CDrvBusInfo = (sI2CDrvBusInfoType *) arg;
            
            /* Check the structure address is OK */
            if(!access_ok(VERIFY_WRITE, (void *) psI2CDrvBusInfo, 
                          sizeof(sI2CDrvBusInfoType)))
            {
                DRV_ERROR("ioctl AESS_I2CDRV_RESET struct access error!\n");
                
                /* error occur */
                return -EFAULT;
            }
            
            /* retrieve bus number */
            bus_num = psI2CDrvBusInfo->u8Channel;

            if (bus_num >= I2C_DRV_MAX_BUS)
                return -EFAULT;
            
            /* get semaphore */
            /* if(down_interruptible(&i2c_buffer[bus_num].sem))
            {
                return -ERESTARTSYS;
            } */
            
            result = aess_i2c_reset(&i2c_buffer[bus_num]);
            
            if (result)
            {
                /* release semaphore */
                /* up(&i2c_buffer[bus_num].sem); */
                
                return (result);
            }
            
            /* release semaphore */
            /* up(&i2c_buffer[bus_num].sem); */
            
            break;
        }
        
        case AESS_I2CDRV_GET_STATUS:
            
            DRV_MSG("ioctl AESS_I2CDRV_GET_STATUS\n");
            
            psI2CDrvBusInfo = (sI2CDrvBusInfoType *) arg;
            
            /* Check the structure address is OK */
            if(!access_ok(VERIFY_WRITE, (void *) psI2CDrvBusInfo, 
                          sizeof(sI2CDrvBusInfoType)))
            {
                DRV_ERROR("ioctl AESS_I2CDRV_GET_STATUS struct access error!\n");
                
                /* error occur */
                return -EFAULT;
            }
            
            /* retrieve bus number */
            bus_num = psI2CDrvBusInfo->u8Channel;

            if (bus_num >= I2C_DRV_MAX_BUS)
                return -EFAULT;
            
            if (i2c_buffer[bus_num].u8ErrorStatus == I2C_DRV_ERROR_NONE)
            {
                /* if i2c bus is busy */
                reg_reading = ioread32((void *) i2c_reg[bus_num].cmd_status);
                if (reg_reading & BUS_BUSY_STATUS)
                {
                    DRV_ERROR("Bus:%d ERROR_BUSY\n", bus_num);
                    
                    /* report error status */
                    psI2CDrvBusInfo->u8ErrorStatus = I2C_DRV_ERROR_BUSY;
                }
                else
                {
                    psI2CDrvBusInfo->u8ErrorStatus = I2C_DRV_ERROR_NONE;
                }
            }
            else
            {
                DRV_ERROR("AESS_I2CDRV_GET_STATUS Bus:%d u8ErrorStatus:%x\n",bus_num, i2c_buffer[bus_num].u8ErrorStatus);
                psI2CDrvBusInfo->u8ErrorStatus = i2c_buffer[bus_num].u8ErrorStatus;
                /* We set error status to none for next read/write command */
                i2c_buffer[bus_num].u8ErrorStatus  = I2C_DRV_ERROR_NONE;
            }
            
            break;
            
        case AESS_I2CDRV_GET_HW_STATUS:
        {
            DRV_MSG("ioctl AESS_I2CDRV_GET_HW_STATUS\n");
            
            psI2CDrvBusInfo = (sI2CDrvBusInfoType *) arg;
            
            /* Check the structure address is OK */
            if(!access_ok(VERIFY_WRITE, (void *) psI2CDrvBusInfo, 
                          sizeof(sI2CDrvBusInfoType)))
            {
                DRV_ERROR("ioctl AESS_I2CDRV_GET_HW_STATUS struct access error!\n");
                
                /* error occur */
                return -EFAULT;
            }
            
            /* retrieve bus number */
            bus_num = psI2CDrvBusInfo->u8Channel;
            current_reg = (i2c_reg_type *)&i2c_reg[bus_num];
            
            /* reset the variable */
            u8BusStatus = 0;
            
            /* get current start tracking count */
            psI2CDrvBusInfo->u16CurStartCount = i2c_buffer[bus_num].u16CurStartCount;
            
            /* get current stop tracking count */
            psI2CDrvBusInfo->u16CurStopCount = i2c_buffer[bus_num].u16CurStopCount;
            
            /* u8BusStatus 
               bit 7 -> reserved
                   6 -> reserved
                   5 -> reserved
                   4 -> reserved
                   3 -> bus error
                   2 -> bus busy
                   1 -> clock status
                   0 -> data status */
            
            /* if i2c bus is busy */
            reg_reading = ioread32((void *) current_reg->cmd_status);
            
            if (reg_reading & BUS_BUSY_STATUS)
            {
                u8BusStatus |= I2C_SET_BUS_BUSY_BIT;
            }
            
            /* check clock status */
            if (reg_reading & SAMPLED_SCL_LINE_STATE)
            {
                u8BusStatus |= I2C_SET_BUS_SCL_HIGH_BIT;
            }           

            /* check data status */
            if (reg_reading & SAMPLED_SDA_LINE_STATE)
            {
                u8BusStatus |= I2C_SET_BUS_SDA_HIGH_BIT;
            }

            psI2CDrvBusInfo->u8BusStatus = u8BusStatus;
            
            break;
        }
        
        case AESS_I2CDRV_CTRL_HW:
        {
            DRV_MSG("ioctl AESS_I2CDRV_CTRL_HW\n");
            
            psI2CDrvBusInfo = (sI2CDrvBusInfoType *) arg;
            
            /* Check the structure address is OK */
            if(!access_ok(VERIFY_WRITE, (void *) psI2CDrvBusInfo, 
                          sizeof(sI2CDrvBusInfoType)))
            {
                DRV_ERROR("ioctl AESS_I2CDRV_CTRL_HW struct access error!\n");
                
                /* error occur */
                return -EFAULT;
            }

			      /* retrieve bus number */
			       bus_num = psI2CDrvBusInfo->u8Channel;

            if (bus_num >= I2C_DRV_MAX_BUS)
                return -EFAULT;
			
			      /* get semaphore */
			      if(down_interruptible((struct semaphore *) &i2c_buffer[bus_num].sem))
			      {
			      	DRV_ERROR("ioctl AESS_I2CDRV_CTRL_HW get semaphore fail!!\n");
				      return -ERESTARTSYS;
			      }
			
			      DRV_MSG("aess_i2c_driver AESS_I2CDRV_CTRL_HW\n");
			      aess_i2c_ctrlhw((sI2CDrvBusInfoType *) psI2CDrvBusInfo);
			
			      /* release semaphore */
			      up((struct semaphore *) &i2c_buffer[bus_num].sem);
			
            break;
        }
        
        default:
            break;
    }
    
    return 0;
}

#ifdef I2C_POLLING_STATUS
/******************************************************************************
*   FUNCTION        :   aess_i2c_polling_func
******************************************************************************/
/**
 *  @brief      The function is used to polling I2C bus status.
 *
 *  @return     None
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   None \n
 *
 *****************************************************************************/
void aess_i2c_polling_func(
                            /** Dummy Data */
                            unsigned long data
                          )
{
    UINT8 bus_num;
    
    DRV_MSG("aess_i2c_polling_func()...\n");
    
    for (bus_num = 0; bus_num < I2C_DRV_MAX_BUS; bus_num++)
    {
        if (i2c_buffer[bus_num].u8HasInit == 1)
        {

            DRV_MSG("-----\n");         
            DRV_MSG("Bus:%d base_add:0x%08x dev_int:0x%08x buf_status:%d\n",
                   bus_num,
                   I2C_GLB_REG,
                   ioread32((void *) I2C_GLB_REG),
                   i2c_buffer[bus_num].u8State);
            
            DRV_MSG("Bus:%d func_ctrl:0x%08x clk1:%08x clk2:0x%08x int_ctrl:0x%08x int_status:0x%08x\n",
                   bus_num,                
                   ioread32((void *) i2c_reg[bus_num].func_ctrl),
                   ioread32((void *) i2c_reg[bus_num].clk_n_ac_timing_1),
                   ioread32((void *) i2c_reg[bus_num].clk_n_ac_timing_2),
                   ioread32((void *) i2c_reg[bus_num].int_ctrl),
                   ioread32((void *) i2c_reg[bus_num].int_status));

            DRV_MSG("Bus:%d cmd_status:0x%08x dev_addr:0x%08x buf_ctrl:0x%08x trans_buf:0x%08x\n",
                   bus_num,
                   ioread32((void *) i2c_reg[bus_num].cmd_status),
                   ioread32((void *) i2c_reg[bus_num].dev_addr),
                   ioread32((void *) i2c_reg[bus_num].buf_ctrl),
                   ioread32((void *) i2c_reg[bus_num].trans_recv_byte_buf));
        }
    }
    
    /* timer interval is 10s */
    i2c_polling.expires = jiffies + 1000;
    add_timer(&i2c_polling);
    
    return;
}
#endif

/******************************************************************************
*   FUNCTION        :   aess_i2c_release
******************************************************************************/
/**
 *  @brief      The API is invoked when user space closed the Linux I2C driver.
 *
 *  @return     always return 0
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   None \n
 *
 *****************************************************************************/
static int aess_i2c_release(
                             /** inode */
                             struct inode *inode,
                             /** File pointer */
                             struct file *filp
                           )
{
    UINT8 bus_num;
    
    DRV_MSG("aess_i2c_release\n");

#ifdef I2C_BUS_RECOVER
        del_timer(&i2c_bus_recover);aess_i2c_ctrlhw
#endif
    
#ifdef I2C_POLLING_STATUS
    del_timer(&i2c_polling);
#endif
    
    for (bus_num = 0; bus_num < I2C_DRV_MAX_BUS; bus_num++)
    {
        if (i2c_buffer[bus_num].u8HasInit == 1)
        {
            /* reset */
            aess_i2c_reset(&i2c_buffer[bus_num]);
            
            if (i2c_buffer[bus_num].u8InitMode == I2C_DRV_MODE_IPMB)
            {
                /* free allocated memory */
                kfree(i2c_buffer[bus_num].psFIFO);
            }
            
            /* clear the initial status flag */
            i2c_buffer[bus_num].u8HasInit = 0;
        }
    }
    
    /* reset to power up state */
    i2c_init_state = 0;
    
    return 0;
}

/******************************************************************************
*   FUNCTION        :   aess_i2c_open
******************************************************************************/
/**
 *  @brief      The function is invoked when user space opens the Linux I2C driver.
 *
 *  @return     always return 0
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   None \n
 *
 *****************************************************************************/
static int aess_i2c_open(
                          /** inode */
                          struct inode *inode,
                          /** File pointer */
                          struct file *filp
                        )
{
    UINT8 bus_num;
    
    DRV_MSG("aess_i2c_open\n");
    
    /* memory allocate */
    
    /* reset i2c modules if reopen */
    if (i2c_init_state == 1)
    {
        DRV_MSG("i2c reopen\n");

#ifdef I2C_BUS_RECOVER
        del_timer(&i2c_bus_recover);
#endif
        
#ifdef I2C_POLLING_STATUS
        del_timer(&i2c_polling);
#endif
        
        for (bus_num = 0; bus_num < I2C_DRV_MAX_BUS; bus_num++)
        {
            if (i2c_buffer[bus_num].u8HasInit == 1)
            {
                /* reset */
                aess_i2c_reset(&i2c_buffer[bus_num]);
                
                if (i2c_buffer[bus_num].u8InitMode == I2C_DRV_MODE_IPMB)
                {
                    /* free allocated memory */
                    kfree(i2c_buffer[bus_num].psFIFO);
                }
                
                /* clear the initial status flag */
                i2c_buffer[bus_num].u8HasInit = 0;
            }
        }

#ifdef I2C_BUS_RECOVER
        /* initiate polling timer, interval is 10s */
        i2c_bus_recover.expires = jiffies + 5000;
        
        add_timer(&i2c_bus_recover);
#endif
        
#ifdef I2C_POLLING_STATUS
        /* initiate polling timer, interval is 10s */
        i2c_polling.expires = jiffies + 1000;
        
        add_timer(&i2c_polling);
#endif
    }
    else
    {
        DRV_MSG("init variables\n");
        for (bus_num = 0; bus_num < I2C_DRV_MAX_BUS; bus_num++)
        {
            /* initiate semaphores */
            init_MUTEX(&i2c_buffer[bus_num].sem);
        }

#ifdef I2C_BUS_RECOVER
        /* initiate polling timer */
        i2c_bus_recover.data = 0;
        i2c_bus_recover.function = &aess_i2c_bus_recover_polling_func;
        
        /* timer interval is 10s */
        i2c_bus_recover.expires = jiffies + 5000;
        
        init_timer(&i2c_bus_recover);
        add_timer(&i2c_bus_recover);
#endif
        
#ifdef I2C_POLLING_STATUS
        /* initiate polling timer */
        i2c_polling.data = 0;
        i2c_polling.function = &aess_i2c_polling_func;
        
        /* timer interval is 10s */
        i2c_polling.expires = jiffies + 1000;
        
        init_timer(&i2c_polling);
        add_timer(&i2c_polling);
#endif
        
        i2c_init_state = 1;
    }
    
    return 0;
}


struct file_operations aess_i2c_fops = {
    .owner = THIS_MODULE,
    .open = aess_i2c_open,
    .release = aess_i2c_release,
    .ioctl = aess_i2c_ioctl,
};

/******************************************************************************
*   FUNCTION        :   aess_i2c_init
******************************************************************************/
/**
 *  @brief      The function is invoked when I2C module is loaded into the kernel.
 *
 *  @return     STATUS_OK if driver initial success.
 *              Others if initial fail.                   
 *
 *  @dependency dynamicirq  Driver
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   None \n
 *
 *****************************************************************************/
int __init aess_i2c_init(void)
{
    int result;
    volatile UINT32 reg_reading;
#ifdef AST2100

    /* Clear reset of LPC and start clocking for it */
    spin_lock_init(&my_lock);
    
    __REG(SCU_WR_PROT_KEY) = SCU_WR_KEY_EN;    /* Write enable key */

#if 0
    DEBUGP(KERN_DEBUG "%s: SCU_WR_PROT_KEY=0x%08lx\n", 
            __FUNCTION__, __REG(SCU_WR_PROT_KEY)); 
    DEBUGP(KERN_DEBUG "%s: SCU_RESET_CTRL=0x%08lx\n", 
            __FUNCTION__, __REG(SCU_RESET_CTRL));
    DEBUGP(KERN_DEBUG "%s: SCU_CLOCK_STOP_CTRL=0x%08lx\n", 
            __FUNCTION__, __REG(SCU_CLOCK_STOP_CTRL));
#endif
    __REG(SCU_RESET_CTRL) &= ~SCU_RESET_I2C; /* Clear reset I2C */
    __REG(SCU_WR_PROT_KEY) = SCU_WR_KEY_DIS;   /* Write disable key */
    
#endif  
    DEBUGP("KN: init_aess_i2c_module\n");
    
    /* allocate a char device */
    i2c_cdev = cdev_alloc();
    
    i2c_cdev->owner = THIS_MODULE;
    i2c_cdev->ops = &aess_i2c_fops;
    
    /* allocate a device number */
    result = alloc_chrdev_region(&i2c_dev, 0, 1, driver_name);
    
    DEBUGP("KN: Major number %d\n", MAJOR(i2c_dev));
    /* show this message for i2c install shell used */
    printk("mknod /dev/aess_i2cdrv c %d 0\n", MAJOR(i2c_dev));

    if (result)
    {
        printk(KERN_ERR "KN: Registering the character device failed with %d\n", 
               MAJOR(i2c_dev));
        
        return result;
    }
    
    /* add a char device */
    cdev_add(i2c_cdev, i2c_dev, 1);
    
#ifdef AST2100  
    /** Set the IRQ to be active high */
    IRQ_SET_HIGH_LEVEL(IRQ_I2C);
    
    reg_reading = ioread32((void *) (ASPEED_VIC_VIRTUAL + 0x24));
    
    iowrite32((UINT32) (reg_reading | 0x00001000),
              (void *) (ASPEED_VIC_VIRTUAL + 0x24) );         
    
#endif
    /* request a IRQ */                             
    result = request_irq(IRQ_I2C,    
                         aess_i2c_isr,            
                         IRQF_SHARED,                
                         "I2C",   
                         (void *) &i2c_buffer[0]);
    
    if (result)
    {
        DEBUGP("KN: %d request IRQ error\n", result);
    }
    
    aess_debugfs_default_create(driver_name,NULL,0,0);
    aess_debugfs_create_file(driver_name,"u32",(u32 *)&u32tmpcounter,DBG_TYPE32,0);
    aess_debugfs_create_file(driver_name,"u16",(u16 *)&u16tmpcounter,DBG_TYPE16,0);
    aess_debugfs_create_file(driver_name,"u8",(u8 *)&u8tmpcounter,DBG_TYPE8,0);
    return 0;
}

/******************************************************************************
*   FUNCTION        :   aess_i2c_exit
******************************************************************************/
/**
 *  @brief      The function is invoked when application close the driver.
 *
 *  @return     None
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   None \n
 *
 *****************************************************************************/
void __exit aess_i2c_exit(void)
{
    /* free the irq */
    free_irq(IRQ_I2C, &i2c_buffer[0]);
    
    /* release the char device */
    cdev_del(i2c_cdev);
    
    /* release the device number */
    unregister_chrdev_region(i2c_dev, 1);
    aess_debugfs_remove(driver_name,NULL,0,0);
    
    return;
}

MODULE_DESCRIPTION("AESS I2C Driver");
MODULE_AUTHOR("AESS");
MODULE_VERSION("0.1.0.0");
MODULE_LICENSE("GPL");
module_param(driver_name, charp, S_IRUGO);
module_init(aess_i2c_init);
module_exit(aess_i2c_exit);

/* End of code */
