/**
 *
 *               Avocent Corporation. Proprietary Information.
 *      This software is supplied under the terms of a license agreement or
 *      nondisclosure agreement with Avocent Corporation, or one of its affiliates, and
 *      may not be used, copied, disseminated, or distributed except in accordance
 *      with the terms of that agreement.
 *
 *      3541 Gateway Boulevard, Fremont, California, 94538 U.S.A.
 *                  US Phone: 510-771-6100
 *      Copyright 2001-2009 Avocent Corporation. All rights reserved.
 *
 *----------------------------------------------------------------------------
 *  MODULES     AESS Hardware Watchdog Driver
 *----------------------------------------------------------------------------
 * @file    aspeed_gucwdt.c
 * @brief   c source file for hardware watchdog driver
 *----------------------------------------------------------------------------
 */
/******************************************************************************
* Content
* ----------------
*   wdt_isr()			-   watchdog ISR function
*   wdt_disable()		-   disable watchdog timer
*   wdt_sel_clk_src()		-   set clock source
*   wdt_set_timeout_action()	-   set timeout action
*   wdt_enable()		-   enable watchdog timer
*   wdt_restart_new()		-   restart watchdog timer with new clock source and timer
*   wdt_restart()		-   restart watchdog timer
*   aess_wdt_keepalive()	-   restart timer
*   wdt_set_heartbeat()		-   set a new heartbeat value
*   umvp2500_wdt_write()	-   send keep alive signal
*   umvp2500_wdt_ioctl()	-   perform IO control functions
*   umvp2500_wdt_open()		-   open wdt
*   umvp2500_wdt_release()	-   release wdt
*   umvp2500_wdt_notify_sys()	-   called on system shutdowns, disable wdt
*   wdt_exit()			-   module exit function
*   wdt_init()			-   module init function
******************************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <linux/interrupt.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include <linux/fs.h>
#include <linux/notifier.h>
#include <linux/reboot.h>
#include <linux/init.h>
#include <asm/hardware.h>
#include <asm/semaphore.h>
#include <asm/uaccess.h>
#include <asm/arch/platform.h>
#include <asm/arch/vic.h>
#include <asm/arch-ast2050/hw_regs.h>

/* implement debug message with debugfs */
#include <linux/aess_debugfs.h>

typedef unsigned char bool_T;

#ifdef TRUE
#undef TRUE
#endif

#ifdef FALSE
#undef FALSE
#endif

#define TRUE	1
#define FALSE	0


#define UMVP_READ_REG(r)		(*((volatile unsigned int *) (r)))
#define UMVP_WRITE_REG(r,v)		(*((volatile unsigned int *) (r)) = ((unsigned int)   (v)))


#define WDT_CLK_SRC_EXT		0
#define WDT_CLK_SRC_PCLK	1

//Global Variables

/* Default heartbeat = 10 seconds */
#define WD_TIMO 10

static int heartbeat = WD_TIMO;
module_param(heartbeat, int, 0);
MODULE_PARM_DESC(heartbeat, "Watchdog heartbeat in seconds. (0<heartbeat<65536, default=" __MODULE_STRING(WD_TIMO) ")");

static int nowayout = WATCHDOG_NOWAYOUT;
module_param(nowayout, int, 0);
MODULE_PARM_DESC(nowayout, "Watchdog cannot be stopped once started (default=CONFIG_WATCHDOG_NOWAYOUT)");

static char *driver_name="aess_wdtdrv";
static u8   u8tmpcounter = 0;
static u16 u16tmpcounter = 0;
static u32 u32tmpcounter = 0;

static unsigned long wdt_is_open;
static char expect_close;

/** Function Declaration */
int __init wdt_init(void);

/******************************************************************************
 *       FUNCTION        :       wdt_isr
 ******************************************************************************/
/**
 *      @brief          watchdog ISR function
 *
 *      @return         0 if success
 *
 *      @dependency	None
 *
 *      @limitation     None
 *
 *      @warning        None
 *
 *      @note           None
 *
 *      @internal       Function Type: Public function
 *
 *****************************************************************************/
static irqreturn_t wdt_isr(
			    /** irq number */
			    int irq,
			    /** device id */
			    void *devid
			  )
{
	DRV_MSG("Enter %s.\n", __FUNCTION__);

	/** clear timeout */
	UMVP_WRITE_REG(WDT_Clr, 1);

	vic_clear_intr(irq);

	return (IRQ_HANDLED);
}

/******************************************************************************
 *       FUNCTION        :       wdt_disable
 ******************************************************************************/
/**
 *      @brief          disable watchdog timer
 *
 *      @return         None.
 *
 *      @dependency	None
 *
 *      @limitation     None
 *
 *      @warning        None
 *
 *      @note           None
 *
 *      @internal       Function Type: Private function
 *
 *****************************************************************************/
void wdt_disable(void)
{
    register unsigned int regVal;
                                                                                     
    /* reset WDT_Ctrl[0] as 0 */
    regVal = UMVP_READ_REG(WDT_Ctrl);
    regVal &= 0xFFFFFFFE;
    UMVP_WRITE_REG(WDT_Ctrl, regVal);
}

/******************************************************************************
 *       FUNCTION        :       wdt_sel_clk_src
 ******************************************************************************/
/**
 *      @brief          set clock source
 *
 *      @return         None
 *
 *      @dependency	None
 *
 *      @limitation     None
 *
 *      @warning        None
 *
 *      @note           None
 *
 *      @internal       Function Type: Private function
 *
 *****************************************************************************/
void wdt_sel_clk_src(
			/** clock source register value */
			unsigned char sourceClk
		    )
{
    register unsigned int regVal;
                                                                                     
    regVal = UMVP_READ_REG(WDT_Ctrl);
    if (sourceClk == WDT_CLK_SRC_PCLK)
    {
        /* reset WDT_Ctrl[4] as 0 */
        regVal &= 0xFFFFFFEF;
    }
    else
    {
        /* set WDT_Ctrl[4] as 1 */
        regVal |= 0x00000010;
    }
    UMVP_WRITE_REG(WDT_Ctrl, regVal);
}

/******************************************************************************
 *       FUNCTION        :       wdt_set_timeout_action
 ******************************************************************************/
/**
 *      @brief          set timeout action
 *
 *      @return         None
 *
 *      @dependency	None
 *
 *      @limitation     None
 *
 *      @warning        None
 *
 *      @note           None
 *
 *      @internal       Function Type: Private function
 *
 *****************************************************************************/
void wdt_set_timeout_action(
			    /** external signal enable after timeout */
			    bool_T bResetOut,
			    /** interrupt enable after timeout */
			    bool_T bIntrSys,
			    /** reset system after timeout */
			    bool_T bResetSys
			   )
{
	register unsigned int regVal;

	regVal = UMVP_READ_REG(WDT_Ctrl);

	if (bResetOut)
	{
		/* set WDT_Ctrl[3] = 1 */
		regVal |= 0x00000008;
	}
	else
	{
		/* reset WDT_Ctrl[3] = 0 */
		regVal &= 0xFFFFFFF7;
	}

	if (bIntrSys)
	{
		/* set WDT_Ctrl[2] = 1 */
		regVal |= 0x00000004;
	}
	else
	{
		/* reset WDT_Ctrl[2] = 0 */
		regVal &= 0xFFFFFFFB;
	}

	if (bResetSys)
	{
		/* set WDT_Ctrl[1] = 1 */
		regVal |= 0x00000002;
	}
	else
	{
		/* reset WDT_Ctrl[1] = 0 */
		regVal &= 0xFFFFFFFD;
	}

	UMVP_WRITE_REG(WDT_Ctrl, regVal);
}

/******************************************************************************
 *       FUNCTION        :       wdt_enable
 ******************************************************************************/
/**
 *      @brief          enable watchdog timer
 *
 *      @return         None
 *
 *      @dependency	None
 *
 *      @limitation     None
 *
 *      @warning        None
 *
 *      @note           None
 *
 *      @internal       Function Type: Private function
 *
 *****************************************************************************/
void wdt_enable(void)
{
	register unsigned int regVal;

	/* set WDT_Ctrl[0] as 1 */
	regVal = UMVP_READ_REG(WDT_Ctrl);
	regVal |= 1;
	UMVP_WRITE_REG(WDT_Ctrl, regVal);
}

/******************************************************************************
 *       FUNCTION        :       wdt_restart_new
 ******************************************************************************/
/**
 *      @brief          restart watchdog timer with new clock source and timer
 *
 *      @return         None
 *
 *      @dependency	None
 *
 *      @limitation     None
 *
 *      @warning        None
 *
 *      @note           None
 *
 *      @internal       Function Type: Private function
 *
 *****************************************************************************/
void wdt_restart_new(
			/** counter reload value */
			unsigned int nPeriod,
			/** select clock source */
			int sourceClk,
			/** external signal enable after timeout*/
			bool_T bResetOut,
			/** interrupt enable after timeout */
			bool_T bIntrSys,
			/** reset system after timeout */
			bool_T bResetSys,
			/** enable wdt */
			bool_T bUpdated
		    )
{
    DRV_INFO("Enter %s.\n", __FUNCTION__);

    wdt_disable();

    UMVP_WRITE_REG(WDT_Reload, nPeriod);

    wdt_sel_clk_src(sourceClk);

    wdt_set_timeout_action(bResetOut, bIntrSys, bResetSys);

    UMVP_WRITE_REG(WDT_Restart, 0x4755);	/* reload! */

    if (!bUpdated)
        wdt_enable();
}

/******************************************************************************
 *       FUNCTION        :       wdt_restart
 ******************************************************************************/
/**
 *      @brief          restart watchdog timer
 *
 *      @return         None
 *
 *      @dependency	None
 *
 *      @limitation     None
 *
 *      @warning        None
 *
 *      @note           None
 *
 *      @internal       Function Type: Private function
 *
 *****************************************************************************/
void wdt_restart(void)
{
    DRV_INFO("Enter %s.\n", __FUNCTION__);

    wdt_disable();
    UMVP_WRITE_REG(WDT_Restart, 0x4755);	/* reload! */
    wdt_enable();
}

/******************************************************************************
 *       FUNCTION        :       aess_wdt_keepalive
 ******************************************************************************/
/**
 *      @brief          restart timer
 *
 *      @return         0 if success
 *
 *      @dependency	None
 *
 *      @limitation     None
 *
 *      @warning        None
 *
 *      @note           None
 *
 *      @internal       Function Type: Public function
 *
 *****************************************************************************/
int aess_wdt_keepalive(
				/** file handle to the watchdog */
				struct file *file
			     )
{
    DRV_INFO("Enter %s.\n", __FUNCTION__);

    wdt_restart();

    return 0;
}

/******************************************************************************
 *       FUNCTION        :       wdt_set_heartbeat
 ******************************************************************************/
/**
 *      @brief          Set a new heartbeat value for the watchdog device
 *
 *      @return         0 if success, -EINVAL otherwise
 *
 *      @dependency	None
 *
 *      @limitation     None
 *
 *      @warning        None
 *
 *      @note           None
 *
 *      @internal       Function Type: Private function
 *
 *****************************************************************************/
static int wdt_set_heartbeat(
				/** new heartbeat value */
				int t
			    )
{
    if ((t < 1) || (t > 1000))
        return -EINVAL;
      
    DRV_INFO("%s: Set heartbeat to %d.\n", __FUNCTION__, t);

    heartbeat=t;
      
    wdt_restart_new(TICKS_PER_uSEC*1000000*t, WDT_CLK_SRC_EXT, FALSE, TRUE, FALSE, FALSE);

    return 0;
}

/******************************************************************************
 *       FUNCTION        :       umvp2500_wdt_write
 ******************************************************************************/
/**
 *      @brief          A write to a watchdog device is defined as a keepalive
 *			signal. Any write of data will do, as we we don't define
 *			content meaning.
 *
 *      @return         0 if success
 *
 *      @dependency	None
 *
 *      @limitation     None
 *
 *      @warning        None
 *
 *      @note           None
 *
 *      @internal       Function Type: Public function
 *
 *****************************************************************************/
static ssize_t umvp2500_wdt_write(
				    /** file handle to the watchdog */
				    struct file *file,
				    /** buffer to write (unused) */
				    const char __user *buf,
				    /** count of bytes */
				    size_t count,
				    /** pointer to the position to write */
				    loff_t *ppos
				 )
{
    DRV_INFO("Enter %s.\n", __FUNCTION__);

    if(count) 
    {
        if (!nowayout)
        {
            size_t i;
             
            /* In case it was set long ago */
            expect_close = 0;
                         
            for (i = 0; i != count; i++) 
            {
                char c;
                if (get_user(c, buf + i))
                    return -EFAULT;
                if (c == 'V')
                    expect_close = 42;
            }
        }
        wdt_restart();
    }

    return count; 
}

/******************************************************************************
 *       FUNCTION        :       umvp2500_wdt_ioctl
 ******************************************************************************/
/**
 *      @brief          defines IO functions for watchdogs driver
 *
 *      @return         0 if success
 *
 *      @dependency	None
 *
 *      @limitation     None
 *
 *      @warning        None
 *
 *      @note           None
 *
 *      @internal       Function Type: Public function
 *
 *****************************************************************************/
static int umvp2500_wdt_ioctl(
				/** inode of the device */
				struct inode *inode,
				/** file handle to the device */
				struct file *file,
				/** watchdog command */
				unsigned int cmd,
				/** argument pointer */
				unsigned long arg
			     )
{
    void __user *argp = (void __user *)arg;
    int __user *p = argp;
    int new_heartbeat;
      
    static struct watchdog_info ident = 
    {
        .options	= WDIOF_SETTIMEOUT|
                          WDIOF_MAGICCLOSE|
                          WDIOF_KEEPALIVEPING,
        .firmware_version = 1,
        .identity 	= "UMVP2500 WDT",
    };
  
    switch(cmd)
    {
        default:
            return -ENOIOCTLCMD;
        case WDIOC_GETSUPPORT:
            return copy_to_user(argp, &ident, sizeof(ident))?-EFAULT:0;                      
        case WDIOC_GETSTATUS:
        case WDIOC_GETBOOTSTATUS:
            return put_user(0, p);
        case WDIOC_KEEPALIVE:
            wdt_restart();
            return 0;
        case WDIOC_SETTIMEOUT:
            if (get_user(new_heartbeat, p))
                return -EFAULT;

        if (wdt_set_heartbeat(new_heartbeat))
            return -EINVAL;

        /* Fall */
        case WDIOC_GETTIMEOUT:
            return put_user(heartbeat, p);      
    }
}

/******************************************************************************
 *       FUNCTION        :       umvp2500_wdt_open
 ******************************************************************************/
/**
 *      @brief          The watchdog device has been opened. The watchdog
 *			device is single open and on opening we load the counters.
 *			Counter zero is a 100Hz cascade, into counter 1 which
 *			downcounts to reboot. When the counter triggers counter 2
 *			downcounts the length of the reset pulse which set
 *			to be as long as possible.
 *
 *      @return         0 if success
 *
 *      @dependency	None
 *
 *      @limitation     None
 *
 *      @warning        None
 *
 *      @note           None
 *
 *      @internal       Function Type: Public function
 *
 *****************************************************************************/
static int umvp2500_wdt_open(
				/** inode of device */
				struct inode *inode,
				/** file handle to device */
				struct file *file
			    )
{
    if(test_and_set_bit(0, &wdt_is_open))
        return -EBUSY;
    /*
     *	Activate
     */
    wdt_restart();

    return nonseekable_open(inode, file);
} 

/******************************************************************************
 *       FUNCTION        :       umvp2500_wdt_release
 ******************************************************************************/
/**
 *      @brief          The watchdog has a configurable API. iThere is a religious
 *			dispute between people who want their watchdog to be able
 *			to shut down and those who want to be sure if the watchdog
 *			manager dies the machine reboots. In the former case we
 *			disable the counters, in the latter case you have to
 *			open it again very soon.
 *
 *      @return         0 if success
 *
 *      @dependency	None
 *
 *      @limitation     None
 *
 *      @warning        None
 *
 *      @note           None
 *
 *      @internal       Function Type: Public function
 *
 *****************************************************************************/
static int umvp2500_wdt_release(
				/** inode to board */
				struct inode *inode,
				/** file handle to board */
				struct file *file
			       )
{
    if (expect_close == 42 || !nowayout) 
    {
        wdt_disable();
        clear_bit(0, &wdt_is_open);
    } 
    else 
    {
        printk(KERN_CRIT "wdt: WDT device closed unexpectedly.  WDT will not stop!\n");
        wdt_restart();
    }
    expect_close = 0;

    return 0;
}

/******************************************************************************
 *       FUNCTION        :       umvp2500_wdt_notify_sys
 ******************************************************************************/
/**
 *      @brief          Our notifier is called on system shutdowns. We want to turn
 *			the card off at reboot otherwise the machine will during
 *			memory test or worse yet during the following fsck.
 *
 *      @return         0 if success
 *
 *      @dependency	None
 *
 *      @limitation     None
 *
 *      @warning        None
 *
 *      @note           None
 *
 *      @internal       Function Type: Public function
 *
 *****************************************************************************/
static int umvp2500_wdt_notify_sys(
				    /** our notifier block */
				    struct notifier_block *this,
				    /** the event being reported */
				    unsigned long code,
				    /** unused */
				    void *unused
				  )
{
    if(code==SYS_DOWN || code==SYS_HALT) 
    {
        /* Turn the WDT off */
        wdt_disable();
    }

    return NOTIFY_DONE;
}

static struct file_operations umvp2500_wdt_fops = 
{
  .owner	= THIS_MODULE,
  .llseek	= no_llseek,
  .write	= umvp2500_wdt_write,
  .ioctl	= umvp2500_wdt_ioctl,
  .open		= umvp2500_wdt_open,
  .release	= umvp2500_wdt_release,
};

static struct miscdevice umvp2500_wdt_miscdev = 
{
   .minor	= WATCHDOG_MINOR,
   .name	= "aess_wdtdrv",
   .fops	= &umvp2500_wdt_fops,
};
     
static struct notifier_block umvp2500_wdt_notifier = 
{
   .notifier_call=umvp2500_wdt_notify_sys,
};

/******************************************************************************
 *       FUNCTION        :       wdt_exit
 ******************************************************************************/
/**
 *      @brief          module exit function
 *
 *      @return         None
 *
 *      @dependency	None
 *
 *      @limitation     None
 *
 *      @warning        None
 *
 *      @note           None
 *
 *      @internal       Function Type: Public function
 *
 *****************************************************************************/
void __exit wdt_exit(void)
{
    misc_deregister(&umvp2500_wdt_miscdev);
    disable_irq(INT_WDT);
    free_irq(INT_WDT, NULL);

    /* remove debugfs module */
    aess_debugfs_remove(driver_name, NULL, 0, 0);
}

/******************************************************************************
 *       FUNCTION        :       wdt_init
 ******************************************************************************/
/**
 *      @brief          module init function
 *
 *      @return         0 if success
 *
 *      @dependency	None
 *
 *      @limitation     None
 *
 *      @warning        None
 *
 *      @note           None
 *
 *      @internal       Function Type: Public function
 *
 *****************************************************************************/
int __init wdt_init(void)
{
    int ret;  
    
    /* set INT_WDT as "raising_edge trigger, irq mode" sensitivity */
    vic_set_intr_trigger(INT_WDT, vicc_edge_activeRaising);


    wdt_disable();
    wdt_sel_clk_src(WDT_CLK_SRC_EXT);
    wdt_set_timeout_action(FALSE, FALSE, FALSE);
	

    /* register ISR */
    if (request_irq(INT_WDT, wdt_isr, IRQF_DISABLED, "WDT", NULL))
    {
        printk("unable to register interrupt INT_WDT = %d\n", INT_WDT);
        return (-1);
    }
    else
        printk("success to register interrupt for INT_WDT (%d)\n", INT_WDT);
     
    ret = register_reboot_notifier(&umvp2500_wdt_notifier);
    if(ret) 
    {
        printk(KERN_ERR "wdt: cannot register reboot notifier (err=%d)\n", ret);
        free_irq(INT_WDT, NULL);
        return ret;
    }
    ret = misc_register(&umvp2500_wdt_miscdev);
    if (ret) 
    {
        printk(KERN_ERR "wdt: cannot register miscdev on minor=%d (err=%d)\n",WATCHDOG_MINOR, ret);
        unregister_reboot_notifier(&umvp2500_wdt_notifier);   
        return ret;
    }

    disable_irq(INT_WDT);

    wdt_restart_new(TICKS_PER_uSEC*1000000*heartbeat, WDT_CLK_SRC_EXT, FALSE, TRUE, TRUE, TRUE);

    /* enable INT */
    enable_irq(INT_WDT);
   
    printk(KERN_INFO "ASPEED WDT is installed.(irq = %d, heartbeat = %d secs, nowayout = %d)\n",INT_WDT,heartbeat,nowayout);
   
    /* add this for script to create device node */
    printk("mknod /dev/aess_wdtdrv c 10 %d\n", WATCHDOG_MINOR);

    /* initial debugfs related resource */
    aess_debugfs_default_create(driver_name, NULL, 0, 0);
    aess_debugfs_create_file(driver_name, "u32", (u32 *)&u32tmpcounter, DBG_TYPE32, 0);
    aess_debugfs_create_file(driver_name, "u16", (u16 *)&u16tmpcounter, DBG_TYPE16, 0);
    aess_debugfs_create_file(driver_name, "u8", (u8 *)&u8tmpcounter, DBG_TYPE8, 0);

    return (0);
}

module_init(wdt_init);
module_exit(wdt_exit);
MODULE_ALIAS_MISCDEV(WATCHDOG_MINOR);
MODULE_AUTHOR("Andy Chang");
MODULE_DESCRIPTION("AESS Hardware Watchdog Driver");
MODULE_VERSION("0.1.0.1");
MODULE_LICENSE("GPL");

/* End of code */

