/*
 *	An implementation of the NCSI protocol handler.
 *	Bill Liu <bill.liu@avocent.com>
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *
 */

//#include <linux/config.h>
#include <linux/module.h>

#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/in.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/if_ether.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <linux/route.h>
#include <linux/inet.h>
#include <linux/etherdevice.h>
#include <linux/if_arp.h>
#include <linux/wireless.h>
#include <linux/skbuff.h>
#include <linux/udp.h>
#include <net/sock.h>
#include <net/inet_common.h>
#include <linux/stat.h>
#include <linux/init.h>
#include <linux/if_ec.h>
#include <net/udp.h>
#include <net/ip.h>
#include <linux/spinlock.h>
#include <linux/rcupdate.h>
#include <linux/bitops.h>
#include <linux/mii.h>


#include <asm/uaccess.h>
#include <asm/system.h>

#include <linux/delay.h> 
#include "ncsi_protocol.h"

#include "../arm/aess_ftgmac100.h"


/* Normally the address family number is defined in include/linux/socket.h
   However, to get everything squeezed into one file without touching too
   many kernel files, we define it here */
#define AF_NCSI		27	/* NCSI Address Family */
#define PF_NCSI		AF_NCSI
#define NCSI_PROTOCOL 0x88F8
#define MAX_NCSI_QUEUE_DEPTH 20
#define MAX_NCSI_SOCKET_NUM  4
#define MAX_OPEN_SOCKET_NUMBER 4                  
static int S_Socket_Num = 0;
static sNCSI_IF_Info ncsi_if;

static int S_isNCSIInited = 0;

/* Error Type Define*/
typedef enum
{
	/* No error */
	STATUS_OK = 0,

	/* Common error */
	STATUS_FAIL,

	/* Too many NCSI sockets being opened */
	STATUS_TOOMANYSOCKET,

	/* Invalid port number specified for BMC MAC */
	STATUS_INVALID_PORT,

	/* The specified network device is not registered with the kernel */
	STATUS_DEVICE_NO_EXIST,

	/* The specified network device is registered, but currently down */
	STATUS_DEVICE_DOWN,
	 
	/* The NCSI queue is currently empty, no data for user app */ 
	STATUS_NO_DATA
	
} eReturnErrorType;

static struct sk_buff_head ncsi_queue;
static wait_queue_head_t inq;
static const struct proto_ops ncsi_ops;
static atomic_t ncsi_queued_depth = ATOMIC_INIT(0);
static struct net_device *ncsi_dev[2] = {NULL, NULL};



static int ncsi_cmd_clear_initial_state(sNCSI_IF_Info *ncsi_if);
static int ncsi_cmd_select_package(sNCSI_IF_Info *ncsi_if);
static int ncsi_cmd_deselect_package(sNCSI_IF_Info *ncsi_if);
static int ncsi_cmd_enable_channel(sNCSI_IF_Info *ncsi_if);
static int ncsi_cmd_disable_channel(sNCSI_IF_Info *ncsi_if);
static int ncsi_cmd_reset_channel(sNCSI_IF_Info *ncsi_if);
int ncsi_cmd_enable_channel_network_tx(sNCSI_IF_Info *ncsi_if);
int ncsi_cmd_disable_channel_network_tx(sNCSI_IF_Info *ncsi_if);
static int ncsi_cmd_aen_enable(sNCSI_IF_Info *ncsi_if);
int ncsi_cmd_set_link(sNCSI_IF_Info *ncsi_if);
int ncsi_cmd_get_link_status(sNCSI_IF_Info *ncsi_if);
static int ncsi_cmd_set_vlan_filters(sNCSI_IF_Info *ncsi_if);
static int ncsi_cmd_enable_vlan(sNCSI_IF_Info *ncsi_if);
static int ncsi_cmd_disable_vlan(sNCSI_IF_Info *ncsi_if);
static int ncsi_cmd_set_mac_address(sNCSI_IF_Info *ncsi_if);
static int ncsi_cmd_enable_broadcast_filtering(sNCSI_IF_Info *ncsi_if);
static int ncsi_cmd_disable_broadcast_filtering(sNCSI_IF_Info *ncsi_if);
static int ncsi_cmd_enable_global_multicast_filtering(sNCSI_IF_Info *ncsi_if);
static int ncsi_cmd_disable_global_multicast_filtering (sNCSI_IF_Info *ncsi_if);
static int ncsi_cmd_set_ncsi_flow_control(sNCSI_IF_Info *ncsi_if);
int ncsi_cmd_get_version_id(sNCSI_IF_Info *ncsi_if);
static int ncsi_cmd_get_capabilities(sNCSI_IF_Info *ncsi_if);
static int ncsi_cmd_get_parameters(sNCSI_IF_Info *ncsi_if);
static int ncsi_cmd_get_controller_packet_statistics(sNCSI_IF_Info *ncsi_if);
static int ncsi_cmd_get_ncsi_statistics(sNCSI_IF_Info *ncsi_if);
int ncsi_cmd_get_ncsi_passthrough_statistics(sNCSI_IF_Info *ncsi_if);

static int ncsi_aen_link_status_change(sNCSI_IF_Info *ncsi_if, struct sNCSI_AEN_HDR *ncsi_aenhdr);
static int ncsi_aen_reconfig_required(sNCSI_IF_Info *ncsi_if, struct sNCSI_AEN_HDR *ncsi_aenhdr);
static int ncsi_aen_os_driver_change(sNCSI_IF_Info *ncsi_if, struct sNCSI_AEN_HDR *ncsi_aenhdr);


static int ncsi_queue_read(sNCSI_IF_Info *ncsi_if, struct sk_buff **res_skb, 
						   enum ncsi_queue_type type);
static int ncsi_queue_write(sNCSI_IF_Info *ncsi_if, struct sk_buff *skb,
							enum ncsi_queue_type type);
static void ncsi_rx_aen_packet(sNCSI_IF_Info *ncsi_if, struct sk_buff *aen_skb);
static void ncsi_rx_response_packet(sNCSI_IF_Info *ncsi_if, struct sk_buff *res_skb);
void ncsi_rx_packets_dispatcher(sNCSI_IF_Info *ncsi_if, struct sk_buff *skb);
static int ncsi_wait_response_packet(sNCSI_IF_Info *ncsi_if, 
									 struct sk_buff **res_skb, u8 res_cmd);
static void ncsi_eth_header(sNCSI_IF_Info *ncsi_if, struct sk_buff *skb);
static void ncsi_header_payload(sNCSI_IF_Info *ncsi_if, struct sk_buff *req_skb, 
								u8 cmd, u8 *payload, u16 payload_len);
static int ncsi_send_packet(sNCSI_IF_Info *ncsi_if,
							u8 cmd, u8 *payload, u16 payload_len);
static int ncsi_send_and_wait(sNCSI_IF_Info *ncsi_if, struct sk_buff **res_skb, 
							  u8 cmd, u8 *payload, u16 payload_len);
static int ncsi_aen_thread(void *arg);
#define NCSI_ETH_HEADER(skb) ((struct ethhdr *)((skb)->data))
#define NCSI_HEADER(skb) ((struct sNCSI_HDR *)((skb)->data + ETH_HLEN))

/* the response packet is already removed the Ethernet header */
#define NCSI_RES_HEADER(skb) ((struct sNCSI_RES_HDR *)((skb)->data))
#define NCSI_AEN_HEADER(skb) ((struct sNCSI_AEN_HDR *)((skb)->data))


#define IS_NCSI_PACKET(ether_type)	\
	(NCSI_ETHER_TYPE == ntohs(ether_type))

#define IS_NCSI_GOOD_PACKET(ncsi_hdr)	\
	(NCSI_HEADER_REVISION == ncsi_hdr->hdr_rev)

#define IS_NCSI_AEN_PACKET(ncsi_hdr)	\
	((ncsi_hdr->cmd_iid == NCSI_AEN_PACKET_IID) && \
	(ncsi_hdr->cmd == 0xFF))

#define IS_NCSI_RES_PACKET(ncsi_hdr)	\
	(((ncsi_hdr->cmd) & NCSI_RES_PACKET_MASK) && \
    (ncsi_hdr->cmd != 0xFF))

#define IS_VALID_NCSI_RES_PACKET(ncsi_reshdr, ncsi_if, res_cmd)					\
	(((ncsi_reshdr->ncsi_hdr.chnl_id == ncsi_if->chnl_id)) &&		\
	(ncsi_reshdr->ncsi_hdr.cmd == res_cmd))// &&					\
//mark by sonata: follow Bill's comment	(ncsi_reshdr->ncsi_hdr.cmd_iid == ncsi_if->cmd_iid))	
	

#define NCSI_RECEIVE_RES_PACKET_OK	0
#define NCSI_RECEIVE_RES_PACKET_INVALID	1
#define NCSI_NEED_REINIT             0xFE
#define PERIODIC_NIC_STATE_POLL_SEC	 15
#define BCM5709C

//#define NCSI_DEBUG
#ifdef NCSI_DEBUG
	#define PDEBUG(fmt, args...) printk(KERN_WARNING fmt, ## args)
#else
	#define PDEBUG(fmt, args...) /* not debugging: nothing */
#endif

extern void ResetMACRxSimple(struct net_device * dev);

#if 0
struct  wpcm450_priv
{
	struct net_device_stats stats;	
	unsigned long which;
	unsigned long rx_mode;
	unsigned long ncsi_nic;
	volatile unsigned long cur_tx_entry;
	volatile unsigned long cur_rx_entry;
	volatile unsigned long is_rx_all;
	
	unsigned long bInit;
	unsigned long rx_packets;
	unsigned long rx_bytes;
	unsigned long start_time;
	spinlock_t 		lock;
	struct sk_buff *skb;
	
#ifdef AUTO_SENSE
	struct 	timer_list timer;
	char	plugout;
#endif
	
	volatile unsigned long tx_ptr;
	unsigned long tx_finish_ptr;
	volatile unsigned long rx_ptr;
	
	unsigned long start_tx_ptr;
	unsigned long start_tx_buf;
	
	
	unsigned long mcmdr;
	volatile unsigned long start_rx_ptr;
	volatile unsigned long start_rx_buf;
    struct mii_if_info mii_if;
	char 		      mac_address[ETH_ALEN];
	volatile  	  RXBD   rx_desc[RX_DESC_SIZE]  __attribute__ ((aligned (16)));
	volatile      TXBD   tx_desc[TX_DESC_SIZE]	__attribute__ ((aligned (16)));
	volatile char rx_buf[RX_DESC_SIZE][PACKET_BUFFER_SIZE]	__attribute__ ((aligned (16)));
	volatile char tx_buf[TX_DESC_SIZE][PACKET_BUFFER_SIZE]	__attribute__ ((aligned (16)));
  
    
//	sNCSI_IF_Info ncsi_if;

};
#endif

static char *ncsi_cmd_str[] = 
{
	"CMD_CLEAR_INITIAL_STATE",
	"CMD_SELECT_PACKAGE",
	"CMD_DESELECT_PACKAGE",
	"CMD_ENABLE_CHANNEL",
	"CMD_DISABLE_CHANNEL",
	"CMD_RESET_CHANNEL",
	"CMD_ENABLE_CHANNEL_NETWORK_TX",
	"CMD_DISABLE_CHANNEL_NETWORK_TX",
	"CMD_AEN_ENABLE",
	"CMD_SET_LINK",
	"CMD_GET_LINK_STATUS",
	"CMD_SET_VLAN_FILTERS",
	"CMD_ENABLE_VLAN",
	"CMD_DISABLE_VLAN",
	"CMD_SET_MAC_ADDRESS",
	"CMD_RESERVED_0xF",
	"CMD_ENABLE_BROADCAST_FILTERING",
	"CMD_DISABLE_BROADCAST_FILTERING",
	"CMD_ENABLE_GLOBAL_MULTICAST_FILTERING",
	"CMD_DISABLE_GLOBAL_MULTICAST_FILTERING ",
	"CMD_SET_NCSI_FLOW_CONTROL",
	"CMD_GET_VERSION_ID",
	"CMD_GET_CAPABILITIES",
	"CMD_GET_PARAMETERS",
	"CMD_GET_CONTROLLER_PACKET_STATISTICS",
	"CMD_GET_NCSI_STATISTICS",
	"CMD_GET_NCSI_PASSTHROUGH_STATISTICS",
};


int ncsi_reinit_channel(sNCSI_IF_Info *ncsi_if)
{
	int err = 0;
	
    /* 1. Discover and Get Capabilities for each Channel in the Package */
	err = ncsi_cmd_clear_initial_state(ncsi_if);
	if (err != 0)
	{
		PDEBUG("NCSI Network Controller Fail\n");
		return err;
    }
	
	err = ncsi_cmd_get_capabilities(ncsi_if);
	if (err != 0)
	{
		PDEBUG("NCSI Network Controller Fail\n");
		return err;
    }  

	/* 2. Initialize each Channel in the Package */
	/* TBD: support only one MAC addr filter now. Maybe can expand */
    memcpy((u8 *)ncsi_if->ncsi_mac_info.mac_addr, (u8 *)ncsi_if->dev->dev_addr, 6);
	ncsi_if->ncsi_mac_info.mac_num = 1;
	ncsi_if->ncsi_mac_info.addr_type_enable = CMD_SET_MAC_ADDRESS_PAYLOAD_ADDR_TYPE_UNICAST | CMD_SET_MAC_ADDRESS_PAYLOAD_ENABLE_MAC_ADDR_FILTER;
	err = ncsi_cmd_set_mac_address(ncsi_if);
	if (err != 0)
	{
		PDEBUG("NCSI Network Controller Fail\n");
		return err;
    }  
    
	ncsi_if->ncsi_chnl_parameters.bcast_filter_flag = ncsi_if->ncsi_chnl_caps.bcast_filter_flag;
	err = ncsi_cmd_enable_broadcast_filtering(ncsi_if);
	if (err != 0)
	{
		PDEBUG("NCSI Network Controller Fail\n");
		return err;
  }  
	
	/* NCSI AEN control */
	if(ncsi_if->ncsi_chnl_caps.aen_ctrl != 0)
	{
    ncsi_if->ncsi_aen_ctrl = ncsi_if->ncsi_chnl_caps.aen_ctrl;
		/* Enable AEN */
		err = ncsi_cmd_aen_enable(ncsi_if);
		if (err != 0)
		{
			PDEBUG("NCSI Network Controller Fail\n");
			return err;
    }  
	}
	
	
	//If ncsi_reinit_channel is used to test opening multi-channels with the same address
	//,then this command must be marked 
  err = ncsi_cmd_enable_channel_network_tx(ncsi_if);
  if (err != 0)
	{
		PDEBUG("NCSI Network Controller Fail\n");
		return err;
  }  
	
	/* 3. Start Pass-through packet and AEN operation on the channels */
	err = ncsi_cmd_enable_channel(ncsi_if);
    if (err != 0)
	{
		PDEBUG("NCSI Network Controller Fail\n");
		return err;
    }  
    
	return err;
}

int ncsi_cmd_error_handler(sNCSI_IF_Info *ncsi_if,struct sNCSI_RES_HDR *ncsi_reshdr)
{
	ncsi_reshdr->response_code = ntohs(ncsi_reshdr->response_code);
	ncsi_reshdr->reason_code = ntohs(ncsi_reshdr->reason_code);
	
	if(ncsi_reshdr->response_code != NCSI_RES_CODE_CMD_OK)
	{
	    if(ncsi_reshdr->reason_code == NCSI_REASON_CODE_INTERFACE_INIT_REQUIRED)
	    {   
		    ncsi_if->error_recovery_enable_flag = 1;
		    PDEBUG("NCSI_REASON_CODE_INTERFACE_INIT_REQUIRED \n");
		    return(NCSI_NEED_REINIT);
      }
        
        PDEBUG("NCSI COMMAND ERROR (%x, %x)\n",ncsi_reshdr->response_code, ncsi_reshdr->reason_code);
        return (ncsi_reshdr->reason_code);
  }	
	return 0;
}


static u32 ncsi_do_checksum(u8 *buf, u16 len)
{
	u32 checksum = 0;
	u16 cnt = 0;
	u16 temp = 0; 

	while (len > 0)
	{
		temp = htons(*(((u16 *) buf) + cnt)); 
		checksum += temp;
		cnt++;
		len -= sizeof(u16);
	}

	return (htonl((~checksum) + 1));
}

//static inline int IS_VALID_NCSI_CHECKSUM(struct sNCSI_RES_HDR *ncsi_reshdr)
//{
//	u32 reschecksum = *(u32 *) (((u8 *) ncsi_reshdr) + NCSI_HLEN + ncsi_reshdr->ncsi_hdr.payload_len);
//	u32 expchecksum = 0;
//	
//
//
//	expchecksum = ncsi_do_checksum((unsigned char *)ncsi_reshdr, 
//							NCSI_HLEN + ncsi_reshdr->ncsi_hdr.payload_len);
//	
//	PDEBUG("CHECKSUM: RES = 0x%lx EXP = 0x%lx\n", reschecksum, expchecksum);
//	
//	if (reschecksum == expchecksum)
//	{
//		return (1);
//	}
//	else
//	{
//		return (0);
//	}
//
//
//	
//}
		
/****** NCSI COMMAND HANDLER ******/
static int ncsi_cmd_clear_initial_state(sNCSI_IF_Info *ncsi_if)
{
	int res;
	struct sk_buff *res_skb = NULL;
	struct sNCSI_RES_HDR *ncsi_reshdr;

	/* Issue the command */
	res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_CLEAR_INITIAL_STATE,
							 NULL, CMD_CLEAR_INITIAL_STATE_PAYLOAD_LEN);

	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		/* proces the command */
		PDEBUG("%s: Start to proces the command\n", ncsi_cmd_str[CMD_CLEAR_INITIAL_STATE]);
    
		ncsi_reshdr = NCSI_RES_HEADER(res_skb);
        
		if ((NCSI_RES_CODE_CMD_OK == ncsi_reshdr->response_code))
		{
			/* Once Select Package success, change the state to the Initial state */
			
			
			
		}
		else
		{
			res = ncsi_cmd_error_handler(ncsi_if,ncsi_reshdr);
		}
	}

	if (res_skb)
	{
		PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);
    
		dev_kfree_skb(res_skb);
		res_skb = NULL;
	}

	return (res);
}

static int ncsi_cmd_select_package(sNCSI_IF_Info *ncsi_if)
{
	int res;
	struct sk_buff *res_skb = NULL;
	u32 payload;
	struct sNCSI_RES_HDR *ncsi_reshdr;
		
    ///* need sent deselect package first */
	//ncsi_cmd_deselect_package(ncsi_if);

	/* Issue the command */
	payload = htonl(ncsi_if->ncsi_hardware_arb);
	
	res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_SELECT_PACKAGE,
							 (u8 *)&payload, CMD_SELECT_PACKAGE_PAYLOAD_LEN);

	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		return (res);
	}

	if (res_skb)
	{
		PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);

		dev_kfree_skb(res_skb);
		res_skb = NULL;
	}

	return (res);
}

static int ncsi_cmd_deselect_package(sNCSI_IF_Info *ncsi_if)
{
	int res;
	struct sk_buff *res_skb = NULL;
	struct sNCSI_RES_HDR *ncsi_reshdr;
	
	/* Issue the command */
	


	


	res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_DESELECT_PACKAGE,
							 NULL, CMD_DESELECT_PACKAGE_PAYLOAD_LEN);

	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		/* proces the command */
		PDEBUG("%s: Start to proces the command\n", ncsi_cmd_str[CMD_DESELECT_PACKAGE]);

		ncsi_reshdr = NCSI_RES_HEADER(res_skb);

		if ((NCSI_RES_CODE_CMD_OK == ncsi_reshdr->response_code))
		{
			
			
			
			
			


			

		}
		else
		{      
	        ncsi_reshdr->response_code = ntohs(ncsi_reshdr->response_code);
	        ncsi_reshdr->reason_code = ntohs(ncsi_reshdr->reason_code);    
			PDEBUG("NCSI COMMAND ERROR (%x, %x)", ncsi_reshdr->response_code, ncsi_reshdr->reason_code);
				
	      //Deselect command cannot use error handler function, because it will get 
	      //"Interface Initialization Required" at first time, and ncsi_cmd_error_handler
	      //will restart the channel.
	      //ncsi_cmd_error_handler(ncsi_if,ncsi_reshdr);
				
		}
	}

	if (res_skb)
	{
		PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);

		dev_kfree_skb(res_skb);
		res_skb = NULL;
	}

	return (res);
}

static int ncsi_cmd_enable_channel(sNCSI_IF_Info *ncsi_if)
{
	int res;
	struct sk_buff *res_skb = NULL;
	struct sNCSI_RES_HDR *ncsi_reshdr;

	/* Issue the command */
	res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_ENABLE_CHANNEL,
							 NULL, CMD_ENABLE_CHANNEL_PAYLOAD_LEN);
   							 

	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		/* proces the command */
		PDEBUG("%s: Start to proces the command\n", ncsi_cmd_str[CMD_ENABLE_CHANNEL]);

		ncsi_reshdr = NCSI_RES_HEADER(res_skb);

		if ((NCSI_RES_CODE_CMD_OK == ncsi_reshdr->response_code))
		{   
//			PDEBUG("\n NCSI is in Operational state!! \n"); 

		}
		else
		{
			res = ncsi_cmd_error_handler(ncsi_if,ncsi_reshdr);
		}
	}

	if (res_skb)
	{
		PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);

		dev_kfree_skb(res_skb);
		res_skb = NULL;
	}

	return (res);
}
	
static int ncsi_cmd_disable_channel(sNCSI_IF_Info *ncsi_if)
{
	int res;
	struct sk_buff *res_skb = NULL;
	struct sNCSI_RES_HDR *ncsi_reshdr;
	u32 payload;

	/* Issue the command */
	payload = CMD_DISABLE_CHANNEL_PAYLOAD_KEEP_LINK_UP;
	res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_DISABLE_CHANNEL,
							 (u8 *)&payload, CMD_DISABLE_CHANNEL_PAYLOAD_LEN);

	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		/* proces the command */
		PDEBUG("%s: Start to proces the command\n", ncsi_cmd_str[CMD_DISABLE_CHANNEL]);

		ncsi_reshdr = NCSI_RES_HEADER(res_skb);

		if ((NCSI_RES_CODE_CMD_OK == ncsi_reshdr->response_code))
		{
			/* nothing to do */
		}
		else
		{
			res = ncsi_cmd_error_handler(ncsi_if,ncsi_reshdr);
		}
	}

	if (res_skb)
	{
		PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);

		dev_kfree_skb(res_skb);
		res_skb = NULL;
	}

	return (res);
}

static int ncsi_cmd_reset_channel(sNCSI_IF_Info *ncsi_if)
{
	int res;
	u32 payload;
	struct sk_buff *res_skb = NULL;
	struct sNCSI_RES_HDR *ncsi_reshdr;
    
    
    payload = 0;
	/* Issue the command */
	res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_RESET_CHANNEL,
							 (u8 *)&payload, CMD_RESET_CHANNEL_PAYLOAD_LEN);

	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		/* proces the command */
		//PDEBUG("%s: Start to proces the command\n", ncsi_cmd_str[CMD_RESET_CHANNEL]);
        //
		//ncsi_reshdr = NCSI_RES_HEADER(res_skb);
        //
		//if ((NCSI_RES_CODE_CMD_OK == ncsi_reshdr->response_code))
		//{
		//	/* nothing to do */
		//}
		//else
		//{
		//	res = ncsi_cmd_error_handler(ncsi_if,ncsi_reshdr);
		//}
	}

	//if (res_skb)
	//{
	//	PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);
    //
	//	dev_kfree_skb(res_skb);
	//}

	return (res);
}

int ncsi_cmd_enable_channel_network_tx(sNCSI_IF_Info *ncsi_if)
{
	int res;
	struct sk_buff *res_skb = NULL;
	struct sNCSI_RES_HDR *ncsi_reshdr;

	/* Issue the command */
	res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_ENABLE_CHANNEL_NETWORK_TX,
							 NULL, CMD_ENABLE_CHANNEL_NETWORK_TX_PAYLOAD_LEN);
   							 
	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		/* proces the command */
		PDEBUG("%s: Start to proces the command\n", ncsi_cmd_str[CMD_ENABLE_CHANNEL_NETWORK_TX]);
  
		ncsi_reshdr = NCSI_RES_HEADER(res_skb);
  
		if ((NCSI_RES_CODE_CMD_OK == ncsi_reshdr->response_code))
		{
			/* nothing to do */
		}
		else
		{
			res = ncsi_cmd_error_handler(ncsi_if,ncsi_reshdr);
		}
	}

	if (res_skb)
	{
		PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);

		dev_kfree_skb(res_skb);
		res_skb = NULL;
	}

	return (res);
}


int ncsi_cmd_disable_channel_network_tx(sNCSI_IF_Info *ncsi_if)
{
	int res;
	struct sk_buff *res_skb = NULL;
	struct sNCSI_RES_HDR *ncsi_reshdr;

	/* Issue the command */
	res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_DISABLE_CHANNEL_NETWORK_TX,
							 NULL, CMD_DISABLE_CHANNEL_NETWORK_TX_PAYLOAD_LEN);

	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		/* proces the command */
		PDEBUG("%s: Start to proces the command\n", ncsi_cmd_str[CMD_DISABLE_CHANNEL_NETWORK_TX]);

		ncsi_reshdr = NCSI_RES_HEADER(res_skb);

		if ((NCSI_RES_CODE_CMD_OK == ncsi_reshdr->response_code))
		{
			/* nothing to do */
		}
		else
		{
			res = ncsi_cmd_error_handler(ncsi_if,ncsi_reshdr);
		}
	}

	if (res_skb)
	{
		PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);

		dev_kfree_skb(res_skb);
		res_skb = NULL;
	}

	return (res);
}



static int ncsi_cmd_aen_enable(sNCSI_IF_Info *ncsi_if)
{
	int res;
	struct sk_buff *res_skb = NULL;
	struct sNCSI_RES_HDR *ncsi_reshdr;
	u32 payload[2];
	
	payload[0] = htonl( (u32) NCSI_MC_ID);
    payload[1] = htonl(ncsi_if->ncsi_aen_ctrl);
	/* Issue the command */
	res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_AEN_ENABLE,
							 (u8 *)payload, CMD_AEN_ENABLE_PAYLOAD_LEN);

	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		/* proces the command */
		PDEBUG("%s: Start to proces the command\n", ncsi_cmd_str[CMD_AEN_ENABLE]);
  
		ncsi_reshdr = NCSI_RES_HEADER(res_skb);
  
		if ((NCSI_RES_CODE_CMD_OK == ncsi_reshdr->response_code))
		{
			/* nothing to do */
		}
		else
		{
			res = ncsi_cmd_error_handler(ncsi_if,ncsi_reshdr);
		}
	}

	if (res_skb)
	{
		PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);

		dev_kfree_skb(res_skb);
		res_skb = NULL;
	}

	return (res);
}

int ncsi_cmd_set_link(sNCSI_IF_Info *ncsi_if)
{
	int res;
	struct sk_buff *res_skb = NULL;
	struct sNCSI_RES_HDR *ncsi_reshdr;
    u32 payload[2];
    payload[0] = htonl(ncsi_if->ncsi_chnl_parameters.link_setting.link_settings);
    payload[1] = 0;
    
	/* Issue the command */
	res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_SET_LINK,
							 (u8 *)payload, CMD_SET_LINK_PAYLOAD_LEN);

	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		/* proces the command */
		PDEBUG("%s: Start to proces the command\n", ncsi_cmd_str[CMD_SET_LINK]);

		ncsi_reshdr = NCSI_RES_HEADER(res_skb);

		if ((NCSI_RES_CODE_CMD_OK == ncsi_reshdr->response_code))
		{
			/* nothing to do */
		}
		else
		{
			res = ncsi_cmd_error_handler(ncsi_if,ncsi_reshdr);
		}
	}

	if (res_skb)
	{
		PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);

		dev_kfree_skb(res_skb);
		res_skb = NULL;
	}

	return (res);
}

int ncsi_cmd_get_link_status(sNCSI_IF_Info *ncsi_if)
{
	int res;
	struct sk_buff *res_skb = NULL;
	struct sNCSI_RES_HDR *ncsi_reshdr;

	/* Issue the command */
	res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_GET_LINK_STATUS,
							 NULL, CMD_GET_LINK_STATUS_PAYLOAD_LEN);

	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		/* proces the command */
		PDEBUG("%s: Start to proces the command\n", ncsi_cmd_str[CMD_GET_LINK_STATUS]);

		ncsi_reshdr = NCSI_RES_HEADER(res_skb);

		if ((NCSI_RES_CODE_CMD_OK == ncsi_reshdr->response_code))
		{
			
			memcpy((u8 *)&ncsi_if->ncsi_link_status, 
					(u8 *)&ncsi_reshdr->payload_data, 
					(ncsi_reshdr->ncsi_hdr.payload_len - 4));
					
					ncsi_if->ncsi_link_status.link_status = ntohl(ncsi_if->ncsi_link_status.link_status);
					//ncsi_if->ncsi_link_status.other_indications = ntohl(ncsi_if->ncsi_link_status.other_indications);
					//ncsi_if->ncsi_link_status.oem_link_status = ntohl(ncsi_if->ncsi_link_status.oem_link_status);
					
                if((ncsi_if->ncsi_link_status.link_status) & 0x00000001)
                {
                    printk("ncsi_cmd_get_link_status: Channel[%d] Link status is Up\n", ncsi_if->chnl_id);
                    netif_carrier_on(ncsi_if->dev);    
                }
                else
                {
                    printk("ncsi_cmd_get_link_status: Channel[%d] Link status is Down\n", ncsi_if->chnl_id);
                    netif_carrier_off(ncsi_if->dev);    
                }     
            }
            else
            {
                printk("%s: Receive command error\n", ncsi_cmd_str[CMD_GET_LINK_STATUS]);
            }
	}
	else
	{
		res = ncsi_cmd_error_handler(ncsi_if,ncsi_reshdr);
	}

	if (res_skb)
	{
		PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);

		dev_kfree_skb(res_skb);
		res_skb = NULL;
	}
    
	return (res);
}

static int ncsi_cmd_set_vlan_filters(sNCSI_IF_Info *ncsi_if)
{
	int res;
	struct sk_buff *res_skb = NULL;
	struct sNCSI_RES_HDR *ncsi_reshdr;

	/* Issue the command */
	res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_SET_VLAN_FILTERS,
							 (u8 *)&ncsi_if->ncsi_vlan_info, CMD_SET_VLAN_FILTERS_PAYLOAD_LEN);

	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		/* proces the command */
		PDEBUG("%s: Start to proces the command\n", ncsi_cmd_str[CMD_SET_VLAN_FILTERS]);

		ncsi_reshdr = NCSI_RES_HEADER(res_skb);

		if ((NCSI_RES_CODE_CMD_OK == ncsi_reshdr->response_code))
		{
			/* nothing to do */
		}
		else
		{
			res = ncsi_cmd_error_handler(ncsi_if,ncsi_reshdr);
		}
	}

	if (res_skb)
	{
		PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);

		dev_kfree_skb(res_skb);
		res_skb = NULL;
	}

	return (res);
}

static int ncsi_cmd_enable_vlan(sNCSI_IF_Info *ncsi_if)
{
	int res;
	struct sk_buff *res_skb = NULL;
	struct sNCSI_RES_HDR *ncsi_reshdr;
	u32 payload;
    
    payload = htonl(ncsi_if->ncsi_vlan_mode);
    

	/* Issue the command */
	res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_ENABLE_VLAN,
							 //(u8 *)&ncsi_if->ncsi_chnl_caps.vlan_mode, CMD_ENABLE_VLAN_PAYLOAD_LEN);
							 (u8 *)&payload, CMD_ENABLE_VLAN_PAYLOAD_LEN);

	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		/* proces the command */
		PDEBUG("%s: Start to proces the command\n", ncsi_cmd_str[CMD_ENABLE_VLAN]);

		ncsi_reshdr = NCSI_RES_HEADER(res_skb);

		if ((NCSI_RES_CODE_CMD_OK == ncsi_reshdr->response_code))
		{
			/* nothing to do */
		}
		else
		{
			res = ncsi_cmd_error_handler(ncsi_if,ncsi_reshdr);
		}
	}

	if (res_skb)
	{
		PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);

		dev_kfree_skb(res_skb);
		res_skb = NULL;
		
	}

	return (res);
}

static int ncsi_cmd_disable_vlan(sNCSI_IF_Info *ncsi_if)
{
	int res;
	struct sk_buff *res_skb = NULL;
	struct sNCSI_RES_HDR *ncsi_reshdr;

	/* Issue the command */
	res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_DISABLE_VLAN,
							 NULL, CMD_DISABLE_VLAN_PAYLOAD_LEN);

	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		/* proces the command */
		PDEBUG("%s: Start to proces the command\n", ncsi_cmd_str[CMD_DISABLE_VLAN]);

		ncsi_reshdr = NCSI_RES_HEADER(res_skb);

		if ((NCSI_RES_CODE_CMD_OK == ncsi_reshdr->response_code))
		{
			/* nothing to do */
		}
		else
		{
			res = ncsi_cmd_error_handler(ncsi_if,ncsi_reshdr);
		}
	}

	if (res_skb)
	{
		PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);

		dev_kfree_skb(res_skb);
		res_skb = NULL;
	}

	return (res);
}

static int ncsi_cmd_set_mac_address(sNCSI_IF_Info *ncsi_if)
{
	int res;
	int i;
	struct sk_buff *res_skb = NULL;
	struct sNCSI_RES_HDR *ncsi_reshdr;
	u8 payload[CMD_SET_MAC_ADDRESS_PAYLOAD_LEN];
	
	for(i=0;i<6;i++) 
	{
	    payload[i] = ncsi_if->ncsi_mac_info.mac_addr[i];
	}	
  payload[6] = ncsi_if->ncsi_mac_info.mac_num;
  payload[7] = ncsi_if->ncsi_mac_info.addr_type_enable; 
	/* Issue the command */
    res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_SET_MAC_ADDRESS,
							 payload, CMD_SET_MAC_ADDRESS_PAYLOAD_LEN);
    
	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		/* proces the command */
		PDEBUG("%s: Start to proces the command\n", ncsi_cmd_str[CMD_SET_MAC_ADDRESS]);
  
		ncsi_reshdr = NCSI_RES_HEADER(res_skb);
  
		if ((NCSI_RES_CODE_CMD_OK == ncsi_reshdr->response_code))
		{
			/* nothing to do */
		}
		else
		{
			res = ncsi_cmd_error_handler(ncsi_if,ncsi_reshdr);
		}
	}

	if (res_skb)
	{
		PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);

		dev_kfree_skb(res_skb);
		res_skb = NULL;
	}

	return (res);
}

static int ncsi_cmd_enable_broadcast_filtering(sNCSI_IF_Info *ncsi_if)
{
	int res;
	struct sk_buff *res_skb = NULL;
	struct sNCSI_RES_HDR *ncsi_reshdr;
	u32 payload;
    
  payload = htonl(ncsi_if->ncsi_chnl_caps.bcast_filter_flag);
    
	/* Issue the command */
	res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_ENABLE_BROADCAST_FILTERING,
							 (u8 *)&payload, CMD_ENABLE_BROADCAST_FILTERING_PAYLOAD_LEN);

	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		/* proces the command */
		PDEBUG("%s: Start to proces the command\n", ncsi_cmd_str[CMD_ENABLE_BROADCAST_FILTERING]);
  
		ncsi_reshdr = NCSI_RES_HEADER(res_skb);
  
		if ((NCSI_RES_CODE_CMD_OK == ncsi_reshdr->response_code))
		{
			/* nothing to do */
		}
		else
		{
			res = ncsi_cmd_error_handler(ncsi_if,ncsi_reshdr);
		}
	}

	if (res_skb)
	{
		PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);

		dev_kfree_skb(res_skb);
		res_skb = NULL;
	}

	return (res);
}

static int ncsi_cmd_disable_broadcast_filtering(sNCSI_IF_Info *ncsi_if)
{
	int res;
	struct sk_buff *res_skb = NULL;
	struct sNCSI_RES_HDR *ncsi_reshdr;

	/* Issue the command */
	res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_DISABLE_BROADCAST_FILTERING,
							 NULL, CMD_DISABLE_BROADCAST_FILTERING_PAYLOAD_LEN);

	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		/* proces the command */
		PDEBUG("%s: Start to proces the command\n", ncsi_cmd_str[CMD_DISABLE_BROADCAST_FILTERING]);

		ncsi_reshdr = NCSI_RES_HEADER(res_skb);

		if ((NCSI_RES_CODE_CMD_OK == ncsi_reshdr->response_code))
		{
			/* nothing to do */
		}
		else
		{
			res = ncsi_cmd_error_handler(ncsi_if,ncsi_reshdr);
		}
	}

	if (res_skb)
	{
		PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);

		dev_kfree_skb(res_skb);
		res_skb = NULL;
	}

	return (res);
}

static int ncsi_cmd_enable_global_multicast_filtering(sNCSI_IF_Info *ncsi_if)
{
    int res = 0;
	struct sk_buff *res_skb = NULL;
	struct sNCSI_RES_HDR *ncsi_reshdr;
	  u32 payload;
    payload = htonl(ncsi_if->ncsi_chnl_caps.mcast_filter_flag);

	/* Issue the command */
	res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_ENABLE_GLOBAL_MULTICAST_FILTERING,
							 (u8 *)&payload, CMD_ENABLE_GLOBAL_MULTICAST_FILTERING_PAYLOAD_LEN);

	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		/* proces the command */
		PDEBUG("%s: Start to proces the command\n", ncsi_cmd_str[CMD_ENABLE_GLOBAL_MULTICAST_FILTERING]);

		ncsi_reshdr = NCSI_RES_HEADER(res_skb);

		if ((NCSI_RES_CODE_CMD_OK == ncsi_reshdr->response_code))
		{
			/* nothing to do */
		}
		else
		{
			res = ncsi_cmd_error_handler(ncsi_if,ncsi_reshdr);
		}
	}

	if (res_skb)
	{
		PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);

		dev_kfree_skb(res_skb);
		res_skb = NULL;
	}


	return (res);
}

static int ncsi_cmd_disable_global_multicast_filtering (sNCSI_IF_Info *ncsi_if)
{
	int res;
	struct sk_buff *res_skb = NULL;
	struct sNCSI_RES_HDR *ncsi_reshdr;

	/* Issue the command */
	res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_DISABLE_GLOBAL_MULTICAST_FILTERING,
							 NULL, CMD_DISABLE_GLOBAL_MULTICAST_FILTERING_PAYLOAD_LEN);

	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		/* proces the command */
		PDEBUG("%s: Start to proces the command\n", ncsi_cmd_str[CMD_DISABLE_GLOBAL_MULTICAST_FILTERING]);

		ncsi_reshdr = NCSI_RES_HEADER(res_skb);

		if ((NCSI_RES_CODE_CMD_OK == ncsi_reshdr->response_code))
		{
			/* nothing to do */
		}
		else
		{
			res = ncsi_cmd_error_handler(ncsi_if,ncsi_reshdr);
		}
	}

	if (res_skb)
	{
		PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);

		dev_kfree_skb(res_skb);
		res_skb = NULL;
	}

	return (res);
}

static int ncsi_cmd_set_ncsi_flow_control(sNCSI_IF_Info *ncsi_if)
{
	int res;
	struct sk_buff *res_skb = NULL;
	struct sNCSI_RES_HDR *ncsi_reshdr;

	/* Issue the command */
	res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_SET_NCSI_FLOW_CONTROL,
							 (u8 *)&ncsi_if->ncsi_flow_ctrl, CMD_SET_NCSI_FLOW_CONTROL_PAYLOAD_LEN);

	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		/* proces the command */
		PDEBUG("%s: Start to proces the command\n", ncsi_cmd_str[CMD_SET_NCSI_FLOW_CONTROL]);

		ncsi_reshdr = NCSI_RES_HEADER(res_skb);

		if ((NCSI_RES_CODE_CMD_OK == ncsi_reshdr->response_code))
		{
			/* nothing to do */
		}
		else
		{
			res = ncsi_cmd_error_handler(ncsi_if,ncsi_reshdr);
		}
	}

	if (res_skb)
	{
		PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);

		dev_kfree_skb(res_skb);
		res_skb = NULL;
	}

	return (res);
}

int ncsi_cmd_get_version_id(sNCSI_IF_Info *ncsi_if)
{
	int res;
	struct sk_buff *res_skb = NULL;
	struct sNCSI_RES_HDR *ncsi_reshdr;

	/* Issue the command */
	res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_GET_VERSION_ID,
							 NULL, CMD_GET_VERSION_ID_PAYLOAD_LEN);

	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		/* proces the command */
		PDEBUG("%s: Start to proces the command\n", ncsi_cmd_str[CMD_GET_VERSION_ID]);

		ncsi_reshdr = NCSI_RES_HEADER(res_skb);

		if ((NCSI_RES_CODE_CMD_OK == ncsi_reshdr->response_code))
		{
			memcpy((u8 *)&ncsi_if->ncsi_version, 
					(u8 *)&ncsi_reshdr->payload_data, 
					(ncsi_reshdr->ncsi_hdr.payload_len - 4));
			
			{
			u8 tmpstr[13];
			
			PDEBUG("\n===== NCSI Version =====\n");
			PDEBUG("NCSI Version:");

			if((ncsi_if->ncsi_version.ncsi_version[0]&0xF0) == 0xf0)
			{
				PDEBUG ( "%X.", (ncsi_if->ncsi_version.ncsi_version[0] & 0x0F));
			}
			else
			{ 
				PDEBUG ( "%X.", (ncsi_if->ncsi_version.ncsi_version[0]));
		    }
		    
		    if((ncsi_if->ncsi_version.ncsi_version[1]&0xF0) == 0xf0)
			{   
				PDEBUG ("%X.", (ncsi_if->ncsi_version.ncsi_version[1] & 0x0F));
			}
			else
			{ 
				PDEBUG ("%X.", (ncsi_if->ncsi_version.ncsi_version[1]));
		    }		
		    
		    if((ncsi_if->ncsi_version.ncsi_version[2]&0xF0) == 0xf0)
			{   
				PDEBUG("%X", (ncsi_if->ncsi_version.ncsi_version[2] & 0x0F));
			}
			else
			{   
				PDEBUG("%X", (ncsi_if->ncsi_version.ncsi_version[2]));
		    }
		    PDEBUG("%c \n", (ncsi_if->ncsi_version.ncsi_version[3]));


			memset(tmpstr, 0, 13);
			memcpy(tmpstr, ncsi_if->ncsi_version.fw_name, 12);
			
			PDEBUG("FW Name String: %s \n", tmpstr); 
			PDEBUG("FW Version:%02X.%02X.%02X.%02X\n",
				*(((u8 *)(&(ncsi_if->ncsi_version.fw_version))) + 0),
				*(((u8 *)(&(ncsi_if->ncsi_version.fw_version))) + 1),
				*(((u8 *)(&(ncsi_if->ncsi_version.fw_version))) + 2),
				*(((u8 *)(&(ncsi_if->ncsi_version.fw_version))) + 3)); 
            
            ncsi_if->ncsi_version.pci_did = ntohs(ncsi_if->ncsi_version.pci_did);
            ncsi_if->ncsi_version.pci_vid = ntohs(ncsi_if->ncsi_version.pci_vid);
            ncsi_if->ncsi_version.pci_ssid = ntohs(ncsi_if->ncsi_version.pci_ssid);
            ncsi_if->ncsi_version.pci_svid = ntohs(ncsi_if->ncsi_version.pci_svid);
             
			PDEBUG("PCI ID Infromation: DID: 0x%X, VID: 0x%X, SSID: 0x%X, SVID: 0x%X\n",
				ncsi_if->ncsi_version.pci_did,
				ncsi_if->ncsi_version.pci_vid,
				ncsi_if->ncsi_version.pci_ssid,
				ncsi_if->ncsi_version.pci_svid);
            
            ncsi_if->ncsi_version.iana = ntohl(ncsi_if->ncsi_version.iana);
			PDEBUG("IANA: 0x%lX\n\n", ncsi_if->ncsi_version.iana);
		    }
			
		}
		else
		{    
			res = ncsi_cmd_error_handler(ncsi_if,ncsi_reshdr);
		}
	}
	
	if (res_skb)
	{       
		PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);
             
		dev_kfree_skb(res_skb);
		res_skb = NULL;
	}       
             
	return (res);
}

static int ncsi_cmd_get_capabilities(sNCSI_IF_Info *ncsi_if)
{
	int res;
	struct sk_buff *res_skb = NULL;
	struct sNCSI_RES_HDR *ncsi_reshdr;

	/* Issue the command */
	res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_GET_CAPABILITIES,
							 NULL, CMD_GET_CAPABILITIES_PAYLOAD_LEN);

	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		/* proces the command */
		PDEBUG("%s: Start to proces the command\n", ncsi_cmd_str[CMD_GET_CAPABILITIES]);
  
		ncsi_reshdr = NCSI_RES_HEADER(res_skb);
  
		if ((NCSI_RES_CODE_CMD_OK == ncsi_reshdr->response_code))
		{   
			memcpy((u8 *)&ncsi_if->ncsi_chnl_caps, 
					(u8 *)&ncsi_reshdr->payload_data, 
					(ncsi_reshdr->ncsi_hdr.payload_len - 4));
		    ncsi_if->ncsi_chnl_caps.capabilities_flags = ntohl(ncsi_if->ncsi_chnl_caps.capabilities_flags);
			ncsi_if->ncsi_chnl_caps.bcast_filter_flag = ntohl(ncsi_if->ncsi_chnl_caps.bcast_filter_flag);
			ncsi_if->ncsi_chnl_caps.mcast_filter_flag = ntohl(ncsi_if->ncsi_chnl_caps.mcast_filter_flag);
			ncsi_if->ncsi_chnl_caps.buf_cap = ntohl(ncsi_if->ncsi_chnl_caps.buf_cap);
			ncsi_if->ncsi_chnl_caps.aen_ctrl = ntohl(ncsi_if->ncsi_chnl_caps.aen_ctrl);	
		PDEBUG(" ncsi_if->ncsi_chnl_caps.capabilities_flags = %x   \n", ncsi_if->ncsi_chnl_caps.capabilities_flags);
		PDEBUG(" ncsi_if->ncsi_chnl_caps.bcast_filter_flag = %x\n", ncsi_if->ncsi_chnl_caps.bcast_filter_flag);
		PDEBUG(" ncsi_if->ncsi_chnl_caps.mcast_filter_flag = %x\n", ncsi_if->ncsi_chnl_caps.mcast_filter_flag);
		PDEBUG(" ncsi_if->ncsi_chnl_caps.buf_cap = %x\n", ncsi_if->ncsi_chnl_caps.buf_cap);
		PDEBUG(" ncsi_if->ncsi_chnl_caps.aen_ctrl = %x\n", ncsi_if->ncsi_chnl_caps.aen_ctrl);
		PDEBUG(" ncsi_if->ncsi_chnl_caps.vlan_filter_cnt = %x\n", ncsi_if->ncsi_chnl_caps.vlan_filter_cnt);
		PDEBUG(" ncsi_if->ncsi_chnl_caps.mixed_filter_cnt = %x\n", ncsi_if->ncsi_chnl_caps.mixed_filter_cnt);
		PDEBUG(" ncsi_if->ncsi_chnl_caps.mcast_filter_cnt = %x\n", ncsi_if->ncsi_chnl_caps.mcast_filter_cnt);
		PDEBUG(" ncsi_if->ncsi_chnl_caps.ucast_filter_cnt = %x\n", ncsi_if->ncsi_chnl_caps.ucast_filter_cnt);
		PDEBUG(" ncsi_if->ncsi_chnl_caps.reserved2 = %x\n", ncsi_if->ncsi_chnl_caps.reserved2);
		PDEBUG(" ncsi_if->ncsi_chnl_caps.vlan_mode = %x\n", ncsi_if->ncsi_chnl_caps.vlan_mode);
		PDEBUG(" ncsi_if->ncsi_chnl_caps.chnl_cnt = %x\n", ncsi_if->ncsi_chnl_caps.chnl_cnt);				
		}
		else
		{
			res = ncsi_cmd_error_handler(ncsi_if,ncsi_reshdr);
		}
	}

	if (res_skb)
	{
		PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);

		dev_kfree_skb(res_skb);
		res_skb = NULL;
	}

	return (res);
}

static int ncsi_cmd_get_parameters(sNCSI_IF_Info *ncsi_if)
{
	int res;
	struct sk_buff *res_skb = NULL;
	struct sNCSI_RES_HDR *ncsi_reshdr;

	/* Issue the command */
	res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_GET_PARAMETERS,
							 NULL, CMD_GET_PARAMETERS_PAYLOAD_LEN);

	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		/* proces the command */
		PDEBUG("%s: Start to proces the command\n", ncsi_cmd_str[CMD_GET_PARAMETERS]);

		ncsi_reshdr = NCSI_RES_HEADER(res_skb);

		if ((NCSI_RES_CODE_CMD_OK == ncsi_reshdr->response_code))
		{
			u8 *payload_mac_entry_start;
			u8 *payload_vlan_entry_start;

			memcpy((u8 *)&ncsi_if->ncsi_chnl_parameters, 
					(u8 *)&ncsi_reshdr->payload_data, 
					(ncsi_reshdr->ncsi_hdr.payload_len - 4));
			ncsi_if->ncsi_chnl_parameters.mac_addr_cnt_flag = ntohl(ncsi_if->ncsi_chnl_parameters.mac_addr_cnt_flag);
			ncsi_if->ncsi_chnl_parameters.vlan_tag_cnt_flag = ntohl(ncsi_if->ncsi_chnl_parameters.vlan_tag_cnt_flag);
			ncsi_if->ncsi_chnl_parameters.link_setting.link_settings = ntohl(ncsi_if->ncsi_chnl_parameters.link_setting.link_settings);
			ncsi_if->ncsi_chnl_parameters.link_setting.oem_link_settings = ntohl(ncsi_if->ncsi_chnl_parameters.link_setting.oem_link_settings);
			ncsi_if->ncsi_chnl_parameters.bcast_filter_flag = ntohl(ncsi_if->ncsi_chnl_parameters.bcast_filter_flag);
            ncsi_if->ncsi_chnl_parameters.config_flag = ntohl(ncsi_if->ncsi_chnl_parameters.config_flag);
            ncsi_if->ncsi_chnl_parameters.aen_ctrl = ntohl(ncsi_if->ncsi_chnl_parameters.aen_ctrl);
	     	PDEBUG(" ncsi_if->ncsi_chnl_parameters.mac_addr_cnt_flag = %x   \n", ncsi_if->ncsi_chnl_parameters.mac_addr_cnt_flag);
		    PDEBUG(" ncsi_if->ncsi_chnl_parameters.vlan_tag_cnt_flag  = %x\n", ncsi_if->ncsi_chnl_parameters.vlan_tag_cnt_flag );
		    PDEBUG(" ncsi_if->ncsi_chnl_parameters.link_setting.link_settings = %x\n", ncsi_if->ncsi_chnl_parameters.link_setting.link_settings);
		    PDEBUG(" ncsi_if->ncsi_chnl_parameters.link_setting.oem_link_settings  = %x\n", ncsi_if->ncsi_chnl_parameters.link_setting.oem_link_settings );		
            PDEBUG(" ncsi_if->ncsi_chnl_parameters.config_flag  = %x\n", ncsi_if->ncsi_chnl_parameters.config_flag );							
			PDEBUG(" ncsi_if->ncsi_chnl_parameters.aen_ctrl  = %x\n", ncsi_if->ncsi_chnl_parameters.aen_ctrl );									
			
			/* TBD: copy the MAC entry */
			payload_mac_entry_start = (u8 *)&ncsi_reshdr->payload_data 
									  + sizeof(struct sNCSI_CHNL_PARAMETER);
			
			/* TBD: copy the VLAN entry */
			//payload_vlan_entry_start = (u8 *)&ncsi_reshdr->payload_data 
			//						   + sizeof(struct sNCSI_CHNL_PARAMETER)
			//						   + (ncsi_if->ncsi_chnl_parameters.mac_addr_cnt * 6);
		}
		else
		{
			res = ncsi_cmd_error_handler(ncsi_if,ncsi_reshdr);
		}
	}

	if (res_skb)
	{
		PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);

		dev_kfree_skb(res_skb);
		res_skb = NULL;
	}

	return (res);
}

static int ncsi_cmd_get_controller_packet_statistics(sNCSI_IF_Info *ncsi_if)
{
	int res;
	struct sk_buff *res_skb = NULL;
	struct sNCSI_RES_HDR *ncsi_reshdr;

	/* Issue the command */
	res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_GET_CONTROLLER_PACKET_STATISTICS,
							 NULL, CMD_GET_CONTROLLER_PACKET_STATISTICS_PAYLOAD_LEN);

	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		/* proces the command */
		PDEBUG("%s: Start to process the command\n", ncsi_cmd_str[CMD_GET_CONTROLLER_PACKET_STATISTICS]);

		ncsi_reshdr = NCSI_RES_HEADER(res_skb);

		if ((NCSI_RES_CODE_CMD_OK == ncsi_reshdr->response_code))
		{
			/* TBD: need to save the satistics */
		}
		else
		{
			res = ncsi_cmd_error_handler(ncsi_if,ncsi_reshdr);
		}
	}

	if (res_skb)
	{
		PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);

		dev_kfree_skb(res_skb);
		res_skb = NULL;
	}

	return (res);
}

static int ncsi_cmd_get_ncsi_statistics(sNCSI_IF_Info *ncsi_if)
{
	int res;
	struct sk_buff *res_skb = NULL;
	struct sNCSI_RES_HDR *ncsi_reshdr;

	/* Issue the command */
	res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_GET_NCSI_STATISTICS,
							 NULL, CMD_GET_NCSI_STATISTICS_PAYLOAD_LEN);

	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		/* proces the command */
		PDEBUG("%s: Start to proces the command\n", ncsi_cmd_str[CMD_GET_NCSI_STATISTICS]);

		ncsi_reshdr = NCSI_RES_HEADER(res_skb);

		if ((NCSI_RES_CODE_CMD_OK == ncsi_reshdr->response_code))
		{
			/* TBD: need to save the satistics */
		}
		else
		{
			res = ncsi_cmd_error_handler(ncsi_if,ncsi_reshdr);
		}
	}

	if (res_skb)
	{
		PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);

		dev_kfree_skb(res_skb);
		res_skb = NULL;
	}

	return (res);
}

int ncsi_cmd_get_ncsi_passthrough_statistics(sNCSI_IF_Info *ncsi_if)
{
	int res;
	struct sk_buff *res_skb = NULL;
	struct sNCSI_RES_HDR *ncsi_reshdr;

	/* Issue the command */
	res = ncsi_send_and_wait(ncsi_if, &res_skb, CMD_GET_NCSI_PASSTHROUGH_STATISTICS,
							 NULL, CMD_GET_NCSI_PASSTHROUGH_STATISTICS_PAYLOAD_LEN);

	if (NCSI_RECEIVE_RES_PACKET_OK == res)
	{
		/* proces the command */
		PDEBUG("%s: Start to proces the command\n", ncsi_cmd_str[CMD_GET_NCSI_PASSTHROUGH_STATISTICS]);

		ncsi_reshdr = NCSI_RES_HEADER(res_skb);

		if ((NCSI_RES_CODE_CMD_OK == ncsi_reshdr->response_code))
		{
			/* TBD: need to save the satistics */
		}
		else
		{
			res = ncsi_cmd_error_handler(ncsi_if,ncsi_reshdr);
		}
	}

	if (res_skb)
	{
		PDEBUG("NCSI Free res_skb = 0x%lX\n", res_skb);

		dev_kfree_skb(res_skb);
		res_skb = NULL;
	}

	return (res);
}

static int ncsi_aen_link_status_change(sNCSI_IF_Info *ncsi_if, struct sNCSI_AEN_HDR *ncsi_aenhdr)
{
	int res = 0;
    u32	link_stat;
    u8 chnl_id;
    struct ncsi_aen_link_hdr *aen_link_hdr;
    struct net_device *dev;

	aen_link_hdr = (struct ncsi_aen_link_hdr *)ncsi_aenhdr;
	link_stat = ntohl(aen_link_hdr->link_stat);	
						
	
	chnl_id = aen_link_hdr->aenh.ncsi_hdr.chnl_id ;
	//dev = ncsi_if->ncsi_slave_dev[PKG_ID(chnl_id)*2 +INTERNAL_CHNL_ID(chnl_id)];
	  dev = ncsi_if->dev;
	
	if ( 0 == (link_stat & BNCSI_LINK_UP) ) /* link down */
	{
		printk(KERN_ALERT "Link down AEN\n");
		if (netif_carrier_ok(dev))  /* if current state is up */
		{
			printk(KERN_ALERT  "Calling netif_carrier_off()\n");
			netif_carrier_off(dev);	
		}
	}
	else
	{
		printk(KERN_ALERT  "Link up AEN\n");
		if (!netif_carrier_ok(dev)) /* if current state is down */
		{
			printk(KERN_ALERT  "Calling netif_carrier_on()\n");
			netif_carrier_on(dev);	
		}
	}	
								
	return res;
}

static int ncsi_aen_reconfig_required(sNCSI_IF_Info *ncsi_if, struct sNCSI_AEN_HDR *ncsi_aenhdr)
{   
	int err = 0;
	ncsi_if->error_recovery_enable_flag = 1;
    return err;
}

static int ncsi_aen_os_driver_change(sNCSI_IF_Info *ncsi_if, struct sNCSI_AEN_HDR *ncsi_aenhdr)
{
	int res = 0;

	/* TBD: */
	return (res);
}

static int ncsi_queue_read(sNCSI_IF_Info *ncsi_if, 
						   struct sk_buff **res_skb, enum ncsi_queue_type type)
{
	struct skb_queue *skbq_res;
  unsigned long flags;
    
	if (type == NCSI_RES_QUEUE)
	{
		skbq_res = &ncsi_if->ncsi_res_quene;    
		
		if (skb_queue_len(&skbq_res->ncsi_queue) == 0)
		{								
			return -ETIMEDOUT;
		}
		else
		{
			*res_skb = skb_dequeue(&skbq_res->ncsi_queue);	
		}
	}
	else 
	{   
		return -EINVAL;
	}
    
	return 0;
}

static int ncsi_queue_write(sNCSI_IF_Info *ncsi_if, 
							struct sk_buff *skb, enum ncsi_queue_type type)
{
	struct skb_queue *skbq;
	u8 queue_len;

	if (type == NCSI_RES_QUEUE)
	{
		skbq = &ncsi_if->ncsi_res_quene;
		queue_len = NCSI_RES_PACKET_QUEUE_LEN;
	}
	else
	{
		if(skb)
    {
        dev_kfree_skb(skb);
        skb = NULL;
    }
		return -EINVAL;
	}
     
	/* if queue is full, just return */
	if (skb_queue_len(&skbq->ncsi_queue) >= queue_len)
	{ 
		
		PDEBUG(KERN_ALERT "ncsi_queue_write: The response queue is error! \n");  
		if(skb)
		{
		    dev_kfree_skb(skb);
		    skb = NULL;
	  }
		return -ENOSPC;
	}

	skb_queue_head(&skbq->ncsi_queue, skb);
	return 0;

}
     
static void ncsi_rx_aen_packet(sNCSI_IF_Info *ncsi_if, struct sk_buff *aen_skb)
{
	struct sNCSI_AEN_HDR	*ncsi_aenhdr;
	
	
  if (aen_skb != NULL) //aen_skb is a pointer, cannot be NULL
	{
         /* read the AEN packet in */
		 ncsi_aenhdr = NCSI_AEN_HEADER(aen_skb);

		 /* start to process AEN command */
			switch (ncsi_aenhdr->aen_type)
			{
				case AEN_LINK_STATUS_CHANGE:
				{   
					PDEBUG(KERN_ALERT"\n Receive a AEN_LINK_STATUS_CHANGE packet !! \n");
					ncsi_aen_link_status_change(ncsi_if, ncsi_aenhdr);
					break;
				}					
				
				case AEN_RECONFIG_REQUIRED:
				{   
					PDEBUG(KERN_ALERT  "\n Receive a AEN_RECONFIG_REQUIRED packet !! \n");
					ncsi_aen_reconfig_required(ncsi_if, ncsi_aenhdr);
					break;
				}

				case AEN_OS_DRIVER_CHANGE:
				{   
					PDEBUG(KERN_ALERT  "\n Receive a AEN_OS_DRIVER_CHANGE packet !! \n");
					//ncsi_aen_os_driver_change(ncsi_if, ncsi_aenhdr);
					break;
				}

				default:
					break;
			}
			
			if (aen_skb)
			{
				PDEBUG("NCSI Free aen_skb = 0x%lX\n", aen_skb);
				dev_kfree_skb(aen_skb);
				aen_skb = NULL;
			}
    }
	
	
	
	return;
}

static void ncsi_rx_response_packet(sNCSI_IF_Info *ncsi_if, struct sk_buff *res_skb)
{
	int res;

	res = ncsi_queue_write(ncsi_if, res_skb, NCSI_RES_QUEUE);
	
	if (res < 0)
	{
		PDEBUG(KERN_WARNING "NCSI response packet in-queue error (errno=%d)!!\n", res);
	}
	
	return;
}




static int ncsi_wait_response_packet(sNCSI_IF_Info *ncsi_if, 
									 struct sk_buff **res_skb, u8 res_cmd)
{
	struct sNCSI_RES_HDR *ncsi_reshdr;
	int res;

	PDEBUG("Enter ncsi_wait_response_packet\n");

    
	res = ncsi_queue_read(ncsi_if, res_skb, NCSI_RES_QUEUE);
	
	PDEBUG(KERN_ALERT "ncsi_wait_response_packet: Finish queue read!!! \n");

	PDEBUG("Finish ncsi_queue_read\n");
	
	if (unlikely(res < 0))
	{
		/* faid to get the response packet */
		PDEBUG("ncsi_wait_response_packet: Fail to get the response packet\n");
		
		return res;
	}
	else
	{
		if (*res_skb != NULL)
		{
			PDEBUG("ncsi_wait_response_packet: res_skb != NULL \n");
			
			ncsi_reshdr = NCSI_RES_HEADER(*res_skb);
			
			/* Is a valid response packet and have the vaild checksum*/
			if (IS_VALID_NCSI_RES_PACKET(ncsi_reshdr, ncsi_if, res_cmd))
//mark by sonata don't do checksum check now && IS_VALID_NCSI_CHECKSUM(ncsi_reshdr))
			{
				PDEBUG("IS_VALID_NCSI_RES_PACKET\n");
				
				return (NCSI_RECEIVE_RES_PACKET_OK);
			}
			else
			{
				PDEBUG("NCSI_RECEIVE_RES_PACKET_INVALID\n");

				return (NCSI_RECEIVE_RES_PACKET_INVALID);
			}
		}
		else
		{
			PDEBUG("ncsi_wait_response_packet res_skb is NULL\n");

			return (NCSI_RECEIVE_RES_PACKET_INVALID);
		}
	}
}

static void ncsi_eth_header(sNCSI_IF_Info *ncsi_if, struct sk_buff *skb)
{
	static const unsigned char ncsi_dest_mac[ETH_ALEN] 
		= {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
	struct ethhdr *eth = NCSI_ETH_HEADER(skb);
    int i;
    
    for(i=0;i<ETH_ALEN;i++)
    {    
         eth->h_source[i] = ncsi_if->dev->dev_addr[i];
    }
	memcpy(eth->h_dest, ncsi_dest_mac, 6);

	eth->h_proto = htons(NCSI_ETHER_TYPE);

	/* TBD: add VLAN tag if support VLAN */
	return;
}


/* This function build the NCSI header, payload and checksum */
static void ncsi_header_payload(sNCSI_IF_Info *ncsi_if, struct sk_buff *req_skb,
								u8 cmd, u8 *payload, u16 payload_len)
{
	struct sNCSI_HDR *ncsi_hdr;
	struct ethhdr *eth = NCSI_ETH_HEADER(req_skb); 
	
	u8 *ncsi_payload_start;
	u32 checksum = 0;
	

	/* build NCSI header */
	ncsi_hdr = NCSI_HEADER(req_skb);

	ncsi_hdr->mc_id = NCSI_MC_ID;
	ncsi_hdr->hdr_rev = NCSI_HEADER_REVISION;
	ncsi_hdr->cmd_iid = ncsi_if->cmd_iid;
	ncsi_hdr->cmd = cmd;
	ncsi_hdr->chnl_id = ncsi_if->chnl_id;
	ncsi_hdr->payload_len = payload_len;
     
	/* build NCSI payload */
	ncsi_payload_start = (req_skb->data + ETH_HLEN + NCSI_HLEN);

	if (0 != ncsi_hdr->payload_len)
	{
		/* the payload is always in big-endian byte order */
		memcpy(ncsi_payload_start, payload, ncsi_hdr->payload_len);
	}


//mark by sonata: keep checksum to zero 
//    checksum = ncsi_do_checksum((unsigned char *)eth, 
//							ETH_HLEN+NCSI_HLEN + ncsi_hdr->payload_len);

    


	/* add 2's complement checksum */
	*((u32 *) (ncsi_payload_start + ncsi_hdr->payload_len)) = checksum;


	return;
}

/* This function send NCSI packet */
static int ncsi_send_packet(sNCSI_IF_Info *ncsi_if, u8 cmd, 
							u8 *payload, u16 payload_len)
{
	int xmit_res = 0;
	u32 i;
	u32 ncsi_packet_len;
/* the payload length should be the 32-bit boundary */
#define PAYLOAD_PAD(len)	(4 - (len % 4))
	
	struct sk_buff *req_skb = NULL;
	unsigned int skb_len ;	

	skb_len = ETH_HLEN + NCSI_HLEN + 
			  payload_len + PAYLOAD_PAD(payload_len) + NCSI_CHECKSUM_LEN;
    
    ncsi_packet_len = skb_len;
    
	if (skb_len < ETH_ZLEN)
	{
		skb_len = ETH_ZLEN;
	}
	
	req_skb = alloc_skb(skb_len + NET_IP_ALIGN, GFP_ATOMIC);
	PDEBUG("NCSI allocate req_skb = 0x%lX\n", req_skb);

	if (likely(req_skb != NULL))
	{
		req_skb->dev = ncsi_if->dev;
		req_skb->len = skb_len;
		skb_reserve(req_skb, NET_IP_ALIGN);

		//It doesn't need skb_pull. jasonhung 090212
		//skb_pull(req_skb, ETH_HLEN);
		


		memset(req_skb->data, 0, skb_len);
		
		ncsi_eth_header(ncsi_if, req_skb);


		ncsi_header_payload(ncsi_if, req_skb, cmd, payload, payload_len);
        
//        #ifdef NCSI_DEBUG
	    for(i=0;i<ncsi_packet_len;i+=4)
	    {
	    	if((ncsi_packet_len-i)>=4)
	    	{
    	        PDEBUG("%2x %2x %2x %2x \n",*(req_skb->data+i),*(req_skb->data+i+1),
    		    *(req_skb->data+i+2),*(req_skb->data+i+3));
    		}
    		else if((ncsi_packet_len%4)==3)
    		{
    			PDEBUG("%2x %2x %2x \n",*(req_skb->data+i),*(req_skb->data+i+1),
    		    *(req_skb->data+i+2));
    	    }
    	    else if((ncsi_packet_len%4)==2)
    		{
    			PDEBUG("%2x %2x \n",*(req_skb->data+i),*(req_skb->data+i+1));
    	    }	
    		else if((ncsi_packet_len%4)==1)
    		{
    			PDEBUG("%2x \n",*(req_skb->data+i));
    	    }    
    	}	
	    PDEBUG("\n");  
//	    #endif
	     
	     
	     
        
        
		/* send NCSI packet */
		PDEBUG(" \n Sonata Test: 980114: Send NCSI packet!!! \n");
		if ( 0 == dev_queue_xmit(req_skb) )
		return STATUS_OK;
	    else
		return STATUS_FAIL;
// Sonata mark		xmit_res = ncsi_if->ncsi_tx_func(req_skb, ncsi_if->dev);
  
        if(likely(xmit_res == 0))
        {   
        	return 0; 
        	//do nothing
        }
        else
        {	
		    if (req_skb)
		    { 
			    PDEBUG("NCSI Free req_skb = 0x%lX\n", req_skb);
    
	    	    dev_kfree_skb(req_skb);
	    	    req_skb = NULL;
		    }
		}    
		return 0;
	}
	else
	{
		PDEBUG(KERN_WARNING "Memory squeeze when alloc_skb in dropping ncsi_send_packet!!\n");
		return -ENOMEM;
	}

#undef PAYLOAD_PAD
}

/* This function send NCSI command and wait the response*/
static int ncsi_send_and_wait(sNCSI_IF_Info *ncsi_if, struct sk_buff **res_skb, 
							  u8 cmd, u8 *payload, u16 payload_len)
{
	int res;
	u8 retrycnt = 0;
	u8 res_cmd = (cmd | NCSI_RES_PACKET_MASK);
	
	do 
	{
		/* send command */
		PDEBUG("[%d]SEND %s!\n", retrycnt, ncsi_cmd_str[cmd]);
	    
		res = ncsi_send_packet(ncsi_if, cmd, payload, payload_len);

		PDEBUG("[%d]End ncsi_send_packet\n", res);
		
		if (res < 0)
		{   
			PDEBUG("Memory squeeze in ncsi_send_and_wait \n\n");
			/* Memory squeeze, return directly */
			return res;
		}
		

        
    mdelay(50);
      
      
		/* wait response, if error it will be re-sended, after that return result */
		res = ncsi_wait_response_packet(ncsi_if, res_skb, res_cmd);
  
		PDEBUG("[%d]End ncsi_wait_response_packet\n", res);
  
		if (likely(NCSI_RECEIVE_RES_PACKET_OK == res))
		{
			/* add the IID */
			ncsi_if->cmd_iid++;
  
			if (0xFF == ncsi_if->cmd_iid)
			{
				ncsi_if->cmd_iid = 1;
			}
			
			break;
		}
		else
		{
			
			if (*res_skb)
			{
				PDEBUG( "NCSI Free wrong res_skb = 0x%lX\n", *res_skb);
  
				dev_kfree_skb(*res_skb);
				*res_skb = NULL;
			}
          
			/* faid to get the response packet, do the next retry */		
			continue;
		}
  } 
	while (++retrycnt < MAX_CMD_RETRY_COUNT);	/* 2 retry */
    
   
    
    
	return res;
}


int ncsi_open(sNCSI_IF_Info *ncsi_if)
{
    int err=0;
    PDEBUG(KERN_ALERT "ncsi_open sonata 20090115 new!!\n");
    ncsi_if->ncsi_hardware_arb = CMD_SELECT_PACKAGE_PAYLOAD_HW_ARB_ENABLE;
    //err = ncsi_cmd_select_package(ncsi_if); 		
    //err = ncsi_cmd_reset_channel(ncsi_if);

    ncsi_if->chnl_id = 0;
    err = ncsi_cmd_clear_initial_state(ncsi_if);
    err = ncsi_cmd_get_capabilities(ncsi_if);
    memcpy((u8 *)ncsi_if->ncsi_mac_info.mac_addr, (u8 *)ncsi_if->dev->dev_addr, 6);
    ncsi_if->ncsi_mac_info.mac_num = 1;
    ncsi_if->ncsi_mac_info.addr_type_enable = CMD_SET_MAC_ADDRESS_PAYLOAD_ADDR_TYPE_UNICAST | CMD_SET_MAC_ADDRESS_PAYLOAD_ENABLE_MAC_ADDR_FILTER;
    err = ncsi_cmd_set_mac_address(ncsi_if);
    ncsi_if->ncsi_chnl_parameters.bcast_filter_flag = ncsi_if->ncsi_chnl_caps.bcast_filter_flag;
    err = ncsi_cmd_enable_broadcast_filtering(ncsi_if);
    
    err = ncsi_cmd_enable_global_multicast_filtering(ncsi_if);

    /* NCSI AEN control */	
    if(ncsi_if->ncsi_chnl_caps.aen_ctrl != 0)
    {
        ncsi_if->ncsi_aen_ctrl = ncsi_if->ncsi_chnl_caps.aen_ctrl;
        /* Enable AEN */
        err = ncsi_cmd_aen_enable(ncsi_if);	
    }

    ncsi_if->ncsi_vlan_mode = VLAN_PROMISCUOUS_MODE;
    err = ncsi_cmd_enable_vlan(ncsi_if);

    err = ncsi_cmd_enable_channel_network_tx(ncsi_if); //we only enable the first function by default 
    err = ncsi_cmd_enable_channel(ncsi_if);

    return err;
}


//int ncsi_open(sNCSI_IF_Info *ncsi_if, u8 chnl_id)
//{
//	int err = 0;
//	int i = 0;
//	int internal_chnl_id = 0;
//    u8 original_chnl_id = 0;
//   
//	/* 1. Discover Package */
//	/* Use the internal channel id (0x1F) to select the package */
//	//ncsi_if->chnl_id = CHNL_ID(PKG_ID(chnl_id), NCSI_ENTIRE_PKG_INTERNAL_CHNL_ID);
//    //Due to Broadcom NIC Firmware issue, the channel id cannot use 0x1f (from Bill)  
//    
////    ncsi_if->chnl_id = CHNL_ID(PKG_ID(chnl_id), 0);
////    
//   // err = ncsi_cmd_select_package(ncsi_if);
//   // if (err != 0)
//   // {
//   //	PDEBUG("NCSI Network Controller Fail\n");
//   //	return err;
//   // }
//   
//    if(ncsi_if->chnl_id == CHNL_ID(1, 0)) //To detect if the second package exist or not
//    {   
//    	ncsi_if->ncsi_hardware_arb = CMD_SELECT_PACKAGE_PAYLOAD_HW_ARB_ENABLE;
//	    err = ncsi_cmd_select_package(ncsi_if); 
//        if (err != 0)
//        {
//        	ncsi_if->second_lom_exist = 0;
//            PDEBUG("Only one package here!!! \n\n");
//   	        return 0;
//        }
//        ncsi_if->second_lom_exist = 1;
//        printk("Have second package!!! \n\n");
//    }   
//    
//	/* 2. Discover and Get Capabilities for each Channel in the Package */
//	/* Use the real channel id to clear initial state */
//    original_chnl_id = ncsi_if->chnl_id;
//    
//for(internal_chnl_id = 0; internal_chnl_id < 2; internal_chnl_id ++)
//{ 
//	    ncsi_if->chnl_id = CHNL_ID(PKG_ID(ncsi_if->chnl_id), internal_chnl_id);
//	            
//	    err = ncsi_cmd_reset_channel(ncsi_if);
//	    if( (err != 0) && (err != NCSI_NEED_REINIT) )
//	    {
//	    	return err;
//	    }	
//	   
//	    err = ncsi_cmd_clear_initial_state(ncsi_if);
//	    if (err != 0)
//	    {
//     		PDEBUG("NCSI Network Controller Fail\n");
//		    return err;
//      }
//     
//    
//	    err = ncsi_cmd_get_capabilities(ncsi_if);
//	    if (err != 0)
//	    {
//		      PDEBUG("NCSI Network Controller Fail\n");
//		      return err;
//      }  
//    
//	    /* 3. Initialize each Channel in the Package */
//	    /* TBD: support only one MAC addr filter now. Maybe can expand */
//      memcpy((u8 *)ncsi_if->ncsi_mac_info.mac_addr, (u8 *)ncsi_if->dev->dev_addr, 6);
//	    ncsi_if->ncsi_mac_info.mac_num = 1;
//	    ncsi_if->ncsi_mac_info.addr_type_enable = CMD_SET_MAC_ADDRESS_PAYLOAD_ADDR_TYPE_UNICAST | CMD_SET_MAC_ADDRESS_PAYLOAD_ENABLE_MAC_ADDR_FILTER;
//	    err = ncsi_cmd_set_mac_address(ncsi_if);
//	    if (err != 0)
//	    {
//		     PDEBUG("NCSI Network Controller Fail\n");
//		     return err;
//      }  
//    
//	    ncsi_if->ncsi_chnl_parameters.bcast_filter_flag = ncsi_if->ncsi_chnl_caps.bcast_filter_flag;
//	    err = ncsi_cmd_enable_broadcast_filtering(ncsi_if);
//	    if (err != 0)
//	{
//		PDEBUG("NCSI Network Controller Fail\n");
//		return err;
//    }  
//	
//	ncsi_if->ncsi_vlan_mode = VLAN_PROMISCUOUS_MODE;
//	err = ncsi_cmd_enable_vlan(ncsi_if);
//	if (err != 0)
//	{
//		PDEBUG("NCSI Network Controller Fail\n");
//		return err;
//    }  
//    
//	/* NCSI AEN control */	
//	if(ncsi_if->ncsi_chnl_caps.aen_ctrl != 0)
//	{
//	    ncsi_if->ncsi_aen_ctrl = ncsi_if->ncsi_chnl_caps.aen_ctrl;
//		/* Enable AEN */
//		err = ncsi_cmd_aen_enable(ncsi_if);
//		if (err != 0)
//	    {
//		    PDEBUG("NCSI Network Controller Fail\n");
//		    return err;
//        } 
//	}
//	
//}
//    ncsi_if->chnl_id = original_chnl_id;
//    
//    if( ncsi_if->chnl_id == 0) 
//	{       
//        err = ncsi_cmd_enable_channel_network_tx(ncsi_if); //we only enable the first function by default 
//        if (err != 0)
//	    {
//		    PDEBUG("NCSI Network Controller Fail\n");
//		    return err;
//        }  
//    }
//	/* 4. Enable Hardware Arbitration for the Package */
//	//ncsi_if->ncsi_hardware_arb = CMD_SELECT_PACKAGE_PAYLOAD_HW_ARB_ENABLE;
//	//ncsi_cmd_select_package(ncsi_if);	
//    
//for(internal_chnl_id = 0; internal_chnl_id < 2; internal_chnl_id ++)
//{     
//	ncsi_if->chnl_id = CHNL_ID(PKG_ID(ncsi_if->chnl_id), internal_chnl_id);
//	/* 5. Start Pass-through packet and AEN operation on the channels */
//	err = ncsi_cmd_enable_channel(ncsi_if);
//    if (err != 0)
//	{
//		PDEBUG("NCSI Network Controller Fail\n");
//		return err;
//    }  
//    
//    err = ncsi_cmd_get_link_status(ncsi_if);
//    if (err != 0)
//	{
//		PDEBUG("NCSI Network Controller Fail\n");
//		return err;
//    }  
//    if((ncsi_if->ncsi_link_status.link_status)&0x00000001) //Initial carrier status for every channel 
//    {   
//    	PDEBUG(" channel %x is carrier on \n\n",ncsi_if->chnl_id);
//    	netif_carrier_on(ncsi_if->ncsi_slave_dev[PKG_ID(ncsi_if->chnl_id)*2 +INTERNAL_CHNL_ID(ncsi_if->chnl_id)]);
//    }
//    else
//    {    
//    	PDEBUG(" channel %x is carrier off \n\n",ncsi_if->chnl_id);
//    	netif_carrier_off(ncsi_if->ncsi_slave_dev[PKG_ID(ncsi_if->chnl_id)*2 +INTERNAL_CHNL_ID(ncsi_if->chnl_id)]);
//    }		
//    
//}   
//
//   ncsi_if->chnl_id = original_chnl_id;
//   
//if(ncsi_if->chnl_id == CHNL_ID(1, 0))
//{
//	ncsi_cmd_disable_channel_network_tx(ncsi_if);
//}	
//   
//	return err;
//}

int ncsi_close(sNCSI_IF_Info  *ncsi_if)
{
	int err;
	

	skb_queue_purge(&ncsi_if->ncsi_res_quene.ncsi_queue);
	skb_queue_purge(&ncsi_if->ncsi_aen_quene.ncsi_queue);

	return 0;
}

void ncsi_init(sNCSI_IF_Info *ncsi_if)
{
	PDEBUG("Enter ncsi_init HZ = %ld\n", HZ);
	ncsi_if->cmd_iid = 1;	/* the comand IID begins with 1 */

	init_MUTEX(&ncsi_if->ncsi_res_quene.sem);
	skb_queue_head_init(&ncsi_if->ncsi_res_quene.ncsi_queue);
	init_waitqueue_head(&ncsi_if->ncsi_res_quene.inq);
	
}




 
static int ncsi_recvmsg(struct kiocb *iocb, struct socket *sock,
			  struct msghdr *msg, size_t len, int flags)
{
	struct sk_buff *skb=NULL;
	int port_number, wait_time;

//	printk(KERN_ALERT "recvmsg() request received from user mode, total length = %d\n", len);
	

	len = len - sizeof(int) - sizeof(int *);	//subtract the length of port_number and wait_time variables
/*	
	
*/	

/* First, get the wait_time parameter */	
	if ( get_user(wait_time, (int *)msg->msg_iov->iov_base) )
	    return STATUS_FAIL;
	
//	printk(KERN_ALERT "wait_time value seen by kernel is %d\n", wait_time);
	    
/* Next, go to the next iovec structure to get the receive buffer */
    msg->msg_iov++;	   
	
	if(!access_ok(VERIFY_WRITE, msg->msg_iov->iov_base, len) )
	{		
		/* error occur */
		return STATUS_FAIL;
	}
	
	while ( 0 == atomic_read(&ncsi_queued_depth) )
	{
	    if (wait_event_interruptible_timeout(inq,
                                            (0!=atomic_read(&ncsi_queued_depth)),
											wait_time*HZ/1000  ) < 0)
				return STATUS_FAIL;
	    
	    if ( 0 == (atomic_read(&ncsi_queued_depth)) )
	    {
		    PDEBUG(KERN_ALERT "NCSI queue is currently empty\n");
		    return STATUS_NO_DATA;
	    }
	}
	
	skb=skb_dequeue(&ncsi_queue);
	if(NULL == skb)
	{
		PDEBUG(KERN_ALERT "NCSI queue is currently empty\n");
		return STATUS_NO_DATA;
	}
	
	atomic_dec(&ncsi_queued_depth); //A sk_buff structure is successfully retrieved
//	PDEBUG(KERN_ALERT "Current NCSI queue depth = %d\n", atomic_read(&ncsi_queued_depth));
	
	
	if (skb->len >= len)
	{
		if (copy_to_user(msg->msg_iov->iov_base, (void *)skb->data, len))
		{ 
			if(skb)
			{
			    kfree_skb(skb);
			    skb = NULL;
			}
			return STATUS_FAIL;
		}
	}
	else
	{				
		if (copy_to_user(msg->msg_iov->iov_base, (void *)skb->data, skb->len))
		{
			if(skb)
			{
			    kfree_skb(skb);
			    skb = NULL;
			}    
			return STATUS_FAIL;
		}
	}

/* Finally, go to the last iovect structure to return receive port info */	
	msg->msg_iov++;
	if ( 0 == strcmp("eth0", skb->dev->name) )
	{
		port_number = 0;	////Packet comes from BMC MAC 0
		if ( put_user(port_number, (int *)msg->msg_iov->iov_base) )
		{ 
			if(skb)
			{
			    kfree_skb(skb);
			    skb = NULL;
			}
			return STATUS_FAIL;			
		}
	}
	else if ( 0 == strcmp("eth1", skb->dev->name) )
	{
		port_number = 1;	////Packet comes from BMC MAC 1
		if ( put_user(port_number, (int *)msg->msg_iov->iov_base) )
		{ 
			if(skb)
	    { 
			    kfree_skb(skb);
			    skb = NULL;
			}    
			return STATUS_FAIL;			
		}
	}
	
	if(skb)
	{
	    kfree_skb(skb);
	    skb = NULL;
	}    
	return STATUS_OK;
}

/*
 *	Send a NCSI packet out to the NIC.
 */

static int ncsi_sendmsg(struct kiocb *iocb, struct socket *sock,
			  struct msghdr *msg, size_t len)
{
	struct net_device *dev;
	struct sk_buff *skb = NULL;
	int port_number;
	int err;
	
	 /*
	 *	Check the flags. 
	 */
	PDEBUG(KERN_ALERT "sendmsg() request received from user mode app\n");

	if ( get_user(port_number, (int *)msg->msg_iov->iov_base) )
		return STATUS_FAIL;
	
	if ( 0 == port_number )   // User specified BMC MAC 0
	{
		if (NULL == ncsi_dev[0])
			return STATUS_DEVICE_NO_EXIST;
			
		dev = ncsi_dev[0];
		
		if (!(dev->flags & IFF_UP))
			return STATUS_DEVICE_DOWN;
	}
	else if  (1 == port_number )  // User specified BMC MAC 1
	{
		if (NULL == ncsi_dev[1])
			return STATUS_DEVICE_NO_EXIST;
			
		dev = ncsi_dev[1];
		
		if (!(dev->flags & IFF_UP))
			return STATUS_DEVICE_DOWN;
	}
	else
		return STATUS_INVALID_PORT;
		
	msg->msg_iov++;
		
	if(!access_ok(VERIFY_READ, msg->msg_iov->iov_base, ETH_ZLEN))
	{		
		/* error occur */
		return STATUS_FAIL;
	}
	
	skb = alloc_skb(ETH_ZLEN + NET_IP_ALIGN, GFP_KERNEL);
	
	if (NULL == skb)
		return STATUS_FAIL;
		
	memset(skb->data, 0, ETH_ZLEN + NET_IP_ALIGN);		
	skb->dev = dev;
	skb->len = ETH_ZLEN;
	skb_reserve(skb, NET_IP_ALIGN);
	skb->protocol = NCSI_PROTOCOL;
	skb->priority = 0;
	
	err = copy_from_user(skb->data, msg->msg_iov->iov_base, ETH_ZLEN);
	if (err)
	{
		if(skb)
		{
		    kfree_skb(skb);
		    skb = NULL;
		}
		return STATUS_FAIL;
	}
	
//	printk(KERN_ALERT "Before calling dev_queue_xmit(), packet length = %d\n",skb->len);
	
	if ( 0 == dev_queue_xmit(skb) )
		return STATUS_OK;
	else
		return STATUS_FAIL;
}

/*
 *	Close a NCSI socket.
 */
static int ncsi_release(struct socket *sock)
{
//	printk(KERN_ALERT "ncsi_release() function called\n");
	
	return 0;
}

/*
 *	Create a NCSI socket
 */
static int ncsi_create(struct socket *sock, int protocol)
{
//	printk(KERN_ALERT "ncsi_create() socket create function called\n");
	sock->ops = &ncsi_ops;

	return(0);
}


static struct net_proto_family ncsi_family_ops = {
	.family =	PF_NCSI,
	.create =	ncsi_create,
	.owner	=	THIS_MODULE,
};

static const struct proto_ops ncsi_ops = {
	.family =	PF_NCSI,
	.owner =	THIS_MODULE,
	.release =	ncsi_release,
	.bind =		sock_no_bind,
	.connect =	sock_no_connect,
	.socketpair =	sock_no_socketpair,
	.accept =	sock_no_accept,
	.getname =	sock_no_getname, 
	.poll =		sock_no_poll,
	.ioctl =	sock_no_ioctl,
	.listen =	sock_no_listen,
	.shutdown =	sock_no_shutdown,
	.setsockopt =	sock_no_setsockopt,
	.getsockopt =	sock_no_getsockopt,
	.sendmsg =	ncsi_sendmsg,
	.recvmsg =	ncsi_recvmsg,
	.mmap =		sock_no_mmap,
	.sendpage =	sock_no_sendpage,
};

/*
 *	Received a NCSI packet from the NIC.
 */
static int ncsi_rcv(struct sk_buff *skb, struct net_device *dev, struct packet_type *pt, struct net_device *orig_dev)
{
	int i;
	struct sNCSI_RES_HDR *ncsi_reshdr;
	struct sNCSI_HDR *ncsi_hdr;
	
	PDEBUG(KERN_ALERT "ncsi_rcv()20090114: A NCSI packet is received from the NIC\n");
	
	
	for(i=0 ; i<skb->len ; i++)
	{
		PDEBUG(KERN_ALERT "Rx: skb->data[%d] = %x \n",i,skb->data[i]);
  }	
	PDEBUG(KERN_ALERT "ncsi_rcv()20090115: Write the response packets into Queue!! \n");
	
//	ncsi_queue_write(&ncsi_if, skb, NCSI_RES_QUEUE); 
	
	
	
	
		    ncsi_reshdr = NCSI_RES_HEADER(skb);
				ncsi_hdr = &(ncsi_reshdr->ncsi_hdr);
				
				if (IS_NCSI_GOOD_PACKET(ncsi_hdr))
				{
					if (IS_NCSI_RES_PACKET(ncsi_hdr))
					{
						PDEBUG(KERN_ALERT "SUCCESS to receive the response packet!! \n");
						
						ncsi_rx_response_packet(&ncsi_if, skb);
					}
					else if (IS_NCSI_AEN_PACKET(ncsi_hdr))
					{
						PDEBUG(KERN_ALERT "SUCCESS to receive the AEN packet!! \n");
						
						ncsi_rx_aen_packet(&ncsi_if, skb);
					}
					else
					{   
						    if(skb != NULL)
				        {   
				        	 
				                dev_kfree_skb(skb);
				                skb = NULL;
				               
				        }  
						    /* drop packet directly */
						    PDEBUG(KERN_ALERT  "unknown NC-SI packet!\n");
					}
				}
				else
				{   
					  if(skb != NULL)
				    {
				        dev_kfree_skb(skb);
				        skb = NULL;
				    }  
					/* drop packet directly */
					  PDEBUG(KERN_ALERT "unknown NC-SI packet!\n");
				}
	
	//if ( (atomic_read(&ncsi_queued_depth)) >= MAX_NCSI_QUEUE_DEPTH )
	//{
	//	printk(KERN_ALERT "Error! Maximum NCSI queue depth of %d reached, dropping this packet\n",MAX_NCSI_QUEUE_DEPTH);
	//	kfree_skb(skb);
	//	return NET_RX_DROP;
	//}

	//skb_queue_tail(&ncsi_queue, skb);
	//atomic_inc(&ncsi_queued_depth);
//printk(KERN_ALERT "Current NCSI queue depth = %d\n", atomic_read(&ncsi_queued_depth));
	//wake_up_interruptible(&inq);
	
	return 0;
}

static struct packet_type ncsi_packet_type = {
	.type =		__constant_htons(NCSI_PROTOCOL),
	.func =		ncsi_rcv,
};


static int ncsi_notifier(struct notifier_block *this, unsigned long msg, void *data)
{
	struct net_device *dev = (struct net_device *)data;
      struct ftgmac100_local *lp = (struct ftgmac100_local *)dev->priv;

	int err;

//	printk(KERN_ALERT "ncsi_notifier() called on %s, msg=%d\n", dev->name,msg);
	
	switch (msg)
	{
		case NETDEV_REGISTER:  /* A new network device has appeared */
            
            if (NULL == lp)
            {
                break;
            }

                if (lp->mac_intf & MAC_INTF_NCSI)
                {
                    PDEBUG(KERN_ALERT "ncsi_notifier: eth%d device registered\n", lp->mac_num);
                    ncsi_dev[lp->mac_num] = dev;
                }
			
                break;
			
		case NETDEV_UP:

                if (NULL == lp)
                {
                    break;
                }

                if (lp->mac_intf & MAC_INTF_NCSI)
                {
                    PDEBUG(KERN_ALERT "ncsi_notifier: eth%d is NCSI device!!\n", lp->mac_num);
                    ncsi_if.dev = dev;

                    netif_carrier_on(dev);    

                    if (0 == S_isNCSIInited)
                    {
                      
                        ncsi_open(&ncsi_if);
                        
                        S_isNCSIInited = 1;
                    }
                    
                    printk("Begin to get the link status\n");
                    ncsi_cmd_get_link_status(&ncsi_if);
                }

                break;
		
		case NETDEV_DOWN:	

                if (NULL == lp)
                {
                    break;
                }

                if (lp->mac_intf & MAC_INTF_NCSI)
                {
                    PDEBUG(KERN_ALERT "ncsi_notifier: eth%d device downed\n", lp->mac_num);
                }
                
                break;
			
		case NETDEV_UNREGISTER:  /* A network device is being unloaded */

                if (NULL == lp)
                {
                    break;
                }

                if (lp->mac_intf & MAC_INTF_NCSI)
                {
                    PDEBUG(KERN_ALERT "ncsi_notifier: eth%d device unregistered\n", lp->mac_num);
                    ncsi_dev[lp->mac_num] = NULL;
                    ncsi_if.dev = NULL;
                    
                    S_isNCSIInited = 0;
                }
			
                break;			
	}

	return NOTIFY_DONE;
}

static struct notifier_block ncsi_netdev_notifier = {
	.notifier_call =ncsi_notifier,
};

static void __exit ncsi_proto_exit(void)
{
	PDEBUG(KERN_ALERT "ncsi_proto_exit() called\n");
	
	skb_queue_purge(&ncsi_queue);
	unregister_netdevice_notifier(&ncsi_netdev_notifier);
	dev_remove_pack(&ncsi_packet_type);
	sock_unregister(ncsi_family_ops.family);
}

static int __init ncsi_proto_init(void)
{
	PDEBUG(KERN_ALERT "ncsi_proto_init() called\n");
	
	skb_queue_head_init(&ncsi_queue);
	init_waitqueue_head(&inq);
	if ( sock_register(&ncsi_family_ops) != 0)
	{
		PDEBUG(KERN_ALERT "sock_register() failed\n");
		return -1;
	}
	
	ncsi_init(&ncsi_if); 
	dev_add_pack(&ncsi_packet_type);
	if (register_netdevice_notifier(&ncsi_netdev_notifier) != 0)
	{
		PDEBUG(KERN_ALERT "register_netdevice_notifier() failed\n");
		return -1;
	}
	
	return 0;
}

module_init(ncsi_proto_init);
module_exit(ncsi_proto_exit);
MODULE_LICENSE("GPL");
