/**
 * $Id: aess_memdrv.c,v 1.0 $
 *
 *<center>
 *               Avocent Corporation. Proprietary Information.
 * \n<em>
 *      This software is supplied under the terms of a license agreement or
 *      nondisclosure agreement with Avocent Corporation, or its subsidiary, and
 *      may not be copied, disseminated, or distributed except in accordance
 *      with the terms of that agreement.
 *
 *      2001 Gateway Place, Suite 520W, San Jose, California, 95110 U.S.A.
 *\n
 *                  US phone: 408.436.6333
 *
 *        Copyright &copy; 2008 Avocent Corporation.
 *</em> </center>
 *----------------------------------------------------------------------------\n
 *  MODULES     linux driver
 *----------------------------------------------------------------------------\n
 *  @file   aess_memdrv.c
 *  @brief  provide a function for user or kener driver to access a specific memory
 *
 *  @internal
 *  $RCSfile: aess_memdrv.c,v $
 *  $Revision: 1.0 $
 *  $Date: 2009/03/06 17:12:00 $
 *  $Author: Ken Lin $
 *  $Source:  $
 *  $Name:  $
 *----------------------------------------------------------------------------*/
/******************************************************************************
* Content
* ----------------
*   aess_mem_open()         - Open memory device.
*   aess_mem_close()        - Close memory device.
*   aess_mem_read()         - read data from specific address.
*   aess_mem_write()        - write data to specific address.
*   aess_mem_request()      - ioremap a physical address to virtual address.
*   aess_mem_release()      - iounmap virtual address request before.
*   aess_mem_ioctl()        - The I/O control function.
*   aess_mem_init()         - insmod module to kernel will call this function.
*   aess_mem_exit()         - rmmod module from kernel will call this function.
******************************************************************************/
#define AESSMEMDRV_C

#include <linux/module.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/signal.h>
#include <linux/spinlock.h>
#include <linux/cdev.h>
#include <linux/kernel.h>
#include <linux/aess_debugfs.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/ioctl.h>
#include <asm/memory.h>
#include "aess_memdrv.h"

/* driver name will passed by insmod command */
static char *driver_name="aess_memdrv"; 

/* Init flag */
static UINT8 S_u8init_flag = 0;

/* Used for protect write command */
struct rw_semaphore mem_drv_rw_sem;

/* Mutex to protect memory driver information structure */
static struct semaphore mem_drv_mutex;

dev_t mem_dev;
struct cdev *mem_cdev;

/******************************************************************************
*   STRUCT      :   mem_info_type
******************************************************************************/
/**
 *  @brief   Structure to memory driver related data parameter.
 *
 *****************************************************************************/
typedef struct 
{
    /** keep file structure */
    UINT32 u32Filp;         
    
    /** virtual address */
    UINT32 u32VirtAddr; 
    
    /** max valid offset */   
    UINT16 u16MaxOffset; 
    
    /** reserve */   
    UINT16 u16reserve;
    
    /** status */      
    UINT8  u8Status;  
    
    /** reserve */      
    UINT8  u8reserve1; 
    
    /** reserve */     
    UINT8  u8reserve2;      
    
    /** reserve */     
    UINT8  u8reserve3;      
} mem_info_type;

/* memory driver related information */
static mem_info_type mem_info[MEM_DRV_MAX_REGION_NUM];


/******************************************************************************
*   FUNCTION        :   aess_mem_open
******************************************************************************/
/**
 *  @brief      Open memory device. 
 *
 *  @return     None
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Private inline function\n
 *
 *****************************************************************************/
static int aess_mem_open (

                /** device node */
                struct inode *inode, 
                
                /** file pointer */
                struct file *filp
                
                                    )
{
    DRV_INFO("aess_mem_open\n");
    
    if (S_u8init_flag == false)
    {
        DRV_INFO("start init aess_mem_driver\n");
	    
        /* clear memory driver structure */
        memset(mem_info, 0, MEM_DRV_MAX_REGION_NUM * sizeof(mem_info_type));
	    
        init_rwsem(&mem_drv_rw_sem);
	    
        init_MUTEX(&mem_drv_mutex);
		
        S_u8init_flag = true;
		
        DRV_INFO("finish init aess_mem_driver\n");
    }
	
    return 0;
}


/******************************************************************************
*   FUNCTION        :   aess_mem_close
******************************************************************************/
/**
 *  @brief      Close memory device. 
 *
 *  @return     None
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Private inline function\n
 *
 *****************************************************************************/
static int aess_mem_close(

                /** device node */
                struct inode* inode, 
                
                /** file pointer */
                struct file *filp
                
                                    )
{
    int i;
    
    DRV_INFO("aess_mem_close\n");
	
    /* get semaphore */
    if (down_interruptible(&mem_drv_mutex))
    {
        DRV_ERROR("aess_mem_close get semaphore fail\n");
        return -ERESTARTSYS;
    }
	
    /* release resources that relate to this file */
    for (i = 0; i < MEM_DRV_MAX_REGION_NUM; i++)
    {
        if (mem_info[i].u32Filp == (UINT32) filp)
        {
            DRV_INFO("mem_info[%d] released\n", i);
	        
            iounmap((void *) mem_info[i].u32VirtAddr);
	        
            mem_info[i].u32VirtAddr = 0;
            mem_info[i].u32Filp = 0;
            mem_info[i].u16MaxOffset = 0;
            mem_info[i].u8Status = 0;
        }
    }
    
    /* release semaphore */
    up(&mem_drv_mutex);
	
    return 0;
}


/******************************************************************************
*   FUNCTION        :   aess_mem_read
******************************************************************************/
/**
 *  @brief      read process, read data with specific size from specific address. 
 *
 *  @return     None
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Private inline function\n
 *
 *****************************************************************************/
static int aess_mem_read(

                /** get parameter from user space */
                sMemDrvInfoType *psMemDrvInfo
                
                                                )
{
    UINT16  u16i;
	
    UINT8  *u8ByteBuff;
    UINT16 *u16WordBuff;
    UINT32 *u32DWordBuff;
	
    void *pSrcAddr;
	
    DRV_INFO("aess_mem_read\n");
    
    if ((psMemDrvInfo->u8ID >= MEM_DRV_MAX_REGION_NUM) ||
            (mem_info[psMemDrvInfo->u8ID].u8Status == 0))
    {
        DRV_ERROR("Read():specific address is not illegal \n");
        return -EINVAL;
    }
            
	
    /* check the offset */
    if ((psMemDrvInfo->u16Offset 
            + (psMemDrvInfo->u16DataSize << psMemDrvInfo->u8DataWidth)) 
            > mem_info[psMemDrvInfo->u8ID].u16MaxOffset)
    {
        DRV_ERROR("Read():Data buffer is out of range \n");
        return -EINVAL;
    }
	
    /* calculate the virtual address of source */
    pSrcAddr = (void *) (mem_info[psMemDrvInfo->u8ID].u32VirtAddr 
                    + psMemDrvInfo->u16Offset);
	
    DRV_INFO("Read(): u8ID %d \n", psMemDrvInfo->u8ID);
    DRV_MSG("Read(): u16Offset 0x%x \n", psMemDrvInfo->u16Offset);
    DRV_MSG("Read(): u16DataSize %d \n", psMemDrvInfo->u16DataSize);
    DRV_MSG("Read(): u8DataWidth %d \n", psMemDrvInfo->u8DataWidth);
	
    switch(psMemDrvInfo->u8DataWidth)
    {
        case MEM_DRV_BYTE:
		    
            u8ByteBuff = (UINT8 *) psMemDrvInfo->pDataPtr;
			
            for (u16i = 0; u16i < psMemDrvInfo->u16DataSize; u16i++)
            {
                *(u8ByteBuff + u16i) = ioread8(pSrcAddr + u16i);
            }
			
            break;
			
        case MEM_DRV_WORD:
		    
            u16WordBuff = (UINT16 *) psMemDrvInfo->pDataPtr;
			
            for (u16i = 0; u16i < psMemDrvInfo->u16DataSize; u16i++)
            {
                *(u16WordBuff + u16i) = ioread16(pSrcAddr + (u16i << 1));
            }
            break;
			
        case MEM_DRV_DWORD:
		    
            u32DWordBuff = (UINT32 *)psMemDrvInfo->pDataPtr;
			
            for (u16i = 0; u16i < psMemDrvInfo->u16DataSize; u16i++)
            {
                *(u32DWordBuff + u16i) = ioread32(pSrcAddr + (u16i << 2));
            }
			
        break;
			
        default:
            DRV_ERROR("Read():Data width error \n");
            return -EINVAL;
    }
	
    return 0;
}


/******************************************************************************
*   FUNCTION        :   aess_mem_write
******************************************************************************/
/**
 *  @brief      write process, write data with specific size to specific address. 
 *
 *  @return     None
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Private inline function\n
 *
 *****************************************************************************/
static int aess_mem_write(

                /** get parameter from user space */
                sMemDrvInfoType *psMemDrvInfo
                                                
                                                )
{
    UINT16  u16i;
	
    UINT8  *u8ByteBuff;
    UINT16 *u16WordBuff;
    UINT32 *u32DWordBuff;
	
    void *pDestAddr;
    DRV_INFO("aess_mem_write\n");
	
    if ((psMemDrvInfo->u8ID >= MEM_DRV_MAX_REGION_NUM) ||
            (mem_info[psMemDrvInfo->u8ID].u8Status == 0))
    {
        DRV_ERROR("Wri():specific address is not illegal \n");
        return -EINVAL;
    }
            
	
    /* check the offset */
    if ((psMemDrvInfo->u16Offset 
            + (psMemDrvInfo->u16DataSize << psMemDrvInfo->u8DataWidth)) 
            > mem_info[psMemDrvInfo->u8ID].u16MaxOffset)
    {
        DRV_ERROR("Write():Data buffer is out of range \n");
        return -EINVAL;
    }
	
    /* calculate the virtual address of destination */
    pDestAddr = (void *) (mem_info[psMemDrvInfo->u8ID].u32VirtAddr 
                        + psMemDrvInfo->u16Offset);
	
    DRV_INFO("Write(): u8ID %d \n", psMemDrvInfo->u8ID);
    DRV_MSG("Write(): u16Offset 0x%x \n", psMemDrvInfo->u16Offset);
    DRV_MSG("Write(): u16DataSize %d \n", psMemDrvInfo->u16DataSize);
    DRV_MSG("Write(): u8DataWidth %d \n", psMemDrvInfo->u8DataWidth);
	
    switch(psMemDrvInfo->u8DataWidth)
    {
        case MEM_DRV_BYTE:
		    
            u8ByteBuff = (UINT8 *) psMemDrvInfo->pDataPtr;
			
			
            for (u16i = 0; u16i < psMemDrvInfo->u16DataSize; u16i++)
            {
                iowrite8(*(u8ByteBuff + u16i), pDestAddr + u16i);
            }
			
            break;
			
        case MEM_DRV_WORD:
		    
            u16WordBuff = (UINT16 *) psMemDrvInfo->pDataPtr;
			
            for (u16i = 0; u16i < psMemDrvInfo->u16DataSize; u16i++)
            {
                iowrite16(*(u16WordBuff + (u16i)), pDestAddr + (u16i << 1));
            }
			
            break;
			
        case MEM_DRV_DWORD:
		    
            u32DWordBuff = (UINT32 *) psMemDrvInfo->pDataPtr;
			
            for (u16i = 0; u16i < psMemDrvInfo->u16DataSize; u16i++)
            {
                iowrite32(*(u32DWordBuff + (u16i)), pDestAddr + (u16i << 2));
            }
			
            break;
			
        default:
		    
            DRV_ERROR("Write():Data width error \n");
			
            return -EINVAL;
    }
	
    return 0;
}


/******************************************************************************
*   FUNCTION        :   aess_mem_request
******************************************************************************/
/**
 *  @brief      request process, ioremap a physical address to a virtual address. 
 *
 *  @return     None
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Private inline function\n
 *
 *****************************************************************************/
static int aess_mem_request(

                /** file pointer */
                struct file *filp, 
                
                /** get parameter from user space */
                sMemDrvInfoType *psMemDrvInfo
                
                                                )
{
    int i;
    
    DRV_INFO("aess_mem_request\n");
    
    for (i = 0; i < MEM_DRV_MAX_REGION_NUM; i++)
    {
        if (mem_info[i].u8Status == 0)
        {
            mem_info[i].u32VirtAddr 
                = (UINT32) ioremap(psMemDrvInfo->u32BaseAddr, 
                                    psMemDrvInfo->u16RegionSize);
            
            if (mem_info[i].u32VirtAddr == 0)
            {
                
                DRV_ERROR("Request():ioremap fail \n");
                return -ENOMEM;
            }
            
            psMemDrvInfo->u8ID = i;
            
            mem_info[i].u8Status = 1;
            mem_info[i].u16MaxOffset = psMemDrvInfo->u16RegionSize;
            mem_info[i].u32Filp = (UINT32) filp;
            
            DRV_INFO("\n%d: u32BaseAddr=%08x, u32VirtAddr=%08x, u16RegionSize=%x\n", 
                    i,
                    psMemDrvInfo->u32BaseAddr,
                    mem_info[i].u32VirtAddr,
                    psMemDrvInfo->u16RegionSize);
            
            return 0;
        }
    }
    
    DRV_ERROR("Request():ioremap fail, the request quota is full \n");
    return -ENOMEM;
}


/******************************************************************************
*   FUNCTION        :   aess_mem_release
******************************************************************************/
/**
 *  @brief      release process, iounmap a virtural address which request before. 
 *
 *  @return     None
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Private inline function\n
 *
 *****************************************************************************/
static int aess_mem_release(

                /** file pointer */
                struct file *filp, 
                
                /** get parameter from user space */
                sMemDrvInfoType *psMemDrvInfo
                
                                                )
{
    DRV_INFO("aess_mem_release\n");
    
    /* check if index is valid */
    if (psMemDrvInfo->u8ID >= MEM_DRV_MAX_REGION_NUM)
    {
        DRV_ERROR("Release(): u8ID out of range \n");
        return -EINVAL;
    }
    
    if (mem_info[psMemDrvInfo->u8ID].u8Status == 1)
    {
        DRV_INFO("mem_info[%d] released\n", psMemDrvInfo->u8ID);
        
        iounmap((void *) mem_info[psMemDrvInfo->u8ID].u32VirtAddr);
        
        mem_info[psMemDrvInfo->u8ID].u32VirtAddr = 0;
        mem_info[psMemDrvInfo->u8ID].u32Filp = 0;
        mem_info[psMemDrvInfo->u8ID].u16MaxOffset = 0;
        mem_info[psMemDrvInfo->u8ID].u8Status = 0;
    }
    else
    {
        /* release semaphore */
        DRV_ERROR("Release():give wrong u8ID, this array is empty \n");
        
        return -EINVAL;
    }
    
    return 0;
}


/******************************************************************************
*   FUNCTION        :   aess_mem_ioctl
******************************************************************************/
/**
 *  @brief      ioctl process, handle four command, request / read / write / release. 
 *
 *  @return     None
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Private inline function\n
 *
 *****************************************************************************/
static int aess_mem_ioctl(

                /** device node */
                struct inode * inode, 
                
                /** file pointer */
                struct file *filp, 
						  
                /** command */
						    unsigned int cmd, 
						    
                /** parameters */
						    unsigned long arg
						    
						                        )
{
    int err_check;
    sMemDrvInfoType *psMemDrvInfo = (sMemDrvInfoType *) arg;
    
    DRV_INFO("aess_mem_ioctl\n");
	
    /* Check the structure address is OK */
    if(!access_ok(VERIFY_WRITE, (void *) psMemDrvInfo, sizeof(sMemDrvInfoType)))
    {
        DRV_ERROR("ioctl(): struct access error\n");
        /* error occur */
        return -EFAULT;
    }
	
    switch(cmd)
    {
        case AESS_MEMDRV_READ:
		    
        /* Check the buffer address exist in the structure is OK */
        if(!access_ok(VERIFY_READ, psMemDrvInfo->pDataPtr, sizeof(UINT8)))
        {
            DRV_ERROR("ioctl(): read - buffer access error\n");
            /* error occur */
            return -EFAULT;
        }
		    
        down_read(&mem_drv_rw_sem);
        err_check = aess_mem_read(psMemDrvInfo);
        up_read(&mem_drv_rw_sem);
		    
        break;
		    
		case AESS_MEMDRV_WRITE:
		    
        /* Check the buffer address exist in the structure is OK */
        if(!access_ok(VERIFY_WRITE, psMemDrvInfo->pDataPtr, sizeof(UINT8)))
        {
            DRV_ERROR("ioctl(): write - buffer access error\n");
            /* error occur */
            return -EFAULT;
        }
		    
        down_write(&mem_drv_rw_sem);
        err_check = aess_mem_write(psMemDrvInfo); 
        up_write(&mem_drv_rw_sem);
		    
        break;
		    
		case AESS_MEMDRV_REQUEST:
		    
        /* get semaphore */
        if (down_interruptible(&mem_drv_mutex))
        {
            DRV_ERROR("ioctl(): request - get semaphore error \n");
            return -ERESTARTSYS;
        }
    
        err_check = aess_mem_request(filp, psMemDrvInfo);
		    
        /* release semaphore */
        up(&mem_drv_mutex);
    
        break;
		    
		case AESS_MEMDRV_RELEASE:
		    
        /* get semaphore */
        if (down_interruptible(&mem_drv_mutex))
        {
            DRV_ERROR("ioctl(): release - get semaphore error \n");
            return -ERESTARTSYS;
        }
    
        err_check = aess_mem_release(filp, psMemDrvInfo);
		    
        /* release semaphore */
        up(&mem_drv_mutex);
    
        break;
			
    default:
		    
        DRV_ERROR("aess_mem_ioctl, command error\n");
        err_check = -EINVAL;
    }
	
    /* 0->ok, minus->fail */
    return err_check;
}


struct file_operations aess_mem_fops = {
    .open = aess_mem_open, 
    .ioctl = aess_mem_ioctl,
    .release = aess_mem_close, 
};


/******************************************************************************
*   FUNCTION        :   aess_mem_init
******************************************************************************/
/**
 *  @brief      init process when insert this module to kernel. 
 *
 *  @return     None
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Private inline function\n
 *
 *****************************************************************************/
int __init aess_mem_init(void)
{
    int result;
    
    DRV_INFO("aess_mem_init\n");
    
    mem_cdev = cdev_alloc();
    mem_cdev->ops = &aess_mem_fops;
	
    result = alloc_chrdev_region(&mem_dev, 0, 1, driver_name);
    
    if (result < 0) 
    {
        DRV_ERROR ("Registering the character device failed with %d\n", 
                    MAJOR(mem_dev));
		
        return result;
    }
	
    cdev_add(mem_cdev, mem_dev, 1);
	
    /* show this message for memory driver install shell used */
    printk("mknod /dev/aess_memdrv c %d 0\n", MAJOR(mem_dev));
	
    aess_debugfs_default_create(driver_name,NULL,0,0);

    return 0;
}


/******************************************************************************
*   FUNCTION        :   aess_mem_exit
******************************************************************************/
/**
 *  @brief      exit process when remove this module from kernel. 
 *
 *  @return     None
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Private inline function\n
 *
 *****************************************************************************/
static void __exit aess_mem_exit(void)
{
    int i;
    
    DRV_INFO("aess_memdrv_exit\n");
	
    cdev_del(mem_cdev);
    unregister_chrdev_region(mem_dev, 1);
	
    for (i = 0; i < MEM_DRV_MAX_REGION_NUM; i++)
    {
        if (mem_info[i].u8Status == 1)
        {
            iounmap((void *) mem_info[i].u32VirtAddr);
            
            mem_info[i].u32VirtAddr = 0;
            mem_info[i].u32Filp = 0;
            mem_info[i].u16MaxOffset = 0;
            mem_info[i].u8Status = 0;
        }
    }
	
    aess_debugfs_remove(driver_name,NULL,0,0);  
    return;
}


MODULE_DESCRIPTION("AESS Memory Driver");
MODULE_AUTHOR("Ken Lin <ken.lin@avocent.com>");
MODULE_LICENSE("GPL");
MODULE_VERSION("1.0");
module_init(aess_mem_init);
module_exit(aess_mem_exit);
module_param(driver_name, charp, S_IRUGO);

/* End of code */ 
