/**
 * $Id$
 *
 *<center>
 *               Avocent Corporation. Proprietary Information.
 * \n<em>
 *      This software is supplied under the terms of a license agreement or
 *      nondisclosure agreement with Avocent Corporation, or  one of its 
 *      affiliates, andaffiliates, and may not be used, disseminated, or 
 *      distributed except in accordance with the terms of that agreement.
 *
 *      3541 Gateway Boulevard Fremont, San Jose, CA 94538 U.S.A.
 *\n
 *                  US phone: 510-771-6100
 *
 *        Copyright &copy; 2001-2009 Avocent Corporation. All rights reserved.
 *</em> </center>
 *----------------------------------------------------------------------------\n
 *  MODULES     Linux device Driver                                           \n
 *----------------------------------------------------------------------------\n
 *  @file   aess_pecisensordrv.c
 *  @brief  peci sensor driver 
 *  @internal
 *  $RCSfile: aess_pecisensordrv.c,v $
 *  $Revision: 1.3 $
 *  $Date: 2008/03/30 20:07:25 $
 *  $Author: Ken Lin $
 *  $Source: $
 *  $Name:  $
 *----------------------------------------------------------------------------*/
 /******************************************************************************
* Content
* ----------------
*   aess_pecisensor_command()      - This API support to issue PECI command.
*   aess_pecisensor_ioctl()        - Handl IOCTL commands
*   aess_pecisensor_open()         - Open peci sensor driver device node
*   aess_pecisensor_release()      - This function will be called when device node is closed. Set init flag to INIT_FAIL
*   aess_peci_isr()                - Interrupt service routine of PECI driver.
*   aess_pecisensor_init()         - Module init function. Initialize everything for PECI sensor driver.
*   aess_pecisensor_exit()         - Module exit function. Release everything for PECI sensor driver.
******************************************************************************/
#define AESSPECISENSORDRV_C
#include <linux/module.h>
#include <linux/types.h>
#include <asm/arch/hardware.h>
#include <asm/arch/irqs.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include <linux/slab.h>
#include <linux/signal.h>
#include <linux/spinlock.h>
#include <linux/cdev.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/delay.h>
#include <asm/arch/hw_regs.h>
#include <linux/aess_debugfs.h>
#include <linux/interrupt.h>
#include <asm/irq.h>
#include "aess_pecisensordrv.h"

volatile                UINT32 PECIDrv;
volatile                UINT8 WaitQueueFlag = 0;
#define IRQ_PECI        INT_PECI

/* initiate wait queue */
DECLARE_WAIT_QUEUE_HEAD (WaitQueue);

static u8   u8tmpcounter = 0;
static u16 u16tmpcounter = 0;
static u32 u32tmpcounter = 0;

/* driver name will passed by insmod command */
static char *driver_name = "aess_pecisensordrv"; 

/* Mutex to protect PECI driver information structure */
static struct semaphore peci_sem;
static dev_t dev;
struct cdev *pecisensor_cdev;
static UINT8 S_u8InitFlag = 0;
volatile    UINT32 u32PECI_INT_STS = 0 ;

/******************************************************************************
*   FUNCTION        :   aess_pecisensor_open
******************************************************************************/
/**
 *  @brief      Open PECI sensor device node
 *
 *  @return     0 for success.
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Static function\n
 *
 *****************************************************************************/
static int aess_pecisensor_open(
                                /**Node*/
                                struct inode *inode, 
                                
                                /**File descriptor*/
                                struct file *filp
                               )
{	
    UINT32 u32Data = 0; 

    DRV_MSG("Enter aess_pecisensor_open...\n");
    if (S_u8InitFlag == NOT_INIT)
    {
        DRV_MSG("aess_pecisensor_open, init register\n");
        DRV_INFO("SCU_SYSTEM_RESET_CONTROL = %08x\n", ioread32((void *) SCU_SYSTEM_RESET_CONTROL));

        /* set scu un_lock */
        iowrite32((UINT32)UN_LOCKKEY, (void *) SCU_PKR);

        /* enable the peci pin */
        u32Data = ioread32((void *) SCU_MULTI_FUNTION1_ADDR) | ENABLE_PECI_DEVICE;
        iowrite32(u32Data, (void *) SCU_MULTI_FUNTION1_ADDR);

        /*  stop to reset PECI controller */
        u32Data = ioread32((void *) SCU_SYSTEM_RESET_CONTROL) & ~RESET_PECI_CONTROLLER;
        iowrite32(u32Data, (void *) SCU_SYSTEM_RESET_CONTROL);

        /* set scu lock */
        iowrite32((UINT32)LOCKKEY, (void *) SCU_PKR);		 

        /* PECI Control Setting */
        u32Data = ENABLE_CLOCK + ENABLE_PECI 
                  + (CLOCK_DIVIDED_VALUE << PECI_CLOCK) 
                  + (MODE_SELECT << PECI_READ_MODE)
                  + (SAMPLING_POINT_SELECT << PECI_READ_SAMPLING);	

        iowrite32(u32Data, (void *) PECI_REG_CONTROL);
        DRV_INFO("aess_pecisensor_open PECI control reg write data:%08x\n", u32Data);

        /* PECI Timing Setting */
        u32Data = (Message_Timing << TIMING_OF_MESSAGE)
                  + (Address_Timing << TIMING_OF_ADDRESS);
        iowrite32(u32Data, (void *) PECI_REG_TIMING_NEGOTIATION);
        DRV_INFO("aess_pecisensor_open PECI timing negotiation write data:%08x\n", u32Data);

        /* PECI Negotiation Selection */
        u32Data = (Timing_Negotiation_Selection << PECI_TIMING_SELECTION);

        iowrite32(u32Data, (void *) PECI_REG_INTERRUPT);	  
        DRV_INFO("aess_pecisensor_open PECI interrupt reg write data:%08x\n", u32Data);

        S_u8InitFlag = INIT_DONE;
    }
    
    return (STATUS_OK);
}

/******************************************************************************
*   FUNCTION        :   aess_pecisensor_command
******************************************************************************/
/**
 *  @brief      This API support to issue PECI command.
 *
 *  @return     Return STATUS_OK if fucntion execute success.
 *              Return others if fucntion execute fail.
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Global function \n
 *
 *****************************************************************************/
static int aess_pecisensor_command(
                                    /** PECI data structure pointe */
                                    sPECICommandData *pPECICommandData
                                  )
{
    UINT8 u8Index = 0;
    UINT8 u8ClientAddr = 0;
    UINT8 u8ReadLength = 0;
    UINT8 u8WriteLength = 0;
    UINT8 u8CommandCode = 0;
    UINT32 u32Data = 0;
    
    DRV_MSG("Enter aess_pecisensor_command...\n");
    /* Check the buffer address exist in the structure is OK */
    if (!access_ok(VERIFY_READ, (void *) pPECICommandData, sizeof(sPECICommandData)))
    {
        DRV_ERROR("aess_pecisensor_command() buffer access error Exit!\n");
        /* error occur */
        return (-EFAULT);
    }

    if(down_interruptible(&peci_sem))
    {
        return (-ERESTARTSYS);
    }    

    u8ClientAddr = pPECICommandData->u8ClientAddress;
    u8ReadLength = pPECICommandData->u8ReadLength;
    u8WriteLength = pPECICommandData->u8WriteLength;
    u8CommandCode = pPECICommandData->u8CommandCode;

    DRV_MSG("u8ClientAddr:%x\n", u8ClientAddr);
    DRV_MSG("u8ReadLength:%x\n", u8ReadLength);
    DRV_MSG("u8WriteLength:%x\n", u8WriteLength);
    DRV_MSG("u8CommandCode:%x\n", u8CommandCode);

    if ((u8ReadLength > PECI_DATA_INOUT_REG_NUMBER) ||
        (u8WriteLength > PECI_DATA_INOUT_REG_NUMBER))
    {
      DRV_ERROR("aess_pecisensor_command Read/Write Length out of buffer\n");
      up(&peci_sem);
      return (-EFAULT);
    }
        /*Write PECI command data to register*/
    iowrite8(u8CommandCode, (void *) PECI_REG_WRITE_DATA0);

        if (u8WriteLength > 1)
        {
          for (u8Index = 0; u8Index < (u8WriteLength - 1); u8Index++)
          {
            iowrite8(pPECICommandData->u8CommandData[u8Index], 
                     (void *)(PECI_REG_WRITE_DATA0 + u8Index + 1));
             DRV_MSG("aess_pecisensor_command Command Data[%d]:%x\n", u8Index, pPECICommandData->u8CommandData[u8Index]);
          }
        }

    /* PECI Command Setting */
    u32Data = (u8ClientAddr << COMMAND_TARGET_ADDRESS) + 
              (u8WriteLength << COMMAND_WRITE_LENGTH) +
              (u8ReadLength << COMMAND_READ_LENGTH);
    iowrite32(u32Data, (void *) PECI_REG_READ_WRITE_LENGTH);
    DRV_INFO("aess_pecisensor_command PECI_REG_READ_WRITE_LENGTH:%x\n", ioread32((void *) PECI_REG_READ_WRITE_LENGTH));

    /* Clear command register */
    iowrite32(0, (void *) PECI_REG_COMMAND);
    DRV_INFO("aess_pecisensor_command PECI_REG_COMMAND:%x\n", ioread32((void *) PECI_REG_COMMAND));
  
    /* Enable Interrupt */
    u32Data = ioread32((void *) PECI_REG_INTERRUPT) + ENABLE_DONE_INT + ENABLE_WRITE_FCS_ABORT_INT +
              ENABLE_WRITE_FCS_BAD_INT + ENABLE_BUS_CONTENT_INT;
    iowrite32(u32Data, (void *) PECI_REG_INTERRUPT);
    DRV_INFO("aess_pecisensor_command PECI_REG_INTERRUPT:%x\n", ioread32((void *) PECI_REG_INTERRUPT));

    udelay(100);
    /* Fire command */
    u32Data = FIRE_COMMAND;	
    iowrite32(u32Data, (void *) PECI_REG_COMMAND);

    if (0 == wait_event_interruptible_timeout(WaitQueue, 
                                              WaitQueueFlag, 
                                              20))
    {
            DRV_ERROR("aess_pecisensor_command command time out error, PECI_REG_COMMAND = 0x%08x\n", 
                      ioread32((void *) PECI_REG_COMMAND));
            up(&peci_sem);
            return (-EPROTO);
    }	

        WaitQueueFlag = 0;
    
    /* Check if PECI error */
     if (u32PECI_INT_STS & CLEAR_WRITE_FCS_ABORT_INT)
     {
        DRV_ERROR("aess_pecisensor_command write FCS abort error\n");
        up(&peci_sem);
        return (-EPROTO); 
     }
     if (u32PECI_INT_STS & CLEAR_WRITE_FCS_BAD_INT)
     {
        DRV_ERROR("aess_pecisensor_command write FCS bad error\n");
        up(&peci_sem);
        return (-EPROTO); 
     }
     if (u32PECI_INT_STS & CLEAR_BUS_CONTENT_INT)
     {
        DRV_ERROR("aess_pecisensor_command bus connection error\n");
        up(&peci_sem);
        return (-EPROTO); 
     }
     if (u32PECI_INT_STS & CLEAR_DONE_INT)
     {
     	/* read the PECI data from the data register */
        for (u8Index = 0 ; u8Index < u8ReadLength; u8Index++)
        {
          pPECICommandData->u8CommandData[u8Index] = ioread8((void *)(PECI_REG_READ_DATA0 + u8Index));
          DRV_INFO("pPECICommandData->u8CommandData[%x]:%x\n", u8Index, pPECICommandData->u8CommandData[u8Index]);
        }
        up(&peci_sem);
        return (STATUS_OK); 
     }
     
    DRV_ERROR("aess_pecisensor_command PECI unknow error\n");
    up(&peci_sem);
    return (-EPROTO); 
}

/******************************************************************************
*   FUNCTION        :   aess_peci_isr
******************************************************************************/
/**
 *  @brief      Interrupt service routine of PECI driver.
 *
 *  @return     IRQ_HANDLED
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Global function\n
 *
 *****************************************************************************/
irqreturn_t aess_peci_isr(
                           /** irq */
                           int irq, 

                           /** device id */
                           void *dev_id, 

                           /** registers stored on the stack during a system call. */
                           struct pt_regs *regs
                         )
{
    DRV_MSG("aess_peci_isr Enter!\n");
    u32PECI_INT_STS = ioread32((void *) PECI_REG_INTERRUPT_STATUS);
    DRV_MSG("aess_peci_isr PECI_REG_INTERRUPT_STATUS:%08x\n", u32PECI_INT_STS);

    /* Disable Interrupt  */
    iowrite8(0, (void *) PECI_REG_INTERRUPT);
    DRV_MSG("aess_peci_isr PECI_REG_INTERRUPT:%x\n", ioread8((void *) PECI_REG_INTERRUPT));

    /* Clear all interrupt status */
    iowrite32(CLEAR_ALL_INT, (void *) PECI_REG_INTERRUPT_STATUS);
    DRV_MSG("aess_peci_isr PECI_REG_INTERRUPT_STATUS:%08x\n", ioread32((void *) PECI_REG_INTERRUPT_STATUS));

    WaitQueueFlag = 1;
    wake_up_interruptible(&WaitQueue);

    return (IRQ_HANDLED);
}

/******************************************************************************
*   FUNCTION        :   aess_pecisensor_ioctl
******************************************************************************/
/**
 *  @brief      The API is invoked when user space application issues an IOCTL command. 
 *              The following is the list of commands that supported currently.
 *              1.  AESS_PECIDRV_COMMAND
 *                  Issue PECI command.
 *
 *  @return     STATUS_OK if driver execute the command success.
 *              Others if execute the command fail.                   
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Global function \n
 *
 *****************************************************************************/
static int aess_pecisensor_ioctl(
                                 /** inode structure point */
                                 struct inode *inode, 
           
                                 /** file structure point */
                                 struct file *filp,
           
                                 /** command from user space */
                                 unsigned int cmd, 
    
                                 /** sBIOSPostInfo structure point */
                                 unsigned long arg
                                )
{
	
  int err_check = 0;
  int temp1 = 0;
  int temp2 = 0;
  sPECICommandData *pPECICommandData = (sPECICommandData *) arg;

  DRV_MSG("Enter aess_pecisensor_ioctl...\n");
  /* Check the structure address is OK */
  if(!access_ok(VERIFY_WRITE, (void *) pPECICommandData, sizeof(pPECICommandData)))
  {
    DRV_ERROR("aess_pecisensor_ioctl struct VERIFY_WRITE access error\n");
    /* error occur */
    return -EFAULT;
  }
	
	switch(cmd)
	{   
       case AESS_PECIDRV_COMMAND:
       	    DRV_MSG("aess_pecisensor_ioctl AESS_PECIDRV_COMMAND\n");
            err_check = aess_pecisensor_command(pPECICommandData);
            break;
       case AESS_PECIDRV_R:
       	    /* This command is a bogus workaround for Intel bogus temperature reading.*/
            DRV_MSG("aess_pecisensor_ioctl AESS_PECIDRV_R\n");
            /* Get first reading. */
            err_check = aess_pecisensor_command(pPECICommandData);  
            if (err_check != STATUS_OK)
            {
                return (err_check);
            }
            /* Convert reading to one byte. */
            temp1 = (((pPECICommandData->u8CommandData[0] + (pPECICommandData->u8CommandData[1] << 8)) & PECI_READING_VALUE_MASK) >> SHIFT_FRACTIONAL);

            /* Get second reading. */
            err_check = aess_pecisensor_command(pPECICommandData);  
            if (err_check != STATUS_OK)
            {
                return (err_check);
            }
            /* Convert reading to one byte. */
            temp2 = (((pPECICommandData->u8CommandData[0] + (pPECICommandData->u8CommandData[1] << 8)) & PECI_READING_VALUE_MASK) >> SHIFT_FRACTIONAL);
            /* if the difference of two readings is greater than 3 , then return fail. */
            if (abs(temp1 - temp2) > 3)
            {
                DRV_ERROR("AESS_PECIDRV_R: Wrong reading ![0x%x,0x%x]\n", temp1, temp2);
                return (-EFAULT);
            }
            pPECICommandData->u8CommandData[0] = (u8) temp2;
            DRV_MSG("AESS_PECIDRV_R reading:%x\n", temp2);
            break;
       default:
            DRV_ERROR("aess_pecisensor_ioctl, command error! \n");
            err_check = -EINVAL;
	}
	
	DRV_INFO("aess_pecisensor_ioctl Error no is %d\n", err_check);
	
	/* 0->ok, minus->fail */
	return err_check;
}

/******************************************************************************
*   FUNCTION        :   aess_pecisensor_release
******************************************************************************/
/**
 *  @brief      This function will be called when device node is closed. Set init flag to INIT_FAIL
 *
 *  @return     0
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Static function\n
 *
 *****************************************************************************/
static int aess_pecisensor_release(
                                   /**Node*/
                                   struct inode* inode, 
                                 
                                   /**File descriptor*/
                                   struct file *flip
                                  )
{   
	return (STATUS_OK);
}

struct file_operations aess_pecisensor_fops = {
	.open = aess_pecisensor_open, 
	.ioctl = aess_pecisensor_ioctl,
	.release = aess_pecisensor_release,
};

/******************************************************************************
*   FUNCTION        :   aess_pecisensor_init
******************************************************************************/
/**
 *  @brief      The function is invoked when biospost module is loaded into the kernel.
 *
 *  @return     STATUS_OK if driver initial success.
 *              Others if initial fail.                   
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Global function  \n
 *
 *****************************************************************************/
int __init aess_pecisensor_init(void)
{
	int result;
	pecisensor_cdev = cdev_alloc();
	
	printk("Init_aess_pecisensor_module Sucess! \n");
	pecisensor_cdev->ops = &aess_pecisensor_fops;
	result = alloc_chrdev_region(&dev, 0, 1, driver_name);

	if (result < 0) {
		printk ("PECI_KN:Registering the character device failed with %d\n", MAJOR(dev));
		return result;
	}
		   
	cdev_add(pecisensor_cdev, dev, 1);

    /* Request PECI IRQ */                             
    result = request_irq(IRQ_PECI,
                         (irq_handler_t)aess_peci_isr,
                         IRQF_SHARED,
                         "PECI",(void *) &PECIDrv);

    /* Set high level trigger */
    IRQ_SET_HIGH_LEVEL(IRQ_PECI);
    IRQ_SET_LEVEL_TRIGGER(IRQ_PECI);
  
    if (result)
    {
        printk("KN: %d request IRQ error\n", result);
    }

    init_MUTEX(&peci_sem);

    printk("mknod /dev/aess_pecisensordrv c %d 0\n", MAJOR(dev));
	
    aess_debugfs_default_create(driver_name,NULL,0,0);
    aess_debugfs_create_file(driver_name,"u32",(u32 *)&u32tmpcounter,DBG_TYPE32,0);
    aess_debugfs_create_file(driver_name,"u16",(u16 *)&u16tmpcounter,DBG_TYPE16,0);
    aess_debugfs_create_file(driver_name,"u8",(u8 *)&u8tmpcounter,DBG_TYPE8,0);
    
	return (STATUS_OK);
}

/******************************************************************************
*   FUNCTION        :   aess_pecisensor_exit
******************************************************************************/
/**
 *  @brief      The function is invoked when biospost driver be removed from kernal.
 *
 *  @return     None.                   
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Global function \n
 *
 *****************************************************************************/
void __exit aess_pecisensor_exit(void)
{
   /* free the irq */
   free_irq(IRQ_PECI,(UINT32 *) &PECIDrv);

   cdev_del(pecisensor_cdev);
   unregister_chrdev_region(dev,1);
   DRV_MSG ("aess_pecisensor_exit \n");
   aess_debugfs_remove(driver_name,NULL,0,0);
	
   return;
}


MODULE_DESCRIPTION("AESS peci Sensor Driver");
MODULE_AUTHOR("Ken Lin <ken.lin@avocent.com>");
MODULE_LICENSE("GPL");
MODULE_VERSION("0.1.0.0");
module_param(driver_name, charp, S_IRUGO);
module_init(aess_pecisensor_init);
module_exit(aess_pecisensor_exit);

/* End of code */
