/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* AST2050 KVM video driver */

#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/fs.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/irq.h>
#include <linux/delay.h>
#include <linux/jiffies.h>
#include <linux/socket.h>
#include <linux/tcp.h>
#include <linux/proc_fs.h>

#include "aess_video.h"
#include "aess_video_i2c.h"

// Compile time support flags
#define INCLUDE_EXTERNAL_VGA_SUPPORT 1
#define INCLUDE_DOWNSCALING 0
#define PROCFS 1

MODULE_LICENSE("GPL");
DECLARE_WAIT_QUEUE_HEAD(my_queue);

int ioaccess_ioctl(struct inode *inode, struct file *filp, unsigned cmd,
		   unsigned long arg);
int ioaccess_open(struct inode *inode, struct file *filp);
int ioaccess_release(struct inode *inode, struct file *filp);

u8 aspeedi2c_read(u8, u8, u8);
u8 aspeedi2c_write(u8, u8, u8, u8);

// HWCursor Entry routines
static void dirtyHWCursorCache(void);
static void clearHWCursorCache(void);
static void removeHWCursorCacheEntry(PHWCURSORDESCRIPTOR pHWCursor);
static PHWCURSORCACHEENTRY getHWCursorCacheEntry(int iCursorNumber);
static char getHWCursorCacheNumber(PHWCURSORDESCRIPTOR pHWCursor,
				   unsigned char cCursorType, u16, u16);
static int isHWCursorCacheEntryDirty(int iCacheNumber);

// Ring setup
void InitializeRing(void);

// Video Engine Control
void VideoEngineUnlock(void);
void VideoEngineReset(void);
void VideoEngineHalt(void);
void VideoEngineStall(void);

// Video Engine Initialization
void InitializeVideoAll(void);
void InitializeVideoBuffers(void);
void InitializeVideoCompressionBuffer(void);
void SwapInAndRefBuffers(void);
void InitializeVideoSource(void);
#if INCLUDE_DOWNSCALING
void HorizontalDownScaling(void);
void InitializeVideoScaling(void);
#endif
void InitializeVideoScaling(void);
void SetFrameRate(int frameRate);
void InitializeVideoFrameRate(void);
void InitializeVideoBCD(void);
void InitializeVideoCompression(void);
void InitializeVideoDirectMode(void);
int IsCompressBufAvailable(void);
void RequestSingleFrame(void);
void SetupDirectMode(void);
// Video Engine Mode Detect
int GetModeDetect(void);
int AutoPosition_Adjust(void);
int DoModeDetection(void);
#if INCLUDE_EXTERNAL_VGA_SUPPORT
u32 JudgeMode(u32 HorizontalCount, u32 VerticalLine);
void AutoPhase_Adjust(void);
#endif
// Frame Descriptor Processing
int GetNextWritableRingEntry(void);
int GetCurrentWritableRingEntry(void);
int GetNextReadableRingEntry(void);
void FreeRingEntry(int curEntry);
int IncrementRingEntry(int curEntry);
int GetRingCount(void);
void UpdateRingCounts(void);

// Debug Stuff
u32 RegisterRead(u32 regNo, u32 fromLine);
void RegisterWrite(u32 regNo, u32 regVal, u32 fromLine);
char *getVideoRegisterName(u32 iRegNo);
void DumpVideoEngineRegisters(void);
void PrintVideoConfig(void);
int gDebugFlags;
int nmask = 0;
int intmasks[MAX_INTS];

#ifdef	PROCFS
void create_proc_files(void);
void remove_proc_files(void);
#endif

int Device_open = 0;
int bFirstConfig = 1;
int ioaccess_major = 0;
u8 *Data_Linear_address = NULL;
u8 *VGA_Linear_address = NULL;

int flag = 0;
int Mode_Watchdog = 0,
    Mode_Detect_Ready = 0,
    Capture_Complete = 0,
    Compression_Complete = 0,
    gPrevRingEndOffset,
    gPrevFrameNumber = -1,
    gSegmentNumber,
    gframeDataCount,
    gCaptureActive = 0,
    gVideoEngineDisabled = 0,
    gSOF, gFullFrameRequested = 0, gVideoEngineIdle = 1, gbHavePrevFrame;

FRAMERINGDESCRIPTOR theRing;

// Hardware cursor related data
#define MAX_HWCURSOR_CACHE 16
HWCURSORCACHEENTRY hwCursorCache[MAX_HWCURSOR_CACHE];
PHWCURSORDESCRIPTOR g_pLastHWCursor = NULL;
u32 g_uiLastHWCursorChecksum = 0;

char noChangeFrame[12] =
    { '\0', '\0', '\0', 0x90, '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', };

#if INCLUDE_EXTERNAL_VGA_SUPPORT
// Note: Modified for modes which have border issue
VESA_MODE ModeTable[] = {
////////////////////////// 60Hz mode
// 720x480 done
	{1056, 497, 720, 480, 60, 29900, 88, 104, 3, 13, 0, 0, 0, 0},
// 848x480 done
	{1064, 517, 848, 480, 60, 31160, 88, 91, 3, 26, 0, 0, 0, 0},
	{800, 525, 640, 480, 60, 31469, 96, 40, 2, 25, 8, 8, 8, 8},
// 720x576
	{912, 597, 720, 576, 60, 35920, 72, 88, 3, 17, 0, 0, 0, 0},
// 960x600 GTF done
	{1232, 622, 960, 600, 60, 37320, 96, 136, 3, 18, 0, 0, 0, 0},
	{1056, 628, 800, 600, 60, 37879, 128, 88, 4, 23, 0, 0, 0, 0},
// 1088x612 GTF done
	{1392, 634, 1088, 612, 60, 38040, 112, 152, 3, 18, 0, 0, 0, 0},
// 1280x720 GTF done
	{1664, 746, 1280, 720, 60, 44760, 136, 192, 3, 22, 0, 0, 0, 0},
// 1360x768 GTF done
	{1776, 795, 1360, 768, 60, 47700, 144, 208, 3, 23, 0, 0, 0, 0},
// 1280x768 done
	{1664, 798, 1280, 768, 60, 47700, 128, 184, 7, 20, 0, 0, 0, 0},
	{1344, 806, 1024, 768, 60, 48363, 136, 160, 6, 29, 0, 0, 0, 0},
// 1280x800 GTF done
	{1680, 828, 1280, 800, 60, 49680, 136, 200, 3, 24, 0, 0, 0, 0},
// 1152x864 GTF done
	{1520, 895, 1152, 864, 60, 53700, 120, 184, 3, 27, 0, 0, 0, 0},
// 1600x900 GTF done
	{2128, 932, 1600, 900, 60, 55920, 168, 264, 3, 28, 0, 0, 0, 0},
	{1800, 1000, 1280, 960, 60, 60000, 112, 312, 3, 36, 0, 0, 0, 0},
// 1600x1024 GTF done
	{2144, 1060, 1600, 1024, 60, 63600, 168, 272, 3, 32, 0, 0, 0, 0},
	{1688, 1066, 1280, 1024, 60, 63981, 112, 248, 3, 38, 0, 0, 0, 0},
	{2160, 1250, 1600, 1200, 60, 75000, 192, 304, 3, 46, 0, 0, 0, 0},
//////////////////////  Not 60Hz mode
	{900, 449, 720, 400, 70, 31469, 108, 45, 2, 25, 8, 8, 8, 8},
	{832, 520, 640, 480, 72, 37861, 40, 120, 3, 20, 8, 8, 8, 8},
	{840, 500, 640, 480, 75, 37500, 64, 120, 3, 16, 0, 0, 0, 0},
	{832, 509, 640, 480, 85, 43269, 56, 80, 3, 25, 0, 0, 0, 0},
	{1024, 625, 800, 600, 56, 35156, 72, 128, 2, 22, 0, 0, 0, 0},
	{1040, 666, 800, 600, 72, 48077, 120, 64, 6, 23, 0, 0, 0, 0},
	{1056, 625, 800, 600, 75, 46875, 80, 160, 3, 21, 0, 0, 0, 0},
	{1048, 631, 800, 600, 85, 53674, 64, 152, 3, 27, 0, 0, 0, 0},
	{1328, 806, 1024, 768, 70, 56476, 136, 144, 6, 29, 0, 0, 0, 0},
	{1312, 800, 1024, 768, 75, 60023, 96, 176, 3, 28, 0, 0, 0, 0},
	{1376, 808, 1024, 768, 85, 68677, 96, 208, 3, 36, 0, 0, 0, 0},
	{1600, 900, 1152, 864, 75, 67500, 128, 256, 3, 32, 0, 0, 0, 0},
	{1728, 1011, 1280, 960, 85, 85938, 160, 224, 3, 47, 0, 0, 0, 0},
	{1688, 1066, 1280, 1024, 75, 79976, 144, 248, 3, 38, 0, 0, 0, 0},
	{1728, 1072, 1280, 1024, 85, 91146, 160, 224, 3, 44, 0, 0, 0, 0},
	{2160, 1250, 1600, 1200, 65, 81250, 192, 304, 3, 46, 0, 0, 0, 0},
	{2160, 1250, 1600, 1200, 70, 87500, 192, 304, 3, 46, 0, 0, 0, 0},
	{2160, 1250, 1600, 1200, 75, 93750, 192, 304, 3, 46, 0, 0, 0, 0},
	{2160, 1250, 1600, 1200, 85, 106250, 192, 304, 3, 46, 0, 0, 0, 0},
	{2080, 1235, 1920, 1200, 60, 74038, 32, 80, 6, 26, 0, 0, 0, 0}
};
u16 ModeNumberCount = sizeof(ModeTable) / sizeof(VESA_MODE);

// ADC_MODE table from old adc.h
ADC_MODE ADCModeTable[] = {
////////////////////////// 60Hz mode
	{720, 480, 60, 0x41, 0xF0, 0x48, 0x05, 0x20, 0x58, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, 6, 2}
	,
	{848, 480, 60, 0x42, 0x70, 0x48, 0x05, 0x20, 0x58, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, 6, 2}
	,
	{640, 480, 60, 0x31, 0xF0, 0x48, 0x05, 0x20, 0x60, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, 6, 2}
	,
	{720, 576, 60, 0x38, 0xF0, 0x48, 0x05, 0x20, 0x48, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, 6, 2}
	,
	{960, 600, 60, 0x4C, 0xF0, 0x60, 0x05, 0x20, 0x60, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, -1, 0}
	,
	{800, 600, 60, 0x41, 0xF0, 0x60, 0x05, 0x20, 0x80, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, -1, 0}
	,
	{1088, 612, 60, 0x56, 0xF0, 0x60, 0x05, 0x20, 0x70, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, -1, 0}
	,
	{1280, 720, 60, 0x67, 0xF0, 0xA8, 0x05, 0x20, 0x88, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, -1, 0}
	,
	{1360, 768, 60, 0x6E, 0xF0, 0xA8, 0x05, 0x20, 0x90, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, 7, 1}
	,
	{1280, 768, 60, 0x67, 0xF0, 0xA8, 0x05, 0x20, 0x80, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, 7, 1}
	,
	{1024, 768, 60, 0x53, 0xF0, 0xA8, 0x05, 0x20, 0x88, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, -1, 7}
	,
	{1280, 800, 60, 0x68, 0xF0, 0xA8, 0x05, 0x20, 0x88, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, -1, 0}
	,
	{1152, 864, 60, 0x5E, 0xF0, 0xA8, 0x05, 0x20, 0x78, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, -1, 0}
	,
	{1600, 900, 60, 0x84, 0xF0, 0xA8, 0x05, 0x20, 0xA8, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, -1, 0}
	,
	{1280, 960, 60, 0x70, 0x70, 0xA8, 0x05, 0x20, 0x70, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, -1, 0}
	,
	{1600, 1024, 60, 0x85, 0xF0, 0xE8, 0x05, 0x20, 0xA8, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, -1, 0}
	,
	{1280, 1024, 60, 0x69, 0x70, 0xA8, 0x05, 0x20, 0x70, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, -1, 0}
	,
	{1600, 1200, 60, 0x86, 0xF0, 0xE8, 0x05, 0x20, 0xC0, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, -1, 0}
	,
//////////////////////  Not 60Hz mode
	{720, 400, 70, 0x38, 0x30, 0x48, 0x05, 0x20, 0x6C, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, 6, 1}
	,
	{640, 480, 72, 0x33, 0xF0, 0x48, 0x05, 0x20, 0x28, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, 6, 3}
	,
	{640, 480, 75, 0x34, 0x70, 0x48, 0x05, 0x20, 0x40, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, -1, 3}
	,
	{640, 480, 85, 0x33, 0xF0, 0x48, 0x05, 0x20, 0x38, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, -1, 3}
	,
	{800, 600, 56, 0x3F, 0xF0, 0x60, 0x05, 0x20, 0x48, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, -1, 0}
	,
	{800, 600, 72, 0x40, 0xF0, 0x60, 0x05, 0x20, 0x78, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, -1, 0}
	,
	{800, 600, 75, 0x41, 0xF0, 0x60, 0x05, 0x20, 0x50, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, -1, 0}
	,
	{800, 600, 85, 0x41, 0x70, 0x60, 0x05, 0x20, 0x40, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, -1, 0}
	,
	{1024, 768, 70, 0x52, 0xF0, 0xA8, 0x05, 0x20, 0x88, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, -1, 7}
	,
	{1024, 768, 75, 0x51, 0xF0, 0xA8, 0x05, 0x20, 0x60, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, -1, 1}
	,
	{1024, 768, 85, 0x55, 0xF0, 0xA8, 0x05, 0x20, 0x60, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, -1, 1}
	,
	{1152, 864, 75, 0x63, 0xF0, 0xA8, 0x05, 0x20, 0x80, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, -1, 0}
	,
	{1280, 1024, 75, 0x69, 0x70, 0xE8, 0x05, 0x20, 0x90, 0x60, 0x60, 0x60,
	 0x5E, 0xFE, -1, 0}
};
u16 ADCModeNumberCount = sizeof(ADCModeTable) / sizeof(ADC_MODE);
#endif				// INCLUDE_EXTERNAL_VGA_SUPPORT

INTERNAL_MODE Internal_Mode[] = {
// 1024x768
	{1024, 768, 0, 65.000}
	,
	{1024, 768, 1, 65.000}
	,
	{1024, 768, 2, 75.000}
	,
	{1024, 768, 3, 78.750}
	,
	{1024, 768, 4, 94.500}
	,
// 1280x1024
	{1280, 1024, 0, 108.000}
	,
	{1280, 1024, 1, 108.000}
	,
	{1280, 1024, 2, 135.000}
	,
	{1280, 1024, 3, 157.500}
	,
// 1600x1200
	{1600, 1200, 0, 162.000}
	,
	{1600, 1200, 1, 162.000}
	,
	{1600, 1200, 2, 175.500}
	,
	{1600, 1200, 3, 189.000}
	,
	{1600, 1200, 4, 202.500}
	,
	{1600, 1200, 5, 229.500}
	,
// 1920x1200 reduce blank
	{1920, 1200, 0, 157.000}
	,
	{1920, 1200, 1, 157.000}
	,
};
u16 InternalEnhanceModeNumberCount = sizeof(Internal_Mode) / sizeof(INTERNAL_MODE);

struct _packetCountBits {
	int packetCount;
	int packetSetting;
} packetCountBits[4] = {
	{
	4, 0}, {
	8, 1}, {
	16, 2}, {
	32, 3}
};

struct _packetSizeBits {
	int packetSize;
	int packetSetting;
} packetSizeBits[8] = {
	{
	1, 0}, {
	2, 1}, {
	4, 2}, {
	8, 3}, {
	16, 4}, {
	32, 5}, {
	64, 6}, {
	128, 7}
};

// Storage for global video engine parameters
VIDEOENGINECONFIG videoConfig;
VIDEOENGINESTATUS videoStatus;
struct memory_buffers videoBuffers;

// debug stuff

struct file_operations ioaccess_fops = {
      open:ioaccess_open,
      ioctl:ioaccess_ioctl,
      release:ioaccess_release
};

int ioaccess_open(struct inode *inode, struct file *filp)
{
	if (Device_open)
		return (-EBUSY);	/* Return negative value for error. (chtsai) */
	Device_open++;
	// Disable all debug on open...
	gDebugFlags = 0;
	// Clear the hardware cursor cache
	clearHWCursorCache();
	// Be sure regiosters are unlocked and halt the video engine
	VideoEngineUnlock();
	VideoEngineHalt();
	return (0);
}

int ioaccess_ioctl(struct inode *inode, struct file *filp,
		   unsigned int cmd, unsigned long arg)
{
	int ret = -1;

	IO_ACCESS_DATA Kernal_IO_Data;

	memset(&Kernal_IO_Data, 0, sizeof(IO_ACCESS_DATA));

	Kernal_IO_Data = *(IO_ACCESS_DATA *) arg;

	switch (cmd) {
	case IOCTL_IO_READ:
		{
			Kernal_IO_Data.Data =
			    *(u32 *) (IO_ADDRESS(Kernal_IO_Data.Address));
			*(IO_ACCESS_DATA *) arg = Kernal_IO_Data;
			ret = 0;
		}
		break;

	case IOCTL_IO_WRITE:
		{
			*(u32 *) (IO_ADDRESS(Kernal_IO_Data.Address)) =
			    Kernal_IO_Data.Data;
			ret = 0;
		}
		break;

	case IOCTL_I2C_READ:
		{
			ret =
			    aspeedi2c_read((u8) Kernal_IO_Data.Address,
					   (u8) Kernal_IO_Data.Data,
					   (u8) Kernal_IO_Data.Value);
		}
		break;

	case IOCTL_I2C_WRITE:
		{
			ret =
			    aspeedi2c_write((u8) Kernal_IO_Data.Address,
					    (u8) Kernal_IO_Data.Data,
					    (u8) Kernal_IO_Data.Value,
					    (u8) Kernal_IO_Data.I2CValue);
		}
		break;

	case IOCTL_REQ_MODE_DETECT:
		{
			// Stop the video engine to do mode detection.
			VideoEngineReset();
			InitializeVideoAll();
			ret = DoModeDetection();
			Mode_Detect_Ready = 0;
			if (ret == 0) {
				if (copy_to_user((u32 *) Kernal_IO_Data.Address,
					     (u32 *) & videoStatus,
					     Kernal_IO_Data.Data))
					return -EFAULT;
			}
		}
		break;

		// IOCTL to read the videoStatus struct to obtain current video status.
	case IOCTL_GET_VIDEOSTATUS:
		{
			if (copy_to_user((u32 *) Kernal_IO_Data.Address,
				     (u32 *) & videoStatus,
				     Kernal_IO_Data.Data))
				return -EFAULT;
			ret = 0;
		}
		break;

		// IOCTL to store the videoConfig struct that we need to set up 
		// capture modes.
	case IOCTL_SET_VIDEOCONFIG:
		{
			u32 lastAddress;
			if (copy_from_user((u32 *) & videoConfig,
				       (char *)Kernal_IO_Data.Address,
				       Kernal_IO_Data.Data))
				return -EFAULT;
			if (bFirstConfig) {
				int i, maxPixels;

				bFirstConfig = 0;

				// Force a full set of all parameters
				videoConfig.bVideoCaptureChanges = 1;

				// Set up default values in case user specifies invalid stream buffer settings
				videoStatus.ulPacketCountBits = 2;
				videoStatus.ulPacketSizeBits = 5;

				// First, use MAX parameters to compute buffer sizes and addresses.  These can ONLY
				// be set on the first call and will never be updated again...
				videoStatus.VideoEngineTopAddress = (videoConfig.TotalMemory - videoConfig.VGAMemory) * 1024 * 1024;	// VGA buffer starts here
#ifndef	EMS
				if (videoConfig.TotalMemory == 128) {
					videoStatus.MaxModeInfo.X =
					    MAX_HORIZONTAL_X_RESOLUTION_128MB;
					videoStatus.MaxModeInfo.Y =
					    MAX_VERTICAL_Y_RESOLUTION_128MB;
					videoStatus.VideoEngineBottomAddress =
					    LOWEST_AVAILABLE_DRAM_ADDRESS_128MB;
				} else {
					videoStatus.MaxModeInfo.X =
					    MAX_HORIZONTAL_X_RESOLUTION_64MB;
					videoStatus.MaxModeInfo.Y =
					    MAX_VERTICAL_Y_RESOLUTION_64MB;
					videoStatus.VideoEngineBottomAddress =
					    LOWEST_AVAILABLE_DRAM_ADDRESS_64MB;
				}

				printk
				    ("%s[%d] Tot Mem %dMB Max Res = %d X %d\n",
				     __FUNCTION__, __LINE__,
				     (int)videoConfig.TotalMemory,
				     videoStatus.MaxModeInfo.X,
				     videoStatus.MaxModeInfo.Y);
#else
				videoStatus.MaxModeInfo.X = MAX_HORIZONTAL_X_RESOLUTION_64MB;
				videoStatus.MaxModeInfo.Y = MAX_VERTICAL_Y_RESOLUTION_64MB;
				lastAddress = (u32) videoBuffers.capture_phys;
				printk
				    ("%s Max Res = %d X %d Ref Mem (%d) @ %p Cap Mem (%d) @ %p\n",
				     __FUNCTION__, 
				     videoStatus.MaxModeInfo.X,
				     videoStatus.MaxModeInfo.Y,
				     videoBuffers.reference_size,
				     videoBuffers.reference_phys,
				     videoBuffers.capture_size,
				     videoBuffers.capture_phys);
#endif

				maxPixels = videoStatus.MaxModeInfo.X * videoStatus.MaxModeInfo.Y;
				videoStatus.VideoFlagBufferSize = maxPixels / 64 / 2;	// 4 bits per tile, multiple of 8
				videoStatus.VideoFlagBufferSize = (videoStatus.VideoFlagBufferSize +
				     7) & 0xFFFFFFF8;
				videoStatus.VideoCRCBufferSize = (maxPixels / 64) * 4;	// 4 bytes per tile of 64 bits, multiple of 8
				videoStatus.VideoCRCBufferSize = (videoStatus.VideoCRCBufferSize +
				     7) & 0xFFFFFFF8;
				videoStatus.VideoCaptureBufferSize = ((maxPixels * 4 + 1023) / 1024) * 1024;	// 4 bytes per pixel, multiple of 1K
				for (i = 0; i < 4; i++) {
					if (packetCountBits[i].packetCount ==
					    videoConfig.ulPacketCount)
						videoStatus.ulPacketCountBits =
						    packetCountBits[i].
						    packetSetting;
				}
				for (i = 0; i < 8; i++) {
					if (packetSizeBits[i].packetSize ==
					    videoConfig.ulPacketSize)
						videoStatus.ulPacketSizeBits =
						    packetSizeBits[i].
						    packetSetting;
				}

				// Compute the compression buffer size from the settings that were just found.  These
				// are recomputed in case the user specified bad values.  In this case, the default
				// values of 16 packets of size 32KB is used.
				for (i = 0; i < 4; i++) {
					if (packetCountBits[i].packetSetting ==
					    videoStatus.ulPacketCountBits)
						videoConfig.ulPacketCount =
						    packetCountBits[i].
						    packetCount;
				}
				for (i = 0; i < 8; i++) {
					if (packetSizeBits[i].packetSetting ==
					    videoStatus.ulPacketSizeBits)
						videoConfig.ulPacketSize =
						    packetSizeBits[i].
						    packetSize;
				}

				// OK, now compute the actual compression buffer size to use
				videoStatus.VideoCompressBufferSize =
				    videoConfig.ulPacketCount *
				    videoConfig.ulPacketSize * 1024;

				// Time to set up the video engine buffer addresses.    
				// Required and dedicated SDRAM values for video engine.  Buffer
				// space is reserved memory (outside of kernel) and possibly allocated
				// memory.
				// Do the capture buffers first so that they will align to a multiple
				// of 256.
#ifndef	EMS
				videoStatus.VideoReferenceBufferAddress =
				    videoStatus.VideoEngineTopAddress -
				    videoStatus.VideoCaptureBufferSize;
				videoStatus.VideoINBufferAddress =
				    videoStatus.VideoReferenceBufferAddress -
				    videoStatus.VideoCaptureBufferSize;
#else
				videoStatus.VideoReferenceBufferAddress = (u32) videoBuffers.reference_phys;
				videoStatus.VideoINBufferAddress = (u32) videoBuffers.capture_phys +
					(videoBuffers.capture_size - videoStatus.VideoCaptureBufferSize);
#endif

				// Now do the compression buffer(s) 
				lastAddress =
				    videoStatus.VideoINBufferAddress -
				    (videoStatus.VideoCompressBufferSize *
				     videoConfig.VideoCompressBufferCount);
				lastAddress &= 0xFFFFF000;	// Round down to a 1K boundary for user mmap call.
				for (i = 0; i < MAX_COMPBUF; i++) {
					videoStatus.VideoCompressBufStatus[i] =
					    COMPRESSION_BUFFER_STATUS_UNUSED;
					videoStatus.
					    VideoCompressBufferAddress[i] = 0;
				}
				for (i = 0;
				     i < videoConfig.VideoCompressBufferCount;
				     i++) {
					videoStatus.VideoCompressBufStatus[i] =
					    COMPRESSION_BUFFER_STATUS_IDLE;
					videoStatus.
					    VideoCompressBufferAddress[i] =
					    lastAddress +
					    (i *
					     videoStatus.
					     VideoCompressBufferSize);
				}

				// IOmap the compression buffers
				Data_Linear_address =
				    ioremap_nocache(lastAddress,
						    videoStatus.
						    VideoCompressBufferSize *
						    videoConfig.
						    VideoCompressBufferCount);
				printk
				    ("%s Data_Linear_address = %8.8X[%d] mapped to %8.8X\n",
				     __FUNCTION__, (int)lastAddress,
				     (int)(videoStatus.VideoCompressBufferSize *
					   videoConfig.
					   VideoCompressBufferCount),
				     (int)Data_Linear_address);

				// IOmap the entire VGA Memory 
				VGA_Linear_address =
				    ioremap_nocache(videoStatus.
						    VideoEngineTopAddress,
						    videoConfig.VGAMemory *
						    1024 * 1024);
				printk
				    ("%s VGA_Linear_address = %8.8X mapped to %8.8X\n",
				     __FUNCTION__, 
				     (int)videoStatus.VideoEngineTopAddress,
				     (int)VGA_Linear_address);

				// Finish up with Flag and CRC buffers
				videoStatus.VideoFlagBufferAddress =
				    lastAddress -
				    videoStatus.VideoFlagBufferSize;
				videoStatus.VideoCRCBufferAddress =
				    videoStatus.VideoFlagBufferAddress -
				    videoStatus.VideoCRCBufferSize;

#ifndef	EMS
				if (videoStatus.VideoCRCBufferAddress <
				    videoStatus.VideoEngineBottomAddress) {
					printk
					    ("%s video compression buffers %8.8X below limit %8.8X\n",
					     __FUNCTION__, 
					     (int)videoStatus.
					     VideoCRCBufferAddress,
					     (int)videoStatus.
					     VideoEngineBottomAddress);
				} else {
					printk
					    ("%s video compression buffers can be reduced by %dK\n",
					     __FUNCTION__,
					     (int)((videoStatus.
						    VideoCRCBufferAddress -
						    videoStatus.
						    VideoEngineBottomAddress) /
						   1024));
				}
#endif
			}	// end if first config call

			if (videoConfig.ulRingMax > MAX_RING_COUNT) {
				if (gDebugFlags & DEBUG_RING_STATUS)
					printk
					    ("drv-%s[%d] requested ring count %d too large, set to %d\n",
					     __FUNCTION__, __LINE__,
					     (int)videoConfig.ulRingMax,
					     MAX_RING_COUNT);
				videoConfig.ulRingMax = MAX_RING_COUNT;
			}

			if (videoConfig.bVideoCaptureChanges) {
				// Changes to capture parameters need to halt capture and then restart
				// We halt the capture on purpose for these changes.  Caller must
				// init for capture again.
				VideoEngineHalt();
				gVideoEngineIdle = 1;
				InitializeVideoAll();
			} else {
				// Only set up the affected registers when the capture is not affected....
				if (videoConfig.bVideoScalingChanges)
					InitializeVideoScaling();
				if (videoConfig.bVideoFrameRateChanges)
					InitializeVideoFrameRate();
			}
			ret = 0;

		}
		break;

	case IOCTL_INIT_CAPTURE:
		{
			int i;
			// NOTE WELL: Do not modify B7 of VIDEO_SEQUENCE_CONTROL enable mode watchdog
			//            Do not modify B0 of VIDEO_INTERRUPT_CONTROL leaving mode watchdog interrupt enabled.

			if (gVideoEngineDisabled) {
				ret = -EINVAL;
				break;
			}

			// Halt the video engine to ensure no capture is in progress
			VideoEngineHalt();

			// Force the compression engine idle
			REGISTERANDWRITE(VIDEOENGINE_SEQUENCE_CONTROL,
					 0xFFFFFF81, 0x04);
			mdelay(2);
			// Clear force compression engine idle and reset capture modes
			REGISTERANDWRITE(VIDEOENGINE_SEQUENCE_CONTROL,
					 0xFFFFFF81, 0);
			mdelay(2);

			if (videoConfig.CaptureMode == CAPTUREMODE_SINGLE) {
				// Indicate there is no previous frame for a BCD reference
				gbHavePrevFrame = 0;
				// Initialize compression buffer status and the current processing buffer..
				for (i = 0;
				     i < videoConfig.VideoCompressBufferCount;
				     videoStatus.VideoCompressBufStatus[i++] =
				     COMPRESSION_BUFFER_STATUS_IDLE) ;
				// The first request will increment the compression buffer index so set it to the max
				videoStatus.CurrentCompressBufIndex =
				    videoConfig.VideoCompressBufferCount - 1;
			}

			if (videoConfig.CaptureMode == CAPTUREMODE_MULTIPLE) {
				// Configure automatic compression 
				// Set B9 which is undocumented anti-tearing mode bit 
				REGISTERANDWRITE(VIDEOENGINE_SEQUENCE_CONTROL,
						 0xFFFFFD81, 0x220);
				mdelay(2);
				// Enable multiple frame capture
				REGISTERANDWRITE(VIDEOENGINE_SEQUENCE_CONTROL,
						 0xFFFFFFA1, 0x08);
				mdelay(2);
				// Request a full frame 
				REGISTERANDWRITE(VIDEOENGINE_SEQUENCE_CONTROL,
						 0xFFFFFFBF, 0x40);
			}
			ret = 0;
		}
		break;

		// Call to setup for multiple frame capture using a ring of frame descriptors
	case IOCTL_TRIGGER_CAPTURE:
		{
			// NOTE WELL: Do not modify B7 of VIDEO_SEQUENCE_CONTROL enable mode watchdog
			//            Do not modify B0 of VIDEO_INTERRUPT_CONTROL leaving mode watchdog interrupt enabled.

			if (gVideoEngineDisabled) {
				ret = -EINVAL;
				break;
			}

			// Initialize for first video segment
			gPrevRingEndOffset = 0;
			videoStatus.CurrentCompressBufIndex = 0;

			// Reset Force Compression Idle and Trigger bits
			REGISTERANDWRITE(VIDEOENGINE_SEQUENCE_CONTROL,
					 0xFFFFFFF9, 0);
			barrier();

			// Initialize the Frame Ring
			InitializeRing();
			// Indicate first frame will be SOF
			gSOF = 1;

			// Single frame capture mode
			if (videoConfig.CaptureMode == CAPTUREMODE_SINGLE) {
				// Request compression of a single frame into the next compression buffer
				RequestSingleFrame();

				// Clear previous interrupt status for single frame
				Compression_Complete = Capture_Complete = 0;

				// Re-enable interrupts on first trigger, but after engine trigger.  This 
				// sequence appears to prevent endless compressed packet ready interrupts
				// when first starting the engine.
				REGISTERANDWRITE(VIDEO_INTERRUPT_CONTROL,
						 0xFFFFFFC1,
						 (VIDEO_INTERRUPT_COMPRESSION_COMPLETE
						  |
						  VIDEO_INTERRUPT_CAPTURE_COMPLETE));
			}
			// Multiple frame capture mode
			if (videoConfig.CaptureMode == CAPTUREMODE_MULTIPLE) {
				// Trigger the capture engine
				REGISTERANDWRITE(VIDEOENGINE_SEQUENCE_CONTROL,
						 0xFFFFFFFD, 0x12);
				*(IO_ACCESS_DATA *) arg = Kernal_IO_Data;

				// Re-enable interrupts on first trigger, but after engine trigger.  This 
				// sequence appears to prevent endless compressed packet ready interrupts
				// when first starting the engine.
				if (videoConfig.PacketMode == PACKETMODE_ENABLE) {
					REGISTERANDWRITE
					    (VIDEO_INTERRUPT_CONTROL,
					     0xFFFFFFC1,
					     (VIDEO_INTERRUPT_FRAME_COMPLETE |
					      VIDEO_INTERRUPT_PACKET_READY));
				} else {
					REGISTERANDWRITE
					    (VIDEO_INTERRUPT_CONTROL,
					     0xFFFFFFC1,
					     VIDEO_INTERRUPT_FRAME_COMPLETE);
				}
			}
			ret = 0;
		}
		break;

	case IOCTL_WAIT_CAPTURE:
		{
			unsigned long interruptFlags;
			int readFrame;
			PFRAMEDESCRIPTOR pRingEntry = NULL;
			char cFrameStatus = 0;
			int iTimeout = (HZ * Kernal_IO_Data.Data / 1000);

			if (gVideoEngineDisabled) {
				ret = -EINVAL;
				break;
			}

			if (iTimeout) {
				// User requests timeout
				wait_event_interruptible_timeout(my_queue,
								 ((readFrame =
								   GetNextReadableRingEntry
								   ()) >= 0),
								 iTimeout);
			} else {
				wait_event_interruptible(my_queue,
							 ((readFrame =
							   GetNextReadableRingEntry
							   ()) >= 0));
			}

			// Hold interrupts to get out current read entry.  Prevents interrupts
			// level from updating it with new video segment information while 
			// getting data to return to the user and then freeing it.
			local_irq_save(interruptFlags);

			readFrame = GetNextReadableRingEntry();

			if (readFrame >= 0) {
				// Note that we have a previous frame now
				gbHavePrevFrame = 1;

				pRingEntry = &theRing.RingEntry[readFrame];
				// Update IOData from the ring entry and then free it 
				Kernal_IO_Data.ulDataStartOffset =
				    pRingEntry->ulDataStartOffset;
				// Adjust the real offset address using the compression buffer index in this ring entry.
				Kernal_IO_Data.ulDataBaseOffset =
				    pRingEntry->ulCompressBufIndex *
				    videoStatus.VideoCompressBufferSize;
				Kernal_IO_Data.ulCompressBufIndex =
				    pRingEntry->ulCompressBufIndex;
				Kernal_IO_Data.ulDataCount =
				    pRingEntry->ulDataCount;
				Kernal_IO_Data.ulFrameNumber =
				    pRingEntry->ulFrameNumber;
				Kernal_IO_Data.ulSegmentNumber =
				    pRingEntry->ulSegmentNumber;
				Kernal_IO_Data.ulRingCount = theRing.ulRingCount - 1;	// Subtract off current entry!
				Kernal_IO_Data.ulFrameCount =
				    theRing.ulFrameCount;
				Kernal_IO_Data.ulCursorEntry =
				    pRingEntry->cCursorCacheNumber;
				Kernal_IO_Data.usCursorX = pRingEntry->usCursorX;	// Current h/w cursor address and offset
				Kernal_IO_Data.usCursorY =
				    pRingEntry->usCursorY;
				Kernal_IO_Data.usCursorXOffset =
				    pRingEntry->usCursorXOffset;
				Kernal_IO_Data.usCursorYOffset =
				    pRingEntry->usCursorYOffset;
				Kernal_IO_Data.ulRepeatCount = pRingEntry->ulRepeatCount;	// This could be a 'no change' frame count
				Kernal_IO_Data.Jiffies = pRingEntry->ulJiffies;
				Kernal_IO_Data.ulFrameStatus =
				    pRingEntry->cFrameStatus;
				Kernal_IO_Data.ulMode = pRingEntry->cMode;
				Kernal_IO_Data.ulY_JPEGTableSelector =
				    pRingEntry->cY_JPEGTableSelector;
				Kernal_IO_Data.ulUV_JPEGTableSelector =
				    pRingEntry->cUV_JPEGTableSelector;
				cFrameStatus = pRingEntry->cFrameStatus;
				FreeRingEntry(readFrame);
			}
			// Release Interrupts
			local_irq_restore(interruptFlags);

			if (pRingEntry
			    && (gDebugFlags & DEBUG_TRACK_RING_RETURNS)) {
				printk
				    ("drv-%s[%d] returning ring entry %d: buf %d seg %d-%d @ %6.6X count %d\n",
				     __FUNCTION__, __LINE__, readFrame,
				     (int)Kernal_IO_Data.ulCompressBufIndex,
				     (int)Kernal_IO_Data.ulFrameNumber,
				     (int)Kernal_IO_Data.ulSegmentNumber,
				     (int)Kernal_IO_Data.Address,
				     (int)Kernal_IO_Data.ulDataCount);
			}

			if (Mode_Watchdog == 1) {
				theRing.ulRingStatus |= VIDEO_RESOLUTION_CHANGE;
				Mode_Watchdog = 0;
			}

			*(IO_ACCESS_DATA *) arg = Kernal_IO_Data;

			// Format the return ring status
			ret =
			    (theRing.
			     ulRingStatus & VIDEO_ENGINE_STATUS_BITS) |
			    ((readFrame << 8) & 0x0000FF00) | cFrameStatus;

			// Clear all of the one-time status bits now that they have been returned
			theRing.ulRingStatus &= (~VIDEO_ENGINE_STATUS_BITS);
		}
		break;

	case IOCTL_HALT_CAPTURE:
		{
			gCaptureActive = 0;
			VideoEngineHalt();
			ret = 0;
		}
		break;

	case IOCTL_GET_HWCURSOR:
		{
			// Returns the requested hardware cursor from the hardware
			// cursor cache if HW cursor is enabled
			u32 iCur = Kernal_IO_Data.Value;
			PHWCURSORCACHEENTRY pReqHWCursor =
			    getHWCursorCacheEntry(iCur);
			ret = -1;
			if (pReqHWCursor != NULL) {
				printk
				    ("drv-%s[%d] Getting cursor %d, type %d @ 0x%8.8X\n",
				     __FUNCTION__, __LINE__, (int)iCur,
				     (int)pReqHWCursor->cCursorType,
				     (int)pReqHWCursor->theCursor.
				     cCursorPattern);
				if (copy_to_user((u32 *) Kernal_IO_Data.Address,
					     (u32 *) (pReqHWCursor->theCursor.
						      cCursorPattern), 8192))
					return -EFAULT;
				Kernal_IO_Data.Type = pReqHWCursor->cCursorType;
				Kernal_IO_Data.usCursorX = 64;	// size of the cursor area...
				Kernal_IO_Data.usCursorY = 64;
				*(IO_ACCESS_DATA *) arg = Kernal_IO_Data;
				// Clear this cursor's dirty bit now that application has it
				pReqHWCursor->cCursorCacheDirty = 0;
				ret = 8192;
			} else {
				printk("drv-%s[%d] Cursor %d not found\n",
				       __FUNCTION__, __LINE__, (int)iCur);
			}
		}
		break;

	case IOCTL_ACK_RING_ENTRY:
		{
			// OK, if we are still processing for this video compression buf, then ack  the ring entry
			// to the chip.  If not, then the compression completed before hitting the end of the
			// current compression buffer and we never needed to ACK the ring entry anyway.  NOTE
			// that in stream mode with one compression buffer, this will always be true.  In automatic
			// single capture frame mode, this will depend on how large the compressed frame was.
			if (Kernal_IO_Data.ulCompressBufIndex ==
			    videoStatus.CurrentCompressBufIndex)
				REGISTERANDWRITE(VIDEO_COMPRESS_READ_OFFSET,
						 0xFFC0007F,
						 (Kernal_IO_Data.
						  Address & 0xFFFFFF80));

			// In automatic single frame capture mode, need to check on compression buffer status and see
			// if we can start a new compression.
			if (videoConfig.CaptureMode == CAPTUREMODE_SINGLE) {
				// Next, if the ring entry was an end of frame, then we can set the associated compression
				// buffer to IDLE.
				if (Kernal_IO_Data.
				    ulFrameStatus & VIDEO_FRAME_STATUS_EOF)
					videoStatus.
					    VideoCompressBufStatus
					    [Kernal_IO_Data.
					     ulCompressBufIndex] =
					    COMPRESSION_BUFFER_STATUS_IDLE;
				// Now, if the video compression engine is IDLE, we can Request another single frame.
				if (gVideoEngineIdle) {
					if (gDebugFlags &
					    DEBUG_SINGLE_FRAME_BUFFERS)
						printk
						    ("drv-%s[%d] video engine is idle\n",
						     __FUNCTION__, __LINE__);
					if (IsCompressBufAvailable())
						RequestSingleFrame();
				}
			}
			ret = 0;
		}
		break;

	case IOCTL_FORCE_FULL_FRAME:
		{
			gFullFrameRequested = 1;
			REGISTERANDWRITE(VIDEOENGINE_SEQUENCE_CONTROL,
					 0xFFFFFFFF, 0x40);
			ret = 0;
		}
		break;

	case IOCTL_BIT_STREAM_BASE:
		{
			if (copy_to_user((u32 *) Kernal_IO_Data.Address,
				     &Data_Linear_address[Kernal_IO_Data.Value],
				     Kernal_IO_Data.Data))
				return -EFAULT;
			Kernal_IO_Data.Value += (int)Data_Linear_address;
			*(IO_ACCESS_DATA *) arg = Kernal_IO_Data;
			ret = 0;
		}
		break;

	case IOCTL_READ_INTERRUPTS:
		{
			if (copy_to_user((u32 *) Kernal_IO_Data.Address,
				     &intmasks[0], MAX_INTS * sizeof(long)))
				return -EFAULT;
			Kernal_IO_Data.Value = nmask;
			*(IO_ACCESS_DATA *) arg = Kernal_IO_Data;
			nmask = 0;
			ret = 0;
		}
		break;

	case IOCTL_CLEAR_INTERRUPTS:
		{
			nmask = 0;
			ret = 0;
		}
		break;

	case IOCTL_SET_DEBUG:
		{
			gDebugFlags = Kernal_IO_Data.Value;
			ret = 0;
		}
		break;

	case IOCTL_DISABLE_VIDEO:
		{
			extern void free_mem_chunk(void* addr, size_t size);

			gVideoEngineDisabled = 1;
			VideoEngineHalt();

			if (videoBuffers.reference_virt) {
				free_mem_chunk(videoBuffers.reference_virt,
					REFERENCE_BUFFER_SIZE_EMS);
				videoBuffers.reference_virt = NULL;
			}

			if (videoBuffers.capture_virt) {
				free_mem_chunk(videoBuffers.capture_virt,
					CAPTURE_BUFFER_SIZE_EMS);
				videoBuffers.capture_virt = NULL;
			}

			ret = 0;
		}
		break;

	case IOCTL_ENABLE_VIDEO:
		{
			extern void * alloc_mem_chunk(size_t size, gfp_t gfp_mask);

			if (videoBuffers.reference_virt == NULL) {
				videoBuffers.reference_virt = alloc_mem_chunk(REFERENCE_BUFFER_SIZE_EMS, GFP_KERNEL);
				if (videoBuffers.reference_virt == NULL) {
					printk(KERN_ERR "%s: unable to mem_chunk %d bytes memory\n",
		       				__FUNCTION__, REFERENCE_BUFFER_SIZE_EMS);
					return -ENOMEM;
				}	
				videoBuffers.reference_phys = (void *) __pa(videoBuffers.reference_virt);
			}

			if (videoBuffers.capture_virt == NULL) {
				videoBuffers.capture_virt = alloc_mem_chunk(CAPTURE_BUFFER_SIZE_EMS, GFP_KERNEL);
				if (videoBuffers.capture_virt == NULL) {
					printk(KERN_ERR "%s: unable to mem_chunk %d bytes memory\n",
		       				__FUNCTION__, CAPTURE_BUFFER_SIZE_EMS);
					return -ENOMEM;
				}	
				videoBuffers.capture_phys = (void *) __pa(videoBuffers.capture_virt);
			}

			gVideoEngineDisabled = 0;
			VideoEngineReset();
			ret = 0;
		}
		break;

	default:
		{
			ret = 3;
		}
	}
	return ret;
}

int ioaccess_release(struct inode *inode, struct file *filp)
{
	// standard release...
	Device_open--;
	VideoEngineHalt();
	return (0);
}

u8 aspeedi2c_read(u8 channel, u8 device_address, u8 register_index)
{
	int status;

	*(u32 *) (IO_ADDRESS(0x1e6e2000)) = 0x1688a8a8;
	status = *(u32 *) (IO_ADDRESS(0x1e6e2004));
//    *(u32 *)(IO_ADDRESS(0x1e6e2004)) = status | 0x04;
//    *(u32 *)(IO_ADDRESS(0x1e6e2004)) = status & ~(0x04);
	*(u32 *) (IO_ADDRESS(0x1e6e2004)) = 0;
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_FUNCTION_CONTROL_REGISTER +
			 0x40 * channel)) = 0;
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_AC_TIMING_REGISTER_1 +
			 0x40 * channel)) = AC_TIMING;
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_AC_TIMING_REGISTER_2 +
			 0x40 * channel)) = 0;
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_INTERRUPT_STATUS_REGISTER +
			 0x40 * channel)) = ALL_CLEAR;
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_FUNCTION_CONTROL_REGISTER +
			 0x40 * channel)) = MASTER_ENABLE;
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_INTERRUPT_CONTROL_REGISTER +
			 0x40 * channel)) = 0xAF;

// Start and Send device address
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_BYTE_BUFFER_REGISTER +
			 0x40 * channel)) = device_address;
	barrier();
//    mdelay (10);
//    status = *(u32 *)(IO_ADDRESS(I2C_BASE + I2C_BYTE_BUFFER_REGISTER + 0x40 * channel));
	*(u32 *) (IO_ADDRESS(I2C_BASE + I2C_COMMAND_REGISTER + 0x40 * channel))
	    = MASTER_START_COMMAND | MASTER_TX_COMMAND;
	do {
		status =
		    *(u32
		      *) (IO_ADDRESS(I2C_BASE + I2C_INTERRUPT_STATUS_REGISTER +
				     0x40 * channel)) & TX_ACK;
	} while (status != TX_ACK);
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_INTERRUPT_STATUS_REGISTER +
			 0x40 * channel)) = ALL_CLEAR;
// Send register index
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_BYTE_BUFFER_REGISTER +
			 0x40 * channel)) = register_index;
	*(u32 *) (IO_ADDRESS(I2C_BASE + I2C_COMMAND_REGISTER + 0x40 * channel))
	    = MASTER_TX_COMMAND;
	do {
		status =
		    *(u32
		      *) (IO_ADDRESS(I2C_BASE + I2C_INTERRUPT_STATUS_REGISTER +
				     0x40 * channel)) & TX_ACK;
	} while (status != TX_ACK);
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_INTERRUPT_STATUS_REGISTER +
			 0x40 * channel)) = ALL_CLEAR;
// Start, Send device address + 1 (read mode)
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_BYTE_BUFFER_REGISTER +
			 0x40 * channel)) = device_address + 1;
	*(u32 *) (IO_ADDRESS(I2C_BASE + I2C_COMMAND_REGISTER + 0x40 * channel))
	    =
	    MASTER_START_COMMAND | MASTER_TX_COMMAND | MASTER_RX_COMMAND |
	    RX_COMMAND_LIST;
	do {
		status =
		    *(u32
		      *) (IO_ADDRESS(I2C_BASE + I2C_INTERRUPT_STATUS_REGISTER +
				     0x40 * channel)) & RX_DONE;
	} while (status != RX_DONE);
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_INTERRUPT_STATUS_REGISTER +
			 0x40 * channel)) = ALL_CLEAR;
// Stop    
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_INTERRUPT_CONTROL_REGISTER +
			 0x40 * channel)) = 0xBF;
	*(u32 *) (IO_ADDRESS(I2C_BASE + I2C_COMMAND_REGISTER + 0x40 * channel))
	    = MASTER_STOP_COMMAND;
	do {
		status =
		    *(u32
		      *) (IO_ADDRESS(I2C_BASE + I2C_INTERRUPT_STATUS_REGISTER +
				     0x40 * channel)) & STOP_DONE;
	} while (status != STOP_DONE);
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_INTERRUPT_CONTROL_REGISTER +
			 0x40 * channel)) = 0xAF;
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_INTERRUPT_STATUS_REGISTER +
			 0x40 * channel)) = ALL_CLEAR;
	status =
	    (*(u32 *)
	     (IO_ADDRESS(I2C_BASE + I2C_BYTE_BUFFER_REGISTER + 0x40 * channel))
	     & 0xFF00) >> 8;

	return status;
}

u8 aspeedi2c_write(u8 channel, u8 device_address, u8 register_index,
		   u8 register_value)
{
	int status;

	*(u32 *) (IO_ADDRESS(0x1e6e2000)) = 0x1688a8a8;
	status = *(u32 *) (IO_ADDRESS(0x1e6e2004));
	*(u32 *) (IO_ADDRESS(0x1e6e2004)) = status | 0x04;
	*(u32 *) (IO_ADDRESS(0x1e6e2004)) = status & ~(0x04);
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_FUNCTION_CONTROL_REGISTER +
			 0x40 * channel)) = 0;
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_AC_TIMING_REGISTER_1 +
			 0x40 * channel)) = AC_TIMING;
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_AC_TIMING_REGISTER_2 +
			 0x40 * channel)) = 0;
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_INTERRUPT_STATUS_REGISTER +
			 0x40 * channel)) = ALL_CLEAR;
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_FUNCTION_CONTROL_REGISTER +
			 0x40 * channel)) = MASTER_ENABLE;
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_INTERRUPT_CONTROL_REGISTER +
			 0x40 * channel)) = 0xAF;

// Start and Send device address
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_BYTE_BUFFER_REGISTER +
			 0x40 * channel)) = device_address;
	*(u32 *) (IO_ADDRESS(I2C_BASE + I2C_COMMAND_REGISTER + 0x40 * channel))
	    = MASTER_START_COMMAND | MASTER_TX_COMMAND;
	do {
		status =
		    *(u32
		      *) (IO_ADDRESS(I2C_BASE + I2C_INTERRUPT_STATUS_REGISTER +
				     0x40 * channel)) & TX_ACK;
	} while (status != TX_ACK);
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_INTERRUPT_STATUS_REGISTER +
			 0x40 * channel)) = ALL_CLEAR;
// Send register index
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_BYTE_BUFFER_REGISTER +
			 0x40 * channel)) = register_index;
	*(u32 *) (IO_ADDRESS(I2C_BASE + I2C_COMMAND_REGISTER + 0x40 * channel))
	    = MASTER_TX_COMMAND;
	do {
		status =
		    *(u32
		      *) (IO_ADDRESS(I2C_BASE + I2C_INTERRUPT_STATUS_REGISTER +
				     0x40 * channel)) & TX_ACK;
	} while (status != TX_ACK);
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_INTERRUPT_STATUS_REGISTER +
			 0x40 * channel)) = ALL_CLEAR;
// Send register value
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_BYTE_BUFFER_REGISTER +
			 0x40 * channel)) = register_value;
	*(u32 *) (IO_ADDRESS(I2C_BASE + I2C_COMMAND_REGISTER + 0x40 * channel))
	    = MASTER_TX_COMMAND;
	do {
		status =
		    *(u32
		      *) (IO_ADDRESS(I2C_BASE + I2C_INTERRUPT_STATUS_REGISTER +
				     0x40 * channel)) & TX_ACK;
	} while (status != TX_ACK);
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_INTERRUPT_STATUS_REGISTER +
			 0x40 * channel)) = ALL_CLEAR;
// Stop    
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_INTERRUPT_CONTROL_REGISTER +
			 0x40 * channel)) = 0xBF;
	*(u32 *) (IO_ADDRESS(I2C_BASE + I2C_COMMAND_REGISTER + 0x40 * channel))
	    = MASTER_STOP_COMMAND;
	do {
		status =
		    *(u32
		      *) (IO_ADDRESS(I2C_BASE + I2C_INTERRUPT_STATUS_REGISTER +
				     0x40 * channel)) & STOP_DONE;
	} while (status != STOP_DONE);
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_INTERRUPT_CONTROL_REGISTER +
			 0x40 * channel)) = 0xAF;
	*(u32
	  *) (IO_ADDRESS(I2C_BASE + I2C_INTERRUPT_STATUS_REGISTER +
			 0x40 * channel)) = ALL_CLEAR;
	status =
	    (*(u32 *)
	     (IO_ADDRESS(I2C_BASE + I2C_BYTE_BUFFER_REGISTER + 0x40 * channel))
	     & 0xFF00) >> 8;

	return status;
}

static void dirtyHWCursorCache(void)
{
	int i;
	for (i = 0; i < MAX_HWCURSOR_CACHE;
	     hwCursorCache[i++].cCursorCacheDirty = 1) ;
}

static void clearHWCursorCache(void)
{
	int i;
	for (i = 0; i < MAX_HWCURSOR_CACHE;
	     hwCursorCache[i++].cCursorCacheNumber = 0) ;
	g_pLastHWCursor = NULL;
	g_uiLastHWCursorChecksum = 0;

}

static void removeHWCursorCacheEntry(PHWCURSORDESCRIPTOR pHWCursor)
{
	int i;
	PHWCURSORCACHEENTRY pHWCursorCacheEntry;
	for (i = 0; i < MAX_HWCURSOR_CACHE; i++) {
		pHWCursorCacheEntry = &hwCursorCache[i];
		if (pHWCursorCacheEntry->cCursorCacheNumber != 0 &&
		    pHWCursor == pHWCursorCacheEntry->theCursorAddress) {
			printk
			    ("drv - %s[%d] DEL HWCursorCache %d = 0x%8.8X checksum 0x%8.8X\n",
			     __FUNCTION__, __LINE__,
			     (int)pHWCursorCacheEntry->cCursorCacheNumber,
			     (int)pHWCursorCacheEntry->theCursorAddress,
			     (int)pHWCursorCacheEntry->theCursor.ulChecksum);
			pHWCursorCacheEntry->cCursorCacheNumber = 0;
		}
	}
}

static PHWCURSORCACHEENTRY getHWCursorCacheEntry(int iCursorNumber)
{
	PHWCURSORCACHEENTRY pthisHWCursor = NULL;
	if (iCursorNumber >= 1 &&
	    iCursorNumber <= MAX_HWCURSOR_CACHE &&
	    hwCursorCache[iCursorNumber - 1].cCursorCacheNumber ==
	    iCursorNumber)
		pthisHWCursor = &hwCursorCache[iCursorNumber - 1];
	return (pthisHWCursor);
}

static char getHWCursorCacheNumber(PHWCURSORDESCRIPTOR pHWCursor,
				   unsigned char cCursorType, u16 usCursorXOffset, u16 usCursorYOffset)
{
	char cRet = 0;
	int i;
	int iFirstAvailable = -1;
	int iLRU = -1;
	unsigned long ulLRUCount = 0xFFFFFFFF;
	PHWCURSORCACHEENTRY pHWCursorCacheEntry;
	for (i = 0; i < MAX_HWCURSOR_CACHE; i++) {
		pHWCursorCacheEntry = &hwCursorCache[i];
		if (pHWCursorCacheEntry->cCursorCacheNumber == 0) {
			// Remember the first available entry in case we need it
			if (iFirstAvailable < 0)
				iFirstAvailable = i;
		} else {
			if (pHWCursorCacheEntry->ulLRUCount < ulLRUCount) {
				// Remember the least recently used entry in case we need it
				iLRU = i;
				ulLRUCount = pHWCursorCacheEntry->ulLRUCount;
			}
			if (pHWCursor == pHWCursorCacheEntry->theCursorAddress) {
				// This entry matches, so return it, increment its LRU count
				cRet = pHWCursorCacheEntry->cCursorCacheNumber;
				pHWCursorCacheEntry->ulLRUCount++;
				break;
			}
		}
	}
	// Check to see if we found the cursor for this frame
	if (cRet == 0) {
		if (iFirstAvailable >= 0) {
			pHWCursorCacheEntry = &hwCursorCache[iFirstAvailable];
			pHWCursorCacheEntry->cCursorCacheNumber =
			    iFirstAvailable + 1;
		} else {
			pHWCursorCacheEntry = &hwCursorCache[iLRU];
			printk("drv-%s[%d] DEL HWCursorCache %d\n",
			       __FUNCTION__, __LINE__,
			       (int)pHWCursorCacheEntry->cCursorCacheNumber);
			pHWCursorCacheEntry->cCursorCacheNumber = iLRU + 1;
		}
		// Here with pHWCursorCacheEntry for this new cursor
		// Mark this entry as dirty and in need of broadcast to clients
		pHWCursorCacheEntry->cCursorType = cCursorType;
		pHWCursorCacheEntry->cCursorCacheDirty = 1;

		pHWCursorCacheEntry->theCursorAddress = pHWCursor;
		pHWCursorCacheEntry->ulLRUCount = 1;
		pHWCursorCacheEntry->usCursorXOffset = usCursorXOffset;
		pHWCursorCacheEntry->usCursorYOffset = usCursorYOffset;

		memcpy((char *)&pHWCursorCacheEntry->theCursor,
		       (char *)pHWCursor,
		       sizeof(pHWCursorCacheEntry->theCursor));
		cRet = pHWCursorCacheEntry->cCursorCacheNumber;
		printk
		    ("drv-%s[%d] ADD HWCursorCache %d = 0x%8.8X checksum 0x%8.8X\n",
		     __FUNCTION__, __LINE__,
		     (int)pHWCursorCacheEntry->cCursorCacheNumber,
		     (int)pHWCursor, (int)pHWCursor->ulChecksum);
	}
	// Return the cursor cache number (NOT INDEX) for this pHWCursor
	return (cRet);
}

static int isHWCursorCacheEntryDirty(int iCacheNumber)
{
	int iRet = 0;
	int i = iCacheNumber - 1;
	if ((i >= 0 && i < MAX_HWCURSOR_CACHE)
	    && hwCursorCache[i].cCursorCacheDirty)
		iRet = 1;
	return (iRet);
}

static void processSOF(PFRAMEDESCRIPTOR pThisEntry)
{
	u32 uiReg;
	PHWCURSORDESCRIPTOR pHWCursor;
	u32 uiHWCursorChecksum;
	u16 usCursorXOffset, usCursorYOffset;
	PHWCURSORCACHEENTRY pReqHWCursor;

	// no previous packets for this frame, so this is the SOF
	if (videoConfig.CaptureMode == CAPTUREMODE_MULTIPLE) {
		// Add 16 to the beginning of the frame when using stream mode
		// to skip the 16-byte frame header
		pThisEntry->ulDataStartOffset += 16;
		// Subtract 4 from the segmentDataCount, too
		pThisEntry->ulDataCount -= 4;
	}
	gSegmentNumber = 1;
	pThisEntry->cFrameStatus |= VIDEO_FRAME_STATUS_SOF;
	pThisEntry->cMode =
	    (REGISTERREAD(VIDEOENGINE_SEQUENCE_CONTROL) >> 10) & 0x01;
	pThisEntry->cY_JPEGTableSelector =
	    (REGISTERREAD(VIDEO_COMPRESS_CONTROL) >> 11) & 0x1f;
	pThisEntry->cUV_JPEGTableSelector =
	    (REGISTERREAD(VIDEO_COMPRESS_CONTROL) >> 6) & 0x1f;

	if ((videoStatus.UsingDirectMode) &&
	    (videoStatus.HaveHostHWCursorSupport)) {
		// Here to do hardware cursor processing

		// If a full frame has been requested, we dirty the entire
		// cursor cache to be sure that the cursor for subsequent frames are
		// retransmitted by the VideoEngine.  This is needed so that
		// cursors are transmitted PRIOR to the frame that they are used in.
		if (gFullFrameRequested)
			dirtyHWCursorCache();

		// Read and store the current h/w cursor address in the ring entry
		uiReg = REGISTERREAD(VIDEO_SCRATCH_344_CURSOR_POS);
		pThisEntry->usCursorX = uiReg & VIDEO_CURSOR_POS_X_MASK;
		pThisEntry->usCursorY =
		    (uiReg & VIDEO_CURSOR_POS_Y_MASK) >>
		    VIDEO_CURSOR_POS_Y_SHIFT;
		uiReg = REGISTERREAD(VIDEO_SCRATCH_340_CURSOR_STATUS);
		pThisEntry->usCursorXOffset =
		    (uiReg & VIDEO_CURSOR_X_OFFSET_MASK) >>
		    VIDEO_CURSOR_X_OFFSET_SHIFT;
		pThisEntry->usCursorYOffset =
		    (uiReg & VIDEO_CURSOR_Y_OFFSET_MASK) >>
		    VIDEO_CURSOR_Y_OFFSET_SHIFT;
		usCursorXOffset = pThisEntry->usCursorXOffset;
		usCursorYOffset = pThisEntry->usCursorYOffset;

		pThisEntry->cCursorType =
		    (uiReg & VIDEO_CURSOR_TYPE_MASK) >> VIDEO_CURSOR_TYPE_SHIFT;
		// As read from the HWDESCRIPTOR VGA memory
		pHWCursor =
		    (PHWCURSORDESCRIPTOR) &
		    VGA_Linear_address[REGISTERREAD
				       (VIDEO_SCRATCH_348_CURSOR_ADDRESS)];

		uiHWCursorChecksum = pHWCursor->ulChecksum;

		// Have current cursor offset address and checksum.  
		// See if the cursor has been changed since the last frame.
		if ((pHWCursor != g_pLastHWCursor) ||
		    (uiHWCursorChecksum != g_uiLastHWCursorChecksum)
		    ) {
			// Locate any existing entry for the pHWCursor address and
			// delete that entry as it is no longer valid
			removeHWCursorCacheEntry(pHWCursor);
		}
		// Get the cache entry for this cursor address and checksum
		// If it is not currently in the cache, it will be added.
		pThisEntry->cCursorCacheNumber =
		    getHWCursorCacheNumber(pHWCursor, pThisEntry->cCursorType, usCursorXOffset, usCursorYOffset);

		pReqHWCursor = getHWCursorCacheEntry(pThisEntry->cCursorCacheNumber);
		if ((pReqHWCursor->usCursorXOffset != usCursorXOffset) ||
		    (pReqHWCursor->usCursorYOffset != usCursorYOffset)) {
			pThisEntry->usCursorXOffset = pReqHWCursor->usCursorXOffset;
			pThisEntry->usCursorYOffset = pReqHWCursor->usCursorYOffset;
		}

		// Save the last known cursor values
		g_uiLastHWCursorChecksum = uiHWCursorChecksum;
		g_pLastHWCursor = pHWCursor;

		// If the cache entry is dirty, so mark the ring entry
		if (isHWCursorCacheEntryDirty(pThisEntry->cCursorCacheNumber))
			pThisEntry->cFrameStatus |= VIDEO_FRAME_CURSOR_DIRTY;
	} else {
		pThisEntry->cCursorCacheNumber = 0;
		pThisEntry->usCursorX = 0;
		pThisEntry->usCursorY = 0;
		pThisEntry->usCursorXOffset = 0;
		pThisEntry->usCursorYOffset = 0;
	}
}

static irqreturn_t video_interrupt(int irq, void *dev_id)
{
	int INTmask,
	    INTstatus,
	    INTorigStatus,
	    curCompressWriteOffset,
	    ackCompressWriteOffset, curFrameEndOffset, curFrameNumber;
	int Ring_Ready = 0;
	FRAMEDESCRIPTOR thisSegment;
	int writeFrame;
	PFRAMEDESCRIPTOR pRingEntry = NULL;

	INTmask = REGISTERREAD(VIDEO_INTERRUPT_CONTROL);

	// Get current interrupt status.  Note that ALL of these status bits will
	// be cleared prior to exiting this interrupt.
	INTorigStatus = INTstatus = REGISTERREAD(VIDEO_INTERRUPT_STATUS);

	// Only process interrupt status bits we are interested in...
	INTstatus &= INTmask;

	if (gDebugFlags & DEBUG_TRACK_INTERRUPTS)
		printk
		    ("drv-%s[%d] INT INTmask %2.2X INTorigStatus %2.2X Intstatus %2.2X\n",
		     __FUNCTION__, __LINE__, INTmask, INTorigStatus, INTstatus);

	// mode detect ready
	if (INTstatus & VIDEO_INTERRUPT_MODE_READY) {
		INTstatus &= ~VIDEO_INTERRUPT_MODE_READY;
		Mode_Detect_Ready = 1;
		// Turn off mode detect interrupt
		REGISTERANDWRITE(VIDEO_INTERRUPT_CONTROL, 0xFFFFFFEF, 0);
	}
	// mode detect watchdog timeout 
	if (INTstatus & VIDEO_INTERRUPT_MODE_WATCHDOG) {
		INTstatus &= ~VIDEO_INTERRUPT_MODE_WATCHDOG;
		Mode_Watchdog = 1;
	} else {
		if (gCaptureActive) {
			curCompressWriteOffset =
			    REGISTERREAD(VIDEO_COMPRESS_WRITE_OFFSET);
			ackCompressWriteOffset = curCompressWriteOffset;
			curFrameEndOffset = REGISTERREAD(COMPRESS_BUFFER_LAST_END_FRAME_OFFSET) & 0x003FFFF8;	// Valid only if EOF
			curFrameNumber =
			    REGISTERREAD(COMPRESS_FRAME_COUNT_REGISTER);
			thisSegment.cFrameStatus = 0;
			thisSegment.ulCompressBufIndex =
			    videoStatus.CurrentCompressBufIndex;

			if (videoConfig.CaptureMode == CAPTUREMODE_SINGLE) {
				// Check out interrupts for Single frame capture mode first
				if (INTstatus &
				    VIDEO_INTERRUPT_COMPRESSION_COMPLETE)
					Compression_Complete = 1;
				if (INTstatus &
				    VIDEO_INTERRUPT_CAPTURE_COMPLETE)
					Capture_Complete = 1;
				if (Compression_Complete && Capture_Complete) {
					Compression_Complete =
					    Capture_Complete = 0;
					// Fack out a frame complete to process this complete single frame below...
					INTstatus |=
					    VIDEO_INTERRUPT_FRAME_COMPLETE;
				}
			}
			// Now, check out interrupts for multiple frame capture mode
			// frame complete
			if (INTstatus & VIDEO_INTERRUPT_FRAME_COMPLETE) {
				// Turn off both frame complete and packet ready so that we don't
				// also process the packet ready interrupt on this frame should
				// that occur.
				INTstatus &=
				    ~(VIDEO_INTERRUPT_FRAME_COMPLETE |
				      VIDEO_INTERRUPT_PACKET_READY);
				Ring_Ready = 1;

				// Get the frame information and store it in the temporary ring entry
				thisSegment.ulDataStartOffset =
				    gPrevRingEndOffset;
				thisSegment.ulDataEndOffset =
				    curCompressWriteOffset = curFrameEndOffset;
				thisSegment.ulDataCount =
				    ((thisSegment.ulDataEndOffset -
				      thisSegment.ulDataStartOffset) %
				     videoStatus.VideoCompressBufferSize) / 4;
				thisSegment.ulFrameNumber = curFrameNumber;
				thisSegment.cFrameStatus = 0;

				// Update ring entry status bits
				// Check to see if this is the beginning of a frame and update the ring entry
				if (gSOF) {
					processSOF(&thisSegment);
				} else {
					gSegmentNumber++;
				}
				// Store update frame status and segment number
				thisSegment.cFrameStatus |=
				    VIDEO_FRAME_STATUS_EOF;
				thisSegment.ulSegmentNumber = gSegmentNumber;

				// The following are only available for a ring entry with VIDEO_FRAME_STATUS_EOF
				thisSegment.usHorizontalXRes =
				    videoStatus.SourceModeInfo.X;
				thisSegment.usVeritcalYRes =
				    videoStatus.SourceModeInfo.Y;
				thisSegment.ulJiffies = jiffies;

				// Mark this compression buffer as FULL
				videoStatus.VideoCompressBufStatus[videoStatus.
								   CurrentCompressBufIndex]
				    = COMPRESSION_BUFFER_STATUS_FULL;
				// If we are in single frame capture mode, then the engine is now Idle.
				if (gCaptureActive == CAPTUREMODE_SINGLE)
					gVideoEngineIdle = 1;
				// The next packet will be the start of a frame
				gSOF = 1;

				if (gDebugFlags &
				    DEBUG_INTERRUPT_LEVEL_RING_CHECK)
					printk
					    ("drv-%s[%d] FRAME_COMPLETE Buf %d segment %d-%d %6.6X-%6.6X count %d %s%s%s%s\n",
					     __FUNCTION__, __LINE__,
					     (int)thisSegment.
					     ulCompressBufIndex,
					     (int)thisSegment.ulFrameNumber,
					     (int)thisSegment.ulSegmentNumber,
					     (int)thisSegment.ulDataStartOffset,
					     (int)thisSegment.ulDataEndOffset,
					     (int)thisSegment.ulDataCount,
					     (thisSegment.
					      cFrameStatus &
					      VIDEO_FRAME_STATUS_SOF) ? "SOF " :
					     "",
					     (thisSegment.
					      cFrameStatus &
					      VIDEO_FRAME_STATUS_EOF) ? "EOF " :
					     "",
					     (thisSegment.
					      cFrameStatus &
					      VIDEO_FRAME_STATUS_RFR) ? "RFR " :
					     "",
					     (thisSegment.
					      cFrameStatus &
					      VIDEO_FRAME_STATUS_NOCHG) ? "NCH "
					     : "");

			}	// end if frame complete interrupt

			// A packet segment (up to a packet boundary) is available
			if (INTstatus & VIDEO_INTERRUPT_PACKET_READY) {
				INTstatus &= ~VIDEO_INTERRUPT_PACKET_READY;
				Ring_Ready = 1;

				// Get the frame information and store it in the temporary ring entry
				thisSegment.ulDataStartOffset =
				    gPrevRingEndOffset;
				thisSegment.ulDataEndOffset =
				    curCompressWriteOffset;
				thisSegment.ulDataCount =
				    ((thisSegment.ulDataEndOffset -
				      thisSegment.ulDataStartOffset) %
				     videoStatus.VideoCompressBufferSize) / 4;
				// NOTE that we increment the frame number, because the packet is for the next frame
				// that will be completed.  The video engine does not increment the frame number (apparently)
				// until it gets to 'frame complete' state.  This makes the earlier segments belong to the
				// next completed frame which is actually what is happening.
				thisSegment.ulFrameNumber = curFrameNumber + 1;
				thisSegment.cFrameStatus = 0;

				// Update ring entry status bits
				// Check to see if this is the beginning of a frame and mark the ring entry
				if (gSOF) {
					processSOF(&thisSegment);
					// The next packet will not be SOF
					gSOF = 0;
				} else {
					gSegmentNumber++;
				}
				// Store the segment number
				thisSegment.ulSegmentNumber = gSegmentNumber;

				if (INTstatus & VIDEO_INTERRUPT_FRAME_COMPLETE) {
					// This should not happen, but mark EOF in case it does.
					thisSegment.cFrameStatus |=
					    VIDEO_FRAME_STATUS_EOF;
					gSOF = 1;
				}

				if (gDebugFlags &
				    DEBUG_INTERRUPT_LEVEL_RING_CHECK)
					printk
					    ("drv-%s[%d] PACKET_READY Buf %d segment %d-%d %6.6X-%6.6X count %d %s%s%s%s\n",
					     __FUNCTION__, __LINE__,
					     (int)thisSegment.
					     ulCompressBufIndex,
					     (int)thisSegment.ulFrameNumber,
					     (int)thisSegment.ulSegmentNumber,
					     (int)thisSegment.ulDataStartOffset,
					     (int)thisSegment.ulDataEndOffset,
					     (int)thisSegment.ulDataCount,
					     (thisSegment.
					      cFrameStatus &
					      VIDEO_FRAME_STATUS_SOF) ? "SOF " :
					     "",
					     (thisSegment.
					      cFrameStatus &
					      VIDEO_FRAME_STATUS_EOF) ? "EOF " :
					     "",
					     (thisSegment.
					      cFrameStatus &
					      VIDEO_FRAME_STATUS_RFR) ? "RFR " :
					     "",
					     (thisSegment.
					      cFrameStatus &
					      VIDEO_FRAME_STATUS_NOCHG) ? "NCH "
					     : "");

			}	// end if packet ready

			// Update the Processing Offset Register to acknowledge we are processing this frame
			REGISTERWRITE(VIDEO_COMPRESS_PROCESSING_OFFSET,
				      ackCompressWriteOffset);
			// Set up for next ring entry
			gPrevRingEndOffset = curCompressWriteOffset;

			if (Ring_Ready) {
				// We have a video segment to store
				if (gFullFrameRequested) {
					if ((REGISTERREAD
					     (VIDEOENGINE_SEQUENCE_CONTROL) &
					     0x40) == 0) {
						thisSegment.cFrameStatus |=
						    VIDEO_FRAME_STATUS_RFR;
						gFullFrameRequested = 0;
					}
				}
				// Init this segment's initial frame count
				thisSegment.ulRepeatCount = 1;
				thisSegment.cFrameStatus |= VIDEO_FRAME_STATUS_PRESENT;	// indicate a frame is present

				// Check here to see if we can just update the current writable ring entry...
				writeFrame = GetCurrentWritableRingEntry();
				if (writeFrame >= 0) {
					// There is a possibility of updating the current ring entry
					pRingEntry =
					    &theRing.RingEntry[writeFrame];

					// Check here to see if we have a 'no change' frame.
					// Need to have SOF, EOF and length of 128 with the characters
					// 00 00 00 90 00 00 00 00 00 00 00 00 as the first 12 chars after
					// the 16-byte ASPEED frame header...
					if (((thisSegment.
					      cFrameStatus &
					      VIDEO_FRAME_STATUS_SINGLE) ==
					     VIDEO_FRAME_STATUS_SINGLE)
					    && (thisSegment.ulDataCount <= 32)
					    &&
					    (memcmp
					     (noChangeFrame,
					      &Data_Linear_address[thisSegment.
								   ulDataStartOffset],
					      12) == 0)
					    ) {
						// This is a no change frame.  We only store one of these for
						// every second's worth of frames.  This reduces the number of
						// ring entries that are required during no changes.
						thisSegment.cFrameStatus |=
						    VIDEO_FRAME_STATUS_NOCHG;

						// Check to see if there is a current ring entry pending (not read
						// by program level) that contains a no change frame.  If so, then
						// check out if we want to simply update the count for the existing
						// ring entry.
						// Now, see if it is a no change frame
						if ((pRingEntry->
						     cFrameStatus &
						     VIDEO_FRAME_STATUS_NOCHG)
						    && (pRingEntry->
							ulRepeatCount <
							videoStatus.FrameRate)
						    ) {
							pRingEntry->
							    ulRepeatCount++;
							// Update the expected frame number
							gPrevFrameNumber++;
							if (gDebugFlags &
							    DEBUG_INTERRUPT_LEVEL_RING_CHECK)
							{
								printk
								    ("drv-%s[%d] Frame %d is a no change frame, Ring %d updated\n",
								     __FUNCTION__,
								     __LINE__,
								     (int)
								     thisSegment.
								     ulFrameNumber,
								     writeFrame);
								printk
								    ("drv-%s[%d] Ring %d Buf %d segment %d-%d %6.6X-%6.6X count %d count %d %s%s%s%s\n",
								     __FUNCTION__,
								     __LINE__,
								     writeFrame,
								     (int)
								     pRingEntry->
								     ulCompressBufIndex,
								     (int)
								     pRingEntry->
								     ulFrameNumber,
								     (int)
								     pRingEntry->
								     ulSegmentNumber,
								     (int)
								     pRingEntry->
								     ulDataStartOffset,
								     (int)
								     pRingEntry->
								     ulDataEndOffset,
								     (int)
								     pRingEntry->
								     ulDataCount,
								     (int)
								     pRingEntry->
								     ulRepeatCount,
								     (pRingEntry->
								      cFrameStatus
								      &
								      VIDEO_FRAME_STATUS_SOF)
								     ? "SOF " :
								     "",
								     (pRingEntry->
								      cFrameStatus
								      &
								      VIDEO_FRAME_STATUS_EOF)
								     ? "EOF " :
								     "",
								     (pRingEntry->
								      cFrameStatus
								      &
								      VIDEO_FRAME_STATUS_RFR)
								     ? "RFR " :
								     "",
								     (pRingEntry->
								      cFrameStatus
								      &
								      VIDEO_FRAME_STATUS_NOCHG)
								     ? "NCH " :
								     "");
							}
						}	// end if we updated a current writable no change ring entry
						else {
							// Didn't use the current writable entry, make a new ring entry below
							pRingEntry = NULL;
						}
					}	// end if this was a no change frame
					else if ((pRingEntry->
						  cFrameStatus &
						  VIDEO_FRAME_STATUS_EOF) ==
						 0) {
						// We can update the current writable frame entry with this
						// segment's data.
						pRingEntry->ulDataEndOffset =
						    thisSegment.ulDataEndOffset;
						pRingEntry->ulDataCount =
						    ((pRingEntry->
						      ulDataEndOffset -
						      pRingEntry->
						      ulDataStartOffset) %
						     videoStatus.
						     VideoCompressBufferSize) /
						    4;
						// Copy the EOF and RFR bits from the new segment
						pRingEntry->cFrameStatus |=
						    (thisSegment.
						     cFrameStatus &
						     (VIDEO_FRAME_STATUS_EOF |
						      VIDEO_FRAME_STATUS_RFR));
						if (gDebugFlags &
						    DEBUG_INTERRUPT_LEVEL_RING_CHECK)
						{
							printk
							    ("drv-%s[%d] Frame %d updated to Ring %d\n",
							     __FUNCTION__,
							     __LINE__,
							     (int)thisSegment.
							     ulFrameNumber,
							     writeFrame);
							printk
							    ("drv-%s[%d] Ring %d buf %d segment %d-%d %6.6X-%6.6X count %d repeat %d %s%s%s%s\n",
							     __FUNCTION__,
							     __LINE__,
							     writeFrame,
							     (int)pRingEntry->
							     ulCompressBufIndex,
							     (int)pRingEntry->
							     ulFrameNumber,
							     (int)pRingEntry->
							     ulSegmentNumber,
							     (int)pRingEntry->
							     ulDataStartOffset,
							     (int)pRingEntry->
							     ulDataEndOffset,
							     (int)pRingEntry->
							     ulDataCount,
							     (int)pRingEntry->
							     ulRepeatCount,
							     (pRingEntry->
							      cFrameStatus &
							      VIDEO_FRAME_STATUS_SOF)
							     ? "SOF " : "",
							     (pRingEntry->
							      cFrameStatus &
							      VIDEO_FRAME_STATUS_EOF)
							     ? "EOF " : "",
							     (pRingEntry->
							      cFrameStatus &
							      VIDEO_FRAME_STATUS_RFR)
							     ? "RFR " : "",
							     (pRingEntry->
							      cFrameStatus &
							      VIDEO_FRAME_STATUS_NOCHG)
							     ? "NCH " : "");
						}
					}	// end if we were able to update the current frame information
					else {
						// Weren't able to use the current writable entry, make a new ring entry below
						pRingEntry = NULL;
					}
				}	// end if there is an entry that we can update

				// Check to see if we did not yet store the current segment data
				if (pRingEntry == NULL) {
					writeFrame = GetNextWritableRingEntry();
					if (writeFrame >= 0) {
						pRingEntry =
						    &theRing.
						    RingEntry[writeFrame];
						*pRingEntry = thisSegment;	// structure copy
						if (gDebugFlags &
						    DEBUG_INTERRUPT_LEVEL_RING_CHECK)
						{
							if (pRingEntry->
							    ulDataCount == 0) {
								printk
								    ("drv-%s[%d] Ring %d Frame %d is a zero frame\n",
								     __FUNCTION__,
								     __LINE__,
								     writeFrame,
								     (int)
								     pRingEntry->
								     ulFrameNumber);
							}

							if (gPrevFrameNumber <
							    0)
								gPrevFrameNumber
								    =
								    pRingEntry->
								    ulFrameNumber
								    - 1;
							printk
							    ("drv-%s[%d] Ring %d Buf %d segment %d-%d %6.6X-%6.6X count %d repeat %d %s%s%s%s\n",
							     __FUNCTION__,
							     __LINE__,
							     writeFrame,
							     (int)pRingEntry->
							     ulCompressBufIndex,
							     (int)pRingEntry->
							     ulFrameNumber,
							     (int)pRingEntry->
							     ulSegmentNumber,
							     (int)pRingEntry->
							     ulDataStartOffset,
							     (int)pRingEntry->
							     ulDataEndOffset,
							     (int)pRingEntry->
							     ulDataCount,
							     (int)pRingEntry->
							     ulRepeatCount,
							     (pRingEntry->
							      cFrameStatus &
							      VIDEO_FRAME_STATUS_SOF)
							     ? "SOF " : "",
							     (pRingEntry->
							      cFrameStatus &
							      VIDEO_FRAME_STATUS_EOF)
							     ? "EOF " : "",
							     (pRingEntry->
							      cFrameStatus &
							      VIDEO_FRAME_STATUS_RFR)
							     ? "RFR " : "",
							     (pRingEntry->
							      cFrameStatus &
							      VIDEO_FRAME_STATUS_NOCHG)
							     ? "NCH " : "");

							if (pRingEntry->
							    ulFrameNumber >
							    gPrevFrameNumber) {
								// Update the expected frame number
								gPrevFrameNumber++;
								if (pRingEntry->
								    ulFrameNumber
								    !=
								    gPrevFrameNumber)
								{
									printk
									    ("drv-%s[%d] Prev Frame %d next frame %d\n",
									     __FUNCTION__,
									     __LINE__,
									     gPrevFrameNumber,
									     (int)
									     pRingEntry->
									     ulFrameNumber);
									gPrevFrameNumber
									    =
									    pRingEntry->
									    ulFrameNumber;
								}
							}
						}
						// While a capture is active, check frame rates and ring depth...
						UpdateRingCounts();

						// Do congestion processing
						if (videoStatus.FrameRate !=
						    videoConfig.
						    ulCongestionFrameRate
						    && theRing.ulFrameCount >=
						    videoConfig.
						    ulHighWaterMark) {
							// Set congestion frame rate...
							videoStatus.FrameRate =
							    videoConfig.
							    ulCongestionFrameRate;
							SetFrameRate
							    (videoStatus.
							     FrameRate);
							theRing.ulRingStatus |=
							    VIDEO_ENGINE_CONGESTED;
						} else if (videoStatus.
							   FrameRate !=
							   videoConfig.
							   ulNormalFrameRate
							   && theRing.
							   ulFrameCount <=
							   videoConfig.
							   ulLowWaterMark) {
							// Set the frame rate for normal
							videoStatus.FrameRate =
							    videoConfig.
							    ulNormalFrameRate;
							SetFrameRate
							    (videoStatus.
							     FrameRate);
							theRing.ulRingStatus &=
							    ~VIDEO_ENGINE_CONGESTED;
						}
						// Check for an engine stall...
						if (theRing.ulRingCount >=
						    videoConfig.ulHaltMark) {
							if (gDebugFlags &
							    DEBUG_RING_STATUS)
								printk
								    ("drv-%s[%d] Stall declared ringCount %d\n",
								     __FUNCTION__,
								     __LINE__,
								     (int)
								     theRing.
								     ulRingCount);
							if (!theRing.bStalled) {
								theRing.
								    ulRingStatus
								    |=
								    VIDEO_CAPTURE_STALLED;
								VideoEngineStall
								    ();
							}
						}
					}	// end if we found a ring entry for the current video segment
					else {
						// Ring Overrun
						theRing.ulRingStatus |=
						    VIDEO_CAPTURE_OVERRUN;
						if (gDebugFlags &
						    DEBUG_RING_STATUS)
							printk
							    ("drv-%s[%d] Ring overrun declared ringCount %d\n",
							     __FUNCTION__,
							     __LINE__,
							     (int)theRing.
							     ulRingCount);
					}	// end else there are no ring entries for this video segment
				}	// end if we had to make a new ring entry for this frame
			}	// end if have a video segment to store
		}		// end if capture is active...

		// Check for unprocessed packet ready interrupt.  If we don't update the processing offset
		// register with the fact that we acknowledge this data, the interrupt will reassert.
		// SO, just read wherever the engine thinks it is and indicate that we have processed it all.
		// TODO seems like restarting a capture ought to clear this, or maybe we should do this before
		// the interrupts are enabled.
		if (INTstatus & VIDEO_INTERRUPT_PACKET_READY) {
			unsigned long ackAddress;
			INTstatus &= ~VIDEO_INTERRUPT_PACKET_READY;
			ackAddress = REGISTERREAD(VIDEO_COMPRESS_WRITE_OFFSET);
			// Update the Processing Offset Register to acknowledge we are processing this frame
			REGISTERWRITE(VIDEO_COMPRESS_PROCESSING_OFFSET,
				      (ackAddress & 0x003FFF80));
		}

	}			// end if this was not a mode watchdog situation

	intmasks[nmask++] =
	    (INTorigStatus & 0x0000003F) | ((gSegmentNumber & 0xFF) <<
					    8) |
	    ((REGISTERREAD(COMPRESS_FRAME_COUNT_REGISTER) & 0x000000FF)
	     << 16);
	if (nmask >= MAX_INTS)
		nmask = 0;

	// Clear ALL status bits for this interrupt...
	REGISTERWRITE(VIDEO_INTERRUPT_STATUS, INTorigStatus);

	// check for wakeup...
	if ((Mode_Detect_Ready == 1) ||
	    (Mode_Watchdog == 1) || (Ring_Ready == 1)
	    ) {
		flag = 1;
		wake_up_interruptible(&my_queue);
	}

	return IRQ_HANDLED;
}

int aess_video_init(void)
{
	int result, retval;
	extern void * alloc_mem_chunk(size_t size, gfp_t gfp_mask);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
        u32 irqFlags = SA_INTERRUPT;
#else
        u32 irqFlags = IRQF_DISABLED | IRQF_SAMPLE_RANDOM;
#endif

//    SET_MODULE_OWNER (&ioaccess_fops);
	result =
	    register_chrdev(ioaccess_major, "aess_video", &ioaccess_fops);
	if (result < 0) {
		printk("drv-%s[%d]<0>" "<0>Can't get major.\n", __FUNCTION__,
		       __LINE__);
		return result;
	}

	if (ioaccess_major == 0) {
		ioaccess_major = result;
	}

	VideoEngineReset();

	memset((char *)&videoConfig, 0, sizeof(videoConfig));
	memset((char *)&videoStatus, 0, sizeof(videoStatus));
	memset((char *)&videoBuffers, 0, sizeof(videoBuffers));

	videoBuffers.reference_size = REFERENCE_BUFFER_SIZE_EMS;
	videoBuffers.reference_virt = alloc_mem_chunk(REFERENCE_BUFFER_SIZE_EMS, GFP_KERNEL);
	if (videoBuffers.reference_virt == NULL) {
		printk(KERN_ERR "%s: unable to mem_chunk %d bytes memory\n",
		       __FUNCTION__, REFERENCE_BUFFER_SIZE_EMS);
		return -1;
	}	
	videoBuffers.reference_phys = (void *) __pa(videoBuffers.reference_virt);

	videoBuffers.capture_size = CAPTURE_BUFFER_SIZE_EMS;
	videoBuffers.capture_virt = alloc_mem_chunk(CAPTURE_BUFFER_SIZE_EMS, GFP_KERNEL);
	if (videoBuffers.capture_virt == NULL) {
		printk(KERN_ERR "%s: unable to mem_chunk %d bytes memory\n",
		       __FUNCTION__, CAPTURE_BUFFER_SIZE_EMS);
		return -1;
	}	
	videoBuffers.capture_phys = (void *) __pa(videoBuffers.capture_virt);

	retval =
	    request_irq(7, &video_interrupt, irqFlags, "aess_video", NULL);
	if (retval) {
		printk("drv-%s[%d] Unable to get IRQ\n", __FUNCTION__,
		       __LINE__);
		return -1;
	}
	IRQ_SET_HIGH_LEVEL(7);
	IRQ_SET_LEVEL_TRIGGER(7);

	init_waitqueue_head(&my_queue);

#ifdef	PROCFS
	create_proc_files();
#endif

	return 0;
}

void aess_video_cleanup(void)
{
	extern void free_mem_chunk(void* addr, size_t size);

#ifdef	PROCFS
	remove_proc_files();
#endif

	unregister_chrdev(ioaccess_major, "aess_video");
	if (Data_Linear_address)
		iounmap(Data_Linear_address);

	if (videoBuffers.reference_virt) {
		free_mem_chunk(videoBuffers.reference_virt,
			REFERENCE_BUFFER_SIZE_EMS);
		videoBuffers.reference_virt = NULL;
	}

	if (videoBuffers.capture_virt) {
		free_mem_chunk(videoBuffers.capture_virt,
			CAPTURE_BUFFER_SIZE_EMS);
		videoBuffers.capture_virt = NULL;
	}

	free_irq(7, NULL);
	return;
}

void InitializeRing(void)
{
	int i;
	// Initialize the Frame Ring
	theRing.ulRingMax = videoConfig.ulRingMax;
	theRing.ulHighWaterMark = videoConfig.ulHighWaterMark;
	theRing.ulLowWaterMark = videoConfig.ulLowWaterMark;
	theRing.ulNormalFrameRate = videoConfig.ulNormalFrameRate;
	theRing.ulCongestionFrameRate = videoConfig.ulCongestionFrameRate;
	theRing.ulHaltMark = videoConfig.ulHaltMark;
	theRing.bStalled = 0;
	theRing.ulRingStatus = 0;
	theRing.ulRingWrite = 0;
	theRing.ulRingRead = 0;

	// Clear all ring entries
	for (i = 0; i < MAX_RING_COUNT; i++)
		memset((char *)&theRing.RingEntry[i], 0,
		       sizeof(theRing.RingEntry[i]));

	// Initilize global capture variables                   
	gPrevRingEndOffset = 0;
	gSegmentNumber = 1;

	// Set up that next video data will be the start of a frame
	gSOF = 1;

	// Indicate that capture is active and that video segments can be
	// stored in the ring
	gCaptureActive = (int)videoConfig.CaptureMode;
}

void VideoEngineUnlock(void)
{
	int i, reg;
	// Read Lock register
	reg = REGISTERREAD(VIDEOENGINE_KEY_CONTROL);
	mdelay(2);
	if (reg == 0) {
		for (i = 0; i < 5; i++) {
			// Video Engine Registers locked, unlock the Video Engine Registers...
			REGISTERWRITE(VIDEOENGINE_KEY_CONTROL, 0x1A038AA8);
			mdelay(2);
			reg = REGISTERREAD(VIDEOENGINE_KEY_CONTROL);
			mdelay(2);
			if (reg == 1)
				break;
		}
	}
	// Output final status of registers...
	reg = REGISTERREAD(VIDEOENGINE_KEY_CONTROL);
	if (reg == 1) {
		printk("drv-%s[%d] Video Registers UNLOCKED\n",
		       __FUNCTION__, __LINE__);
	} else {
		printk("drv-%s[%d] Video Registers LOCKED\n",
		       __FUNCTION__, __LINE__);
	}
}

void VideoEngineReset(void)
{
	int SCU_Data;

	// Unlock SCU and force a reset of the Video Engine...
	REGISTERWRITE(SCU_KEY_CONTROL, 0x1688A8A8);
	mdelay(2);
	// Enable Clock
	REGISTERANDWRITE(0x1E6E200C, 0xFFFFFFF6, 0);
	// Reset the Video engine via SCU
	SCU_Data = REGISTERREAD(SCU_RESET_CONTROL);
	REGISTERWRITE(SCU_RESET_CONTROL,
		      SCU_Data | SCU_RESET_CONTROL_VIDEOENGINE);
	SCU_Data &= ~(SCU_RESET_CONTROL_VIDEOENGINE);
	REGISTERWRITE(SCU_RESET_CONTROL, SCU_Data);
	mdelay(2);

	VideoEngineUnlock();
}

// Temporarily stall the video engine, allowing the current frame to complete
// NOTE WELL: This function is called from interrupt level
void VideoEngineStall(void)
{
	theRing.bStalled = 1;
	// Disable frame capture, allowing current frame to complete
	REGISTERANDWRITE(VIDEOENGINE_SEQUENCE_CONTROL, 0xFFFFFFFD, 0);
}

// Halt the video capture and compression engine, complete restart required
void VideoEngineHalt(void)
{
	int i;
	int controlReg = 0;
	// Stop any frame captures that may be going, but leave the
	// mode resolution detect and watchdog bits alone
	REGISTERANDWRITE(VIDEOENGINE_SEQUENCE_CONTROL, 0xFFFFFF81, 0);
	mdelay(2);
	// Also, reset the compression engine which may have been waiting for
	// available buffer space
	REGISTERANDWRITE(VIDEOENGINE_SEQUENCE_CONTROL, 0xFFFFFF81, 0x04);
	mdelay(2);
	REGISTERANDWRITE(VIDEOENGINE_SEQUENCE_CONTROL, 0xFFFFFF81, 0);

	// Clear all interrupts, except mode detect watchdog
	REGISTERANDWRITE(VIDEO_INTERRUPT_CONTROL, 0xFFFFFFC1, 0);
	REGISTERWRITE(VIDEO_INTERRUPT_STATUS, 0xFFFFFFFF);

	// Wait for capture and compression engine idle status
	for (i = 0; i < 10; i++) {
		controlReg = REGISTERREAD(VIDEOENGINE_SEQUENCE_CONTROL);
		if ((controlReg & VIDEOENGINE_SEQUENCE_CONTROL_COMPRESSION_IDLE)
		    && (controlReg & VIDEOENGINE_SEQUENCE_CONTROL_CAPTURE_IDLE)
		    ) {
			break;
		}
		mdelay(100);
	}
	if (i == 10) {
		printk
		    ("drv-%s[%d] Video engine not idle after halt VIDEOENDING_SEQUENCE_CONTROL=%8.8X\n",
		     __FUNCTION__, __LINE__, controlReg);
	}
}

void InitializeVideoAll(void)
{
	InitializeVideoBuffers();
	InitializeVideoCompressionBuffer();
	InitializeVideoSource();
	InitializeVideoBCD();
	InitializeVideoCompression();
	InitializeVideoScaling();
	InitializeVideoFrameRate();
	SetupDirectMode();
}

void InitializeVideoBuffers(void)
{
// Setup all chip buffer addresses

	// Setup 'normal' use of buffer 1 and buffer 2
	REGISTERANDWRITE(VIDEO_IN_BUFFER_BASEADDRESS,
			 0, videoStatus.VideoINBufferAddress);
	REGISTERANDWRITE(VIDEO_REFERENCE_BUFFER_BASEADDRESS,
			 0, videoStatus.VideoReferenceBufferAddress);
	// Set up the CRC flag buffer
	REGISTERANDWRITE(VIDEO_BCD_FLAG_BUFFER_BASEADDRESS,
			 0, videoStatus.VideoFlagBufferAddress);
	REGISTERANDWRITE(VIDEO_CRC_BUFFER_BASEADDRESS,
			 0, videoStatus.VideoCRCBufferAddress);
	REGISTERANDWRITE(VIDEO_COMPRESS_SIZE_REGISTER,
			 0,
			 ((videoStatus.
			   ulPacketCountBits & 0x03) << 3) | (videoStatus.
							      ulPacketSizeBits &
							      0x07));
}

void InitializeVideoCompressionBuffer(void)
{
	// Set up the Video compression buffer address to the current one to process with
	REGISTERANDWRITE(VIDEO_COMPRESS_DESTINATION_BASEADDRESS,
			 0,
			 videoStatus.VideoCompressBufferAddress[videoStatus.
								CurrentCompressBufIndex]);
}

void SwapInAndRefBuffers(void)
{
	int OldBufferAddress, NewBufferAddress;

	//Swap Buffers
	OldBufferAddress = REGISTERREAD(VIDEO_IN_BUFFER_BASEADDRESS);
	NewBufferAddress = REGISTERREAD(VIDEO_REFERENCE_BUFFER_BASEADDRESS);
	REGISTERANDWRITE(VIDEO_IN_BUFFER_BASEADDRESS, 0, NewBufferAddress);
	REGISTERANDWRITE(VIDEO_REFERENCE_BUFFER_BASEADDRESS, 0,
			 OldBufferAddress);

}

void InitializeVideoSource(void)
{
	int ThresholdSetting = 0;

// Check for video source - external or internal...
	if (videoConfig.InputSignalSource == INPUT_SIGNAL_SOURCE_INTERNAL) {
		// Internal always uses internal timing generation from video 0.60
		REGISTERANDWRITE(VIDEOENGINE_VIDEO_CONTROL, 0xFFFFFEDF, 0);

	} else {
		if (videoStatus.InputSignal == 0) {
			// External digital always uses internal timing generation from video 0.60
			REGISTERANDWRITE(VIDEOENGINE_VIDEO_CONTROL,
					 0xFFFFFFEF, 0x10);
		}
	}

// BCD Thresholds - External analog has higher BCD threshold
	if (videoStatus.InputSignal == 0) {
		ThresholdSetting += (videoConfig.DigitalBCDTolerance << 16);
	} else {
		ThresholdSetting += (videoConfig.AnalogBCDTolerance << 16);
	}
	REGISTERANDWRITE(VIDEO_BCD_CONTROL, 0x0000FFFF, ThresholdSetting);
}

// TODO: Update test to 'driver' mode if we ever decide to support this
#if INCLUDE_DOWNSCALING
void HorizontalDownScaling(void)
{
	u32 HorizontalDownScalingFactor;

//  Calculate scaling factor D / S = 4096 / Factor  ======> Factor = (S / D) * 4096
	if (pVideoEngineInfo->videoConfig.DownScaleMode == DOWNSCALING_ENABLED) {
		HorizontalDownScalingFactor =
		    (uint32_t) ((double)
				(pVideoEngineInfo->videoStatus.SourceModeInfo.
				 X -
				 1) /
				(double)(pVideoEngineInfo->videoConfig.
					 DestinationModeX - 1) * 4096);
		if (HorizontalDownScalingFactor < 4096) {
			HorizontalDownScalingFactor = 4096;
		}
		if ((HorizontalDownScalingFactor *
		     (pVideoEngineInfo->videoConfig.DestinationModeX - 1)) !=
		    (pVideoEngineInfo->videoStatus.SourceModeInfo.X -
		     1) * 4096) {
			HorizontalDownScalingFactor += 1;
		}
		REGISTERANDWRITE(SCALE_FACTOR_REGISTER, 0xFFFF0000,
				 HorizontalDownScalingFactor);
	} else {
		REGISTERANDWRITE(SCALE_FACTOR_REGISTER, 0xFFFF0000, 0x1000);
	}
}

static void VerticalDownScaling(void)
{
	u32 VerticalDownScalingFactor;

//  Calculate scaling factor D / S = 4096 / Factor  ======> Factor = (S / D) * 4096
	if (pVideoEngineInfo->videoConfig.DownScaleMode == DOWNSCALING_ENABLED) {
		VerticalDownScalingFactor =
		    (uint32_t) ((double)
				(pVideoEngineInfo->videoStatus.SourceModeInfo.
				 Y -
				 1) /
				(double)(pVideoEngineInfo->videoConfig.
					 DestinationModeY - 1) * 4096);
		if (VerticalDownScalingFactor < 4096) {
			VerticalDownScalingFactor = 4096;
		}
		if ((VerticalDownScalingFactor *
		     (pVideoEngineInfo->videoConfig.DestinationModeY - 1)) !=
		    (pVideoEngineInfo->videoStatus.SourceModeInfo.Y -
		     1) * 4096) {
			VerticalDownScalingFactor += 1;
		}
		REGISTERANDWRITE(SCALE_FACTOR_REGISTER, 0x0000FFFF,
				 (VerticalDownScalingFactor << 16));
	} else {
		REGISTERANDWRITE(SCALE_FACTOR_REGISTER, 0x0000FFFF, 0x10000000);
	}
}
#endif				// INCLUDE_DOWNSCALING

void InitializeVideoScaling(void)
{
	u32	Buffer_Offset = 0;

	// Initialize the Scale factor register to no scaling factor.
	REGISTERWRITE(SCALE_FACTOR_REGISTER, 0x10001000);

// Check for any scaling that is currently configured and set that up...
// TODO : Seems like we should not call this function unless we 'know' our input mode.  Same
// holds true for the HorizontalDownScaling and VerticalDownScaling calls.
	if (videoConfig.DownScaleMode == DOWNSCALING_ENABLED) {
		if ((videoStatus.SourceModeInfo.X ==
		     videoConfig.DestinationModeX)
		    && (videoStatus.SourceModeInfo.Y ==
			videoConfig.DestinationModeY)
		    ) {
			REGISTERANDWRITE(SCU_CLOCK_SELECTION_REGISTER,
					 0xFFF9FFFF, 0);
			REGISTERANDWRITE(SCU_CLOCK_CONTROL_REGISTER, 0xFFFFF7FF,
					 (0x01 << 11));
			REGISTERANDWRITE(VIDEO_CONTROL_REGISTER, 0xFFFFFFEF,
					 0x00);
			REGISTERANDWRITE(SCALING_FILTER_PARAMETERS_1, 0,
					 0x00200000);
			REGISTERANDWRITE(SCALING_FILTER_PARAMETERS_2, 0,
					 0x00200000);
			REGISTERANDWRITE(SCALING_FILTER_PARAMETERS_3, 0,
					 0x00200000);
			REGISTERANDWRITE(SCALING_FILTER_PARAMETERS_4, 0,
					 0x00200000);
		} else {
			//  Down-Scaling
			REGISTERANDWRITE(SCU_CLOCK_SELECTION_REGISTER,
					 0xFFF9FFFF, 0x40000);
			REGISTERANDWRITE(SCU_CLOCK_CONTROL_REGISTER, 0xFFFFF7FF,
					 (0x00 << 11));
			REGISTERANDWRITE(VIDEO_CONTROL_REGISTER, 0xFFFFFFEF,
					 0x10);
			if (videoStatus.SourceModeInfo.X <=
			    videoConfig.DestinationModeX * 2) {
				REGISTERANDWRITE(SCALING_FILTER_PARAMETERS_1, 0,
						 0x00101000);
				REGISTERANDWRITE(SCALING_FILTER_PARAMETERS_2, 0,
						 0x00101000);
				REGISTERANDWRITE(SCALING_FILTER_PARAMETERS_3, 0,
						 0x00101000);
				REGISTERANDWRITE(SCALING_FILTER_PARAMETERS_4, 0,
						 0x00101000);
			} else {
				REGISTERANDWRITE(SCALING_FILTER_PARAMETERS_1, 0,
						 0x08080808);
				REGISTERANDWRITE(SCALING_FILTER_PARAMETERS_2, 0,
						 0x08080808);
				REGISTERANDWRITE(SCALING_FILTER_PARAMETERS_3, 0,
						 0x08080808);
				REGISTERANDWRITE(SCALING_FILTER_PARAMETERS_4, 0,
						 0x08080808);
			}
		}
	} else {
		// no scaling is being performed so set up for 1:1
		REGISTERANDWRITE(SCU_CLOCK_SELECTION_REGISTER, 0xFFF9FFFF, 0);
		REGISTERANDWRITE(SCU_CLOCK_CONTROL_REGISTER, 0xFFFFF7FF,
				 (0x01 << 11));
		REGISTERANDWRITE(VIDEO_CONTROL_REGISTER, 0xFFFFFFEF, 0x00);
		REGISTERANDWRITE(SCALING_FILTER_PARAMETERS_1, 0, 0x00200000);
		REGISTERANDWRITE(SCALING_FILTER_PARAMETERS_2, 0, 0x00200000);
		REGISTERANDWRITE(SCALING_FILTER_PARAMETERS_3, 0, 0x00200000);
		REGISTERANDWRITE(SCALING_FILTER_PARAMETERS_4, 0, 0x00200000);
	}
	REGISTERANDWRITE(VIDEO_CAPTURE_WINDOW_SETTING,
			 0,
			 (videoStatus.SourceModeInfo.Y +
			  ((videoStatus.SourceModeInfo.X) << 16)));
	if (videoConfig.DownScaleMode == DOWNSCALING_ENABLED) {
		REGISTERANDWRITE(VIDEO_COMPRESS_WINDOW_SETTING,
				 0,
				 (videoConfig.DestinationModeY +
				  ((videoConfig.DestinationModeX) << 16)));
	} else {
		REGISTERANDWRITE(VIDEO_COMPRESS_WINDOW_SETTING,
				 0,
				 (videoStatus.SourceModeInfo.Y +
				  ((videoStatus.SourceModeInfo.X) << 16)));
	}

	/*
	 * When horizontal line count is not a multiplier of 8 or 16
	 * it may cause unexpected problems with aspeed hardware.  Force it to
	 * be multiplier of 16.
	 */
	Buffer_Offset = videoStatus.SourceModeInfo.X;
	if (Buffer_Offset % 16) {
		Buffer_Offset += (16 - Buffer_Offset % 16);
	}

	Buffer_Offset *= 4;
	REGISTERANDWRITE(VIDEO_IN_BUFFER_OFFSET, 0, Buffer_Offset);
}

void SetFrameRate(int frameRate)
{
	REGISTERANDWRITE(VIDEOENGINE_VIDEO_CONTROL,
			 0xFF00FFEF, frameRate << 16);

}

void InitializeVideoFrameRate(void)
{
	// Set the frame rate for normal
	videoStatus.FrameRate = videoConfig.ulNormalFrameRate;
	SetFrameRate(videoStatus.FrameRate);
}

void InitializeVideoBCD(void)
{
	int BCDSetting = 0;
	int CRC1Setting = 0;
	int CRC2Setting = 0;
	switch (videoConfig.BCDControl) {
	case BCDCONTROL_OFF:
		{
			BCDSetting = 0;
		}
		break;
	case BCDCONTROL_PIXEL:
		{
			BCDSetting = 1;
		}
		break;
	case BCDCONTROL_CRC:
		{
			BCDSetting = 1;
			// Setup recommend CRCs from ASPEED
			CRC1Setting = 0x131350FD;
			CRC2Setting = 0x13266500;
		}
		break;
	}
	REGISTERWRITE(CRC_1_REGISTER, CRC1Setting);
	REGISTERWRITE(CRC_2_REGISTER, CRC2Setting);
	REGISTERANDWRITE(VIDEO_BCD_CONTROL, 0xFFFFFFFE, BCDSetting);
	REGISTERWRITE(REDUCE_BIT_REGISTER,
		videoConfig.ReduceBitControl << 6 |
		videoConfig.ReduceBitControl << 3 |
		videoConfig.ReduceBitControl);
}

void InitializeVideoCompression(void)
{
	int CompressSetting = 1;	// Setup for JPEG-only compression

	//  420/444 Setting
	REGISTERANDWRITE(VIDEOENGINE_SEQUENCE_CONTROL,
			 0xFFFFF3FF,
			 (videoConfig.CompressionMode & 0x03) << 10);

	CompressSetting += (videoConfig.Y_JPEGTableSelector) << 11;
	CompressSetting += (videoConfig.UV_JPEGTableSelector) << 6;
	CompressSetting += 0x80000;
	REGISTERANDWRITE(VIDEO_COMPRESS_CONTROL, 0, CompressSetting);

	REGISTERWRITE(VIDEO_COMPRESS_READ_OFFSET, 0);

}

int IsCompressBufAvailable(void)
{
	int nextCompressBufIndex;
	// Check, then set up the next compression buffer if it is idle
	nextCompressBufIndex =
	    (videoStatus.CurrentCompressBufIndex +
	     1) % videoConfig.VideoCompressBufferCount;
	if (videoStatus.VideoCompressBufStatus[nextCompressBufIndex] ==
	    COMPRESSION_BUFFER_STATUS_IDLE) {
		if (gDebugFlags & DEBUG_SINGLE_FRAME_BUFFERS)
			printk("drv-%s[%d] Buf %d is idle\n",
			       __FUNCTION__, __LINE__, nextCompressBufIndex);
		return (1);
	} else {
		if (gDebugFlags & DEBUG_SINGLE_FRAME_BUFFERS)
			printk("drv-%s[%d] Buf %d is not idle %d\n",
			       __FUNCTION__, __LINE__,
			       nextCompressBufIndex,
			       (int)videoStatus.
			       VideoCompressBufStatus[nextCompressBufIndex]);
		return (0);
	}
}

void RequestSingleFrame(void)
{
	int RegSetting;

	//BCD buffer swapping for single frame capture mode
	if (gbHavePrevFrame == 1) {
		SwapInAndRefBuffers();
		RegSetting = 1;
	} else {
		InitializeVideoBuffers();
		RegSetting = 0;
	}
	REGISTERANDWRITE(VIDEO_BCD_CONTROL, 0xFFFFFFFE, RegSetting);

	InitializeVideoCompressionBuffer();

	// OK, now that we are about to trigger for a compression, increment
	// the next available compression buffer index and check for wrap
	videoStatus.CurrentCompressBufIndex =
	    (videoStatus.CurrentCompressBufIndex +
	     1) % videoConfig.VideoCompressBufferCount;
	// Initialize for first video segment of this compression buffer
	gPrevRingEndOffset = 0;
	if (gDebugFlags & DEBUG_SINGLE_FRAME_BUFFERS)
		printk("drv-%s[%d] using buf %d\n",
		       __FUNCTION__, __LINE__,
		       (int)videoStatus.CurrentCompressBufIndex);

	// NOTE WELL: Do not modify B7 of VIDEO_SEQUENCE_CONTROL enable mode watchdog
	//            Do not modify B0 of VIDEO_INTERRUPT_CONTROL leaving mode watchdog interrupt enabled.

	// Set B9 which is undocumented anti-tearing mode bit 
	RegSetting = 0x200;
	// If we have no previous frame, then request a full frame 
	if (gbHavePrevFrame != 1)
		RegSetting |= 0x40;
	REGISTERANDWRITE(VIDEOENGINE_SEQUENCE_CONTROL, 0xFFFFFD81, RegSetting);

	// Clear compression engine and trigger bits
	REGISTERANDWRITE(VIDEOENGINE_SEQUENCE_CONTROL, 0xFFFFFFE9, 0);
	barrier();

	videoStatus.VideoCompressBufStatus[videoStatus.
					   CurrentCompressBufIndex] =
	    COMPRESSION_BUFFER_STATUS_BUSY;
	// Trigger the capture engine
	REGISTERANDWRITE(VIDEOENGINE_SEQUENCE_CONTROL, 0xFFFFFFED, 0x12);
}

#define MODE_DETECT_GUARD 200
#define MODE_DETECT_LIMIT 5

void SetupDirectMode(void)
{
	if (videoStatus.UsingDirectMode == 0) {
		REGISTERANDWRITE(VIDEOENGINE_VIDEO_CONTROL, 0xFFFFFFDF, 0);
		REGISTERANDWRITE(VIDEOENGINE_VIDEO_CONTROL, 0xFFFFFEFF, 0);
		printk
		    ("drv-%s[%d] Disabled direct mode VIDEOENGINE_VIDEO_CONTROL=%8.8X\n",
		     __FUNCTION__, __LINE__,
		     (int)REGISTERREAD(VIDEOENGINE_VIDEO_CONTROL));
	} else {
		REGISTERANDWRITE(VIDEOENGINE_VIDEO_CONTROL, 0xFFFFFFDF, 0x20);
		REGISTERANDWRITE(VIDEOENGINE_VIDEO_CONTROL, 0xFFFFFEFF, 0x100);
		REGISTERANDWRITE(VIDEOENGINE_MODEDETECTIONSETTING_H,
				 0, 0x7800000);

		REGISTERANDWRITE(VIDEOENGINE_MODEDETECTIONSETTING_H,
				 0,
				 (videoConfig.TotalMemory -
				  videoConfig.VGAMemory) << 20);
		REGISTERANDWRITE(VIDEOENGINE_MODEDETECTIONSETTING_V, 0,
				 (videoConfig.DelayControl +
				  (videoConfig.DelayControl << 16) +
				  (videoStatus.SourceModeInfo.X * 4)
				 )
		    );
		printk
		    ("drv-%s[%d] Enabled direct mode VIDEOENGINE_VIDEO_CONTROL=%8.8X\n",
		     __FUNCTION__, __LINE__,
		     (int)REGISTERREAD(VIDEOENGINE_VIDEO_CONTROL));
	}
	return;
}

int GetModeDetect(void)
{
	int iRet;
	int iTimeout = (HZ * MODE_DETECT_GUARD / 1000);
	u32 VR35C;

	// Assume no input signal
	videoStatus.HaveSignal = 0;
	// Check if internal VGA screen is off
	if (videoConfig.InputSignalSource == INPUT_SIGNAL_SOURCE_INTERNAL) {
		VR35C = REGISTERREAD(VIDEO_SCRATCH_REGISTER_35C) >> 24;

		if ((VR35C & 0xC0) || (VR35C & 0x04) || (!(VR35C & 0x3B))) {
			videoStatus.HaveSignal = 0;
		} else {
			videoStatus.HaveSignal = 1;
		}
		// The internal VGA is off, no signal when internal
		if (!videoStatus.HaveSignal) {
			return (0);
		}
	}
	// Reset all interrupts and clear any previous mode detect status
	REGISTERWRITE(VIDEO_INTERRUPT_STATUS, 0xFFFFFFFF);
	// Clear any previous Mode Detected status
	Mode_Detect_Ready = 0;
	// Enable ONLY the Mode Detect Ready interrupt
	REGISTERWRITE(VIDEO_INTERRUPT_CONTROL, 0x10);
	//  1st mode detection for polarity check
	REGISTERANDWRITE(VIDEOENGINE_SEQUENCE_CONTROL, 0xFFFFFF7E, 0);
	mdelay(2);
	// Trigger mode detection
	REGISTERANDWRITE(VIDEOENGINE_SEQUENCE_CONTROL, 0xFFFFFFFE, 0x01);

	if (!wait_event_timeout(my_queue, (Mode_Detect_Ready == 1), iTimeout))
		iRet = 0;
	iRet = Mode_Detect_Ready;
	// Clear this mode detect...
	Mode_Detect_Ready = 0;

	// Now if we did get mode detect, process it and set HaveSignal and InputSignal
	// status values
	if (iRet) {
		unsigned long ulModeDetectReg =
		    REGISTERREAD(MODEDETECTION_STATUS_READBACK);
		unsigned long ulHorizontalCount =
		    ulModeDetectReg & MODEDETECTION_HSCAN_COUNT;
		unsigned long ulVerticalCount =
		    ulModeDetectReg & MODEDETECTION_VSCAN_COUNT;

		if (JudgeMode(ulHorizontalCount, ulVerticalCount) == 1) {
			videoStatus.HaveSignal = 1;
		}

#if INCLUDE_EXTERNAL_VGA_SUPPORT
		if (videoConfig.InputSignalSource ==
		    INPUT_SIGNAL_SOURCE_EXTERNAL) {
			unsigned long ulModeDetectReg =
			    REGISTERREAD(MODEDETECTION_STATUS_READBACK);
			unsigned long ulHorizontalCount =
			    ulModeDetectReg & MODEDETECTION_HSCAN_COUNT;
			unsigned long ulVerticalCount =
			    ulModeDetectReg & MODEDETECTION_VSCAN_COUNT;
			// TODO: SHould we check ulModeDetectReg to see if we detected both HSYNC and VSYNC
			// signals - regartdless of stability as a further check that we have signal.

			// Get External digital/analog input signal type
			videoStatus.InputSignal =
			    (ulModeDetectReg & MODEDETECTION_EXTERNAL_SOURCE) >>
			    12;

			//  Check mode resolution for external analog only, we will check internal 
			//  and external digital input mode resolution at AutoPosition() function
			if (videoStatus.InputSignal == 1)	// Video signal is ADC
			{
				int SCU_Data;
				if (JudgeMode
				    (ulHorizontalCount, ulVerticalCount) == 1) {
					videoStatus.HaveSignal = 1;
					// TODO: Verify this, whiy are we checking this again?  Previous level if already
					// validated we are externa.  Is this trying to really check signal source??
					if (videoConfig.InputSignalSource ==
					    INPUT_SIGNAL_SOURCE_EXTERNAL) {
						SCU_Data =
						    REGISTERREAD(0x1E6E2078);
// TODO: For now, comment out code that processes for 2100 and always run the 1100/2050 code??
//                                              if (VideoEngineInfo->ChipBounding == 3) 
//                                              {
//                                                      SCU_Data |= 0x01;
//                                                      REGISTERANDWRITE(VIDEOENGINE_VIDEO_CONTROL, 0xFFFFCFFF, 0);
//                                              }
//                                              else 
						{
							SCU_Data |= 0x05;
							REGISTERANDWRITE
							    (VIDEOENGINE_VIDEO_CONTROL,
							     0xFFFFCFFF,
							     0x01000);
						}
						REGISTERWRITE(0x1E6E2078,
							      SCU_Data);
						// External analog will use no delay clock
						REGISTERANDWRITE
						    (VIDEOENGINE_VIDEO_CONTROL,
						     0xFFFFF3FF, 0);
						REGISTERANDWRITE
						    (VIDEOENGINE_VIDEO_CONTROL,
						     0xFFFFFFE7,
						     (videoStatus.
						      InputSignal << 3) +
						     (videoStatus.
						      InputSignal << 4));
					}
					return (1);
				} else {
					//  Enable Watchdog detection
					//yriver no signal
					videoStatus.HaveSignal = 0;
//                                      if (videoConfig.InputSignalSource == INPUT_SIGNAL_SOURCE_EXTERNAL) 
//                                      {
					SCU_Data = REGISTERREAD(0x1E6E2078);
//                                              if (VideoEngineInfo->ChipBounding == 3) 
//                                              {
//                                                      SCU_Data |= 0x01;
//                                                      REGISTERANDWRITE(VIDEOENGINE_VIDEO_CONTROL, 0xFFFFCFFF, 0);
//                                              }
//                                              else 
					{
						SCU_Data |= 0x05;
						REGISTERANDWRITE
						    (VIDEOENGINE_VIDEO_CONTROL,
						     0xFFFFCFFF, 0x01000);
					}
					REGISTERWRITE(0x1E6E2078, SCU_Data);
					//External analog will use no delay clock
					REGISTERANDWRITE
					    (VIDEOENGINE_VIDEO_CONTROL,
					     0xFFFFF3FF, 0);
					REGISTERANDWRITE
					    (VIDEOENGINE_VIDEO_CONTROL,
					     0xFFFFFFE7,
					     (videoStatus.InputSignal << 3) +
					     (videoStatus.InputSignal << 4));
//                                      }
					return (0);
				}
			}
		}
#endif				// INCLUDE_EXTERNAL_VGA_SUPPORT
	}
	return (iRet);
}

/* 1024x768@32bpp, 85Hz */
#define	DIRECT_MODE_BANDWIDTH_THRESHOLD	564
	
int AutoPosition_Adjust(void)
{
	int i,
	    ulErr,
	    ulRetry,
	    H_Start = 0,
	    H_End = 0, V_Start = 0, V_End = 0, H_Temp = 0, V_Temp = 0;
	u32 uiReg;

	ulErr = 1;
	REGISTERANDWRITE(MODE_DETECTION_REGISTER, 0xFFFF00FF, (0x65 << 8));
	for (ulRetry = 0; ulRetry < MODE_DETECT_LIMIT; ulRetry++) {
		// Wait up to MODE_DETECT_GUARD ms for mode detection
		ulErr = 1;	// assume the worst!
		if (!GetModeDetect())
			break;
		ulErr = 0;

		H_Start = REGISTERREAD(H_DETECTION_STATUS) & 0x00000FFF;
		H_End = (REGISTERREAD(H_DETECTION_STATUS) & 0x0FFF0000) >> 16;

		if ((videoStatus.SourceModeInfo.X == 720) &&
		    (videoStatus.SourceModeInfo.Y == 400)) {
			if (((H_End - H_Start) > (videoStatus.SourceModeInfo.X + 20)) &&
			    ((H_End - H_Start + 1) != 360)) {
				printk("%s adjust for abnormal video detect at 720x400\n",
					__FUNCTION__);
				H_End = H_Start + videoStatus.SourceModeInfo.X - 1;
			}
		}
		
		V_Start = REGISTERREAD(V_DETECTION_STATUS) & 0x00000FFF;
		V_End = (REGISTERREAD(V_DETECTION_STATUS) & 0x0FFF0000) >> 16;

		//Check if cable quality is too bad. If it is bad then we use 0x65 as threshold
		//Because RGB data is arrived slower than H-sync, V-sync. We have to read more times to confirm RGB data is arrived
		if ((abs(H_Temp - H_Start) > 1)
		    || ((H_Start <= 1) || (V_Start <= 1) || (H_Start == 0x3FF)
			|| (V_Start == 0x3FF))) {
			H_Temp = REGISTERREAD(H_DETECTION_STATUS) & 0x00000FFF;
			V_Temp = REGISTERREAD(V_DETECTION_STATUS) & 0x00000FFF;
			ulErr = 2;
			continue;
		}
		ulErr = 0;
		break;
	}			// end retries on mode detection

	if (ulErr == 0) {
		if (gDebugFlags & DEBUG_MODE_DETECTION) {
			printk("drv-%s[%d] H_Start = %x, H_End = %x\n",
			       __FUNCTION__, __LINE__, H_Start, H_End);
			printk("drv-%s[%d] V_Start = %x, V_End = %x\n",
			       __FUNCTION__, __LINE__, V_Start, V_End);
		}

		if (videoConfig.InputSignalSource ==
		    INPUT_SIGNAL_SOURCE_INTERNAL) {

			REGISTERANDWRITE(VIDEOENGINE_MODEDETECTIONSETTING_H,
					 0xFFFF0000, H_End);
			REGISTERANDWRITE(VIDEOENGINE_MODEDETECTIONSETTING_H,
					 0x0000FFFF, (H_Start - 1) << 16);
			REGISTERANDWRITE(VIDEOENGINE_MODEDETECTIONSETTING_V,
					 0xFFFFF000, (V_End + 1));
			REGISTERANDWRITE(VIDEOENGINE_MODEDETECTIONSETTING_V,
					 0xF000FFFF, (V_Start) << 16);
			videoStatus.SourceModeInfo.X = (H_End - H_Start) + 1;
			videoStatus.SourceModeInfo.Y = (V_End - V_Start) + 1;

			if (gDebugFlags & DEBUG_MODE_DETECTION) {
				printk("drv-%s[%d] Mode.X = %d\n",
				       __FUNCTION__, __LINE__,
				       videoStatus.SourceModeInfo.X);
				printk("drv-%s[%d] Mode.Y = %d\n",
				       __FUNCTION__, __LINE__,
				       videoStatus.SourceModeInfo.Y);
			}
			// If res is greater than the largest supported by buffers, truncate the x and y values now
			if (videoStatus.SourceModeInfo.X >
			    videoStatus.MaxModeInfo.X) {
				videoStatus.SourceModeInfo.X =
				    videoStatus.MaxModeInfo.X;
				if (gDebugFlags & DEBUG_MODE_DETECTION)
					printk
					    ("drv-%s[%d] Mode.X truncated to %d\n",
					     __FUNCTION__, __LINE__,
					     videoStatus.SourceModeInfo.X);
			}
			if (videoStatus.SourceModeInfo.Y >
			    videoStatus.MaxModeInfo.Y) {
				videoStatus.SourceModeInfo.Y =
				    videoStatus.MaxModeInfo.Y;
				if (gDebugFlags & DEBUG_MODE_DETECTION)
					printk
					    ("drv-%s[%d] Mode.Y truncated to %d\n",
					     __FUNCTION__, __LINE__,
					     videoStatus.SourceModeInfo.Y);
			}
			//  Judge if bandwidth is not enough then enable direct mode in internal VGA
			videoStatus.RefreshRateIndex =
			    ((REGISTERREAD(VIDEO_SCRATCH_REGISTER_34C) >> 8) &
			     0x0F);
			videoStatus.Mode_PixelClock = 0;
			videoStatus.ColorDepthIndex =
			    ((REGISTERREAD(VIDEO_SCRATCH_REGISTER_34C) >> 4) & 0x0F);
			if (videoStatus.ColorDepthIndex >= 2) {	
				/* 15bpp/16bpp or higher */
				/* 15bpp (2) provides workaround for aspeed linux video driver bug */
				for (i = 0; i < 17; i++) {
					if ((videoStatus.SourceModeInfo.X ==
					     Internal_Mode[i].HorizontalActive)
					    && (videoStatus.SourceModeInfo.Y ==
						Internal_Mode[i].VerticalActive)
					    && (videoStatus.RefreshRateIndex ==
						Internal_Mode[i].
						RefreshRateIndex)) {
						videoStatus.Mode_PixelClock =
						    Internal_Mode[i].PixelClock;
						if (gDebugFlags &
						    DEBUG_MODE_DETECTION)
							printk
							    ("drv-%s[%d] Mode_PixelClock = %d\n",
							     __FUNCTION__,
							     __LINE__,
							     (int)videoStatus.
							     Mode_PixelClock);
					}
				}

				/*
				 * add workaround for aspeed linux video driver bug.
				 * scratch register will be off by one - i.e.
				 * 32bpp mode will report 3 and 16bpp mode will report 2
				 */
				if (videoStatus.ColorDepthIndex == 2) {
					videoStatus.ColorDepthIndex = 3;
				}

				//  Calculate bandwidth required for this mode
				//  Video requires pixelclock * 3, VGA requires pixelclock * bpp / 8
				videoStatus.Mode_Bandwidth =
				    videoStatus.Mode_PixelClock * (4 +
								   2 *
								   (videoStatus.
								    ColorDepthIndex
								    - 2));
				if (gDebugFlags & DEBUG_MODE_DETECTION)
					printk
					    ("drv-%s[%d] Mode_Bandwidth = %d\n",
					     __FUNCTION__, __LINE__,
					     (int)videoStatus.Mode_Bandwidth);

				// Check to see if the host driver has a hardware cursors enabled.
				videoStatus.HaveHostHWCursorSupport = 0;
				// Check to see if host has hardware cursor enabled
				uiReg =
				    REGISTERREAD
				    (VIDEO_SCRATCH_340_CURSOR_STATUS);
				if (uiReg & VIDEO_CURSOR_ENABLED_MASK) {
					printk
					    ("drv-%s[%d] Host HW Cursor is available\n",
					     __FUNCTION__, __LINE__);
					videoStatus.HaveHostHWCursorSupport = 1;
				} else {
					printk
					    ("drv-%s[%d] Host HW Cursor is NOT available\n",
					     __FUNCTION__, __LINE__);
				}

				// Assume we will not use direct mode and then check to see if we need to
				videoStatus.UsingDirectMode = 0;
				if (videoConfig.DirectMode == DIRECTMODE_AUTO) {
					if ((videoStatus.SourceModeInfo.X > 1024
					    || videoStatus.SourceModeInfo.Y >
					    768) ||
					    (videoStatus.Mode_Bandwidth >=
						DIRECT_MODE_BANDWIDTH_THRESHOLD))  {
						// There is not enough bandwidth, use direct mode if available
						printk
						    ("drv-%s[%d] Resolution %d x %d forces direct mode \n",
						     __FUNCTION__, __LINE__,
						     videoStatus.SourceModeInfo.
						     X,
						     videoStatus.SourceModeInfo.
						     Y);
						videoStatus.UsingDirectMode = 1;
					}
				} else if (videoConfig.DirectMode ==
					   DIRECTMODE_ENABLE) {
					videoStatus.UsingDirectMode = 1;
				}
				if ((!videoStatus.HaveHostHWCursorSupport) &&
				    (videoStatus.UsingDirectMode))
					printk("drv-%s using DirectMode and host does not have HW cursor support\n",
						__FUNCTION__);
			} else {
				if (gDebugFlags & DEBUG_MODE_DETECTION)
					printk
					    ("drv-%s[%d] Color Depth is not 16bpp or higher\n",
					     __FUNCTION__, __LINE__);
				// We will need to clock the bits in because this is not 24-bit color...
				videoStatus.UsingDirectMode = 0;
			}

			// Now, configure direct mode based on videoStatus.UsingDirectMode
			SetupDirectMode();
		}
	} else {
	}
	return (ulErr);
}

int DoModeDetection(void)
{
	int i;
	int ulErr, ulRetry;
	int Status;
	int Sync_Count;

// TODO : These are needed in videoStatus, but it might be nice to know them...
	int HPolarity_Positive;
	int HPolarity_Negative;
	int VPolarity_Positive;
	int VPolarity_Negative;
// 0 for negative, 1 for positive, hardware always use rising edge
	int Horizontal_Polarity;
	int Vertical_Polarity;

	videoStatus.UsingDirectMode = 0;
	if (videoConfig.DirectMode != DIRECTMODE_DISABLE)
		videoStatus.UsingDirectMode = 1;
	// Now, configure direct mode based on videoSatus.UsingDirectMode
	SetupDirectMode();

	Horizontal_Polarity = 0;
	Vertical_Polarity = 0;
	ulErr = 1;
	for (ulRetry = 0; ulRetry < MODE_DETECT_LIMIT; ulRetry++) {
// STEP 1 - Set vertical and horizonal scan bandwidths and get polarity
		Sync_Count = HPolarity_Positive = HPolarity_Negative =
		    VPolarity_Positive = VPolarity_Negative = 0;

//  Set input signal that is configured (internal or external)
		REGISTERANDWRITE(VIDEOENGINE_VIDEO_CONTROL, 0xFFFFFFFB,
				 (videoConfig.InputSignalSource) << 2);
//  Check polarity (zero H an V polarity values prior to generating mode detect)
		REGISTERANDWRITE(VIDEOENGINE_VIDEO_CONTROL, 0xFFFFFFFC, 0);
		REGISTERANDWRITE(MODE_DETECTION_REGISTER, 0xFFF0FFFF,
				 (MODEDETECTION_VERTICAL_STABLE_MAXIMUM << 16));
		REGISTERANDWRITE(MODE_DETECTION_REGISTER, 0xFF0FFFFF,
				 (MODEDETECTION_HORIZONTAL_STABLE_MAXIMUM <<
				  20));
		REGISTERANDWRITE(MODE_DETECTION_REGISTER, 0xF0FFFFFF,
				 (MODEDETECTION_VERTICAL_STABLE_THRESHOLD <<
				  24));
		REGISTERANDWRITE(MODE_DETECTION_REGISTER, 0x0FFFFFFF,
				 (MODEDETECTION_HORIZONTAL_STABLE_THRESHOLD <<
				  28));

// STEP 2 - Wait for Mode Detect Ready
		if (GetModeDetect()) {
			// We got the mode detect, continue on
			ulErr = 0;
			break;
		}
	}			// end while waiting for video and mode

// STEP 3 Read back Vertical and Horizontal polarity
	if (!ulErr) {
		for (i = 0; i <= 10; i++) {
			Status = REGISTERREAD(MODEDETECTION_STATUS_READBACK);
			Status = REGISTERREAD(MODEDETECTION_STATUS_READBACK);
//                      if (gDebugFlags & DEBUG_MODE_DETECTION) 
//                              printk ("drv-%s[%d] MODEDETECTION_STATUS_READBACK VR098 = %8.8X\n",__FUNCTION__,__LINE__, Status);
			if ((Status & 0x10000000) == 0)
				VPolarity_Positive++;
			else
				VPolarity_Negative++;
			if ((Status & 0x20000000) == 0)
				HPolarity_Positive++;
			else
				HPolarity_Negative++;
		}

		// If more positive than negative, use positive as best polairy guess
		if (VPolarity_Positive > VPolarity_Negative)
			Vertical_Polarity = 1;
		else
			Vertical_Polarity = 0;

		// Same for Horizontal polarity
		if (HPolarity_Positive > HPolarity_Negative)
			Horizontal_Polarity = 1;
		else
			Horizontal_Polarity = 0;

		if (gDebugFlags & DEBUG_MODE_DETECTION) {
			printk
			    ("drv-%s[%d] Horizontal Positive=%d, Negative=%d, Polarity=%s\n",
			     __FUNCTION__, __LINE__, HPolarity_Positive,
			     HPolarity_Negative,
			     Horizontal_Polarity ? "POS" : "NEG");
			printk
			    ("drv-%s[%d] Horizontal Positive=%d, Negative=%d, Polarity=%s\n",
			     __FUNCTION__, __LINE__, VPolarity_Positive,
			     VPolarity_Negative,
			     Vertical_Polarity ? "POS" : "NEG");
		}
		//Direct mode always uses positive/positive so we invert them
//              if (videoConfig.DirectMode == DIRECTMODE_ENABLE) 
//              {
//                  REGISTERANDWRITE (VIDEOENGINE_VIDEO_CONTROL, 0xFFFFFFFC, 0x03);
//              }
//              else 
		{
			REGISTERANDWRITE(VIDEOENGINE_VIDEO_CONTROL, 0xFFFFFFFC,
					 ((Vertical_Polarity << 1) +
					  Horizontal_Polarity));
		}

		// STEP 4 - Trigger 2nd mode detection for determining vertical and horizontal screen pixel size
		ulErr = 2;
		if (GetModeDetect()) {
			// We got the mode detect, continue on
			ulErr = 0;
		}		// end if we got the second mode detections and processed it
	}			// end if we got the initial polarity mode detection done

	// If both mode detections suceeded, then we now know that we have
	// input signal and mode detected correctly.  Enable mode detect watchdog.
	if (!ulErr) {
		// Mode detection was successful...
#if INCLUDE_EXTERNAL_VGA_SUPPORT
		if (videoConfig.InputSignalSource ==
		    INPUT_SIGNAL_SOURCE_EXTERNAL)
			AutoPhase_Adjust();
#endif
		AutoPosition_Adjust();
		// Clear all interrupts except mode watchdog
		REGISTERANDWRITE(VIDEO_INTERRUPT_CONTROL, 0xFFFFFC00,
				 VIDEO_INTERRUPT_MODE_WATCHDOG);
		// Enable mode watchdog detection
		REGISTERANDWRITE(VIDEOENGINE_SEQUENCE_CONTROL, 0xFFFFFF7F,
				 0x80);
	} else {
		// Mode detection failed, clear interrupts and disable mode detect
		REGISTERANDWRITE(VIDEO_INTERRUPT_CONTROL, 0xFFFFFC00, 0);
		REGISTERANDWRITE(VIDEOENGINE_SEQUENCE_CONTROL, 0xFFFFFF00, 0);
	}

	return (ulErr);
}

// NOTE WELL: This function is called from interrupt routine
int GetNextWritableRingEntry(void)
{
	int nextFrame = -1;
	if (IncrementRingEntry(theRing.ulRingWrite) != theRing.ulRingRead) {
		nextFrame = theRing.ulRingWrite;
		theRing.ulRingWrite = IncrementRingEntry(theRing.ulRingWrite);
	}
	return (nextFrame);
}

// NOTE WELL: This function is called from interrupt routine
// Returns the current writable ring entry if it is a no change
// and it has been less than a second's worth of frames since we stored
// a frame
int GetCurrentWritableRingEntry(void)
{
	int chkWrite;
	if (theRing.ulRingRead != theRing.ulRingWrite) {
		// There is at least one pending frame, get its ring entry number
		chkWrite = theRing.ulRingWrite - 1;
		if (chkWrite < 0)
			chkWrite += theRing.ulRingMax;
		return (chkWrite);
	}
	return (-1);
}

// NOTE WELL: This function is used by wait_event_interruptible and
// should not do anything to sleep or schedule.
int GetNextReadableRingEntry(void)
{
	int readFrame = -1;
	PFRAMEDESCRIPTOR pRingEntry = &theRing.RingEntry[theRing.ulRingRead];
	if (pRingEntry->cFrameStatus & VIDEO_FRAME_STATUS_PRESENT
	    && pRingEntry->ulDataCount != -4)
		readFrame = theRing.ulRingRead;
	return (readFrame);
}

void FreeRingEntry(int curEntry)
{
	// Clear the entire ring entry
	memset((char *)&theRing.RingEntry[curEntry], 0,
	       sizeof(theRing.RingEntry[curEntry]));
	theRing.ulRingRead = IncrementRingEntry(curEntry);
}

// NOTE WELL: This function is called from interrupt routine
int IncrementRingEntry(int curEntry)
{
	if (++curEntry >= theRing.ulRingMax)
		curEntry = 0;
	return (curEntry);
}

int GetRingCount()
{
	int count;

	// Compute ring count
	count = theRing.ulRingWrite - theRing.ulRingRead;
	if (count < 0)
		count += theRing.ulRingMax;
	return (count);
}

// NOTE WELL: This function is called from interrupt routine
void UpdateRingCounts()
{
	int i, count;

	// Get ring count
	theRing.ulRingCount = GetRingCount();

	// Estimate frame count
	for (i = theRing.ulRingRead, count = 0; i != theRing.ulRingWrite;
	     i = IncrementRingEntry(i))
		if (theRing.RingEntry[i].cFrameStatus & VIDEO_FRAME_STATUS_EOF)
			count++;
	theRing.ulFrameCount = count;

	// See if we are congested
	if (videoStatus.FrameRate == videoConfig.ulCongestionFrameRate)
		theRing.ulRingStatus |= VIDEO_ENGINE_CONGESTED;
}

module_init(aess_video_init);
module_exit(aess_video_cleanup);

/**                                                              
 *                                                                
 */
typedef struct _RegDef {
	char *szRegName;
	int iRegNo;
} REGDEF, *PREGDEF;

#define DEFINEREG(name) {	#name, name }
REGDEF videoRegs[] = {
	DEFINEREG(VIDEOENGINE_SEQUENCE_CONTROL),
	DEFINEREG(VIDEOENGINE_VIDEO_CONTROL),
	DEFINEREG(VIDEOENGINE_MODEDETECTIONSETTING_H),
	DEFINEREG(VIDEOENGINE_MODEDETECTIONSETTING_V),
	DEFINEREG(SCALE_FACTOR_REGISTER),
	DEFINEREG(SCALING_FILTER_PARAMETERS_1),
	DEFINEREG(SCALING_FILTER_PARAMETERS_2),
	DEFINEREG(SCALING_FILTER_PARAMETERS_3),
	DEFINEREG(SCALING_FILTER_PARAMETERS_4),
	DEFINEREG(VIDEO_BCD_CONTROL),
	DEFINEREG(VIDEO_CAPTURE_WINDOW_SETTING),
	DEFINEREG(VIDEO_COMPRESS_WINDOW_SETTING),
	DEFINEREG(VIDEO_COMPRESS_PROCESSING_OFFSET),
	DEFINEREG(VIDEO_COMPRESS_READ_OFFSET),
	DEFINEREG(VIDEO_CRC_BUFFER_BASEADDRESS),
	DEFINEREG(VIDEO_IN_BUFFER_BASEADDRESS),
	DEFINEREG(VIDEO_IN_BUFFER_OFFSET),
	DEFINEREG(VIDEO_REFERENCE_BUFFER_BASEADDRESS),
	DEFINEREG(VIDEO_BCD_FLAG_BUFFER_BASEADDRESS),
	DEFINEREG(VIDEO_COMPRESS_DESTINATION_BASEADDRESS),
	DEFINEREG(VIDEO_COMPRESS_SIZE_REGISTER),
	DEFINEREG(VIDEO_COMPRESS_WRITE_OFFSET),
	DEFINEREG(VIDEO_COMPRESS_CONTROL),
	DEFINEREG(COMPRESS_DATA_COUNT_REGISTER),
	DEFINEREG(COMPRESS_BLOCK_COUNT_REGISTER),
	DEFINEREG(COMPRESS_BUFFER_LAST_END_FRAME_OFFSET),
	DEFINEREG(COMPRESS_FRAME_COUNT_REGISTER),
	DEFINEREG(H_DETECTION_STATUS),
	DEFINEREG(V_DETECTION_STATUS),
	DEFINEREG(MODEDETECTION_STATUS_READBACK),
	DEFINEREG(VIDEO_CONTROL_REGISTER),
	DEFINEREG(VIDEO_INTERRUPT_CONTROL),
	DEFINEREG(VIDEO_INTERRUPT_STATUS),
	DEFINEREG(MODE_DETECTION_REGISTER),
	DEFINEREG(FRONT_BOUND_REGISTER),
	DEFINEREG(END_BOUND_REGISTER),
	DEFINEREG(CRC_1_REGISTER),
	DEFINEREG(CRC_2_REGISTER),
	DEFINEREG(REDUCE_BIT_REGISTER),
	DEFINEREG(BIOS_SCRATCH_REGISTER),
	{NULL, 0}		// Last entry required to be a null entry to mark end of registers  
};

static char gRegName[32];
char *getVideoRegisterName(u32 iRegNo)
{
	int i;
	char *retName = NULL;
	for (i = 0; videoRegs[i].szRegName != NULL; i++) {
		if (iRegNo == videoRegs[i].iRegNo)
			retName = videoRegs[i].szRegName;
	}
	if (retName == NULL) {
		sprintf(gRegName, "%8.8X", (int)iRegNo);
		retName = gRegName;
	}
	return (retName);
}

u32 RegisterRead(u32 reg, u32 fromLine)
{
	u32 thisReg = *((volatile u32 *)(IO_ADDRESS(reg)));
	if (gDebugFlags & DEBUG_REGISTER_IO_READS)
		printk("%s[%d] from %d Read %s = 0x%8.8X\n",
		       __FUNCTION__, __LINE__,
		       (int)fromLine, getVideoRegisterName(reg), (int)thisReg);
	return (thisReg);
}

void RegisterWrite(u32 reg, u32 value, u32 fromLine)
{
	if (gDebugFlags & DEBUG_REGISTER_IO_WRITES)
		printk("%s[%d] from %d Write %s = 0x%8.8X\n",
		       __FUNCTION__, __LINE__,
		       (int)fromLine, getVideoRegisterName(reg), (int)value);
	*((volatile u32 *)(IO_ADDRESS(reg))) = (value);
	if (gDebugFlags & DEBUG_REGISTER_IO_WRITES) {
		unsigned long ulReadBack = *((volatile u32 *)(IO_ADDRESS(reg)));
		if (ulReadBack != value) {
			printk
			    ("%s[%d] from %d %s Readback failed 0x%8.8X S/B 0x%8.8X\n",
			     __FUNCTION__, __LINE__, (int)fromLine,
			     getVideoRegisterName(reg), (int)ulReadBack,
			     (int)value);
		}
	}
}

void DumpVideoEngineRegisters(void)
{
	int i;

	for (i = 0; videoRegs[i].szRegName != NULL; i++) {
		printk("drv- %-35.35s VR%3.3X = 0x%8.8X\n",
		       videoRegs[i].szRegName, videoRegs[i].iRegNo % 0x1000,
		       (int)REGISTERREAD(videoRegs[i].iRegNo));
	}
}

#define PCONFIGCHG(field,desc) printk("drv-%-20.20s     = %d\n",desc,(int)videoConfig.field);

void PrintVideoConfig(void)
{
	printk("drv-%s\n", __FUNCTION__);
	PCONFIGCHG(ulMaxSourceXWidth, "MaxSourceX");
	PCONFIGCHG(ulMaxSourceYHeight, "MaxSourceY");
	PCONFIGCHG(ulPacketCount, "PacketCount");
	PCONFIGCHG(ulPacketSize, "PacketSizeK");
	PCONFIGCHG(ulRingMax, "RingMax");
	PCONFIGCHG(ulHighWaterMark, "HighWaterMark");
	PCONFIGCHG(ulLowWaterMark, "LowWaterMark");
	PCONFIGCHG(ulHaltMark, "HaltMark");
	PCONFIGCHG(CompressionMode, "CompressionMode");
	PCONFIGCHG(Y_JPEGTableSelector, "Y_JPEGTableSelector");
	PCONFIGCHG(UV_JPEGTableSelector, "UV_JPEGTableSelector");
	PCONFIGCHG(ulNormalFrameRate, "NormalFrameRate");
	PCONFIGCHG(ulCongestionFrameRate, "CongestionFrameRate");
	PCONFIGCHG(BCDControl, "BCDControl");
	PCONFIGCHG(AnalogBCDTolerance, "AnalogBCDTolerance");
	PCONFIGCHG(DigitalBCDTolerance, "DigitalBCDTolerance");
	PCONFIGCHG(DownScaleMode, "DownScaleMode");
	PCONFIGCHG(DestinationModeX, "DestinationModeX");
	PCONFIGCHG(DestinationModeY, "DestinationModeY");
	PCONFIGCHG(InputSignalSource, "InputSignalSource");
	PCONFIGCHG(CaptureMode, "CaptureMode");
	PCONFIGCHG(PacketMode, "PacketMode");
	PCONFIGCHG(DirectMode, "DirectMode");
	PCONFIGCHG(ReduceBitControl, "ReduceBitControl");
}

#if INCLUDE_EXTERNAL_VGA_SUPPORT
//  Parameters
#define    SAMPLE_RATE                                 24000000
#define    HORIZONTAL_SCALING_FILTER_PARAMETERS_LOW    0xFFFFFFFF
#define    HORIZONTAL_SCALING_FILTER_PARAMETERS_HIGH   0xFFFFFFFF
#define    VIDEO_WRITE_BACK_BUFFER_THRESHOLD_LOW       0x08
#define    VIDEO_WRITE_BACK_BUFFER_THRESHOLD_HIGH      0x04
#define    VQ_Y_LEVELS                                 0x10
#define    VQ_UV_LEVELS                                0x05
#define    EXTERNAL_VIDEO_HSYNC_POLARITY               0x01
#define    EXTERNAL_VIDEO_VSYNC_POLARITY               0x01
#define    VIDEO_SOURCE_FROM                           0x01
#define    EXTERNAL_ANALOG_SOURCE                      0x01
#define    USE_INTERNAL_TIMING_GENERATOR               0x01
#define    WRITE_DATA_FORMAT                           0x00
#define    SET_BCD_TO_WHOLE_FRAME                      0x01
#define    ENABLE_VERTICAL_DOWN_SCALING                0x01
#define    BCD_TOLERENCE                               0xFF
#define    BCD_START_BLOCK_XY                          0x0
#define    BCD_END_BLOCK_XY                            0x3FFF
#define    COLOR_DEPTH                                 16
#define    BLOCK_SHARPNESS_DETECTION_HIGH_THRESHOLD    0xFF
#define    BLOCK_SHARPNESS_DETECTION_LOE_THRESHOLD     0xFF
#define    BLOCK_SHARPNESS_DETECTION_HIGH_COUNTS_THRESHOLD   0x3F
#define    BLOCK_SHARPNESS_DETECTION_LOW_COUNTS_THRESHOLD    0x1F
#define    VQTABLE_AUTO_GENERATE_BY_HARDWARE           0x0
#define    VQTABLE_SELECTION                           0x0
#define    JPEG_COMPRESS_ONLY                          0x0
#define    DUAL_MODE_COMPRESS                          0x1
#define    BSD_H_AND_V                                 0x0
#define    ENABLE_RC4_ENCRYPTION                       0x1
#define    BSD_ENABLE_HIGH_THRESHOLD_CHECK             0x0
#define    VIDEO_PROCESS_AUTO_TRIGGER                  0x0
#define    VIDEO_COMPRESS_AUTO_TRIGGER                 0x0

// TODO - need to capture one frame
int CaptureSingleFrame(void)
{
	return (1);
}

void AutoPhase_Adjust(void)
{
	int iResHasChanged;
	int i,
	    PhaseSetting,
	    DifferBlock,
	    BestPhase = 0, BestDifferBlock = -1, BCDSetting, DifferSize;
	IO_ACCESS_DATA IO_Data;

	memset(&IO_Data, 0, sizeof(IO_ACCESS_DATA));
	for (i = 0; i < 32; i++) {
		PhaseSetting = (i << 3);
		aspeedi2c_write(2, 0x98, 0x04, PhaseSetting);
		InitializeVideoAll();
#if INCLUDE_DOWNSCALING
		HorizontalDownScaling();
		VerticalDownScaling();
#endif
		REGISTERANDWRITE(VIDEOENGINE_SEQUENCE_CONTROL, 0xFFFFFFDF,
				 0x20);
		iResHasChanged = CaptureSingleFrame();
		if (iResHasChanged != 0) {
			return;
		}
		DifferBlock =
		    (REGISTERREAD(COMPRESS_BLOCK_COUNT_REGISTER) & 0xFFFF0000)
		    >> 16;

		SwapInAndRefBuffers();
		// Now that we have a frame, set BCD using analog tolerance value
		BCDSetting = 1 + (videoConfig.AnalogBCDTolerance << 16);
		REGISTERANDWRITE(VIDEO_BCD_CONTROL, 0, BCDSetting);

		REGISTERANDWRITE(VIDEOENGINE_SEQUENCE_CONTROL, 0xFFFFFFDF,
				 0x20);
		iResHasChanged = CaptureSingleFrame();
		if (iResHasChanged != 0) {
			return;
		}

		DifferBlock =
		    (REGISTERREAD(COMPRESS_BLOCK_COUNT_REGISTER) & 0xFFFF0000)
		    >> 16;
		DifferSize =
		    (REGISTERREAD(COMPRESS_BLOCK_COUNT_REGISTER) & 0xFFFFF);
		if ((DifferBlock == 0) && (DifferSize != 12)) {
			continue;
		}
		if (DifferBlock < BestDifferBlock) {
			BestDifferBlock = (int)DifferBlock;
			BestPhase = PhaseSetting;
			if (BestDifferBlock <= 2) {
				break;
			}
		}
	}
	aspeedi2c_write(2, 0x98, 0x04, (unsigned char)BestPhase);
	if (gDebugFlags & DEBUG_MODE_DETECTION)
		printk("%s[%d] Best Phase = %d, BestDiffBlock = %d\n",
		       __FUNCTION__, __LINE__, BestPhase, BestDifferBlock);
	return;
}

u32 JudgeMode(u32 HorizontalCount, u32 VerticalLine)
{
	int i, ModeIndex = -1;
	int HorizontalFrequencyDiffInHz;
	int VerticalScanLineDiff;

	if ((HorizontalCount == 0xFFF) || (VerticalLine == 0xFFF)) {
		return (0);
	}
	if (gDebugFlags & DEBUG_MODE_DETECTION) {
		printk("drv-%s[%d] Checking for mode nearest to %d x %d\n",
		       __FUNCTION__, __LINE__,
		       (int)HorizontalCount, (int)VerticalLine);
	}
//  Find nearest mode in 60Hz table list
// TODO: Why are we doing this first and then possible resetting it later.  Do we simply
// want to prefer at 60Hz refresh rate.
// TODO: Note that 18 is hard-coded to exactly match the number of 60Hz entries
// in the table.
// TODO: When we find a first entry, we don't brak, but just keep going.
	for (i = 0; i < 18; i++) {
		if (VerticalLine >= ModeTable[i].VerticalTotal) {
			ModeIndex = i;
			videoStatus.SourceModeInfo.ModeIndex = i;
			videoStatus.SourceModeInfo.X =
			    ModeTable[i].HorizontalActive;
			videoStatus.SourceModeInfo.Y =
			    ModeTable[i].VerticalActive;
			videoStatus.SourceModeInfo.RefreshRate = 60;
		}
	}

	for (i = 0; i < ModeNumberCount; i++) {
//  Calculate horizontal frequency rate
		HorizontalFrequencyDiffInHz =
		    (SAMPLE_RATE / HorizontalCount) -
		    ModeTable[i].HorizontalFrequency;
		if (HorizontalFrequencyDiffInHz < 0)
			HorizontalFrequencyDiffInHz =
			    -HorizontalFrequencyDiffInHz;
		VerticalScanLineDiff =
		    VerticalLine - ModeTable[i].VerticalTotal;
		if (VerticalScanLineDiff < 0)
			VerticalScanLineDiff = -VerticalScanLineDiff;
		if (gDebugFlags & DEBUG_MODE_DETECTION) {
			printk
			    ("drv-%s[%d] Entry %d H Hz %d H Hz Diff %d V Lines %d V Line Diff %d\n",
			     __FUNCTION__, __LINE__, i,
			     (int)ModeTable[i].HorizontalFrequency,
			     HorizontalFrequencyDiffInHz,
			     (int)ModeTable[i].VerticalTotal,
			     VerticalScanLineDiff);
		}

		if (HorizontalFrequencyDiffInHz <= 800 &&
		    VerticalScanLineDiff <= 2) {
			ModeIndex = i;
			videoStatus.SourceModeInfo.ModeIndex = i;
			videoStatus.SourceModeInfo.X =
			    ModeTable[i].HorizontalActive;
			videoStatus.SourceModeInfo.Y =
			    ModeTable[i].VerticalActive;
			videoStatus.SourceModeInfo.RefreshRate =
			    ModeTable[i].RefreshRate;
			break;
		}
	}
	if (ModeIndex >= 0) {
		i = ModeIndex;

		if (gDebugFlags & DEBUG_MODE_DETECTION) {
			printk
			    ("drv-%s[%d] Selected Entry %d %d x %d @ %d Hz\n",
			     __FUNCTION__, __LINE__, i,
			     videoStatus.SourceModeInfo.X,
			     videoStatus.SourceModeInfo.Y,
			     videoStatus.SourceModeInfo.RefreshRate);
		}

		REGISTERANDWRITE(VIDEOENGINE_MODEDETECTIONSETTING_H, 0xFFFFE000,
				 ModeTable[i].HBackPorch +
				 ModeTable[i].HLeftBorder +
				 ModeTable[i].HorizontalActive);
		REGISTERANDWRITE(VIDEOENGINE_MODEDETECTIONSETTING_H, 0xE000FFFF,
				 (ModeTable[i].HBackPorch +
				  ModeTable[i].HLeftBorder) << 16);
		REGISTERANDWRITE(VIDEOENGINE_MODEDETECTIONSETTING_V, 0xFFFFF000,
				 ModeTable[i].VBackPorch +
				 ModeTable[i].VTopBorder +
				 ModeTable[i].VerticalActive);
		REGISTERANDWRITE(VIDEOENGINE_MODEDETECTIONSETTING_V, 0xF000FFFF,
				 (ModeTable[i].VBackPorch +
				  ModeTable[i].VTopBorder) << 16);

		//External Analog
		if ((videoConfig.InputSignalSource ==
		     INPUT_SIGNAL_SOURCE_EXTERNAL)
		    && (videoStatus.InputSignal == 1)
		    ) {
			aspeedi2c_write(2, 0x98, 0x01,
					ADCModeTable[i].ADCIndex1);
			aspeedi2c_write(2, 0x98, 0x02,
					ADCModeTable[i].ADCIndex2);
			aspeedi2c_write(2, 0x98, 0x03,
					ADCModeTable[i].ADCIndex3);
			aspeedi2c_write(2, 0x98, 0x05,
					ADCModeTable[i].ADCIndex5);
			aspeedi2c_write(2, 0x98, 0x06,
					ADCModeTable[i].ADCIndex6);
			aspeedi2c_write(2, 0x98, 0x07,
					ADCModeTable[i].ADCIndex7);
			aspeedi2c_write(2, 0x98, 0x08,
					ADCModeTable[i].ADCIndex8);
			aspeedi2c_write(2, 0x98, 0x09,
					ADCModeTable[i].ADCIndex9);
			aspeedi2c_write(2, 0x98, 0x0a,
					ADCModeTable[i].ADCIndexA);
			aspeedi2c_write(2, 0x98, 0x0e, 0x64);
			aspeedi2c_write(2, 0x98, 0x0f,
					ADCModeTable[i].ADCIndexF);
			aspeedi2c_write(2, 0x98, 0x15,
					ADCModeTable[i].ADCIndex15);
		}
		return (1);
	} else {
		if (gDebugFlags & DEBUG_MODE_DETECTION)
			printk("drv-%s[%d] JudgeMode - no mode selected\n",
			       __FUNCTION__, __LINE__);
	}
	return (0);
}
#endif				// INCLUDE_EXTERNAL_VGA_SUPPORT

#ifdef	PROCFS

static const char proc_node_registers_name[] = "driver/ast2050_video_registers";
static const char proc_node_config_name[] = "driver/ast2050_video_config";
static const char proc_node_intmasks_name[] = "driver/ast2050_video_intmasks";
static const char proc_node_ringdesc_name[] = "driver/ast2050_video_ringdescriptor";

static int
proc_read_registers(char *page, char **start, off_t off, int count,
              int *eof, void *_dev)
{
	char *buf = page;
	char *next = buf;
	unsigned size = count;
	int i, t;

        if (off != 0)
                return 0;

	for (i = 0; videoRegs[i].szRegName != NULL; i++) {
		t = scnprintf(next, size, "%-35.35s VR%3.3X = 0x%8.8X\n",
		       videoRegs[i].szRegName, videoRegs[i].iRegNo % 0x1000,
		       (int)REGISTERREAD(videoRegs[i].iRegNo));
		size -= t;
		next += t;
	}

	*eof = 1;
	return count - size;
}

static int
proc_read_config(char *page, char **start, off_t off, int count,
              int *eof, void *_dev)
{
	char *buf = page;
	char *next = buf;
	unsigned size = count;
	int t;

        if (off != 0)
                return 0;

	t = scnprintf(next, size, "%-20.20s     = %d\n", "MaxSourceX",
		(int)videoConfig.ulMaxSourceXWidth);
	size -= t;
	next += t;
	t = scnprintf(next, size, "%-20.20s     = %d\n", "MaxSourceY",
		(int)videoConfig.ulMaxSourceYHeight);
	size -= t;
	next += t;
	t = scnprintf(next, size, "%-20.20s     = %d\n", "PacketCount",
		(int)videoConfig.ulPacketCount);
	size -= t;
	next += t;
	t = scnprintf(next, size, "%-20.20s     = %d\n", "PacketSizeK",
		(int)videoConfig.ulPacketSize);
	size -= t;
	next += t;
	t = scnprintf(next, size, "%-20.20s     = %d\n", "RingMax",
		(int)videoConfig.ulRingMax);
	size -= t;
	next += t;
	t = scnprintf(next, size, "%-20.20s     = %d\n", "HighWaterMark",
		(int)videoConfig.ulHighWaterMark);
	size -= t;
	next += t;
	t = scnprintf(next, size, "%-20.20s     = %d\n", "LowWaterMark",
		(int)videoConfig.ulLowWaterMark);
	size -= t;
	next += t;
	t = scnprintf(next, size, "%-20.20s     = %d\n", "HaltMark",
		(int)videoConfig.ulHaltMark);
	size -= t;
	next += t;
	t = scnprintf(next, size, "%-20.20s     = %d\n", "CompressionMode",
		(int)videoConfig.CompressionMode);
	size -= t;
	next += t;
	t = scnprintf(next, size, "%-20.20s     = %d\n", "Y_JPEGTableSelector",
		(int)videoConfig.Y_JPEGTableSelector);
	size -= t;
	next += t;
	t = scnprintf(next, size, "%-20.20s     = %d\n", "UV_JPEGTableSelector",
		(int)videoConfig.UV_JPEGTableSelector);
	size -= t;
	next += t;
	t = scnprintf(next, size, "%-20.20s     = %d\n", "NormalFrameRate",
		(int)videoConfig.ulNormalFrameRate);
	size -= t;
	next += t;
	t = scnprintf(next, size, "%-20.20s     = %d\n", "CongestionFrameRate",
		(int)videoConfig.ulCongestionFrameRate);
	size -= t;
	next += t;
	t = scnprintf(next, size, "%-20.20s     = %d\n", "BCDControl",
		(int)videoConfig.BCDControl);
	size -= t;
	next += t;
	t = scnprintf(next, size, "%-20.20s     = %d\n", "AnalogBCDTolerance",
		(int)videoConfig.AnalogBCDTolerance);
	size -= t;
	next += t;
	t = scnprintf(next, size, "%-20.20s     = %d\n", "DigitalBCDTolerance",
		(int)videoConfig.DigitalBCDTolerance);
	size -= t;
	next += t;
	t = scnprintf(next, size, "%-20.20s     = %d\n", "DownScaleMode",
		(int)videoConfig.DownScaleMode);
	size -= t;
	next += t;
	t = scnprintf(next, size, "%-20.20s     = %d\n", "DestinationModeX",
		(int)videoConfig.DestinationModeX);
	size -= t;
	next += t;
	t = scnprintf(next, size, "%-20.20s     = %d\n", "DestinationModeY",
		(int)videoConfig.DestinationModeY);
	size -= t;
	next += t;
	t = scnprintf(next, size, "%-20.20s     = %d\n", "InputSignalSource",
		(int)videoConfig.InputSignalSource);
	size -= t;
	next += t;
	t = scnprintf(next, size, "%-20.20s     = %d\n", "CaptureMode",
		(int)videoConfig.CaptureMode);
	size -= t;
	next += t;
	t = scnprintf(next, size, "%-20.20s     = %d\n", "PacketMode",
		(int)videoConfig.PacketMode);
	size -= t;
	next += t;
	t = scnprintf(next, size, "%-20.20s     = %d\n", "DirectMode",
		(int)videoConfig.DirectMode);
	size -= t;
	next += t;
	t = scnprintf(next, size, "%-20.20s     = %d\n", "ReduceBitControl",
		(int)videoConfig.ReduceBitControl);
	size -= t;
	next += t;

	*eof = 1;
	return count - size;
}

static int
proc_read_intmasks(char *page, char **start, off_t off, int count,
              int *eof, void *_dev)
{
	char *buf = page;
	char *next = buf;
	unsigned size = count;
	int i, t, index;

        if (off != 0)
                return 0;

	index = nmask;
	for (i = 0; i<MAX_INTS; i++) {
		index = (index == 0) ? MAX_INTS-1 : index-1;
		t = scnprintf(next, size, "int[%2.2d] : frame=%8.8X packet=%8.8X status=%8.8X\n",
			index, (intmasks[index] >> 16) & 0xFF,
			(intmasks[index] >> 8) & 0xFF,
			intmasks[index] & 0xFF);
		size -= t;
		next += t;
	}

	*eof = 1;
	return count - size;
}

static int
proc_read_ringdesc(char *page, char **start, off_t off, int count,
              int *eof, void *_dev)
{
	char *buf = page;
	char *next = buf;
	unsigned size = count;
	int t;

        if (off != 0)
                return 0;

	t = scnprintf(next, size, "RingDesc: Max=%8.8X, HighWater=%8.8X, LowWater=%8.8X\n", 
		theRing.ulRingMax, theRing.ulHighWaterMark,
		theRing.ulLowWaterMark);
	size -= t;
	next += t;

	t = scnprintf(next, size, "RingDesc: Stalled=%8.8X, RingCnt=%8.8X, FrameCnt=%8.8X\n", 
		theRing.bStalled, theRing.ulRingCount,
		theRing.ulFrameCount);
	size -= t;
	next += t;

	t = scnprintf(next, size, "RingDesc: Status=%8.8X, WriteIdx=%8.8X, ReadIdx=%8.8X\n", 
		theRing.ulRingStatus, theRing.ulRingWrite,
		theRing.ulRingRead);
	size -= t;
	next += t;

	*eof = 1;
	return count - size;
}

void create_proc_files(void)
{
	create_proc_read_entry(proc_node_registers_name, 0777, NULL, proc_read_registers, NULL);
	create_proc_read_entry(proc_node_config_name, 0, NULL, proc_read_config, NULL);
	create_proc_read_entry(proc_node_intmasks_name, 0, NULL, proc_read_intmasks, NULL);
	create_proc_read_entry(proc_node_ringdesc_name, 0, NULL, proc_read_ringdesc, NULL);
}

void remove_proc_files(void)
{
	remove_proc_entry(proc_node_registers_name, NULL);
	remove_proc_entry(proc_node_config_name, NULL);
	remove_proc_entry(proc_node_intmasks_name, NULL);
	remove_proc_entry(proc_node_ringdesc_name, NULL);
}
#endif
