/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* AST2050 KVM video driver */

#ifndef AESS_VIDEO_H
#define AESS_VIDEO_H

// When there is 64MB of memory, support no more than 1280x1024 resolution
// ASpeed driver needs 11MB in addition to the normal 8MB for VGA buffer space
// reserved for the VGA driver
// Note that for EMS - this memory is not contiguous - the 11MB is dynamically
// allocated in two chunks - 5MB for reference buffer and 6MB for capture buffer
#define	MAX_HORIZONTAL_X_RESOLUTION_64MB	1280
#define	MAX_VERTICAL_Y_RESOLUTION_64MB		1024

#define	REFERENCE_BUFFER_SIZE_EMS		(5*1024*1024)	
#define	CAPTURE_BUFFER_SIZE_EMS			(6*1024*1024)	

// When there is 128MB of memory, support up to 1280x1024 resolution.
// ASpeed driver needs 11MB in addition to the normal 8MB for VGA buffer space
// reserved for the VGA driver
#define	MAX_HORIZONTAL_X_RESOLUTION_128MB 	1280
#define	MAX_VERTICAL_Y_RESOLUTION_128MB   	1024

#define	LOWEST_AVAILABLE_DRAM_ADDRESS_128MB_EVB	(109*1024*1024)	// MEM=109M

#define	MODEDETECTION_VERTICAL_STABLE_MAXIMUM       0x6
#define	MODEDETECTION_HORIZONTAL_STABLE_MAXIMUM     0x6
#define	MODEDETECTION_VERTICAL_STABLE_THRESHOLD     0x2
#define	MODEDETECTION_HORIZONTAL_STABLE_THRESHOLD   0x2

#ifdef APP_BUILD
#define REGISTERWRITE(reg,value) WriteMMIOLong(reg,value)
#define REGISTERANDWRITE(reg,mask,value) WriteMMIOLongAndData(reg, mask, value)
#define REGISTERREAD(reg) ReadMMIOLong(reg)
#else
#define REGISTERWRITE(reg,value) RegisterWrite(reg,value,__LINE__)
#define REGISTERANDWRITE(reg,mask,value) RegisterWrite(reg,( (*((u32 *)(IO_ADDRESS(reg)))) & (mask)) | (value),__LINE__)
#define REGISTERREAD(reg) RegisterRead(reg,__LINE__)
#endif

#define IOCTL_IO_READ				0x1101
#define IOCTL_IO_WRITE   			0x1102
#define IOCTL_I2C_READ        			0x1103
#define IOCTL_I2C_WRITE       			0x1104

#define IOCTL_REQ_MODE_DETECT			0x1110
#define IOCTL_GET_VIDEOSTATUS			0x1111
#define IOCTL_SET_VIDEOCONFIG   		0x1112
#define IOCTL_INIT_CAPTURE			0x1116
#define IOCTL_TRIGGER_CAPTURE 			0x1117
#define IOCTL_WAIT_CAPTURE			0x1118
#define IOCTL_ACK_RING_ENTRY  			0x1119
#define IOCTL_HALT_CAPTURE			0x111A
#define IOCTL_GET_HWCURSOR			0x111C

#define IOCTL_FORCE_FULL_FRAME  		0x1120

#define IOCTL_READ_FRAME_DESC  			0x1130
#define IOCTL_BIT_STREAM_BASE			0x1131

#define	IOCTL_DISABLE_VIDEO			0x1140
#define	IOCTL_ENABLE_VIDEO			0x1141

#define IOCTL_READ_INTERRUPTS			0x1190
#define IOCTL_CLEAR_INTERRUPTS			0x1191
#define IOCTL_SET_DEBUG				0x1192

// Driver debug enable options set via IOCTL_SET_DEBUG
#define DEBUG_PRINTK_LEVEL			0x000000FF
#define DEBUG_VIDEO_SETTING_CHANGES 		0x00000100
#define DEBUG_TRACK_RING_RETURNS 		0x00000200
#define DEBUG_TRACK_INTERRUPTS			0x00000400
#define DEBUG_INTERRUPT_LEVEL_RING_CHECK	0x00000800
#define DEBUG_SINGLE_FRAME_BUFFERS		0x00001000
#define DEBUG_MODE_DETECTION			0x00002000
#define DEBUG_RING_STATUS			0x00004000
#define DEBUG_REGISTER_IO_READS			0x00008000
#define DEBUG_REGISTER_IO_WRITES		0x00010000

#define MAX_INTS				50

// Registers for the System Control Unit
#define	SCU_KEY_CONTROL				0x1E6E2000
#define	SCU_RESET_CONTROL			0x1E6E2004
#define SCU_RESET_CONTROL_VIDEOENGINE		0x00000040
#define	SCU_CLOCK_SELECTION_REGISTER		0x1E6E2008
#define	SCU_CLOCK_CONTROL_REGISTER		0x1E6E200C

// Registers for the Video Engine
#define	VIDEOBASE_OFFSET			0x1E700000
#define	VIDEOENGINE_KEY_CONTROL			0x00 + VIDEOBASE_OFFSET
#define	VIDEOENGINE_SEQUENCE_CONTROL		0x04 + VIDEOBASE_OFFSET
#define	VIDEOENGINE_SEQUENCE_CONTROL_COMPRESSION_IDLE	0x00040000
#define	VIDEOENGINE_SEQUENCE_CONTROL_CAPTURE_IDLE	0x00010000

#define	VIDEOENGINE_VIDEO_CONTROL		0x08 + VIDEOBASE_OFFSET
#define	VIDEOENGINE_MODEDETECTIONSETTING_H	0x0C + VIDEOBASE_OFFSET
#define	VIDEOENGINE_MODEDETECTIONSETTING_V	0x10 + VIDEOBASE_OFFSET
#define	SCALE_FACTOR_REGISTER			0x14 + VIDEOBASE_OFFSET
#define	SCALING_FILTER_PARAMETERS_1		0x18 + VIDEOBASE_OFFSET
#define	SCALING_FILTER_PARAMETERS_2		0x1C + VIDEOBASE_OFFSET
#define	SCALING_FILTER_PARAMETERS_3		0x20 + VIDEOBASE_OFFSET
#define	SCALING_FILTER_PARAMETERS_4		0x24 + VIDEOBASE_OFFSET
#define	VIDEO_BCD_CONTROL			0x2C + VIDEOBASE_OFFSET
#define	VIDEO_CAPTURE_WINDOW_SETTING		0x30 + VIDEOBASE_OFFSET
#define	VIDEO_COMPRESS_WINDOW_SETTING		0x34 + VIDEOBASE_OFFSET
#define	VIDEO_COMPRESS_PROCESSING_OFFSET	0x38 + VIDEOBASE_OFFSET
#define	VIDEO_COMPRESS_READ_OFFSET		0x3C + VIDEOBASE_OFFSET
#define	VIDEO_CRC_BUFFER_BASEADDRESS		0x40 + VIDEOBASE_OFFSET
#define	VIDEO_IN_BUFFER_BASEADDRESS		0x44 + VIDEOBASE_OFFSET
#define	VIDEO_IN_BUFFER_OFFSET			0x48 + VIDEOBASE_OFFSET
#define	VIDEO_REFERENCE_BUFFER_BASEADDRESS	0x4C + VIDEOBASE_OFFSET
#define	VIDEO_BCD_FLAG_BUFFER_BASEADDRESS	0x50 + VIDEOBASE_OFFSET
#define	VIDEO_COMPRESS_DESTINATION_BASEADDRESS	0x54 + VIDEOBASE_OFFSET
#define	VIDEO_COMPRESS_SIZE_REGISTER		0x58 + VIDEOBASE_OFFSET
#define	VIDEO_COMPRESS_WRITE_OFFSET		0x5C + VIDEOBASE_OFFSET
#define	VIDEO_COMPRESS_CONTROL			0x60 + VIDEOBASE_OFFSET
#define	COMPRESS_DATA_COUNT_REGISTER		0x70 + VIDEOBASE_OFFSET
#define	COMPRESS_BLOCK_COUNT_REGISTER		0x74 + VIDEOBASE_OFFSET
#define	COMPRESS_BUFFER_LAST_END_FRAME_OFFSET	0x78 + VIDEOBASE_OFFSET
#define	COMPRESS_FRAME_COUNT_REGISTER		0x7C + VIDEOBASE_OFFSET
#define	H_DETECTION_STATUS			0x90 + VIDEOBASE_OFFSET
#define	V_DETECTION_STATUS			0x94 + VIDEOBASE_OFFSET
#define	MODEDETECTION_STATUS_READBACK		0x98 + VIDEOBASE_OFFSET
#define	MODEDETECTION_HSYNC_READY		0x80000000
				// 1 - Means signal detected, but not stable
				// 0 - No HSYNC signal detected
#define MODEDETECTION_VSYNC_READY		0x40000000
				// 1 - Means signal detected, but not stable
				// 0 - No VSYNC signal detected
#define MODEDETECTION_HSYNC_POLARITY		0x20000000
				// 1 - Negative Polarity
				// 0 - Positive Polarity
#define MODEDETECTION_VSYNC_POLARITY		0x10000000
				// 1 - Negative Polarity
				// 0 - Positive Polarity
#define MODEDETECTION_VSCAN_COUNT		0x0FFF0000
				// 0xFFF - no VSYNC signal detected
				// Vertical scan line count
#define MODEDETECTION_OUT_OF_SYNC		0x00008000
				// Note - only valid when mode detect watchdog is enabled
				// 1 - Video is out of sync, need to mode detect
				// 0 - Video mode is still stable after previous mode detection
#define MODEDETECTION_VSYNC_OUT_OF_SYNC		0x00004000
				// 1 - Vertical signal is not stable
				// 0 - Detected Vertical signal is still stable
#define MODEDETECTION_HSYNC_OUT_OF_SYNC 	0x00002000
				// 1 - Horizontal signal is not stable
				// 0 - Detected Horizontal signal is still stable
#define MODEDETECTION_EXTERNAL_SOURCE		0x00001000
				// 1 - External video source is from analog/digital receiver (ADC)
				// 0 - External video source is from DVI receiver
#define MODEDETECTION_HSCAN_COUNT		0x00000FFF
				// 0xFFF - no HSYNC signal detected
				// Horizontal scan line count

#define	VIDEO_CONTROL_REGISTER			0x300 + VIDEOBASE_OFFSET

#define	VIDEO_INTERRUPT_CONTROL			0x304 + VIDEOBASE_OFFSET
#define	VIDEO_INTERRUPT_STATUS			0x308 + VIDEOBASE_OFFSET
#define VIDEO_INTERRUPT_MODE_WATCHDOG		0x01
#define VIDEO_INTERRUPT_CAPTURE_COMPLETE	0x02
#define VIDEO_INTERRUPT_PACKET_READY		0x04
#define VIDEO_INTERRUPT_COMPRESSION_COMPLETE	0x08
#define VIDEO_INTERRUPT_MODE_READY		0x10
#define VIDEO_INTERRUPT_FRAME_COMPLETE		0x20

#define	MODE_DETECTION_REGISTER			0x30C + VIDEOBASE_OFFSET
#define	FRONT_BOUND_REGISTER			0x310 + VIDEOBASE_OFFSET
#define	END_BOUND_REGISTER			0x314 + VIDEOBASE_OFFSET
#define	CRC_1_REGISTER				0x320 + VIDEOBASE_OFFSET
#define	CRC_2_REGISTER				0x324 + VIDEOBASE_OFFSET
#define	REDUCE_BIT_REGISTER			0x328 + VIDEOBASE_OFFSET
#define	VIDEO_SCRATCH_340_CURSOR_STATUS		0x340 + VIDEOBASE_OFFSET
#define VIDEO_CURSOR_X_OFFSET_MASK		0x3F000000
#define VIDEO_CURSOR_X_OFFSET_SHIFT		24
#define VIDEO_CURSOR_Y_OFFSET_MASK		0x003F0000
#define VIDEO_CURSOR_Y_OFFSET_SHIFT		16
#define VIDEO_CURSOR_TYPE_MASK			0x0200
#define VIDEO_CURSOR_TYPE_SHIFT			9
#define VIDEO_CURSOR_ENABLED_MASK		0x0100
#define	VIDEO_SCRATCH_344_CURSOR_POS		0x344 + VIDEOBASE_OFFSET
#define VIDEO_CURSOR_POS_Y_MASK			0x07FF0000
#define VIDEO_CURSOR_POS_Y_SHIFT		16
#define VIDEO_CURSOR_POS_X_MASK			0x00000FFF
#define VIDEO_CURSOR_POS_X_SHIRT		0
#define	VIDEO_SCRATCH_348_CURSOR_ADDRESS	0x348 + VIDEOBASE_OFFSET
#define VIDEO_CURSOR_ADDRESS_MASK		0x03FFFFF8
#define	BIOS_SCRATCH_REGISTER			0x34C + VIDEOBASE_OFFSET
#define	VIDEO_SCRATCH_REGISTER_34C		0x34C + VIDEOBASE_OFFSET
#define	VIDEO_SCRATCH_REGISTER_35C		0x35C + VIDEOBASE_OFFSET
#define	RC4KEYS_REGISTER			0x400 + VIDEOBASE_OFFSET
#define	VQHUFFMAN_TABLE_REGISTER		0x300 + VIDEOBASE_OFFSET

typedef struct _HWCURSORDESRIPTOR {
	char cCursorPattern[8192];		// 64x64 2bpp bitmap of the actual cursor.
						// Cursor is anchored in the lower left of this pixel square
	u32 ulChecksum;				// Checksum of cursor region
	u32 ulCursorXWidth;			// The pixel location within the frame that is the first pixl
	u32 ulCursorYHeight;			// to 'paint' the cursor
	u32 ulCursorXOffset;			// Pixel offset from lower right corner of the 64x64 bit pixel memeory
	u32 ulCursorYOffset;			// that identifies the first pixel to 'paint' at the cursore location
	u32 ulCursorHotspotX;			// Pixel that represents the hotspot of the cursor
	u32 ulCursorHotspotY;
} HWCURSORDESCRIPTOR, *PHWCURSORDESCRIPTOR;

typedef struct _HWCURSORCACHEENTRY {
	u8 cCursorCacheNumber;			// a number from 1 to MAX_HWCURSOR_CACHE to identify
						// the cached value for this cursor
						// When zero, it means that the entry is 'empty'
	u8 cCursorCacheDirty;			// When 1, this entry is new or changed and program level
						// should forward this entry to the remove video viewer client
	u8 cCursorType;				// Cursor Type 0 = mono, 1 = color
	PHWCURSORDESCRIPTOR theCursorAddress;	// Address of this cursor
	u32 ulLRUCount;				// Least recently used count for this entry
	u16 usCursorXOffset;
	u16 usCursorYOffset;
	HWCURSORDESCRIPTOR theCursor;		// Content of ASPEED HW Cursor entry
} HWCURSORCACHEENTRY, *PHWCURSORCACHEENTRY;

typedef struct _VIDEOENGINECONFIG {
// Required and dedicated SDRAM values for video engine
	u32 TotalMemory;
	u32 VGAMemory;
	u32 ulMaxSourceXWidth;			// Maximum resolution
	u32 ulMaxSourceYHeight;			// Maximum resolution
	u32 ulPacketCount;			// must be 4, 8, 16 or 32.
	u32 ulPacketSize;			// in KB - must be 1, 2, 4, 8, 16, 32, 64, 128
	u32 ulRingMax;				// maximum number of ring entries (<= 64)

	u32 ulHighWaterMark;			// When unacked frames reaches this mark, 
						// frame rate is set to usCongestionFrameRate
	u32 ulLowWaterMark;			// When unacked frames falls below this mark, 
						// frame rate is set to usNormalFrameRate
	u32 ulNormalFrameRate;			// Normal frame rate 
	u32 ulCongestionFrameRate;		// Congestion frame rate when ulHighWaterMark 
						// entries are in the ring
	u32 ulHaltMark;				// Halt the video engine capture when we reach 
						// this mark.  Flag the ring as stalled and force 
						// application level to restart the capture.

	char CompressionMode;
#define COMPRESSIONMODE_JPEG_YUV420 1
#define	COMPRESSIONMODE_JPEG_YUV444 0

	char Y_JPEGTableSelector;

	char UV_JPEGTableSelector;

	char DownScaleMode;
#define DOWNSCALING_DISABLED 0
#define DOWNSCALING_ENABLED  1
	u16 DestinationModeX;
	u16 DestinationModeY;

	char BCDControl;
#define BCDCONTROL_OFF   0
#define BCDCONTROL_PIXEL 1
#define BCDCONTROL_CRC   2
	u16 AnalogBCDTolerance;
	u16 DigitalBCDTolerance;

	char InputSignalSource;
#define INPUT_SIGNAL_SOURCE_INTERNAL 0
#define INPUT_SIGNAL_SOURCE_EXTERNAL 1

	char CaptureMode;
#define CAPTUREMODE_SINGLE   1
#define CAPTUREMODE_MULTIPLE 2
	char DirectMode;
#define	DIRECTMODE_DISABLE 0
#define	DIRECTMODE_ENABLE  1
#define DIRECTMODE_AUTO    2

	char PacketMode;
#define	PACKETMODE_DISABLE 0
#define	PACKETMODE_ENABLE  1

	u16 DelayControl;
// Desired number of compression buffers
	u32 VideoCompressBufferCount;

// Change type flags
	u32 bVideoCompressionChanges;
	u32 bVideoFrameRateChanges;
	u32 bVideoScalingChanges;

	u32 bVideoSourceChanges;
	u32 bVideoBCDChanges;
	u32 bVideoCaptureChanges;

	char ReduceBitControl; 
} VIDEOENGINECONFIG, *PVIDEOENGINECONFIG;

typedef struct _VIDEO_MODE_INFO {
	u16 X;
	u16 Y;
	u16 ColorDepth;
	u16 RefreshRate;
	char ModeIndex;
	char bModeValid;
} VIDEO_MODE_INFO, *PVIDEO_MODE_INFO;

#define MAX_COMPBUF 4
//VIDEO Engine Info
typedef struct _VIDEOENGINESTATUS {
// Computed buffer sized and locations
	u32 VideoCompressBufferSize;		// Based on configured ulPackCount and ulPacketSize
	u32 ulPacketCountBits;			// Translation for VR058 Video Stream Buffer Size Register
	u32 ulPacketSizeBits;			// Translation for VR058 Video Stream Buffer Size Register
	u32 VideoCaptureBufferSize;		// Based on maximum resolution 4 bytes per pixel
	u32 VideoCRCBufferSize;			// Based on maximum resolution 4 bytes per pixel
	u32 VideoFlagBufferSize;		// Based on maximum resolution 4 bits per pixel
	u32 VideoEngineBottomAddress;		// Last available address for video engine
	u32 VideoEngineTopAddress;		// (0x3800000) Video Engine Mem ENDS at 56M where VGA ends

	u32 VideoCompressBufferAddress[MAX_COMPBUF];	// This is the address of the compression buffer(s)
	u32 VideoINBufferAddress;		// This is address of capture buffer 1
	u32 VideoReferenceBufferAddress;	// This is address of capture buffer 2
	u32 VideoCRCBufferAddress;		// This is address of the CRC buffer
	u32 VideoFlagBufferAddress;		// This is address of the BCD buffer

// Compression Buffer Status
	u32 VideoCompressBufStatus[MAX_COMPBUF];
#define COMPRESSION_BUFFER_STATUS_UNUSED 0	// This entry and buffer are not being used
#define COMPRESSION_BUFFER_STATUS_IDLE	 1	// This entry is idle, no compression in progress here
#define COMPRESSION_BUFFER_STATUS_BUSY   2	// This entry is busy capturing/compressing a frames
#define COMPRESSION_BUFFER_STATUS_FULL   3	// This entry has an available compressed frame
	u32 CurrentCompressBufIndex;		// This is the entry that has a capture currently running
						// The IOCTL_ACK_RING_ENTRY tracks progress of this index and
						// determines when the last capture/compression completes. 
						// When there is an IDLE entry and the interrupt routine indicates
						// that it is no longer busy with a compression, the next entry
						// will be used to start another frame capture.

// Current rate info
	u32 FrameRate;				// Current Frame rate (will be one of normal or congested)

// Status parameters provided to the caller
	u32 RefreshRateIndex;
	u32 ColorDepthIndex;
	u32 Mode_PixelClock;
	u32 Mode_Bandwidth;

// Maximum mode info 
	VIDEO_MODE_INFO MaxModeInfo;
// Current mode info
	VIDEO_MODE_INFO SourceModeInfo;

// Status and calculated values
	char InputSignal;			// Valid only when input signal configured for external ADC/DVI
						// Remap of VR098[12]
						// 1 - means that video signal is from ADC
						// 0 - means that video signal is from DVI

	char HaveSignal;			// 1 - means that we have a video input signal
						// If configured for internal VGA, this is determined
						// via the VR35C scrtach register remap to VGA registers
						// If configured for external VGA, this is determined
						// via detection (not stability) of HSYNC and VSYNC
						// 0 - no video input signal is detected
	char HaveHostHWCursorSupport;		// 1 - means that hardware cursor processing is available
						//     from the host.
						// 0 - no hardware cursor available.
	char UsingDirectMode;			// 1 - means that direct mode is enabled and that the engine
						//     must forward cursors to the client.
						// 0 - Cursor is in the captured/compressed image already.  Do
						//     not forward hardare cursors to the client.
	unsigned short MemoryBandwidth;

} VIDEOENGINESTATUS, *PVIDEOENGINESTATUS;

typedef struct {
	u16 HorizontalTotal;
	u16 VerticalTotal;
	u16 HorizontalActive;
	u16 VerticalActive;
	u8 RefreshRate;
	u32 HorizontalFrequency;		// *1000 (Hz, not KHz)
	u16 HSyncTime;
	u16 HBackPorch;
	u16 VSyncTime;
	u16 VBackPorch;
	u16 HLeftBorder;
	u16 HRightBorder;
	u16 VBottomBorder;
	u16 VTopBorder;
} VESA_MODE;

typedef struct {
	u16 HorizontalActive;
	u16 VerticalActive;
	u16 RefreshRate;
	u8 ADCIndex1;
	u8 ADCIndex2;
	u8 ADCIndex3;
	u8 ADCIndex5;
	u8 ADCIndex6;
	u8 ADCIndex7;
	u8 ADCIndex8;
	u8 ADCIndex9;
	u8 ADCIndexA;
	u8 ADCIndexF;
	u8 ADCIndex15;
	int HorizontalShift;
	int VerticalShift;
} ADC_MODE;

typedef struct {
	u16 HorizontalActive;
	u16 VerticalActive;
	u16 RefreshRateIndex;
	u16 PixelClock;
} INTERNAL_MODE;

typedef struct FrameDescriptor {
	u32 ulFrameNumber;
	u32 ulSegmentNumber;
	u32 ulCompressBufIndex;
	u32 ulBaseOffset;
	u32 ulDataStartOffset;
	u32 ulDataEndOffset;
	u32 ulDataCount;
	u32 ulRepeatCount;			// No change repeat count
	u16 usCursorX;				// current X cursor position
	u16 usCursorY;				// current Y cursor position
	u16 usCursorXOffset;			// current X cursor offset
	u16 usCursorYOffset;			// current Y cursor offset
	u16 usHorizontalXRes;
	u16 usVeritcalYRes;
	u32 ulJiffies;
	char cMode;				// CompressionMode for frame
	char cY_JPEGTableSelector;		// Y JPEG table selector for frame
	char cUV_JPEGTableSelector;		// UV JPEG table selector for frame
	char cFrameStatus;
#define VIDEO_FRAME_STATUS_SOF	 	0x01	//  Start of Frame
#define VIDEO_FRAME_STATUS_EOF   	0x02	//  End of Frame
#define VIDEO_FRAME_STATUS_RFR   	0x04	//  Rreference) frame
#define VIDEO_FRAME_STATUS_SINGLE (VIDEO_FRAME_STATUS_SOF|VIDEO_FRAME_STATUS_EOF)
	// A SINGLE frame is one that has only one segment,
	// with both SOF and EOF in the same segment.
#define VIDEO_FRAME_CURSOR_DIRTY 	0x08	// This frame's cursor cache is dirty and
	// needs to be sent to clients
#define VIDEO_FRAME_STATUS_NOCHG 	0x40	// This frame has no changes
#define VIDEO_FRAME_STATUS_PRESENT 	0x80	// There is a valid entry in this ring entry
	char cCursorCacheNumber;		// A number from 1 to MAX_CURSOR_CACHE
	char cCursorType;			// Type 0 - mono, 1 = color
} FRAMEDESCRIPTOR, *PFRAMEDESCRIPTOR;

typedef struct _VIDEOENGINEINFO {
	VIDEOENGINECONFIG videoConfig;
	VIDEOENGINESTATUS videoStatus;
} VIDEOENGINEINFO, *PVIDEOENGINEINFO;

#define MAX_RING_COUNT 64
typedef struct FrameRingDescriptor {
	u32 ulRingMax;				// Max number of ring entries to use (<= MAX_RING_COUNT)
	u32 ulHighWaterMark;			// When unacked frames reaches this mark, frame rate is set to usCongestionFrameRate
	u32 ulLowWaterMark;			// When unacked frames falls below this mark, frame rate is set to usNormalFrameRate
	u32 ulNormalFrameRate;			// Normal frame rate
	u32 ulCongestionFrameRate;		// Congestion frame rate when ulHighWaterMark unacked frames are in the ring
	u32 ulHaltMark;				// Temporarily halt frame capture when we reach this high water mark
	u32 bStalled;				// TRUE if the video engine has been stalled due to ring congestion
	u32 ulRingCount;			// Number of occupied ring entries
	u32 ulFrameCount;			// Number of EOF ring entries in the ring (best estimate of frame count)
	u32 ulRingStatus;			// Status of the frame, video engine and capture ring
#define VIDEO_FRAME_STATUS_MASK	 0x000000FF	// Status of this frame (as in cFrameStatus in FRAMEDESCRIPTOR
#define VIDEO_ENGINE_STATUS_BITS 0x0FFF0000	// Mask defines where the 16 status bits are
						// The first nibble status bits 'lock' once they are set and can only be
						// cleared with a new IOCTL_INIT_MULTIPLE call.
#define VIDEO_CAPTURE_STALLED	 0x80000000	//  Video engine halted due to ring count >= ulHaltMark
#define VIDEO_CAPTURE_OVERRUN	 0x40000000	//      Ring overflow, firmware not keeping up with capture rate, capture halted
#define VIDEO_ENGINE_CONGESTED   0x08000000	//  The Video Engine is operating at the Congestion Frame Rate
#define VIDEO_RESOLUTION_CHANGE  0x04000000	//      A mode watchdog interrupt indicates a change in resolution
#define VIDEO_CAPTURE_IDLE       0x00010000	//  The Video engine capture status from h/w VR004
#define VIDEO_COMPRESSION_IDLE   0x00100000	//  The Video engine compression status from h/w VR004
	u32 ulRingWrite;			// The next entry that the interrupt routine will write a frame 
	u32 ulRingRead;				// The last frame entry read by the firmware
	FRAMEDESCRIPTOR RingEntry[MAX_RING_COUNT];
} FRAMERINGDESCRIPTOR, *PFRAMERINGDESCRIPTOR;

typedef struct _IO_ACCESS_DATA {
	u8 Type;
	u32 Address;
	u32 Data;
	u32 Value;
	u32 I2CValue;
	u32 ulDataBaseOffset;
	u32 ulDataStartOffset;
	u32 ulDataCount;
	u32 ulFrameNumber;
	u32 ulSegmentNumber;
	u32 ulRepeatCount;
	u32 ulFrameStatus;
	u32 ulMode;
	u32 ulY_JPEGTableSelector;
	u32 ulUV_JPEGTableSelector;
	u32 ulCompressBufIndex;
	u32 ulRingCount;
	u32 ulFrameCount;
	u32 ulCursorEntry;
	u16 usCursorX;
	u16 usCursorY;
	u16 usCursorXOffset;
	u16 usCursorYOffset;
	u32 Jiffies;
} IO_ACCESS_DATA, *PIO_ACCESS_DATA;

struct memory_buffers {
	void * reference_phys;			// reference buffer memory
	void * reference_virt;
	u32 reference_size;			// reference buffer size
	void * capture_phys;			// capture buffer memory
	void * capture_virt;	
	u32 capture_size;			// capture buffer size
};
#endif
